/* GSI - general sound interface
 *
 * synth_oss.c (MIDI playing)
 *
 * Lower level API dependencies:
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's? 
 *   - OSS sequencer/synth playing (MIDI's)
 *
 *
 * Copyright (C) W.H.Scholten 1998 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsisynth_config.h>
#include <gsi/config.h>
#if  USE_OSS

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSISYNTH_"
#include <gsi/gsi_debug.h>
#include <gsi/whs_defines.h>

#include "oss_instrument.h"

//#include <gsi/device.h>
#include "../gsisynth.h"

#include "../api.h"

#include "synth_oss.h"
#include "oss_includes.h"

#include "../log.h"


/*__________________DEFINITIES___________________*/



/*__________________EXTERNALS___________________*/
extern int  best_synth_device;
extern int  best_synth;


/*__________________VARIABELEN___________________*/
int  oss_seqfd = -1;
SEQ_DEFINEBUF(1024);
static int  synths_attached = False;


/*__________________PROTOTYPES___________________*/

static char *synth_device_type (int type);
static char *synth_device_subtype (int type);

int api_open_synth_device (char *device);
int api_close_synth_device (void);


gsisynth_device_t default_oss_synth_device = {
	NULL,
	"default synth",/* identification string */

	api_init_synth,
	api_close_synth,
	api_write_synth,
	api_synth_set_timeout,
	api_synth_buffer_status,

	NULL, /* api_load_instrument, */
	NULL, /* api_write_instrument, */
	NULL, /* api_clear_instrument, */
	NULL, /* api_clear_instruments, */
	NULL, /* api_sizeof_default_instrument, */
	api_reset_synth,

	NULL, /* instrument_memory_available */

	0, /* total_instrument_memory */
	100, /* synth_rate */

	// Hardware info:
	0, /* type */
	0, /* num_voices */
	0 /* listno */
};


gsisynth_device_t       myhw;
gsisynth_devicedata_t   myhw_status;
char *api_attach_synth_device(char *device, int i) {
	int  status;
	char *full_description;
	int  num_oss_synths;

	status = api_open_synth_device(device);
	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "(oss) %s: api_attach_synth_device: no such device\n", device);
		return NULL;
	}

	DEB(DPRT("api_attach_synth_device %s %d\n", device, i));

	if (ioctl(oss_seqfd, SNDCTL_SEQ_NRSYNTHS, &num_oss_synths) == -1) {
		log_message(GSI_MESSAGE_ERROR, "(oss) %s: unable to determine number of synthesizers");
		return NULL;
	}
	if (i > num_oss_synths) return NULL;

	full_description = api_attach_synth(&myhw, &myhw_status, i);

	DEB(DPRT("full_description: %s\n", full_description));
	api_close_synth_device();
	DEB(DPRT("#api_attach_synth_device#\n"));

	return device;
}


char *api_attach_synth(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status, int i) {
	int  status;
	struct synth_info this_card;

	DEB1(DPRT("("__FILE__"): *oss_init_synth*\n"));

	this_card.device = i;

	status = ioctl(oss_seqfd, SNDCTL_SYNTH_INFO, &this_card);
	if (status == -1) {
		log_message(GSI_MESSAGE_ERROR, "(oss) %s (devno %d): SYNTH_INFO: %s\n", hw->device, i, strerror(errno));
		return NULL;
	}
	DEB1(DPRT("  -> device %d: `%s', type: `%s' (%d), subtype: `%s' (%d), voices: %d\n",
			  i, this_card.name, synth_device_type(this_card.synth_type), this_card.synth_type,
			  synth_device_subtype(this_card.synth_subtype), this_card.synth_subtype, this_card.nr_voices
		));


	gsisynth_rate = 0; /* just read the current value */
	ioctl(oss_seqfd, SNDCTL_SEQ_CTRLRATE, &gsisynth_rate);
	DEB1(DPRT("  synth_rate=%d\n", gsisynth_rate));
	/* currently 100 (on alpha 1000?) */
	hw->synth_rate = gsisynth_rate;

	/* copy default data: */
	memcpy(hw, &default_oss_synth_device, sizeof(gsisynth_device_t));
	hw->type = SYNTH_HW_NONE;
	hw->num_voices = 0; /* in case this synth is not selected with the compile options (otherwise voices[d] not malloced ..) */

	hw_status->fd = oss_seqfd;

	if (this_card.synth_type == SYNTH_TYPE_SAMPLE && this_card.synth_subtype == SAMPLE_TYPE_GUS) {
#if  USE_GUS_SYNTH
		DEB1(DPRT("  -> Found a GUS\n"));
		hw->num_voices = this_card.nr_voices;

		/* Function table init: */
		hw->init = oss_gus_init;
		//hw->close = oss_gus_done;
		hw->load_instrument = oss_gus_load_instrument;
		hw->write_instrument = oss_gus_write_instrument;
		hw->clear_instrument = oss_gus_clear_instrument;
		hw->clear_instruments = oss_gus_clear_instruments;
		hw->sizeof_default_instrument = oss_gus_sizeof_default_instrument;
		hw->instrument_memory_available = oss_gus_instrument_memory_available;
		hw->listno = i;

		hw->type = SYNTH_HW_GUS;
		if (SYNTH_HW_GUS > best_synth_device) {
			best_synth_device = SYNTH_HW_GUS; best_synth = i;
		}
		return "(oss) GUS";
#endif
	}
	else if (this_card.synth_type == SYNTH_TYPE_FM && (this_card.synth_subtype == FM_TYPE_OPL3 || this_card.synth_subtype == FM_TYPE_ADLIB)) {
#if  USE_OPL3_SYNTH
		DEB1(DPRT("  -> Found a OPL\n"));
		hw->num_voices = this_card.nr_voices;

		// Function table init:
		hw->init = oss_opl_init;
		//hw->close = oss_opl_done;
		hw->load_instrument = oss_opl_load_instrument;
		hw->write_instrument = oss_opl_write_instrument;
		hw->clear_instrument = oss_opl_clear_instrument;
		hw->clear_instruments = oss_opl_clear_instruments;
		hw->sizeof_default_instrument = oss_opl_sizeof_default_instrument;
//X			hw->reset = oss_opl_reset;
		hw->instrument_memory_available = oss_opl_instrument_memory_available;

		hw->total_instrument_memory = 256; /* 1 for each instr. */
		hw->listno = i;
		hw->type = SYNTH_HW_OPL;

		if (SYNTH_HW_OPL > best_synth_device) {
			best_synth_device = SYNTH_HW_OPL; best_synth = i;
		}

		return "(oss) OPL3";
#endif
	}
	else if (this_card.synth_type == SYNTH_TYPE_SAMPLE && this_card.synth_subtype == SAMPLE_TYPE_AWE32) {
#if  USE_AWE_SYNTH
		DEB1(DPRT("  -> Found a AWE32\n"));
		hw->num_voices = this_card.nr_voices;

//			mdata->voices[i] = malloc(sizeof(struct voice_status) * this_card.nr_voices);
		hw->type = SYNTH_HW_AWE32;
		if (SYNTH_HW_AWE32 > best_synth_device) {
			best_synth_device = SYNTH_HW_AWE32; best_synth = i;
		}
		return "(oss) AWE";
#endif
	}

	return NULL;
}


int  api_write_synth(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status) {
	seqbuf_dump();
	return 0;
}


void api_synth_set_timeout (gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status, struct timeval *timeout) {
	if (_seqbufptr >300) timeout->tv_usec = 20000; /* 20 ms, so we write often enough to the sequencer */
}


int api_reset_midi(gsi_song_playstatus_t *spdata) {
	int  i;
	int  synth_type;
	gsi_midi_status_t       *mdata;
	gsisynth_device_t      *hw;
	gsisynth_devicedata_t  *hw_status;

	int  dev;


	if (!spdata) return -1;
	mdata = spdata->song;
	hw = spdata->hw;
	hw_status = spdata->hw_status;

	if (!mdata) return -2;

	dev = hw->listno;
	synth_type = hw->type;

	api_reset_synth(hw, hw_status);

#if  USE_GUS_SYNTH || USE_OPL3_SYNTH
	if (synth_type == SYNTH_HW_OPL || synth_type == SYNTH_HW_GUS) {
		for (i = 0; i < gsisynth_devices[dev].num_voices; i++) {
			SEQ_CONTROL(dev, i, SEQ_VOLMODE, VOL_METHOD_LINEAR);
			//SEQ_CONTROL(dev, i, SEQ_VOLMODE, VOL_METHOD_ADAGIO);
			//if (gsisynth_devices[dev].voices[i].note)
			//	SEQ_STOP_NOTE(dev, i, gsisynth_devices[dev].voices[i].note, 64);
			if (mdata->voices[i].note)
				SEQ_STOP_NOTE(dev, i, mdata->voices[i].note, 64);
		}

		for (i = 0; i < gsisynth_devices[dev].num_voices; i++) {
			mdata->voices[i].available = mdata->voices[i].timestamp = -1;
		}
	}
#endif
#if  USE_AWE_SYNTH
	if (synth_type == SYNTH_HW_AWE32) {
		AWE_SET_CHANNEL_MODE(dev, 1);
		AWE_DRUM_CHANNELS(dev, mdata->midi_percussion_chn);
		AWE_TERMINATE_ALL(dev);
		for (i = 0; i < gsisynth_devices[dev].num_voices; i++) {
			mdata->voices[i].available = mdata->voices[i].timestamp = -1;
		}
	}
#endif
	SEQ_DUMPBUF();

	return 0;
}
int api_reset_synth (gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status) {
	int  synth_type = hw->type;
	int  dev = hw->listno; /* synth identifier */

	_seqbufptr = 0;
	ioctl(hw_status->fd, SNDCTL_SEQ_RESET);

#if  USE_OPL3_SYNTH
	if (synth_type == SYNTH_HW_OPL) {
//ALWAYS USE OPL3 MODE		if (use_opl3)
		ioctl(hw_status->fd, SNDCTL_FM_4OP_ENABLE, &dev);
	}
#endif	
	SEQ_DUMPBUF();

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Remove samples from card memory (to get space for a new song)
 *---------------------------------------------------------------------------*/
int api_clear_instruments(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status)
{
#if  USE_GUS_SYNTH
	if (hw->type == SYNTH_HW_GUS)
		hw->clear_instruments(hw, hw_status);
#endif

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Close the synth device.
 *---------------------------------------------------------------------------*/
int  api_close_synth(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status)
{
	DEB1(DPRT("("__FILE__"): *oss_close_synth*\n"));
	//_seqbufptr = 0;
	//close(oss_seqfd);
	//oss_seqfd = -1;
//XXX	gsisynth_devices[dev].initialized = False;

	DEB1(DPRT("#oss_close_synth#\n"));
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Open the synth device.
 *---------------------------------------------------------------------------*/
#define NON_BLOCKING_SEQUENCER_WRITE
int api_open_synth(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status)
{
	int  oldflags;

	DEB1(DPRT("("__FILE__"): *oss_open_synth*\n"));
	if (oss_seqfd <= 0) {
		oss_seqfd = hw_status->fd = open(hw->device, O_WRONLY);
		if (oss_seqfd == -1) {
			log_message(GSI_MESSAGE_ERROR, "(oss) %s: open device: %s\n", hw->device, strerror(errno));
			return -1;
		}
	}

#ifdef NON_BLOCKING_SEQUENCER_WRITE
	/* Non blocking writes */
	oldflags = fcntl(oss_seqfd, F_GETFL);
	if (oldflags == -1) oldflags = 0; /* on error, assume no flags are set */
	fcntl(oss_seqfd, F_SETFL, oldflags | O_NONBLOCK);
#endif
	DEB1(DPRT("#oss_open_synth#\n"));

	return 0;
}


int api_open_synth_device(char *device) {
	oss_seqfd = open(device, O_WRONLY);
	if (oss_seqfd == -1) {
		return -1;
	}
	return 0;
}


int api_close_synth_device(void) {
	if (oss_seqfd > 0) {
		close(oss_seqfd);
		oss_seqfd = -1;
	}
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Initialize all synthesizer devices found.
 *---------------------------------------------------------------------------*/
DEB(
static char *synth_device_type(int type) {
	switch (type) {
	case SYNTH_TYPE_FM:     return("FM");
	case SYNTH_TYPE_SAMPLE: return("Sampling");
	case SYNTH_TYPE_MIDI:   return("MIDI");
	default:                return("unknown");
	}
}
static char *synth_device_subtype(int type) {
	switch (type) {
	case FM_TYPE_ADLIB:   return("Adlib");
	case FM_TYPE_OPL3:    return("OPL-3");
	case SAMPLE_TYPE_GUS: return("GUS");
	case SAMPLE_TYPE_AWE32: return("AWE32");
	default:              return("unknown");
	}
}
);


int api_init_synth (gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status)
{
	return -1;
}


/*---------------------------------------------------------------------------*
 *      Some sequencer buffer management functions.
 *---------------------------------------------------------------------------*/
enum { _EMPTY=0, _USED, _FULL };

int api_synth_buffer_status(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status) {
	if (_seqbufptr == 0) return 0;
	if (_seqbufptr < 400) return 1;
	return 2;
}


/* is the sequencer (or buffer) ready to accept data or not? */
int api_synth_buffer_full(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status) {
//if (oss_seqfd<0) return True;
	if (_seqbufptr>400) return True;

	return False;
}


/*---------------------------------------------------------------------------*
 *      Send the sequencer data to the sound card.
 *---------------------------------------------------------------------------*/
#define WRITE_FD_V1

#ifdef WRITE_FD_V2
int _seqbeginptr = 0;
void seqbuf_dump (void)
{
	int status;

	DEB1(DPRT("("__FILE__"): *seqbuf_dump*\n"));
	DEB1(DPRT("  _seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));
	if (_seqbufptr) {
		status = write (oss_seqfd, ((char*) _seqbuf)+_seqbeginptr, (_seqbufptr-_seqbeginptr));
		DEB1(DPRT("  status=%d (bytes written)\n", status));
		if (status >0) {
			_seqbeginptr +=status;
		}
		if (_seqbeginptr==_seqbufptr) { _seqbufptr=0; _seqbeginptr=0; }
	}
	DEB1(DPRT("  _seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));
	DEB1(DPRT("#seqbuf_dump#\n"));
}
#endif

void seqbuf_dump_soft(void) {
}


//int seq_dump_fd =-1;
#ifdef WRITE_FD_V1
void seqbuf_dump (void) {
	int status;

	DEB1(DPRT("("__FILE__"): *seqbuf_dump*\n"));
	DEB1(DPRT("  _seqbufptr=%d\n", _seqbufptr));
	if (_seqbufptr) {
		status = write (oss_seqfd, _seqbuf, _seqbufptr);
		DEB1(DPRT("  status=%d (bytes written)\n", status));

		if (status>0) {
//#warning seq dump code..
//if (seq_dump_fd<0) {seq_dump_fd=open("seq_dump.bin", O_WRONLY | O_CREAT, 0660); }
//write(seq_dump_fd, _seqbuf, status);
			memmove(_seqbuf, _seqbuf+status, _seqbufptr-status); 
			_seqbufptr -=status;
		}
	}
	DEB1(DPRT("  _seqbufptr=%d\n", _seqbufptr));
	DEB1(DPRT("#seqbuf_dump#\n"));
}
#endif
#ifdef WRITE_FD_V0
void seqbuf_dump (void) {
	int status, tmp;

	DEB1(DPRT("("__FILE__"): *seqbuf_dump*\n"));
	DEB1(DPRT("  _seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));
	//status = (ioctl(oss_seqfd, SNDCTL_SEQ_GETOUTCOUNT, &tmp) == -1);
	//DEB1(DPRT("status=%d, tmp=%d\n", status, tmp));
	if (_seqbufptr)
		if (write (oss_seqfd, _seqbuf, _seqbufptr) == -1) {
			log_message(GSI_MESSAGE_ERROR, "(oss): seq write error %s", strerror(errno));
			exit (-1);
		}
	_seqbufptr = 0;
	DEB1(DPRT("  _seqbufptr=%d (_seqbeginptr=%d)\n", _seqbufptr, _seqbeginptr));
	DEB1(DPRT("#seqbuf_dump#\n"));
}
#endif


#endif /* USE_OSS */


/*      EOF synth_oss.c */
