/* GSI - general sound interface
 *
 * midi_control.c: controls the synthesizer with midi type events
 * requires: libgsif includes
 *
 * Copyright (C) W.H.Scholten 1998 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsisynth_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSISYNTH_"
#include <gsi/gsi_debug.h>

#include <gsi/gsif_midi.h>

#include "gsisynth.h"

#include "api.h"


/*__________________EXTERNALS___________________*/
extern gsisynth_device_t *synth_devices;


/*---------------------------------------------------------------------------*
 *      Universal sysex reset GM synth controllers (not used).
 *---------------------------------------------------------------------------*/
unsigned char gm_reset[5] = {
	0x7e, 0x7f, 0x09, 0x01, 0xf7
};



/*---------------------------------------------------------------------------*
 *      Prepare to play events.
 *---------------------------------------------------------------------------*/
void gsisynth_seq_prepare(gsi_song_playstatus_t *d) {
	api_seq_prepare(d);
}

/*---------------------------------------------------------------------------*
 *      Reset the sequencer and external midi devices.
 *---------------------------------------------------------------------------*/
void gsisynth_seq_reset(gsi_song_playstatus_t *d) {
	int  i;

	gsi_midi_status_t *mdata = d->song;

	gsisynth_device_t *hw = d->hw;
	gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB1(DPRT("("__FILE__"): *gsisynth_seq_reset*\n"));

	hw->reset(hw, hw_status);

	for (i = 0; i < 16; i++) {
		int  j;

		if (IS_MIDI(i)) {
			if (mdata->midi_channel_output [ i ] == hw->listno) {
				gsisynth_seq_control(d, i,  0, 0);
				gsisynth_seq_control(d, i, 32, 0);
			}
		}

		//gsisynth_seq_set_instrument(i, 0);
		for (j = 0; j < 128; j++)
			mdata->midi_note_vel [ i ][ j ] = 0;

		mdata->midi_channel [ i ].bender = 8192;
		mdata->midi_channel [ i ].bender_range = 2;
		mdata->midi_channel [ i ].controller[ MIDICTL_PAN ] = 64;
		mdata->midi_channel [ i ].controller[ MIDICTL_SUSTAIN ] = 0;
	}
	gsisynth_seq_flush(d);

	DEB1(DPRT("#gsisynth_seq_reset#\n"));
}


#warning "Finish gsisynth_reset_voices (move code to API dir)"
int  gsisynth_reset_voices(gsisynth_device_t *hw, gsisynth_devicedata_t *hw_status) {

#if  USE_ULTRASOUND
	gus_midi_reset(); // kill playing voices from previous song etc.
	WARN("this ok? (replace by a good instrument cache algorithm)");
#endif
	return -1;
}


/*---------------------------------------------------------------------------*
 *      Set a instrument for channel chn to pgm.
 *      Returns 0 if ok, -1 if an error occurred (instrument not set).
 *---------------------------------------------------------------------------*/
int gsisynth_seq_set_instrument(gsi_song_playstatus_t *d, int chn, int pgm) {
	gsi_midi_status_t *mdata = d->song;

	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB1(DPRT("("__FILE__"): *gsisynth_seq_set_instrument*\n"));
	DEB1(DPRT("   ch=%d, pgm=%d\n", chn, pgm));

	if (mdata->midi_force_chn_pgm [ chn ])
		pgm = mdata->midi_force_chn_pgm [ chn ] - 1;

	api_seq_set_instrument(d, chn, pgm);
	DEB1(DPRT(" #gsisynth_seq_set_instrument#\n"));

	mdata->midi_channel [ chn ].program = pgm;

	return 0;
}


void gsisynth_seq_stop_note(gsi_song_playstatus_t *d, int chn, int note, int vel) {
	gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB2(DPRT("   ("__FILE__"): *gsisynth_seq_stop_note*\n"));
	DEB2(DPRT("     chn=%d, note=%d, vel=%d\n", chn, note, vel));

	mdata->midi_note_vel [ chn ][ note ] = 0;

	api_seq_stop_note(d, chn, note, vel);

	DEB2(DPRT("   #gsisynth_seq_stop_note#\n"));
}


void gsisynth_seq_key_pressure(gsi_song_playstatus_t *d, int chn, int note, int vel) {
	//gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB2(DPRT("   ("__FILE__"): *gsisynth_seq_key_pressure*\n"));
	DEB2(DPRT("     chn=%d, note=%d, vel=%d\n", chn, note, vel));

	api_seq_key_pressure(d, chn, note, vel);

	DEB2(DPRT("   #gsisynth_seq_key_pressure#\n"));
}


void gsisynth_seq_start_note(gsi_song_playstatus_t *d, int chn, int note, int vel) {
	gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB2(DPRT("   ("__FILE__"): *gsisynth_seq_start_note*\n"));
	DEB2(DPRT("     chn=%d, note=%d, vel=%d\n", chn, note, vel));
	mdata->midi_note_vel [ chn ] [ note ] = vel;

	api_seq_start_note(d, chn, note, vel);

	DEB2(DPRT("   #gsisynth_seq_start_note#\n"));
}


/* REGIST_PARM_NUM_MSB / REGIST_PARM_NUM_LSB */
int rpn_msb[ 16 ] = {
	127, 127, 127, 127, 127, 127, 127, 127,
	127, 127, 127, 127, 127, 127, 127, 127
};
int rpn_lsb[ 16 ] = {
	127, 127, 127, 127, 127, 127, 127, 127,
	127, 127, 127, 127, 127, 127, 127, 127
};


void gsisynth_seq_control(gsi_song_playstatus_t *d, int chn, int p1, int p2) {
	gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB2(DPRT("   ("__FILE__"): *gsisynth_seq_control*\n"));
	DEB2(DPRT("     chn=%d, controller parm1=%d, conroller parm2=%d\n", chn, p1, p2));

	mdata->midi_channel[ chn ].controller[ p1 ] = p2;

	api_seq_control(d, chn, p1, p2);

	DEB2(DPRT("   #gsisynth_seq_control#\n"));
}


void gsisynth_seq_chn_pressure(gsi_song_playstatus_t *d, int chn, int vel) {
	gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	DEB2(DPRT("   seq_chn_pressure:\n"));
	DEB2(DPRT("     chn=%d, vel=%d\n", chn, vel));

	mdata->midi_channel[ chn ].pressure = vel;

	api_seq_chn_pressure(d, chn, vel);

	DEB2(DPRT("   #gsisynth_seq_chn_pressure#\n"));
}


void gsisynth_seq_bender(gsi_song_playstatus_t *d, int chn, int p1, int p2) {
	gsi_midi_status_t *mdata = d->song;
	//gsisynth_device_t *hw = d->hw;
	//gsisynth_devicedata_t *hw_status = d->hw_status;

	int  val;

	DEB2(DPRT("   ("__FILE__"): *gsisynth_seq_bender*\n"));
	DEB2(DPRT("     chn=%d, p1=%d, p2=%d\n", chn, p1, p2));

	val = (p2 << 7) + p1;
	mdata->midi_channel[ chn ].bender = val;

	//api_seq_bender(d, chn, val);
	api_seq_bender(d, chn, p1, p2);

	DEB2(DPRT("  #gsisynth_seq_bender#\n"));
}


/*---------------------------------------------------------------------------*
 *      Various MIDI stuff.
 *---------------------------------------------------------------------------*/
void gsisynth_seq_midi_out(gsi_song_playstatus_t *d, int ch, int data)
{
	api_seq_midi_out(d, ch, data);
}


/*---------------------------------------------------------------------------*
 *      Pause and continue use a workaround for OSS; MIDI timer doesn't seem
 *      to stop.
 *---------------------------------------------------------------------------*/
void gsisynth_seq_pause (gsi_song_playstatus_t *d) {
	api_seq_pause(d);

	gsisynth_seq_flush(d);
}


void gsisynth_seq_continue (gsi_song_playstatus_t *d) {
	api_seq_continue(d);

	gsisynth_seq_flush(d);
}


/*---------------------------------------------------------------------------*
 *      MIDI timing.
 *---------------------------------------------------------------------------*/
void gsisynth_seq_start_timer(gsi_song_playstatus_t *d) {
	api_seq_start_timer(d);

	gsisynth_seq_flush(d);
}


void gsisynth_seq_wait_time(gsi_song_playstatus_t *d, int ticks)
{
	api_seq_wait_time(d, ticks);
}



/*---------------------------------------------------------------------------*
 *      Send the sequencer data to the sound card.
 *---------------------------------------------------------------------------*/
int  gsisynth_seq_flush(gsi_song_playstatus_t *d)
{
	api_seq_flush(d);

	return 0;
}


/*      EOF midi_control.c */

