/* GSI - general sound interface
 *
 * cd_bsd.c - libgsicd: OpenBSD/FreeBSD/NetBSD calls.
 *
 * Lower level API dependencies:
 *   - BSD ioctl's/structs
 *
 * Q: can I use cd_fd in select (in the exception set) when a track is
 *    finished ? Or do I need to use an alarm/timer ? looks like the last case.
 *
 *
 * Copyright (C) 1998 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifdef GSICD_API_BSD
/* See cdio.[ch1]: */
#  include <sys/file.h>
#  include <sys/cdio.h>
#  include <sys/ioctl.h>

#  define CD_AUDIO_STATUS_INVALID    0x00
#  define CD_AUDIO_STATUS_PLAYING    0x11
#  define CD_AUDIO_STATUS_PAUSED     0x12
#  define CD_AUDIO_STATUS_COMPLETED  0x13
#  define CD_AUDIO_STATUS_ERROR      0x14
#  define CD_AUDIO_STATUS_VOID       0x15

#  define CDROM_AUDIO_INVALID        CD_AUDIO_STATUS_INVALID
#  define CDROM_AUDIO_PLAY           CD_AUDIO_STATUS_PLAYING
#  define CDROM_AUDIO_PAUSED         CD_AUDIO_STATUS_PAUSED
#  define CDROM_AUDIO_COMPLETED      CD_AUDIO_STATUS_COMPLETED
#  define CDROM_AUDIO_NO_STATUS      CD_AUDIO_STATUS_VOID
#  define CDROM_AUDIO_ERROR          CD_AUDIO_STATUS_ERROR

struct cd_toc_entry     toc_buffer[ GSICD_MAX_TRACKS+1 ];

int api_cd_can_open_without_medium = False; /* Bad ... */

int api_cd_init(int cd) {
	int  status = GSICD_OK;

	gsicd_status[ cd ]->fd = open (gsicd_devices[ cd ]->device, O_RDONLY | O_NONBLOCK);
	if (gsicd_status[ cd ]->fd < 0) {
		DEB1(DPRT("   -> cd_init error (errno=%d, error=%s)\n", errno, strerror(errno)));
		log_message(GSI_MESSAGE_ERROR, "%s: %s", gsicd_devices[ cd ]->device, strerror(errno));
		if ( (errno == EIO)
			 || (errno == ENXIO)
			 /*   from the OpenBSD cdio.c listing:
			  *   ENXIO has an overloaded meaning here. "Device not configured"
			  *   should be interpreted as "No disc in drive". */
			) {
			gsicd_status[ cd ]->status = GSICD_STATUS_NO_DISK;
			DEB1(DPRT("   -> No disc in drive %s.\n", gsicd_devices[ cd ]->device));
			//status = GSICD_STATUS_NO_DISK;
		}
		else if (errno == EBUSY) {
			gsicd_status[ cd ]->status = GSICD_STATUS_UNAVAILABLE;
			DEB1(DPRT("   -> CD device busy (%s).\n", gsicd_devices[ cd ]->device));
			status = GSICD_ERROR_BUSY;
		}
		else {
			gsicd_status[ cd ]->status = GSICD_STATUS_NO_DRIVE;
			/* Now set cd_ok = False, and then ignore all cd commands, except
			 * eject/change cd .. (OBSOLETE ?) */
//XXX			cd_ok = False;
			status = GSICD_ERROR_NOT_PRESENT;
		}

		return status;
	}

	return status;
}


int api_cd_stop(int cd) {
	int  status = ioctl(gsicd_status[ cd ]->fd, CDIOCSTOP);
	(void) ioctl (gsicd_status[ cd ]->fd, CDIOCALLOW);

	return status;
}


int api_cd_pause(int cd) {
	int  status = ioctl(gsicd_status[ cd ]->fd, CDIOCPAUSE);

	return status;
}


int api_cd_resume(int cd) {
	int status = ioctl(gsicd_status[ cd ]->fd, CDIOCRESUME);

	return status;
}


int api_cd_eject(int cd) {
	int  status;
	(void) ioctl(gsicd_status[ cd ]->fd, CDIOCALLOW);
	status = ioctl (gsicd_status[ cd ]->fd, CDIOCEJECT);

	return status;
}


int api_cd_insert(int cd) {
	int  status;
#  if defined(CDIOCCLOSE)
	(void) ioctl (gsicd_status[ cd ]->fd, CDIOCALLOW);
	status = ioctl (gsicd_status[ cd ]->fd, CDIOCCLOSE);
#  else
	status = -1; /* OK? or just return 0? */
#  endif

	return status;
}


int api_cd_select_slot(int cd, int slot) {
#  warning "Is there an ioctl for this OS to select a cd-changer slot?"
	return -1;
}


int api_cd_set_output_volume(int cd, int volume) {
	int  status;
	struct ioc_vol cd_volume;

	if (gsicd_status[ cd ]->fd < 0)
		return -1; /* open temporarily? */

	cd_volume.vol[0] = volume; /* L */
	cd_volume.vol[1] = volume; /* R */
	cd_volume.vol[2] = 0;
	cd_volume.vol[3] = 0;
	status = ioctl(gsicd_status[ cd ]->fd, CDIOCSETVOL, &cd_volume);

	return status;
}


int api_cd_get_output_volume(int cd) {
	struct ioc_vol cd_volume;
	int  status;

	if (gsicd_status[ cd ]->fd < 0)
		return -1; /* open temporarily? */

	status = ioctl(gsicd_status[ cd ]->fd, CDIOCGETVOL, &cd_volume);

	return cd_volume.vol[0]; /* L */
}


int api_cd_set_channels(int cd, int cmd)
{
	int  status = -1;

	switch (cmd) {
	case GSICD_OUTPUT_MUTE:
		status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETMUTE);
		break;
	case GSICD_OUTPUT_LEFT_CHANNEL:
		status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETLEFT);
		break;
	case GSICD_OUTPUT_RIGHT_CHANNEL:
		status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETRIGHT);
		break;
	case GSICD_OUTPUT_MONO:
		status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETMONO);
		break;
	case GSICD_OUTPUT_STEREO:
		status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETSTEREO);
		break;
	default:
		return -1;
		break;
	}

	return status;
}


int api_cd_get_status(int cd, gsicd_status_t *cd_pos) {
	struct ioc_read_subchannel ch;
	struct cd_sub_channel_info ch_data;
	int  status;

	bzero (&ch, sizeof (ch));
	ch.data = &ch_data;
	ch.data_len = sizeof (ch_data);
	ch.address_format = CD_MSF_FORMAT /* CD_LBA_FORMAT*/ ;
	ch.data_format = CD_CURRENT_POSITION;
	status = ioctl (gsicd_status[ cd ]->fd, CDIOCREADSUBCHANNEL, (char *) &ch);
	if (status == -1) {
		/* => probably no disk in the drive.
		 * Note that a data cd will give valid subchannel info but the track
		 * type is 'data'. */
		log_message(GSI_MESSAGE_ERROR, "%s: READ SUBCHANNEL: %s", gsicd_devices[ cd ]->device, strerror(errno));

		close(gsicd_status[ cd ]->fd);
		gsicd_status[ cd ]->fd = -1;
		cd_pos->track = 0;
		gsicd_status[ cd ]->status = GSICD_STATUS_NO_DISK;

		return -1;
	}

	cd_pos->track = ch.data->what.position.track_number;
	cd_pos->min   = ch.data->what.position.absaddr.msf.minute;
	cd_pos->sec   = ch.data->what.position.absaddr.msf.second;
	cd_pos->frame = ch.data->what.position.absaddr.msf.frame;

	switch (ch.data->header.audio_status) {
	case CDROM_AUDIO_PLAY:
		cd_pos->status = GSICD_STATUS_PLAYING;
		break;
	case CDROM_AUDIO_PAUSED:
		cd_pos->status = GSICD_STATUS_PAUSED;
		break;
	case CDROM_AUDIO_COMPLETED:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	case CDROM_AUDIO_NO_STATUS:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	case CDROM_AUDIO_ERROR:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	default:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	}

	return status;
}


int api_cd_get_contents(int cd) {
	struct ioc_toc_header     toc_h;
	struct ioc_read_toc_entry entry;
	int  i;
	int  status;

	status = ioctl(gsicd_status[ cd ]->fd, CDIOREADTOCHEADER, &toc_h);
	if (status == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: READ TOC HEADER: %s", gsicd_devices[ cd ]->device, strerror(errno));
		return -1;
	}

	gsicd_status[ cd ]->num_tracks = toc_h.ending_track - toc_h.starting_track + 1;
	DEB1(DPRT("  -> tracks=%d (first=%d, last=%d)\n", gsicd_status[ cd ]->num_tracks, toc_h.starting_track, toc_h.ending_track));

	gsicd_status[ cd ]->first_audio_track = toc_h.starting_track;
	gsicd_status[ cd ]->last_audio_track = toc_h.ending_track;

#  if 1
	/* 1. read all toc entries in one go. */
	entry.address_format = CD_MSF_FORMAT;
	entry.starting_track = 1;
	entry.data_len = (gsicd_status[ cd ]->num_tracks+1)*sizeof (struct cd_toc_entry);
	entry.data = toc_buffer;
	status = ioctl(gsicd_status[ cd ]->fd, CDIOREADTOCENTRYS, &entry);
	if (status == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: READ TOC ENTRYS: %s", gsicd_devices[ cd ]->device, strerror(errno));
		return -1;
	}

	for (i=1; i <= gsicd_status[ cd ]->num_tracks+1; i++) {
		gsicd_status[ cd ]->track[i].type = ((toc_buffer[i-1].control & 4) ? GSICD_DATA: GSICD_AUDIO);
		gsicd_status[ cd ]->track[i].min = toc_buffer[i-1].addr.msf.minute;
		gsicd_status[ cd ]->track[i].sec = toc_buffer[i-1].addr.msf.second;
		gsicd_status[ cd ]->track[i].frame = toc_buffer[i-1].addr.msf.frame;
	}
#  else
	/* 2. read toc entries one at a time.
	 *
	 * N.B. to get the leadout (end of last track), you actually need to read
	 * track 170. Another possibilty is reading 2 entries at a time (i.e. per
	 * track) which leaves the leadout in entry 2 of the last track.
	 */
	for (i=1; i <= gsicd_status[ cd ]->num_tracks +1; i++) {
		entry.address_format = CD_MSF_FORMAT;
		entry.starting_track = (i <= num_tracks) ? i : 170;
		entry.data_len = (1)*sizeof (struct cd_toc_entry);
		entry.data = toc_buffer;

		status = ioctl(gsicd_status[ cd ]->fd, CDIOREADTOCENTRYS, &entry);

		gsicd_status[ cd ]->track[i].type = ((toc_buffer[0].control & 4) ? GSICD_DATA: GSICD_AUDIO);
		gsicd_status[ cd ]->track[i].min = toc_buffer[0].addr.msf.minute;
		gsicd_status[ cd ]->track[i].sec = toc_buffer[0].addr.msf.second;
		gsicd_status[ cd ]->track[i].frame = toc_buffer[0].addr.msf.frame;
	}
#  endif
	/* Leadout: */
	//gsicd_status[ cd ]->track[gsicd_status[ cd ]->num_tracks+1].type = _CD_END;

	return 0;
}


int api_cd_play(int cd, gsicd_track_info_t *start, gsicd_track_info_t *end) {
	struct ioc_play_msf msf;
	int  status;

#  define AUDIOCD_PLAY_MSF  CDIOCPLAYMSF
#  define AUDIOCD_PLAY_LBA  CDIOCPLAYBLOCKS
#  define AUDIOCD_START     CDIOCSTART

	msf.start_m   = start->min;
	msf.start_s   = start->sec;
	msf.start_f   = start->frame;
	msf.end_m     = end->min;
	msf.end_s     = end->sec;
	msf.end_f     = end->frame;

	/* Start drive. */
	ioctl(gsicd_status[ cd ]->fd, AUDIOCD_START);

	status = ioctl(gsicd_status[ cd ]->fd, AUDIOCD_PLAY_MSF, &msf);
	return status;
}


int api_cd_get_slots(int cd) {
	int  slots = -1;

	return slots;
}
#endif /* GSICD_API_BSD */

/*      EOF cd_bsd.c */
