/* GSI - general sound interface
 *
 * cd_linux.c - libgsicd: Linux calls (also for Solaris?).
 *
 * Lower level API dependencies:
 *   - Linux ioctl's/structs
 *
 * Q: can I use cd_fd in select (in the exception set) when a track is
 *    finished ? Or do I need to use an alarm/timer ? looks like the last case.
 *
 *
 * Copyright (C) 1998 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifdef GSICD_API_LINUX /* Solaris too? */
#  include <sys/ioctl.h>
#  include <linux/cdrom.h>
/* For Solaris: #  include <sys/cdio.h> */

int api_cd_can_open_without_medium = True;

int api_cd_init(int cd) {
	int  status = GSICD_OK;

	gsicd_status[ cd ]->fd = open (gsicd_devices[ cd ]->device, O_RDONLY | O_NONBLOCK);
	if (gsicd_status[ cd ]->fd < 0) {
		DEB1(DPRT("   -> cd_init error (errno=%d, error=%s)\n", errno, strerror(errno)));
		log_message(GSI_MESSAGE_ERROR, "%s: %s", gsicd_devices[ cd ]->device, strerror(errno));
		if ( (errno == EIO)
			 || (errno == ENOMEDIUM)
			) {
			gsicd_status[ cd ]->status = GSICD_STATUS_NO_DISK;
			DEB1(DPRT("   -> No disc in drive %s.\n", gsicd_devices[ cd ]->device));
			//status = GSICD_STATUS_NO_DISK;
		}
		else if (errno == EBUSY) {
			gsicd_status[ cd ]->status = GSICD_STATUS_UNAVAILABLE;
			DEB1(DPRT("   -> CD device busy (%s).\n", gsicd_devices[ cd ]->device));
			status = GSICD_ERROR_BUSY;
		}
		else {
			gsicd_status[ cd ]->status = GSICD_STATUS_NO_DRIVE;
			/* Now set cd_ok = False, and then ignore all cd commands, except
			 * eject/change cd .. (OBSOLETE ?) */
//XXX			cd_ok = False;
			status = GSICD_ERROR_NOT_PRESENT;
		}

		return status;
	}
	return status;
}


int api_cd_stop(int cd) {
	int status = ioctl(gsicd_status[ cd ]->fd, CDROMSTOP);

	return status;
}


int api_cd_pause(int cd) {
	int status = ioctl(gsicd_status[ cd ]->fd, CDROMPAUSE);

	return status;
}


int api_cd_resume(int cd) {
	int status = ioctl(gsicd_status[ cd ]->fd, CDROMRESUME);

	return status;
}


int api_cd_eject(int cd) {
	int  status;
	(void) ioctl(gsicd_status[ cd ]->fd, CDROMSTOP);
	status = ioctl (gsicd_status[ cd ]->fd, CDROMEJECT);

	return status;
}


int api_cd_insert(int cd) {
	int  status = ioctl (gsicd_status[ cd ]->fd, CDROMCLOSETRAY);

	return status;
}


int api_cd_select_slot(int cd, int slot) {
	int  status;
#  ifdef CDROM_SELECT_DISC /* linux 2.2 */
	status = ioctl (gsicd_status[ cd ]->fd, CDROM_SELECT_DISC, slot);
#  else /* linux <= 2.0 */
	status = ioctl (gsicd_status[ cd ]->fd, CDROMLOADFROMSLOT, slot);
#  endif
	return status;
}


int api_cd_set_output_volume(int cd, int volume) {
	int  status;
	struct cdrom_volctrl cd_volume;

	if (gsicd_status[ cd ]->fd < 0)
		return -1; /* open temporarily? */

	cd_volume.channel0 = volume; /* L */
	cd_volume.channel1 = volume; /* R */
	cd_volume.channel2 = 0;
	cd_volume.channel3 = 0;

	status = ioctl(gsicd_status[ cd ]->fd, CDROMVOLCTRL, &cd_volume);

	return status;
}


int api_cd_get_output_volume(int cd) {
	struct cdrom_volctrl cd_volume;
	int  status;

	if (gsicd_status[ cd ]->fd < 0)
		return -1; /* open temporarily? */

	status = ioctl(gsicd_status[ cd ]->fd, CDROMVOLREAD, &cd_volume);

	return cd_volume.channel0; /* L */
}


int api_cd_set_channels(int cd, int cmd)
{
	int  status = -1;
#warning "Don't know how to select the CD channels in linux"
	switch (cmd) {
	case GSICD_OUTPUT_MUTE:
		//status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETMUTE);
		break;
	case GSICD_OUTPUT_LEFT_CHANNEL:
		//status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETLEFT);
		break;
	case GSICD_OUTPUT_RIGHT_CHANNEL:
		//status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETRIGHT);
		break;
	case GSICD_OUTPUT_MONO:
		//status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETMONO);
		break;
	case GSICD_OUTPUT_STEREO:
		//status = ioctl (gsicd_status[ cd ]->fd, CDIOCSETSTEREO);
		break;
	default:
		return -1;
		break;
	}

	return status;
}


int api_cd_get_status(int cd, gsicd_status_t *cd_pos) {
	struct cdrom_subchnl ch;
	int  status;

	ch.cdsc_format = CDROM_MSF;

	status = ioctl(gsicd_status[ cd ]->fd, CDROMSUBCHNL, &ch);
	if (status == -1) {
		/* => probably no disk in the drive.
		 * Note that a data cd will give valid subchannel info but the track
		 * type is 'data'. */
		log_message(GSI_MESSAGE_ERROR, "%s: READ SUBCHANNEL: %s", gsicd_devices[ cd ]->device, strerror(errno));

		close(gsicd_status[ cd ]->fd);
		gsicd_status[ cd ]->fd = -1;
		cd_pos->track = 0;
		gsicd_status[ cd ]->status = GSICD_STATUS_NO_DISK;

		return -1;
	}

	cd_pos->track = ch.cdsc_trk;
	cd_pos->min   = ch.cdsc_absaddr.msf.minute;
	cd_pos->sec   = ch.cdsc_absaddr.msf.second;
	cd_pos->frame = ch.cdsc_absaddr.msf.frame;

	switch (ch.cdsc_audiostatus) {
	case CDROM_AUDIO_PLAY:
		cd_pos->status = GSICD_STATUS_PLAYING;
		break;
	case CDROM_AUDIO_PAUSED:
		cd_pos->status = GSICD_STATUS_PAUSED;
		break;
	case CDROM_AUDIO_COMPLETED:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	case CDROM_AUDIO_NO_STATUS:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	case CDROM_AUDIO_ERROR:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	default:
		cd_pos->status = GSICD_STATUS_NONE;
		break;
	}
	return status;
}


int api_cd_get_contents(int cd) {
	int  i;
	struct cdrom_tochdr    toc_h;
	struct cdrom_tocentry  entry;
	int  status;

	status = ioctl(gsicd_status[ cd ]->fd, CDROMREADTOCHDR, &toc_h);
	if (status == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: READ TOC HEADER: %s", gsicd_devices[ cd ]->device, strerror(errno));
		return -1;
	}

	gsicd_status[ cd ]->num_tracks = toc_h.cdth_trk1-toc_h.cdth_trk0+1;
	DEB1(DPRT("  -> tracks=%d (start=%d, end=%d)\n", gsicd_status[ cd ]->num_tracks, toc_h.cdth_trk0, toc_h.cdth_trk1));

	gsicd_status[ cd ]->first_audio_track = toc_h.cdth_trk0;
	gsicd_status[ cd ]->last_audio_track = toc_h.cdth_trk1;

	for (i=1; i <= gsicd_status[ cd ]->num_tracks+1; i++) {
		entry.cdte_format = CDROM_MSF;
		entry.cdte_track = i;
		if ( i > gsicd_status[ cd ]->num_tracks) {
			entry.cdte_track = CDROM_LEADOUT; /* end of the CD */
		}
		status = ioctl(gsicd_status[ cd ]->fd, CDROMREADTOCENTRY, &entry);
		if (status == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: READ TOC ENTRY: %s", gsicd_devices[ cd ]->device, strerror(errno));
			return -1;
		}
		gsicd_status[ cd ]->track[i].min = entry.cdte_addr.msf.minute;
		gsicd_status[ cd ]->track[i].sec = entry.cdte_addr.msf.second;
		gsicd_status[ cd ]->track[i].frame = entry.cdte_addr.msf.frame;

		if (entry.cdte_ctrl & CDROM_DATA_TRACK)
			gsicd_status[ cd ]->track[i].type = GSICD_DATA;
		else
			gsicd_status[ cd ]->track[i].type = GSICD_AUDIO;
	}
	/* Leadout: */
	//gsicd_status[ cd ]->track[gsicd_status[ cd ]->num_tracks+1].type = _CD_END;

	return 0;
}


int api_cd_play(int cd, gsicd_track_info_t *start, gsicd_track_info_t *end) {
	struct cdrom_msf msf;
	int  status;

#  define start_m cdmsf_min0
#  define start_s cdmsf_sec0
#  define start_f cdmsf_frame0
#  define end_m   cdmsf_min1
#  define end_s   cdmsf_sec1
#  define end_f   cdmsf_frame1
#  define AUDIOCD_PLAY_MSF  CDROMPLAYMSF
#  define AUDIOCD_PLAY_LBA  CDROMPLAYBLK
#  define AUDIOCD_START     CDROMSTART

	msf.start_m   = start->min;
	msf.start_s   = start->sec;
	msf.start_f   = start->frame;
	msf.end_m     = end->min;
	msf.end_s     = end->sec;
	msf.end_f     = end->frame;

	/* Start drive. */
	ioctl(gsicd_status[ cd ]->fd, AUDIOCD_START);

	status = ioctl(gsicd_status[ cd ]->fd, AUDIOCD_PLAY_MSF, &msf);
	return status;
}


int api_cd_get_slots(int cd) {
	int  slots = -1;
#  ifdef CDROM_CHANGER_NSLOTS
	slots = ioctl(gsicd_status[ cd ]->fd, CDROM_CHANGER_NSLOTS);
	DEB1(fprintf(stderr, "CDROM_CHANGER_NSLOTS =%d", slots));
#  endif
	return slots;
}
#endif /* GSICD_API_LINUX */

/*      EOF cd_linux.c */
