/* GSI - general sound interface
 *
 * cd_functions.c - libgsicd, cd playing, ejecting etc.
 *
 * Lower level API dependencies:
 *   - Linux IOCTL's/structs (Solaris seems to use the same ones)
 *   - *BSD IOCTL's/structs
 *   What are the ioctl commands/include files in other OS'es ?
 *
 * Q: can I use cd_fd in select (in the exception set) when a track is
 *    finished ? Or do I need to use an alarm/timer ? looks like the last case.
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */




/*__________________INCLUDES___________________*/
//#include <gsi/gsicd_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <signal.h>

#include <errno.h>

#include <gsi/device.h>


#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSICD_"
#include <gsi/gsi_debug.h>

#include "gsicd.h"
#include "api.h"

#include "log.h"


/*__________________VARIABLES___________________*/
gsicd_device_t **gsicd_devices = NULL;
int  gsicd_num_cds = 0;
//NEW int  num_cd_devices;

gsicd_devicedata_t **gsicd_status = NULL;


/*---------------------------------------------------------------------------*
 *      Initialize the CD. Returns 0 if OK, -1 if failed
 *---------------------------------------------------------------------------*/
int gsicd_init(int cd)
{
	int  status;
	int  slots = 1;

	DEB1(DPRT("gsicd_init\n"));
	DEB(DPRT("gsicd_status[cd]->fd = %d\n", gsicd_status[ cd ]->fd));

	if (gsicd_status[ cd ]->fd >= 0) return 0;

	gsicd_stop_timer_for(cd);

	status = api_cd_init(cd);

//	gsicd_status[ cd ]->status = GSICD_STATUS_NONE;
	DEB1(DPRT("   -> cd_init done\n"));

	slots = api_cd_get_slots(cd);
	if (slots < 0) slots = 1;

	if (gsicd_update_status(cd)) {
		if (api_cd_can_open_without_medium) {
			gsicd_status[ cd ]->status = GSICD_STATUS_NO_DISK;
		}
fprintf(stderr, "gsicd_init fail update\n");
		return -1;
	}


	_gsicd_get_contents (cd); /* make track entries etc. */
	DEB1(DPRT("-> cd_init\n"));

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Free the CD
 *---------------------------------------------------------------------------*/
int gsicd_done(int cd)
{
	if (gsicd_status[ cd ]->fd < 0)
		return 0;

	gsicd_status[ cd ]->playing = False;
	gsicd_stop_timer_for(cd);

	/* Keep fd open for bad drivers/drives/slow module loading etc */
//	if (!gsicd_status[ cd ]->keep_open) {
	if (!gsicd_devices[ cd ]->flags & GSICD__FLAG_KEEP_OPEN) {
		close(gsicd_status[ cd ]->fd);
		gsicd_status[ cd ]->fd = -1;
	}

	//gsicd_status[ cd ]->status = GSICD_STATUS_UNAVAILABLE;
	gsicd_status[ cd ]->first_audio_track = 0;
	gsicd_status[ cd ]->last_audio_track = 0;

	gsicd_status[ cd ]->current_track = 0;

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Stop the CD (if it was playing)
 *---------------------------------------------------------------------------*/
int gsicd_stop(int cd)
{
	int  status;

	DEB1(DPRT("gsicd_stop\n"));

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	gsicd_status[ cd ]->playing = False;
	gsicd_stop_timer_for(cd);

	/* Now get exact position before stopping as we send the track number at
	 * which we stopped and if the CD was left playing while the client
	 * disconnected, then the timer code will have been disabled, right?
	 */
	gsicd_update_status(cd);

	status = api_cd_stop(cd);

	if (status == 0) {
		gsicd_status[ cd ]->playing = False;
//XXX		_gsicd_event(cd, GSICD_EVENT_STOP, gsicd_status[ cd ]->current_track);
//XXX Or send a GSI_EV_CD_STATUS ..?
		gsicd_generate_events();
		// update timer:
		gsicd_set_timer();
	}

	return status;
}


/*---------------------------------------------------------------------------*
 *      Pause the CD (if it is playing, otherwise no effect)
 *---------------------------------------------------------------------------*/
int gsicd_pause(int cd)
{
	int  status;

	DEB1(DPRT("gsicd_pause\n"));

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	gsicd_status[ cd ]->playing = False;
	gsicd_stop_timer_for(cd);

	status = api_cd_pause(cd);

	if (status == 0) {
		gsicd_status[ cd ]->playing = False;
//XXX		_gsicd_event(cd, GSICD_EVENT_TRACK_PAUSE, gsicd_status[ cd ]->current_track);
//XXX Or send a GSI_EV_CD_STATUS ..?
		gsicd_generate_events();
		// update timer:
		gsicd_set_timer();
	}

	return status;
}


/*---------------------------------------------------------------------------*
 *      Resume playing the CD (if the CD is paused from a 'pause' command,
 *      otherwise no effect).
 *---------------------------------------------------------------------------*/
int gsicd_resume(int cd)
{
	int  status;

	DEB1(DPRT("gsicd_resume\n"));

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	gsicd_update_status(cd); /* get exact position before resuming */
	/* Actually, this is not quite true. when the CD is paused, the position
	 * can vary in a range of lowest - lowest + 12 (experimentally on a
	 * pioneer DRU-16S SCSI player) ... */

	status = api_cd_resume(cd);

	if (status == 0) {
		gsicd_status[ cd ]->playing = True;
//XXX		_gsicd_event(cd, GSICD_EVENT_TRACK_RESUME, gsicd_status[ cd ]->current_track);
		gsicd_generate_events();
		// update timer:
		gsicd_set_timer();
	}

	return status;
}


/*---------------------------------------------------------------------------*
 *      Play a range of tracks, gsicd_first ... gsicd_last.
 *      An interrupt is set to send an event at the start of each track (also
 *      the first track even when that start is with a non zero offset), and
 *      then for the end (when all tracks have been played).
 *
 *      N.B. m1,s1,f1 are the offset from the msf of the start of track t1 for
 *      the first frame to be played, similarly, m2,s2,f2 are the offset from
 *      the msf of the start of track t2 for the last frame to be played.
 *---------------------------------------------------------------------------*/
int  gsicd_play(int cd, int t1, int m1, int s1, int f1,  int t2, int m2, int s2, int f2)
{
	int  status;
	int  lba_offset1, lba_offset2, lba_trk_len;

	gsicd_track_info_t start, end;

	DEB1(DPRT("gsicd_play\n"));

	DEB1(DPRT("   -> from tmsf %d,%d,%d,%d to %d,%d,%d,%d\n", t1,m1,s1,f1, t2,m2,s2,f2));
	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

DEB(DPRT("t1=%d, t2=%d\n", t1, t2));

	if (t1 < gsicd_status[ cd ]->first_audio_track) { t1 = gsicd_status[ cd ]->first_audio_track; m1 = 0; s1 = 0; f1 = 0; }
	if (t2 > gsicd_status[ cd ]->last_audio_track) { t2 = gsicd_status[ cd ]->last_audio_track; m2 = 255; s2 = 255; f2 = 255; }
DEB(DPRT("t1=%d, t2=%d\n", t1, t2));

	/* Skip data tracks (is it possible to have them alternating? multisession
	 * perhaps?). */
	while (gsicd_status[ cd ]->track[ t1 ].type == GSICD_DATA && t1 <= t2) { t1++; m1 = 0; s1 = 0; f1 = 0; }
	while (gsicd_status[ cd ]->track[ t2 ].type == GSICD_DATA && t1 <= t2) { t2--; m2 = 255; s2 = 255; f2 = 255; }
DEB(DPRT("t1=%d, t2=%d\n", t1, t2));

	/* First track to be played: */
	lba_offset1 = gsicd_lba_from_msf(m1, s1, f1);
	lba_trk_len = gsicd_lba_track_len(cd, t1);
	DEB1(DPRT("  first: lba_offset = %d (msf=%d,%d,%d), lba_trk_len=%d\n", lba_offset1, m1, s1, f1, lba_trk_len));

	if (lba_offset1 > lba_trk_len) {
		lba_offset1 = lba_trk_len-1;
		m1 = gsicd_m_from_lba(lba_offset1);
		s1 = gsicd_s_from_lba(lba_offset1);
		f1 = gsicd_f_from_lba(lba_offset1);
	}

	/* Last track to be played: */
	lba_offset2 = gsicd_lba_from_msf(m2, s2, f2);
	lba_trk_len = gsicd_lba_track_len(cd, t2);
	DEB1(DPRT("  last: lba_offset = %d (msf=%d,%d,%d), lba_trk_len=%d\n", lba_offset2, m2, s2, f2, lba_trk_len));

	if (lba_offset2 > lba_trk_len) {
		lba_offset2 = lba_trk_len-1;
		m2 = gsicd_m_from_lba(lba_offset2);
		s2 = gsicd_s_from_lba(lba_offset2);
		f2 = gsicd_f_from_lba(lba_offset2);
	}

	DEB1(DPRT("   -> first=%d (offset %d,%d,%d)\n", t1, m1, s1, f1));
	DEB1(DPRT("   -> last=%d (offset %d,%d,%d)\n", t2, m2, s2, f2));

	if (t2 < t1)
		return -1;

	/* Interrupt drive (Is this really necessary? It slows playing on a pioneer
	 * DR-U16S (36X) a lot... (for cd_next & cd_prev commands))
	 * To be sure, I do it if the drive is in an error status:
	 */
	if (gsicd_status[ cd ]->status > 2) /* (0,1,2 = none, playing, paused). */
		gsicd_stop(cd);

	{
	int lba_start = gsicd_lba_track_pos(cd, t1) + lba_offset1;
	int lba_end   = gsicd_lba_track_pos(cd, t2) + lba_offset2;

	start.min   = gsicd_m_from_lba(lba_start);
	start.sec   = gsicd_s_from_lba(lba_start);
	start.frame = gsicd_f_from_lba(lba_start);
	end.min     = gsicd_m_from_lba(lba_end);
	end.sec     = gsicd_s_from_lba(lba_end);
	end.frame   = gsicd_f_from_lba(lba_end);
	}

	DEB1(DPRT("   -> cd play %d %d %d to %d,%d,%d\n", start.min, start.sec, start.frame, end.min, end.sec, end.frame  ));

	/* For lba we do: */
	//lba.cdlba0 ?? = lba_track_pos(cd, t1) + gsicd_lba_from_msf(m1,s1,f1);
	//lba.cdlba1 ?? = lba_track_pos(cd, t1) + gsicd_lba_from_msf(m2,s2,f2);

	status = api_cd_play(cd, &start, &end);
	DEB1(DPRT("   -> cd_play ioctl done (status=%d)\n", status));

	if (status < 0) /* -405 for data track with sbpcd, linux */
		return -1;

	/* Keep a record of the current playrange. */
	gsicd_status[ cd ]->first.track        = t1;
	gsicd_status[ cd ]->first.offset_min   = m1;
	gsicd_status[ cd ]->first.offset_sec   = s1;
	gsicd_status[ cd ]->first.offset_frame = f1;

	gsicd_status[ cd ]->last.track         = t2;
	gsicd_status[ cd ]->last.offset_min    = m2;
	gsicd_status[ cd ]->last.offset_sec    = s2;
	gsicd_status[ cd ]->last.offset_frame  = f2;

	gsicd_status[ cd ]->current_track = gsicd_status[ cd ]->first.track;

	/* For event generation: */
	gsicd_status[ cd ]->next_track = min(t1+1, t2);
	gsicd_status[ cd ]->playing = True;

	gsicd_status[ cd ]->playing = True;
//XXX	_gsicd_event(cd, GSICD_EVENT_TRACK_START, gsicd_status[ cd ]->current_track);
	gsicd_generate_events();

	/* Update timer: */
	gsicd_set_timer();

	DEB1(DPRT("-> cd_play done\n"));

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Set CD channels: mono/stereo, left channel only etc.
 *      Seems only available in *BSD, didn't find Linux equivalents.
 *---------------------------------------------------------------------------*/
int  gsicd_set_channels(int cd, int cmd)
{
	int  status = -1;

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	status = api_cd_set_channels(cd, cmd);

	return status;
}


/*---------------------------------------------------------------------------*
 *      Disk jockey
 *---------------------------------------------------------------------------*/
int gsicd_eject(int cd)
{
	int  status;

	gsicd_stop_timer_for(cd);

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	status = api_cd_eject(cd);

	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "Eject CD failed (cd mounted?)");
		return GSICD_ERROR_MOUNTED;
	}

	/* Not actually necessary: */
	close(gsicd_status[ cd ]->fd);
	gsicd_status[ cd ]->fd = -1;

	return 0;
}


int gsicd_insert(int cd)
{
	int  status;

	if (gsicd_status[ cd ]->fd < 0 && (status = gsicd_init (cd))) {
fprintf(stderr, "gsicd_insert: err=%d\n", status);
		if (gsicd_status[ cd ]->status == GSICD_STATUS_NO_DISK)
			return GSICD_ERROR_CANT_INSERT_CD;
		else
			return -1;
	}

	status = api_cd_insert(cd);
	if (status < 0)
		return status;

	//close(gsicd_status[ cd ]->fd);
	//gsicd_status[ cd ]->fd = -1;
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Change the CD
 *---------------------------------------------------------------------------*/
int gsicd_select_slot(int cd, int slot)
{
	int  status;
	int  temp_open = False;

	//if (cd < 0 || cd >= gsicd_num_cds) return -1;
	if (slot < 0 || slot >= gsicd_devices[ cd ]->num_slots)
		return GSICD_ERROR_BAD_SLOT;

	//if (slot == gsicd_devices[ cd ]->slot) return 0;

	gsicd_stop_timer_for(cd);

	if (gsicd_status[ cd ]->fd < 0) {
		gsicd_status[ cd ]->fd = open(gsicd_devices[ cd ]->device, O_RDONLY | O_NONBLOCK);
		if (gsicd_status[ cd ]->fd < 0)
			return -1;

		temp_open = True;
	}

	status = api_cd_select_slot(cd, slot);

	if (status != 0) {
		DEB1(DPRT("  CDROM_SELECT_DISC ioctl failed for `%s': %s (error %d), status=%d\n", gsicd_devices[ cd ]->device, strerror (errno), errno, status));
//???		if (errno == EINVAL)
//???			return GSICD_ERROR_BAD_SLOT;
		if (errno == ENOENT) /* tested in linux */
			return GSICD_STATUS_NO_DISK; /* no disk in slot (doesn't matter ? just retry this when we actually start playing a disk ?) */

		/* How do I find out how many slots a changer has ?
		 * I get errno=ENOENT for slots>3 on NEC CDR-C251 4cd changer, but also
		 * for a good slot with no CD.
		 */

		if (errno == ENOSYS) {
			/* Function not implemented (on the device? or the driver?)
			 * In the first case, this means num_slots = 1. The following code
			 * assumes case 1. */
			gsicd_devices[ cd ]->num_slots = 1;
			if (slot == 0) return 0; else return GSICD_ERROR_BAD_SLOT;
		}

		return -1;
	}
	if (temp_open) {
		close(gsicd_status[ cd ]->fd);
		gsicd_status[ cd ]->fd = -1;
	}
	else {
		gsicd_update_status(cd);
		_gsicd_get_contents(cd); /* get new track list */
	}
	gsicd_status[ cd ]->slot = slot;

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Get the CD status.
 *      Read info about current track etc. If no cd was previously inside,
 *      check if a cd was inserted and if so, initialize it.
 *      Returns 0 if OK, -1 otherwise.
 *
 *      Note: should not be called during signals (use _gsicd_get_status).
 *---------------------------------------------------------------------------*/
int gsicd_update_status(int cd)
{
	int  status;
	gsicd_status_t pos;

//	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd)) /* This can give an inifinte loop */
	if (gsicd_status[ cd ]->fd < 0)
		return -1;

	status = _gsicd_get_status(cd, &pos);
	if (status == 0) {
		/* See _gsicd_get_status about the subchannel track # & track-leadout
		 * Do we need to (WANT TO?) correct the track number?:
		 * or do that in event.c?
		 */
		if (gsicd_lba_track_pos(cd, pos.track) > gsicd_lba_from_msf(pos.min, pos.sec, pos.frame)) {
			pos.track--;
DEB(DPRT("correcting track no. from %d to %d\n", pos.track+1, pos.track));
		}

		gsicd_status[ cd ]->current_track = pos.track;

		gsicd_status[ cd ]->current_m = pos.min;
		gsicd_status[ cd ]->current_s = pos.sec;
		gsicd_status[ cd ]->current_f = pos.frame;

		gsicd_status[ cd ]->status = pos.status;
	}

	return status;
}


/*---------------------------------------------------------------------------*
 *      Get the CD status.
 *      Should be used during signals (instead of gsicd_update_status).
 *---------------------------------------------------------------------------*/
int _gsicd_get_status(int cd, gsicd_status_t *cd_pos)
{
	int  status;

	DEB1(DPRT("("__FILE__"): gsicd_status\n"));

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd)) {
		DEB1(DPRT("  -> gsicd_status: quit\n"));
		return -1;
	}

	status = api_cd_get_status(cd, cd_pos);

	DEB1(DPRT("  -> cd_pos->track = %d\n", cd_pos->track));
	/* --------------------------------------------------------------------- *
	 * N.B.
	 * The current track from the subchannel data is actually already the next
	 * track at the non-audio leadout of each track, this can be several, even
	 * many seconds, and is displayed on CD players as a negative time e.g.
	 * -3, -2, -1 s .., and the track really starts at 0.0 s.
	 * --------------------------------------------------------------------- */

	return status;
}


/*---------------------------------------------------------------------------*
 *      Read the tracks, msf and wether they are data or audio.
 *---------------------------------------------------------------------------*/
int _gsicd_get_contents (int cd)
{
	int  status;

	DEB1(DPRT("("__FILE__"): cd_get_contents\n"));

	if (gsicd_status[ cd ]->fd < 0 && gsicd_init (cd))
		return -1;

	status = api_cd_get_contents(cd);

	/* Skip data tracks. */
	while (gsicd_status[ cd ]->track[ gsicd_status[ cd ]->first_audio_track ].type == GSICD_DATA
		   && gsicd_status[ cd ]->first_audio_track <= gsicd_status[ cd ]->last_audio_track)
		gsicd_status[ cd ]->first_audio_track++;
DEB(DPRT("first audio track =%d\n", gsicd_status[ cd ]->first_audio_track));
	while (gsicd_status[ cd ]->track[ gsicd_status[ cd ]->last_audio_track ].type == GSICD_DATA
		   && gsicd_status[ cd ]->first_audio_track <= gsicd_status[ cd ]->last_audio_track)
		gsicd_status[ cd ]->last_audio_track--;
DEB(DPRT("last audio track =%d\n", gsicd_status[ cd ]->last_audio_track));

	if (gsicd_status[ cd ]->first_audio_track > gsicd_status[ cd ]->last_audio_track) {
		DEB1(DPRT("  -> no audio tracks on this CD\n"));
	}

	gsicd_status[ cd ]->last.track        = gsicd_status[ cd ]->last_audio_track;
	gsicd_status[ cd ]->last.offset_min   = gsicd_m_from_lba (gsicd_lba_track_len (cd, gsicd_status[ cd ]->last_audio_track) -1);

	gsicd_status[ cd ]->last.offset_sec   = gsicd_s_from_lba (gsicd_lba_track_len (cd, gsicd_status[ cd ]->last_audio_track) -1);

	gsicd_status[ cd ]->last.offset_frame = gsicd_f_from_lba (gsicd_lba_track_len (cd, gsicd_status[ cd ]->last_audio_track) -1);

	DEB1(DPRT("-> cd_get_contents\n"));
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Set/get the CD player's output volume (0-255)
 *---------------------------------------------------------------------------*/
void gsicd_set_output_volume(int cd, int volume)
{
	int  status;

	status = api_cd_set_output_volume(cd, volume);
}


int gsicd_get_output_volume(int cd) {
//	int  status;

	return api_cd_get_output_volume(cd);
}

#include "api/cd_bsd.c"
#include "api/cd_linux.c"
#include "api/cd_null.c"

/*      EOF cd_functions.c */
