/* GSI - general sound interface
 *
 * event.c - libgsicd, cd events, timer code.
 *
 * Lower level API dependencies:
 *  - p_signal (portable signal) functions.
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsicd_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
//#include <string.h>

#include <limits.h>

#include <gsi/device.h>
#include <gsi/p_signal.h>
#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSICD_"
#include <gsi/gsi_debug.h>

#include "gsicd.h"

#include "log.h"


//static int _gsicd_alarm_cdno = -1;    /* The cd for which the alarm is set */
//static int _gsicd_alarm_trackno = -1; /* The track of that cd for which the alarm is set */

static volatile sig_atomic_t _gsicd_timer_set = False;


int gsicd_stop_timer_for(int cd) {
	/* timer not used for cd, so if all cd's are paused/stopped => timer off
	 * if cd had the first track => set timer for other CD. */

//	if (_gsicd_alarm_cdno == cd) {
		_gsicd_stop_timer();
		gsicd_set_timer();
//	}

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Send an event when starting a new track except, if it was the last
 *      one, upon completion of that track.
 *---------------------------------------------------------------------------*/
/* track timer signal notification */
void (*_gsicd_lowlevel_event) (void) = NULL;
void gsicd_lowlevel_event () {
	DEB(DPRT("lowlevel event\n"));
	if (_gsicd_lowlevel_event) {
		DEB(DPRT("yes, do lowlevel event\n"));

		_gsicd_lowlevel_event();
	}
}

/* Passing NULL will disable event generation */
void gsicd_set_lowlevel_event_function(void *fn) {
	DEB(DPRT("set lowlevel event fn\n"));
	_gsicd_lowlevel_event = fn;
}



/* track event generation */
void (*_gsicd_event) (int event, int cd, int track);
void gsicd_event (int event, int cd, int track) {
	if (_gsicd_event) {
		_gsicd_event(event, cd, track);
	}
}

/* Passing NULL will disable event generation */
void gsicd_set_event_function(void *fn) {
	_gsicd_event = fn;
}


/* This may be called from a signal handler */
void gsicd_set_timer()
{
	int  mycd;
	int  event_time;

	int  earliest_cdno = -1;
	int  earliest_time = INT_MAX;
DEB(DPRT("gsicd_set_timer\n"));

	for (mycd = 0; mycd< gsicd_num_cds; mycd++) {
DEB(DPRT("checking cd %d\n", mycd));
		if (gsicd_status[ mycd ]->playing) {
			int  t_next, t_last;
			int  lba_next, lba_end;

DEB(DPRT("cd %d is playing\n", mycd));
			gsicd_update_status(mycd);
DEB(DPRT("t msf=%d %d,%d,%d\n", gsicd_status[ mycd ]->current_track, gsicd_status[ mycd ]->current_m, gsicd_status[ mycd ]->current_s, gsicd_status[ mycd ]->current_f));

			t_next = gsicd_status[ mycd ]->current_track + 1;

			// we need the earliest of lba_next, lba_end:
			lba_next = gsicd_lba_track_pos(mycd, t_next);

			/* end of play range: */
			t_last = gsicd_status[ mycd ]->last.track;
			lba_end = gsicd_lba_track_pos(mycd, t_last) + gsicd_lba_from_msf (gsicd_status[ mycd ]->last.offset_min, gsicd_status[ mycd ]->last.offset_sec, gsicd_status[ mycd ]->last.offset_frame);

DEB(DPRT("t_next = %d, t_last = %d\n", t_next, t_last));
DEB(DPRT("lba_next = %d, lba_end = %d\n", lba_next, lba_end));

			event_time = (lba_next < lba_end ? lba_next : lba_end) - gsicd_lba_from_msf (gsicd_status[ mycd ]->current_m, gsicd_status[ mycd ]->current_s, gsicd_status[ mycd ]->current_f);
			if (event_time < earliest_time) {
				earliest_time = event_time;
				earliest_cdno = mycd;
			}
		}
	}
DEB(DPRT("earliest_time = %d (frames), for cd %d\n", earliest_time, earliest_cdno));
	if (earliest_time < INT_MAX) {
		_gsicd_set_timer(_gsicd_update, earliest_time/75, 0);
	}
	else {
		/* Now switch off the timer: */
		_gsicd_stop_timer();
		_gsicd_timer_set = False;
		/* to make sure all events are produced, this is necessary: */
		gsicd_lowlevel_event();
	}
}


/*---------------------------------------------------------------------------*
 *      Low level signal/timer handling.
 *---------------------------------------------------------------------------*/
int _gsicd_spurious_alarm = False;
void _gsicd_update(int signum)
{
	if (_gsicd_timer_set) {
		gsicd_lowlevel_event();
		gsicd_set_timer();
	}
	else {
		/* ignore the signal */
		/* shouldn't use stdio or syslog in signal handlers, so: */
		_gsicd_spurious_alarm = True;
		/* and gsi_cd_generate_events will log the problem. */
	}
}


void gsicd_generate_events()
{
	/* 
	 * When playing multiple tracks, only send GSI_EV_CD_TRACK_END after
	 * the last track
	 */

	int  mycd;

	if (_gsicd_spurious_alarm) {
		log_message(GSI_MESSAGE_ERROR, "spurious SIGALRM received");
		_gsicd_spurious_alarm = False;
	}

	for (mycd = 0; mycd< gsicd_num_cds; mycd++) {
		if (gsicd_status[ mycd ]->playing) {
			int  i, t_this, t_last;
			int  lba_current, lba_end;
DEB(DPRT("CD %d\n", mycd));
			gsicd_update_status(mycd);

			t_this = gsicd_status[ mycd ]->current_track;

			lba_current = gsicd_lba_from_msf (gsicd_status[ mycd ]->current_m, gsicd_status[ mycd ]->current_s, gsicd_status[ mycd ]->current_f);

			/* end of play range: */
			t_last = gsicd_status[ mycd ]->last.track;

			lba_end = gsicd_lba_track_pos(mycd, t_last) + gsicd_lba_from_msf (gsicd_status[ mycd ]->last.offset_min, gsicd_status[ mycd ]->last.offset_sec, gsicd_status[ mycd ]->last.offset_frame);


			for (i = gsicd_status[ mycd ]->next_track; i<= min(t_this, t_last); i++) {
DEB(DPRT("gsicd_event START for cd %d track %d\n", mycd, i));
				gsicd_event(GSICD_EVENT_TRACK_START, mycd, i);
			}

			if (lba_current >= lba_end) {
				gsicd_event(GSICD_EVENT_TRACK_END, mycd, t_last);
				gsicd_stop(mycd);
			}
			gsicd_status[ mycd ]->next_track = min(t_this + 1, t_last);
		}
	}
}

void _gsicd_stop_timer() {
	p_signal(SIGALRM, SIG_IGN);
}


struct itimerval timer;
void _gsicd_set_timer(void *update_fn, int sec, int musec)
{
	struct sigaction my_signal;

	_gsicd_stop_timer();
	_gsicd_timer_set = True;
	DEB1(DPRT("  set_timer: %d s %d s\n", sec, musec));

	sigemptyset(&my_signal.sa_mask);

	/* To interrupt system calls: */
#ifdef  SA_INTERRUPT
	my_signal.sa_flags |= SA_INTERRUPT;    /* SunOS 4.x */
#endif

	/* To restart system calls: */
//#ifdef SA_RESTART
//	my_signal.sa_flags = SA_RESTART; /* SVR4, 4.4BSD */
//#endif

	my_signal.sa_handler = update_fn;
	sigaction(SIGALRM, &my_signal, NULL);

	timer.it_interval.tv_sec = sec+10;
	timer.it_interval.tv_usec = 0;
	timer.it_value.tv_sec = sec;
	timer.it_value.tv_usec = musec;

	setitimer(ITIMER_REAL, &(timer), NULL);
}


/* EOF event.c */
