/* GSI - general sound interface
 *
 * mixer_obsd.c - mixer setting/reading for OpenBSD/NetBSD
 *
 * Lower level API dependencies:
 *   - OpenBSD/NetBSD audio.h
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*---------------------------------------------------------------------------*
 *      Get/set sound volumes with the OpenBSD/NetBSD audio system.
 *      Volumes are in range 0-255
 *---------------------------------------------------------------------------*/
#if  USE_OPENBSD_AUDIO
#include <sys/audioio.h>

#include <stdio.h>

static int obsd_mixer_dev[ 6 ];
static int obsd_mixer_type[ 6 ];

char *obsd_mixer_devices[6] = { /* master, pcm, synth, cd, LINE_IN, mic in */
	AudioNmaster, AudioNdac, "", AudioNcd, AudioNline, AudioNmicrophone
};


static int  api_init_mixer(void) {
	int  status;
	mixer_devinfo_t mix_info;
	mixer_ctrl_t    mix_val;
	int ndev;

	if (gsimixer_fd > 0) return 0;

	if (gsimixer_device == NULL) return -1;
	gsimixer_fd = open(gsimixer_device, O_WRONLY | O_NONBLOCK);
	if (gsimixer_fd == -1) {
		gsimixer_ok = False;
		log_message(GSI_MESSAGE_ERROR, "%s: (obsd) open mixer: %s", gsimixer_device, strerror(errno));
		return -1;
	}
	else {
		gsimixer_ok = True;

		emulate_master_volume = True;
		for(ndev = 0; ; ndev++) {
			int  i;

			mix_info.index = ndev;
			if ((status = ioctl(gsimixer_fd, AUDIO_MIXER_DEVINFO, &mix_info)) < 0)
				break;
//DEB(DPRT("OBSD mixer: %d: name %s (type=%d, class=%d, next=%d, prev=%d)\n", ndev, mix_info.label.name, mix_info.type, mix_info.mixer_class, mix_info.next, mix_info.prev));
			for (i=0; i<6; i++) {
				if (!strcmp(mix_info.label.name, obsd_mixer_devices[i])) {
					obsd_mixer_dev[ i ] = mix_info.index;
					obsd_mixer_type[ i ] = mix_info.type;
					break;
				}
			}
			if (!strcmp(mix_info.label.name, AudioNmaster)) {
				emulate_master_volume = False;
			}
			if (!strcmp(mix_info.label.name, AudioNmute)) {
				/* Switch off 'mute' for all devices GSI handles */
				for (i=0; i<6; i++) {
					if (mix_info.prev == obsd_mixer_dev[i]) {
						mix_val.dev = mix_info.index;
						mix_val.type = mix_info.type;
						mix_val.un.ord = 0; /* mute off */
						status = ioctl(gsimixer_fd, AUDIO_MIXER_WRITE, &mix_val);
						if (status < 0) {
							log_message(GSI_MESSAGE_ERROR, "%s: (obsd) unmute mixer failed: %s", gsimixer_device, strerror(errno));
						}
					}
				}
			}
		}

		close(gsimixer_fd);
		gsimixer_fd = -1;

		if (ndev == 0) {
			log_message(GSI_MESSAGE_ERROR, "%s: (obsd) no mixer devices configured", gsimixer_device);
			gsimixer_ok = False;
			return -1;
		}
	}

	return 0;
}


static int  api_close_mixer(void) {
	close(gsimixer_fd);
	gsimixer_fd = -1;

	return 0;
}


static void api_set_volume(int device_type, int device_no, int vol) {
	int  mixer, volume;
	mixer_ctrl_t    mix_val;

	DEB1(DPRT("("__FILE__"): set obsd volume\n"));
	if (gsimixer_ok) {
		mixer = open(gsimixer_device, O_WRONLY | O_NONBLOCK);
		if (mixer < 0) {
			log_message(GSI_MESSAGE_ERROR, "%s: (obsd) open mixer: %s", gsimixer_device, strerror(errno));
		}
		else {
			volume = (vol & 65535); /* maybe use range 0-65535 */
			if (volume > 256) volume = 256;
			volume = (volume * 255) / 256;
			DEB1(DPRT("  -> set volume of device %d:%d to %d (converted to %d for obsd)\n", device_type, device_no, vol, volume));

#if 0
			/* balance type 1: center gives half volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				volume_l = (vol*(256-balance)) >> 9;
				volume_r = (vol*(256+balance)) >> 9;
			}
#endif
#if 0
			/* balance type 2: center gives full volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				volume_l = min((vol*(256-balance)) >> 8, 256);
				volume_r = min((vol*(256+balance)) >> 8, 256);
			}
#endif

			mix_val.dev = obsd_mixer_dev[ device_type ];
			mix_val.type = obsd_mixer_type[ device_type ];
			mix_val.un.value.num_channels = 2;

			/* Left and right channels */
			mix_val.un.value.level[0] = volume;
			mix_val.un.value.level[1] = volume;

			ioctl(mixer, AUDIO_MIXER_WRITE, &mix_val);

			close(mixer);
		}
	}
}


static int  api_get_volume(int device_type, int device_no) {
	int  mixer;
	int  volume = -1;
	int  volume_l, volume_r;
	mixer_ctrl_t    mix_val;

	DEB1(DPRT("("__FILE__"): get_obsd_volume\n"));
	if (gsimixer_ok) {
		mixer = open(gsimixer_device, O_RDONLY | O_NONBLOCK);
		if (mixer < 0) {
			log_message(GSI_MESSAGE_ERROR, "%s: (obsd) open mixer: %s", gsimixer_device, strerror(errno));
		}
		else {
			mix_val.dev = obsd_mixer_dev[ device_type ];
			mix_val.type = obsd_mixer_type[ device_type ];
			ioctl(mixer, AUDIO_MIXER_READ, &mix_val);

			volume_l = mix_val.un.value.level[0];
			volume_r = mix_val.un.value.level[1];

#if 0
			/* balance type 1: center gives half volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				vol = volume_l+volume_r;
				if (vol > 256) vol = 256;
				balance = ((volume_r-volume_l) << 8)/vol;
			}
#endif
#if 0
			/* balance type 2: center gives full volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				vol = max(volume_l, volume_r);
				balance = ((volume_r-volume_l) << 8)/vol;
			}
#endif


			volume = (((volume_l+volume_r)*128)/255); /* We currently return the average channel value */

			/* Should all HW volume settings be made 8-channel? */
			//volume.nch = 2;
			//volume.ch[0] = (volume_l*256)/255;
			//volume.ch[1] = (volume_l*256)/255;

			DEB1(DPRT("  -> get_volume of device %d:%d: sound_volume=%d (from obsd value of %d)\n", device_type, device_no, volume, volume_l));
			close(mixer);
		}
	}

	return volume;
}
#endif /* USE_OPENBSD_AUDIO */


/*      EOF mixer_obsd.c */
