/* GSI - general sound interface
 *
 * mixer_oss.c - mixer setting/reading for OSS
 *
 * Lower level API dependencies:
 *   - OSS include locations are Linux/FreeBSD specific; other OS's?
 *   - OSS mixer constants etc
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*---------------------------------------------------------------------------*
 *      Get/set sound volumes with OSS.
 *
 *      Note that in OSS, the mixer can be opened by multiple devices at the
 *      same time, so opening should always return 0 and we can keep the file
 *      open.
 *      OSS volumes are 0-100.
 *---------------------------------------------------------------------------*/
#if  USE_OSS

#ifdef OSS_INCLUDES

#  include OSS_INCLUDES"/soundcard.h" /* Commercial OSS */

#else

#if defined(__OpenBSD__) || defined(__NetBSD__)
#  include <soundcard.h>
#else /* linux, FreeBSD: OSS-free */
#  include <sys/soundcard.h>
#endif

#endif /* OSS_INCLUDES */

#include <stdio.h>

static int  api_init_mixer(void) {

	if (gsimixer_fd > 0) return 0;
//DEB(DPRT("gsimixer_device = %s\n", gsimixer_device));
	if (gsimixer_device == NULL) return -1;
	gsimixer_fd = open(gsimixer_device, O_WRONLY | O_NONBLOCK);
	if (gsimixer_fd == -1) {
		gsimixer_ok = False;
		log_message(GSI_MESSAGE_ERROR, "%s: (oss) open mixer: %s", gsimixer_device, strerror(errno));
		return -1;
	}
	else {
		gsimixer_ok = True;
		ioctl(gsimixer_fd, SOUND_MIXER_READ_DEVMASK, &gsimixer_channels);

		close(gsimixer_fd);
		gsimixer_fd = -1;

		if (!(gsimixer_channels & SOUND_MASK_VOLUME)) {
			emulate_master_volume = True;
		}
	}

	return 0;
}


static int  api_close_mixer(void) {
	close(gsimixer_fd);
	gsimixer_fd = -1;

	return 0;
}


/* GSI_MIXER_MASTER, ... -> native: */
static int oss_mixer_read[] = {
	SOUND_MIXER_READ_VOLUME, SOUND_MIXER_READ_PCM, SOUND_MIXER_READ_SYNTH,
	SOUND_MIXER_READ_CD, SOUND_MIXER_READ_LINE, SOUND_MIXER_READ_MIC
};
static int oss_mixer_write[] = {
	SOUND_MIXER_WRITE_VOLUME, SOUND_MIXER_WRITE_PCM, SOUND_MIXER_WRITE_SYNTH,
	SOUND_MIXER_WRITE_CD, SOUND_MIXER_WRITE_LINE, SOUND_MIXER_WRITE_MIC
};

static void api_set_volume(int device_type, int device_no, int vol) {
	int  mixer, volume;
	int  device;

	DEB1(DPRT("("__FILE__"): set oss volume\n"));
	if (gsimixer_ok) {
		device = oss_mixer_write[ device_type ];
		mixer = open(gsimixer_device, O_WRONLY | O_NONBLOCK);
		if (mixer < 0) {
			log_message(GSI_MESSAGE_ERROR, "%s: (oss) open mixer: %s", gsimixer_device, strerror(errno));
		}
		else {
			volume = (vol & 65535); /* Maybe use range 0-65535. */
			if (volume > 256) volume = 256;
			volume = (volume*100)/256; /* OSS uses a scale 0-100. */
			DEB1(DPRT(" -> set volume of device x%8x to %d (converted to %d for OSS)\n", device, vol, volume));

#if 0
			/* balance type 1: center gives half volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				volume_l = (vol*(256-balance))>>9;
				volume_r = (vol*(256+balance))>>9;
			}
#endif
#if 0
			/* balance type 2: center gives full volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				volume_l = min((vol*(256-balance))>>8, 256);
				volume_r = min((vol*(256+balance))>>8, 256);
			}
#endif

			volume |= volume<<8; /* Left and right channels */
			ioctl(mixer, device, &volume);
			close(mixer);
		}
	}
}


static int  api_get_volume(int device_type, int device_no) {
	int  mixer;
	int  volume = -1;
	int  device;
	int  volume_l, volume_r;

	DEB1(DPRT("("__FILE__"): get_oss_volume\n"));
	if (gsimixer_ok) {
		device = oss_mixer_read [ device_type ];
		mixer = open(gsimixer_device, O_RDONLY | O_NONBLOCK);
		if (mixer < 0) {
			log_message(GSI_MESSAGE_ERROR, "%s: (oss) open mixer: %s", gsimixer_device, strerror(errno));
		}
		else {
			ioctl(mixer, device, &volume);
			volume_l = volume & 255;
			volume_r = (volume >> 8) & 255;

#if 0
			/* balance type 1: center gives half volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				vol = volume_l+volume_r;
				if (vol > 256) vol = 256;
				balance = ((volume_r-volume_l) << 8)/vol;
			}
#endif
#if 0
			/* balance type 2: center gives full volume on left and on right */
			if (gsimixer_device == SOUND_MASK_VOLUME) {
				vol = max(volume_l, volume_r);
				balance = ((volume_r-volume_l) << 8)/vol;
			}
#endif

			volume = (((volume_l+volume_r)*128)/100); /* We currently return the average channel value */
			DEB1(DPRT("  -> get_volume of device x%8x: sound_volume = %d (from OSS value of %d)\n", device, volume, volume_l));
			close(mixer);
		}
	}

	return volume;
}
#endif /* USE_OSS */


/*      EOF mixer_oss.c */
