/* GSI - general sound interface
 *
 * mixer_ultra.c - mixer setting/reading for ultra
 *
 * Lower level API dependencies:
 *   - Ultrasound mixer constants (same names as OSS constants), + devices
 *
 *
 * Copyright (C) 1998 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#if  USE_ULTRASOUND
#  include <libgus.h>

void *ultra_mixer_handle = NULL;
int  ultra_card = 0;
int  ultra_dev = 0;


/*---------------------------------------------------------------------------*
 *      Get/set sound volumes with ultra.
 *      Ultrasound driver volumes are 0-100.
 *---------------------------------------------------------------------------*/
static int  api_init_mixer(void) {
	int  status;

	// ultra_card = gus_mixer_look_for_card("");
	DEB1(DPRT("ultra_mixer_handle=%x, ultra_card=%d, ultra_dev=%d\n", (int) ultra_mixer_handle, ultra_card, ultra_dev));

	status = gus_mixer_open(&ultra_mixer_handle, ultra_card, ultra_dev);
	if (status < 0 || ultra_mixer_handle == NULL) {
		gsimixer_ok = False;
		return -1;
	}

	gsimixer_channels = gus_mixer_read_devmask(ultra_mixer_handle);
	if (!(gsimixer_channels & SOUND_MASK_VOLUME)) {
		emulate_master_volume = True;
	}

	return 0;
}


static int  api_close_mixer(void) {
	gus_mixer_close(ultra_mixer_handle);
	ultra_mixer_handle = NULL;

	return 0;
}


/* all mixer channels included here, only a few are used. */
struct mixer_channel_info {
	char *name;
	int  id;
	int  mask;
	int  (*read)(void *handle);
	int  (*write)(void *handle, int vol);
} gus_mixer_channels[] = {
	{ "Master",	GUS_MIXER_ID_MASTER,	GUS_MIXER_MASK_MASTER,
	  gus_mixer_read_volume,	gus_mixer_write_volume	},
	{ "Synth",	GUS_MIXER_ID_SYNTH,		GUS_MIXER_MASK_SYNTH,
	  gus_mixer_read_synth,	gus_mixer_write_synth	},
	{ "Effect",	GUS_MIXER_ID_EFFECT,	GUS_MIXER_MASK_EFFECT,
	  gus_mixer_read_effect,	gus_mixer_write_effect },
	{ "PCM",	GUS_MIXER_ID_PCM,		GUS_MIXER_MASK_PCM,
	  gus_mixer_read_pcm,	gus_mixer_write_pcm	},
	{ "PCM1",	GUS_MIXER_ID_PCM1,		GUS_MIXER_MASK_PCM1,
	  gus_mixer_read_pcm1,	gus_mixer_write_pcm1	},
	{ "Line",	GUS_MIXER_ID_LINE,		GUS_MIXER_MASK_LINE,
	  gus_mixer_read_line,	gus_mixer_write_line	},
	{ "MIC",	GUS_MIXER_ID_MIC,		GUS_MIXER_MASK_MIC,
	  gus_mixer_read_mic,	gus_mixer_write_mic	},
	{ "CD",		GUS_MIXER_ID_CD,			GUS_MIXER_MASK_CD,
	  gus_mixer_read_cd,	gus_mixer_write_cd	},
	{ "Soft",	GUS_MIXER_ID_SOFT,		GUS_MIXER_MASK_SOFT,
	  gus_mixer_read_imix,	gus_mixer_write_imix	},
	{ "Gain",	GUS_MIXER_ID_GAIN,		GUS_MIXER_MASK_GAIN,
	  gus_mixer_read_reclev,	gus_mixer_write_reclev	},
	{ "Loop",	GUS_MIXER_ID_LOOP,		GUS_MIXER_MASK_LOOP,
	  gus_mixer_read_loopback, gus_mixer_write_loopback }
};

// GSI_MIXER_MASTER, ... -> native
static int ultra_mixer_ids[] = {
	GUS_MIXER_ID_MASTER, GUS_MIXER_ID_PCM, GUS_MIXER_ID_SYNTH, GUS_MIXER_ID_CD,
	GUS_MIXER_ID_LINE, GUS_MIXER_ID_MIC
};

static void api_set_volume(int device_type, int device_no, int vol) {
	int  volume_l, volume_r;
	int  i;

	DEB1(DPRT("api_set_volume: GSI dev=%d, vol =%d\n", device_type, vol));
	device_type = ultra_mixer_ids [ device_type ];

	DEB1(DPRT("api_set_volume: translated dev=%d\n", device_type));
	for (i=0; i<10; i++) {
		if (gus_mixer_channels[i].id == device_type) break;
	}
	DEB1(DPRT("api_set_volume: i=%d, dev name=%s, gus handle=0x%x\n", i, gus_mixer_channels[i].name, (int) ultra_mixer_handle));

	if (vol > 256) vol = 256;
	vol = (vol*100)/256; /* ultra uses a scale 0-100 */
	/* add balance code here .. */
	volume_l = vol;
	volume_r = vol;
	gus_mixer_channels[i].write(ultra_mixer_handle, volume_l | (volume_r << 8));
}


static int  api_get_volume(int device_type, int device_no) {
	int  volume_l, volume_r;
	int  vol;
	int  i;

	device_type = ultra_mixer_ids[ device_type ];
	for (i=0; i<10; i++) {
		if (gus_mixer_channels[i].id == device_type) break;
	}
	vol = gus_mixer_channels[i].read(ultra_mixer_handle);

	volume_l = (vol & 0xff);
	volume_r = (vol >> 8) & 0xff;

	return volume = (((volume_l+volume_r)*128)/100); /* We currently return the average channel value */
}
#endif /* USE_ULTRASOUND */


/*      EOF mixer_ultrasound.c */
