/* GSI - general sound interface
 *
 * mixer_functions.c - libgsimixer, mixer setting/reading, somewhat abstracted
 * from the lower level api's (e.g. the master volume emulation).
 *
 * Lower level API dependencies:
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsimixer_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

/* For open(2): */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* For the api code: */
#include <sys/ioctl.h>
#include <errno.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIMIXER_"
#include <gsi/gsi_debug.h>

#include "gsimixer.h"

#include "log.h"

#include "api.h"


/*__________________VARIABLES___________________*/
char *gsimixer_device;
int  gsimixer_fd;

int  gsimixer_channels; /* bitmask */


int  gsimixer_ok = False;

static int  emulate_master_volume = False;

int  gsimixer_balance = 0; /* -256 to 256 */
int  gsimixer_muted = False;
static int  master_save;

int  gsimixer_num_devices [ GSIMIXER_NUM_DEVICE_TYPES ] = { 0,0,0,0,0,0 };

int  *gsimixer_volume [ _RECORD_DEVICE_OFFSET + GSIMIXER_NUM_DEVICE_TYPES ];
/*
int  gsimixer_volume [ _RECORD_DEVICE_OFFSET + GSIMIXER_NUM_DEVICE_TYPES ] = {
	-1,  -1,  -1,  -1,  -1,  -1
	// recording devices:
	-1,  -1,  -1,  -1,  -1,  -1
};
*/


/*---------------------------------------------------------------------------*
 *      Initialize mixer, checks whether we have a master volume and reads
 *      the current volumes if there are no previous (saved between open/
 *      close mixer) values.
 *---------------------------------------------------------------------------*/
int gsimixer_init() {
	int  i, j;

	DEB1(DPRT("("__FILE__"): init mixer\n"));

	for (i=0; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
		if (gsimixer_num_devices[i] > 0) {
			gsimixer_volume[i] = malloc(sizeof(int) * gsimixer_num_devices[i]);
			if (gsimixer_volume[i] == NULL) {
				log_message(GSI_MESSAGE_ERROR, "Out of memory in %s", "gsimixer_init");
				return -1;
			}
		}
		else
			gsimixer_volume[i] = NULL;
		/* Init volumes: */
		for (j=0; j<gsimixer_num_devices[i]; j++) {
			gsimixer_volume[i][j] = -1;
		}
	}
	if (api_init_mixer()) return -1;

DEB(fprintf(stderr, "mix x\n"));

	for (i=0; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
		for (j=0; j< gsimixer_num_devices[i]; j++) {

			if (i == GSI_MASTER && emulate_master_volume)
				gsimixer_volume[ GSI_MASTER ][0] = 256;
			else
				gsimixer_volume[i][j] = api_get_volume(GSI_MASTER+i, j);
		}
	}

	DEB1(DPRT("#init mixer#\n"));

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Close the mixer
 *---------------------------------------------------------------------------*/
int gsimixer_close() {
	api_close_mixer();

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Mixer volume settings:
 *
 *      saving volumes is done to remove any rounding problems with repeated
 *      reading/writing (device values read may not be the same as what has
 *      been set, due to range conversions and/or hardware limitations).
 *---------------------------------------------------------------------------*/
int gsimixer_mute() {
	master_save = gsimixer_volume[ GSI_MASTER ][0];
	//set_mixer(GSI_MASTER, 0);
	gsimixer_set_device_volume(GSI_MASTER, 0 /*devno*/, 0);
	gsimixer_muted = True;

	return -1;
}


int gsimixer_unmute() {
	int  i, j;

	if (gsimixer_muted)
		gsimixer_set_device_volume(GSI_MASTER, 0, master_save);
	gsimixer_muted = False;

	for (i=1; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
		for (j=0; j< gsimixer_num_devices[i]; j++) {
			gsimixer_set_device_volume(i, j, gsimixer_volume[i][j]);
		}
	}
//	for (i=0; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
//		set_application_volume(i, gsi_application_volume[i]);
//	}

	return -1;
}


int gsimixer_set_balance(int new_balance) {
	gsimixer_balance = new_balance;
	gsimixer_set_device_volume(GSI_MASTER, 0, gsimixer_get_device_volume(GSI_MASTER, 0));

	return -1;
}


int gsimixer_set_device_volume(int devtype, int devno, int vol) {
#if 0 /* THIS IS DONE IN GSI_SERVER */
	int  hw_volume = -1;

	DEB1(DPRT("set mixer dev=%d vol=%d\n", devtype, vol));
	gsimixer_volume [ devtype ] = vol;

	switch (devtype) {
	case GSI_MASTER:
		hw_volume = vol;
		break;
	case GSI_PCM:
		hw_volume = vol;
		break;
	case GSI_SYNTH:
	case GSI_CD:
	case GSI_LINE_IN:
	case GSI_MICROPHONE:
		if (current == NULL) // Initialization (no applications yet)
			hw_volume = vol;
		else
			hw_volume = (vol * current->application_volume[devtype])>>8;
		break;
	}
#endif
	int hw_volume = vol;

	if (hw_volume >= 0) {
		if (!gsimixer_muted) {
			if (devtype == GSI_MASTER) {
				if (emulate_master_volume) {
					int  i, j;

					for (i=1; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
						for (j=0; j<gsimixer_num_devices[i]; j++) {
							api_set_volume(GSI_MASTER+i, j, (gsimixer_volume[i][j] * hw_volume)>>8);
						}
					}
				}
				else {
					DEB1(DPRT("set master to %d:\n", hw_volume));
					api_set_volume(GSI_MASTER, 0, hw_volume);
					DEB1(DPRT("set master done.\n"));
				}
			}
			else {
				api_set_volume(devtype, devno, hw_volume);
			}
		}

		return 0;
	}

	return -1;
}


int gsimixer_get_device_volume(int devtype, int devno) {
	if (devtype >= GSI_MASTER && devtype <= GSI_MICROPHONE && devno >= 0 && devno < gsimixer_num_devices[ devtype ]) {
		if (emulate_master_volume) {
			if (devtype == GSI_MASTER)
				return gsimixer_volume[ devtype ][ devno ];
			else
				return api_get_volume(devtype, devno);
		}
		else {
			return api_get_volume(devtype, devno);
		}
	}

	return -1;
}


// The underlying API specific bits:
#include "api/mixer_oss.c"
#include "api/mixer_ultra.c"
#include "api/mixer_devaudio.c"
#include "api/mixer_obsd.c"
#include "api/mixer_fbsd.c"


/*      EOF mixer_functions.c */
