/* GSI - general sound interface
 *
 * pcm_devaudio.c - /dev/audio pcm mode setting
 *
 * Lower level API dependencies:
 *
 *
 * Copyright (C) 1996 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#if  USE_DEV_AUDIO
/* Simulated fragment stuff: */
#define _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE   1024 /* = 0.128 s; No implementation should have a smaller buffer ... (we could find out the real size with non-blocking write). */

static int            prev_written;
static int            prev_buf_space;
static struct timeval prev_time_of_day;
/* Initilize at first write to pcm device after opening (or just at opening as
 * we're supposed to write continuously to the pcm device after opening).
 */


/*---------------------------------------------------------------------------*
 *      setup_pcm (devaudio):
 *        opens device (pcmdev; usually /dev/dsp)
 *        sets fragment size = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE
 *        sets samplerate = 8000
 *        sets channels = 1
 *        sets sample type = 8 bit mulaw
 *---------------------------------------------------------------------------*/
int api_setup_pcm (GSIPCM_HW, int samplerate, int channels, int type, int log2_fragsize)
{
	DEB1(DPRT("("__FILE__"): *setup_pcm* (devaudio)\n"));

	if (hw_status->fd < 0) { /* Not yet opened. */
		if (hw->device == NULL) return -1;
		hw_status->fd = open(hw->device, O_WRONLY);
DEB(DPRT("api_setup_pcm: fd=%d\n", hw_status->fd));

		if (hw_status->fd < 0) {
			if (errno == EBUSY) {
				return GSIPCM_BUSY;
			}
			else {
				return GSIPCM_NOT_PRESENT;
			}
		}
	}

	/* We use a 128 byte fragsize for /dev/audio. */
	hw_status->samplerate = 8000;
	hw_status->channels = 1;
	hw_status->type = GSI_8BIT_MULAW;
	hw_status->log2_fragsize = 7;
	hw_status->fragsize = 128;

	hw_status->bps = 1;
	hw_status->samples = 128;
	hw_status->byte_order = GSI_BIG_ENDIAN; /* Endianness actually doesn't matter as we only have mulaw. */

	hw_status->max_write_space = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE;

	/* Initialize fragment/timing stuff: */
	prev_written = 0;
	prev_buf_space = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE;
	gettimeofday(&prev_time_of_day, NULL);

	api_pcm_set_mode (hw, hw_status, hw_status->modus);

	DEB1(DPRT("  pcm settings: bps=%d, samples=%d [ samplerate=%d, channels=%d, type=%d, log2_frag=%d, frag=%d ]\n", hw_status->bps, hw_status->samples, hw_status->samplerate, hw_status->channels, hw_status->type, hw_status->log2_fragsize, hw_status->fragsize));

	DEB1(DPRT("#setup_pcm#\n"));

	return 0;
}


int api_pcm_set_mode (GSIPCM_HW, int modus) {
	int status = -1;

	if (hw_status->fd > 0)
	switch(modus) {
	case GSIPCM_MODE_NON_BLOCKING:
		status = fcntl(hw_status->fd, F_GETFL);
		if (status != -1)
			status = fcntl(hw_status->fd, F_SETFL, status | O_NONBLOCK);
		break;
	case GSIPCM_MODE_BLOCKING:
		status = fcntl(hw_status->fd, F_GETFL);
		if (status != -1)
			status = fcntl(hw_status->fd, F_SETFL, status & ~O_NONBLOCK);
		break;
	case GSIPCM_MODE_DMA:
		break;
	}

	return status;
}


int api_pcm_write (GSIPCM_HW, void *buf, int bytes_to_write) {
	struct timeval current_time_of_day, dt;
	int  status;

	status = write (hw_status->fd, buf, bytes_to_write);

	/* Now set prev_tod = current_tod - written time. */
	if (status > 0) {
		gettimeofday(&current_time_of_day, NULL);

		dt.tv_sec  = current_time_of_day.tv_sec  - prev_time_of_day.tv_sec;
		dt.tv_usec = current_time_of_day.tv_usec - prev_time_of_day.tv_usec;
		if (dt.tv_usec < 0) {
			dt.tv_usec += 1000000;
			dt.tv_sec --;
		}
		if (dt.tv_sec < 0) {
			/* Either date was set, or some bug (on Linux, -3 s is possible!).
			 *
			 * We now use the stored prev_written and assume they were all
			 * handled (reasonable assumption as the timing takes the bufsize/
			 * /fragszie into account).
			 */
			prev_buf_space = prev_buf_space - status + prev_written;

			/* Or should we just use do the same as for underruns and set
			 * buffer = max ?
			 */
		}
		else if (dt.tv_sec > 0) {
			/* Serious underflow (or date changed).
			 * Assume full buffer is available.
			 */
			prev_buf_space = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE;
		}
		else {
			/* tv_sec = 0, tv_usec >= 0; factored out 1000 from samplerate
			 * and time (musec).
			 */
			prev_buf_space = prev_buf_space + (dt.tv_usec * 8)/1000 - status;
		}
		prev_time_of_day.tv_sec  = current_time_of_day.tv_sec;
		prev_time_of_day.tv_usec = current_time_of_day.tv_usec;
		prev_written = status;
	}

	return status;
}


int api_pcm_write_space (GSIPCM_HW) {
	//int  status;
	int  audio_space;
	struct timeval current_time_of_day, dt;


	/* Almost the same code as in api_pcm_write above: */
		gettimeofday(&current_time_of_day, NULL);

		dt.tv_sec  = current_time_of_day.tv_sec  - prev_time_of_day.tv_sec;
		dt.tv_usec = current_time_of_day.tv_usec - prev_time_of_day.tv_usec;
		if (dt.tv_usec < 0) {
			dt.tv_usec += 1000000;
			dt.tv_sec --;
		}
		if (dt.tv_sec < 0) {
			/* Either date was set, or some bug (on Linux, -3 s is possible!).
			 *
			 * We now use the stored prev_written and assume they were all
			 * handled (reasonable assumption as the timing takes the bufsize/
			 * /fragszie into account).
			 */
			audio_space = prev_buf_space /* - status */ + prev_written;

			/* Or should we just use do the same as for underruns and set
			 * buffer = max ?
			 */
		}
		else if (dt.tv_sec > 0) {
			/* Serious underflow (or date changed).
			 * Assume full buffer is available.
			 */
			audio_space = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE;
		}
		else {
			/* tv_sec = 0, tv_usec >= 0; factored out 1000 from samplerate
			 * and time (musec).
			 */
			audio_space = prev_buf_space + (dt.tv_usec * 8)/1000 /* - status*/;
		}

		if (audio_space > _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE) {
			audio_space = _GSI_DEV_AUDIO_KERNEL_BUFFER_SIZE;
		}

	DEB(DPRT("devaudio pcm out space = %d\n", audio_space));

	return audio_space;
}


int api_pcm_buffer_status (GSIPCM_HW, int *writespace, int *readdata) {
	int  status;

	if (writespace) {
		*writespace = api_pcm_write_space(hw, hw_status);
	}
	if (readdata) {
		/* forgot for now */
	}

	return 0;
}

#endif /* USE_DEV_AUDIO */


/*      EOF pcm_devaudio.c */
