/* GSI - general sound interface
 *
 * pcm_fbsd.c - FreeBSD driver pcm mode setting
 *
 * Lower level API dependencies:
 *   - The FreeBSD audio driver.
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#if  USE_FBSD
#include <machine/soundcard.h>

/*---------------------------------------------------------------------------*
 *      setup_pcm (fbsd):
 *        opens device (pcmdev; usually /dev/dsp)
 *        sets fragment size
 *        sets samplerate
 *        sets channels
 *        sets sample type
 *---------------------------------------------------------------------------*/
int api_setup_pcm (GSIPCM_HW, int samplerate, int channels, int type, int log2_fragsize)
{
	int  status;
	int  i;
	int  format;
	int  n_bytes, n_bytes_kernel, j, my_log2_frag, my_num_frags;

	int  snd_formats[] = { AFMT_MU_LAW, AFMT_U8, AFMT_S16_LE };
#ifdef DEBUG
	char *gsi_format_strings[] = { "alaw", "mulaw", "unsigned 8 bit", "signed 16 bit" };
#endif

	DEB1(DPRT("("__FILE__"): *setup_pcm* (fbsd)\n"));

	if (hw_status->fd < 0) { /* Not yet opened. */
		if (hw->device == NULL) return -1;
		hw_status->fd = open(hw->device, O_WRONLY);
		DEB1(fprintf(stderr, "pcm: fd=%d\n", hw_status->fd));
		if (hw_status->fd < 0) {
			if (errno == EBUSY) {
				return GSIPCM_BUSY;
			}
			else {
				return GSIPCM_NOT_PRESENT;
			}
		}
	}

	if (type < GSI_8BIT_MULAW) type = GSI_8BIT_MULAW;
	if (type > GSI_16BIT) type = GSI_16BIT;
	format = snd_formats[ type - 1 ];

	status = ioctl(hw_status->fd, SNDCTL_DSP_RESET);

	status = ioctl(hw_status->fd, SNDCTL_DSP_SETFMT, &format);
DEB(DPRT("status = %d, format=%d\n", status, format));
	channels -= 1;
    status = ioctl(hw_status->fd, SNDCTL_DSP_STEREO, &channels);
	channels += 1;
DEB(DPRT("status = %d, channels=%d\n", status, channels));
//test:
//samplerate = 8000;

    status = ioctl(hw_status->fd, SNDCTL_DSP_SPEED, &samplerate);
DEB(DPRT("status = %d, samplerate=%d\n", status, samplerate));
	// { int value = 9; /* 1 << 9 = 512 b fragment */
	// status = ioctl(hw_status->fd, SNDCTL_DSP_SETFRAGMENT, &value); }

	/* format, channels & samplerate may have been adjusted by the driver: */
	hw_status->samplerate = samplerate;
	hw_status->channels = channels;
	for (i=GSI_8BIT_MULAW; i<=GSI_16BIT; i++) {
		if (format == snd_formats[ i-1 ]) {
			hw_status->type = i;
			break;
		}
	}
	if (i == 4) hw_status->type = GSI_UNKNOWN_FORMAT;

	log2_fragsize = min(log2_fragsize, 16); /* maximum is 16 */
	log2_fragsize = max(log2_fragsize, 7); /* minimum is 7 */

#if 0 /* AIOSSIZE doesn't work as I want it. */
	/* Trickery: determine the size of the internal kernel audio buffer */
	for (j=18; j >= 7; j--) {
		struct snd_size size;

		size.play_size = 1 << j;
		size.rec_size = 0 /* 1 << j */;
		status = ioctl(hw_status->fd, AIOSSIZE, &size);
		DEB(DPRT("AIOSSIZE status = %d\n", status));
		DEB(DPRT("AIO SIZE is now %d,%d\n", size.play_size, size.rec_size));

		if (status < 0) continue;

		hw_status->buf_size = size.play_size;
		hw_status->fragsize = max(128, 1 << log2_fragsize);

		DEB(DPRT("buf_size = %d, fragsize = %d\n", hw_status->buf_size, hw_status->fragsize));

		/* Emulate blocking write & fragments with AIONWRITE */
		size.play_size -= hw_status->fragsize /* *2 */;
		size.rec_size -= hw_status->fragsize /* *2 */;

		status = ioctl(hw_status->fd, AIOSSIZE, &size);

		DEB(DPRT("AIOSSIZE status = %d\n", status));
		DEB(DPRT("AIO SIZE is now %d,%d\n", size.play_size, size.rec_size));

		hw_status->log2_fragsize = log2_fragsize; /* bad ... */
		break;
	}
#else

	/* Another check, I'm setting log2_fragsize such that fragsize/data_rate
	 * > SYS_TIME where SYS_TIME = 1/SYS_Hz, i.e. usually .01 s on unix. */

#  define SYS_TIME 10000 /* microseconds, so .01 s */
	/* Set fragment no/size. */

	/* Software_fragsize:
	 */
	n_bytes = (SYS_TIME * samplerate * channels * gsipcm_bytes_per_sample[ type ] + 1000000) / 1000000;
	DEB1(DPRT("samplerate =%d, channels = %d, gsipcm_bytes_per_sample [ %d ] = %d\n", samplerate, channels, type, gsipcm_bytes_per_sample[ type ]));
	DEB1(DPRT("nbytes: %d\n", n_bytes));
	for (j=0; n_bytes != 0; j++) {
		n_bytes >>= 1;
	}
	DEB1(DPRT("nbytes 2n-rounded up: %d (log2=%d)\n", 1 << j, j));
	n_bytes = 1 << j;

	/* Minimal kernel buffer size = 8 * SYS_TIME, approximated with 8*n_bytes
	 * (for now).
	 */
	n_bytes_kernel = 8 * n_bytes;

	/* We try to set a hardware fragsize 1/2 (-> 1/4?) of this software fragsize.
	 */
	for (;;) {
		my_log2_frag = j;
		my_num_frags = n_bytes_kernel / (1 << my_log2_frag);
		DEB(DPRT("my_log2_frag = %d, my_num_frags = %d\n", my_log2_frag, my_num_frags));
		break;
	}

	/* Set the software fragsize. */
	hw_status->log2_fragsize = j;
	hw_status->fragsize = 1 << j;
#endif


	hw_status->bps = gsipcm_bytes_per_sample[ hw_status->type ];

	hw_status->samples = hw_status->fragsize / (hw_status->bps * hw_status->channels);
#warning "FIX byte order"
	hw_status->byte_order = GSI_LITTLE_ENDIAN;

	hw_status->max_write_space = api_pcm_write_space(hw, hw_status);

	api_pcm_set_mode (hw, hw_status, hw_status->modus);

	DEB1(DPRT("  pcm settings: bps=%d, samples=%d [ samplerate=%d, channels=%d, type=%d, log2_frag=%d, frag=%d ]\n", hw_status->bps, hw_status->samples, hw_status->samplerate, hw_status->channels, hw_status->type, hw_status->log2_fragsize, hw_status->fragsize));

	DEB1(DPRT("  type_string = %s\n", hw_status->type != GSI_UNKNOWN_FORMAT ? gsi_format_strings [ hw_status->type ] : "unknown"));

	DEB1(DPRT("#setup_pcm#\n"));

	return 0;
}


int api_pcm_set_mode (GSIPCM_HW, int modus) {
	int status = -1;

	if (hw_status->fd > 0) {
		switch(modus) {
		case GSIPCM_MODE_NON_BLOCKING:
			status = fcntl(hw_status->fd, F_GETFL);
			if (status != -1)
				status = fcntl(hw_status->fd, F_SETFL, status | O_NONBLOCK);
			break;
		case GSIPCM_MODE_BLOCKING:
			status = fcntl(hw_status->fd, F_GETFL);
			if (status != -1)
				status = fcntl(hw_status->fd, F_SETFL, status & ~O_NONBLOCK);
			break;
		case GSIPCM_MODE_DMA:
			break;
		}
	}

	return status;
}


int api_pcm_write_space (GSIPCM_HW) {
	int  status;
	int  audio_space;

	status = ioctl(hw_status->fd, AIONWRITE, &audio_space);
	DEB1(DPRT("AIONWRITE space = %d (ioctl st=%d)\n", audio_space, status));
	if (status < 0) audio_space = -1;

	return audio_space;
}


int api_pcm_buffer_status (GSIPCM_HW, int *writespace, int *readdata) {
	int  status;

	if (writespace) {
		status = ioctl(hw_status->fd, AIONWRITE, writespace);
		DEB1(DPRT("fbsd AIONWRITE: write space = %d (ioctl st=%d)\n", *writespace, status));
		if (status < 0) *writespace = -1;
	}
	if (readdata) {
		status = ioctl(hw_status->fd, FIONREAD, readdata);
		DEB1(DPRT("fbsd FIONREAD: read data = %d (ioctl st=%d)\n", *readdata, status));
		if (status < 0) *readdata = -1;
	}

	return 0;
}


int api_pcm_write (GSIPCM_HW, void *buf, int bytes_to_write) {
	return write (hw_status->fd, buf, bytes_to_write);
}
#endif /* USE_FBSD */


/*      EOF pcm_fbsd.c */
