/* GSI - general sound interface
 *
 * pcm_obsd.c - open/net bsd pcm mode setting
 *
 * Lower level API dependencies:
 *   - Open/Net BSD audio.h
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#if  USE_OPENBSD_AUDIO
#include <sys/types.h>
#include <sys/audioio.h>

#warning "Should change fragsize allocation: more smaller frags as output space can be determined only in frags"
/*---------------------------------------------------------------------------*
 *      setup_pcm (obsd):
 *        opens device (hw->device; /dev/audio or /dev/sound)
 *        sets fragsize (2 fragments of size 'fragsize', or 3 of fragsize/2).
 *        sets samplerate ('samplerate')
 *        sets mono/stereo ('channels')
 *        sets sample type; 8 bit mulaw/8 bit unsigned/16 bit ('type')
 *---------------------------------------------------------------------------*/
void get_device_info(GSIPCM_HW);
void get_audio_formats(GSIPCM_HW);


int OBSD_encoding_of(int gsi_type) {
	switch(gsi_type) {
	case GSI_8BIT:
		return AUDIO_ENCODING_ULINEAR;

	case GSI_16BIT:
		return AUDIO_ENCODING_SLINEAR_BE;

	case GSI_16BIT_LE:
		return AUDIO_ENCODING_SLINEAR_LE;

	case GSI_8BIT_MULAW:
		return AUDIO_ENCODING_ULAW;

	case GSI_8BIT_ALAW:
		return AUDIO_ENCODING_ALAW;

	default:
		log_message(GSI_MESSAGE_ERROR, "(OBSD) OBSD_encoding_of: Unhandled GSI type\n");
		return AUDIO_ENCODING_NONE;
	}

	return AUDIO_ENCODING_NONE;
}


int outspace = 0;
int api_setup_pcm (GSIPCM_HW, int samplerate, int channels, int type, int log2_fragsize) {
	int  status;

//	int  snd_formats[] = { AUDIO_ENCODING_ULAW, AUDIO_ENCODING_SLINEAR, AUDIO_ENCODING_SLINEAR_LE, AUDIO_ENCODING_SLINEAR_BE };
	char *snd_format_strings[] = { "mulaw", "unsigned 8 bit", "signed 16 bit" };
//	int  first_pcm_type; /* Modesetting, first one to try */
//	int  need_fragsize_change = True;

	audio_info_t audio_settings;


	DEB1(DPRT("("__FILE__"): *setup_pcm* (obsd)\n"));

	/* A few sanity checks */
	samplerate = min(samplerate, 48000);
	samplerate = max(samplerate, 4000);
	type  = min(type, GSI_16BIT);
	type  = max(type, GSI_8BIT_MULAW);
	log2_fragsize = min(log2_fragsize, 13); /* Max 8K buffer at the moment */
	DEB1(DPRT("request samplerate=%d, ch=%d, type=%s\n", samplerate, channels, snd_format_strings[type-1]));

	/* Fragment size */
	DEB1(DPRT("  setting fragment to 2^%d\n", log2_fragsize));
	if (log2_fragsize <= 3) {
		/* Choose the best fragsize
		 *                      1  2  3  4  5  6  7  8  9 10 11 12 */
		int  base_log[ 12 ] = { 0, 1, 2, 2, 2, 3, 3, 3, 3, 3, 3, 4 };
		log2_fragsize += base_log[ (samplerate+2000)/4000 -1 ] + 7 + (channels >> 1) + (gsipcm_bytes_per_sample[ type+1 ] >> 1);
		/* e.g. 8000Hz mono 8 bit = 256, 48000Hz, stereo 16 bit = 8192 */
		DEB1(DPRT("  auto log2_fragsize=%d\n", log2_fragsize));
	}
	log2_fragsize = max(log2_fragsize, 4); /* sanity check */
	log2_fragsize = min(log2_fragsize, 13);


	if (hw_status->fd > 0) {
		/* Do we need to close the PCM in some cases with the BSD audio system?
		 */
		close(hw_status->fd); hw_status->fd = -1;
	}

	if (hw_status->fd < 0) { /* Not yet opened */
		if (hw->device == NULL) return -1;
		hw_status->fd = open(hw->device, O_WRONLY);
		if (hw_status->fd < 0) {
			log_message(GSI_MESSAGE_ERROR, "Open %s: %s", hw->device, strerror(errno));
			if (errno == EBUSY) {
				return GSIPCM_BUSY;
			}
			else {
				return GSIPCM_NOT_PRESENT;
			}
			return -1;
		}
	}

	AUDIO_INITINFO(&audio_settings);
	audio_settings.blocksize = 1 << log2_fragsize;
	audio_settings.lowat = 1; /* low water mark (in audio blocks) */
	if (hw_status->modus == GSIPCM_MODE_BLOCKING)
		audio_settings.hiwat = 2; /* high water mark */
	else
		audio_settings.hiwat = 16384 / audio_settings.blocksize; /* high water mark (make higher for non-blocking writes otherwise the driver overwrites old data...) */
	audio_settings.mode = AUMODE_PLAY | AUMODE_PLAY_ALL;
	audio_settings.play.sample_rate = /*hw_status->*/samplerate;
	audio_settings.play.channels = /*hw_status->*/channels;
	audio_settings.play.precision = 8*gsipcm_bytes_per_sample [ /*hw_status->*/type ];
	audio_settings.play.encoding = OBSD_encoding_of (/*hw_status->*/type);
#if 1
DEB(DPRT("AUDIO_SETINFO ok; values set:\n"));
DEB(DPRT("audio_settings.blocksize=%d\n", audio_settings.blocksize));
DEB(DPRT("audio_settings.lowat=%d\n", audio_settings.lowat));
DEB(DPRT("audio_settings.hiwat=%d\n", audio_settings.hiwat));
DEB(DPRT("audio_settings.mode=%d\n", audio_settings.mode));
DEB(DPRT("audio_settings.play.sample_rate=%d\n", audio_settings.play.sample_rate));
DEB(DPRT("audio_settings.play.precision=%d\n", audio_settings.play.precision));
DEB(DPRT("audio_settings.play.encoding=%d\n", audio_settings.play.encoding));
DEB(DPRT("audio_settings.play.channels=%d\n", audio_settings.play.channels));
#endif

	status = ioctl(hw_status->fd, AUDIO_SETINFO, &audio_settings);
	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "%s: (OBSD) AUDIO_SETINFO: %s", hw->device, strerror(errno));
		exit(1);
	}

	status = ioctl(hw_status->fd, AUDIO_GETINFO, &audio_settings);
#if 1
DEB(DPRT("AUDIO_SETINFO ok; values set:\n"));
DEB(DPRT("audio_settings.blocksize=%d\n", audio_settings.blocksize));
DEB(DPRT("audio_settings.lowat=%d\n", audio_settings.lowat));
DEB(DPRT("audio_settings.hiwat=%d\n", audio_settings.hiwat));
DEB(DPRT("audio_settings.mode=%d\n", audio_settings.mode));
DEB(DPRT("audio_settings.play.sample_rate=%d\n", audio_settings.play.sample_rate));
DEB(DPRT("audio_settings.play.precision=%d\n", audio_settings.play.precision));
DEB(DPRT("audio_settings.play.encoding=%d\n", audio_settings.play.encoding));
DEB(DPRT("audio_settings.play.channels=%d\n", audio_settings.play.channels));
#endif

	/* Now get the settings to see if they really changed: */
	hw_status->samplerate = audio_settings.play.sample_rate;
	hw_status->channels = audio_settings.play.channels;

	hw_status->type = type;

#warning "FIX byte order"
	hw_status->byte_order = GSI_LITTLE_ENDIAN;


	hw_status->log2_fragsize = log2_fragsize;
	hw_status->fragsize = 1 << hw_status->log2_fragsize;

hw_status->max_write_space = audio_settings.play.buffer_size - 1 /* low watermark */ * hw_status->fragsize; //64000; //api_pcm_write_space(hw, hw_status);
DEB(DPRT("init 1: outspace=%d\n", outspace));
 outspace = hw_status->max_write_space + 2 /* low watermark */ * hw_status->fragsize;
DEB(DPRT("init 2: outspace=%d\n", outspace));

	/* Calculate some constants used to play fragments: */
	hw_status->bps = gsipcm_bytes_per_sample [ hw_status->type ];

	hw_status->samples = hw_status->fragsize/(hw_status->bps * hw_status->channels);
	/* Should make sure that hw_status->samples/hw_status->samplerate >.02 s for most unices
	 * (.02s = minimal repeated usleep time for .01 s system resolution)
	 * i.e. call setup_pcm recursively or loop. */

	api_pcm_set_mode (hw, hw_status, hw_status->modus);

	DEB1(DPRT("  pcm settings: bps=%d, samples=%d [ samplerate=%d, channels=%d, type=%d, log2_frag=%d, frag=%d ]\n", hw_status->bps, hw_status->samples, hw_status->samplerate, hw_status->channels, hw_status->type, hw_status->log2_fragsize, hw_status->fragsize));

#if 0
/* DEBUG: */
get_device_info();
get_audio_formats();
#endif
	DEB1(DPRT("#setup_pcm#\n"));

	return 0;
}


/* For testing/information only: */
void get_device_info(GSIPCM_HW) {
	int  status;
	audio_device_t  audio_device;
	int  duplex;

	status = ioctl(hw_status->fd, AUDIO_GETDEV, &audio_device);
	DEB1(DPRT("PCM OBSD audio device\n"));
	DEB1(DPRT("PCM OBSD name = %s\n", audio_device.name));
	DEB1(DPRT("PCM OBSD version = %s\n", audio_device.version));
	DEB1(DPRT("PCM OBSD config = %s\n", audio_device.config));

	status = ioctl(hw_status->fd, AUDIO_GETFD, &duplex);
	DEB1(DPRT("PCM OBSD audio device is %s\n", duplex? "full duplex" : "half-duplex"));
}

void get_audio_formats(GSIPCM_HW) {
	int  status;
	audio_encoding_t  audio_encoding;

	audio_encoding.index = 0;

	for (;;) {
		status = ioctl(hw_status->fd, AUDIO_GETENC, &audio_encoding);
		if (status < 0) break;
		DEB1(DPRT("PCM OBSD audio encoding %d = %s\n", audio_encoding.index, audio_encoding.name));
		audio_encoding.index++;
	}
}


int api_pcm_set_mode (GSIPCM_HW, int modus) {
	int status = -1;
	audio_info_t audio_settings;

	if (hw_status->fd > 0) {
		switch(modus) {
		case GSIPCM_MODE_NON_BLOCKING:
			status = fcntl(hw_status->fd, F_GETFL);
			if (status != -1) {
				status = fcntl(hw_status->fd, F_SETFL, status | O_NONBLOCK);
				AUDIO_INITINFO(&audio_settings);
				audio_settings.hiwat = 16384 / hw_status->fragsize;
				status = ioctl(hw_status->fd, AUDIO_SETINFO, &audio_settings);
				if (status < 0) {
					log_message(GSI_MESSAGE_ERROR, "%s: (OBSD) AUDIO_SETINFO: %s", hw->device, strerror(errno));
				}
			}
			break;
		case GSIPCM_MODE_BLOCKING:
			status = fcntl(hw_status->fd, F_GETFL);
			if (status != -1) {
				status = fcntl(hw_status->fd, F_SETFL, status & ~O_NONBLOCK);
				AUDIO_INITINFO(&audio_settings);
				audio_settings.hiwat = 2;
				status = ioctl(hw_status->fd, AUDIO_SETINFO, &audio_settings);
				if (status < 0) {
					log_message(GSI_MESSAGE_ERROR, "%s: (OBSD) AUDIO_SETINFO: %s", hw->device, strerror(errno));
				}
			}
			break;
		case GSIPCM_MODE_DMA:
			break;
		}
	}

	return status;
}


int api_pcm_write_space (GSIPCM_HW) {
	int  status;
	int  audio_space;
//	audio_info_t   audio_settings;
	audio_offset_t audio_offset;

//	status = ioctl(hw_status->fd, AUDIO_GETINFO, &audio_settings);
	status = ioctl(hw_status->fd, AUDIO_GETOOFFS, &audio_offset);
	DEB(DPRT("obsd pcm write space: samples=%d, deltablks=%d, offset=%d (max write=%d)\n", audio_offset.samples, audio_offset.deltablks, audio_offset.offset, hw_status->max_write_space));
	audio_space = hw_status->max_write_space - audio_offset.offset;

	DEB(DPRT("obsd pcm write space: audio_space = %d (ioctl st=%d):: OUTSPACE=%d\n", audio_space, status, outspace));
	if (status < 0) audio_space = -1;

	if (audio_offset.deltablks < 1000)
		outspace += audio_offset.deltablks * hw_status->fragsize;

//	return audio_space;
	return outspace;
}


int api_pcm_buffer_status (GSIPCM_HW, int *writespace, int *readdata) {
	int  status;
	audio_offset_t audio_offset;

	if (writespace) {
		status = ioctl(hw_status->fd, AUDIO_GETOOFFS, &audio_offset);
//		*writespace = hw_status->max_write_space - audio_offset.offset;
outspace += audio_offset.deltablks * hw_status->fragsize;
*writespace = outspace;

		DEB(DPRT("obsd pcm write space: write space = %d (ioctl st=%d):: OUTSPACE=%d\n", *writespace, status, outspace));
		if (status < 0) *writespace = -1;
	}
	if (readdata) {
		status = ioctl(hw_status->fd, AUDIO_GETOOFFS, &audio_offset);
		*readdata = hw_status->max_write_space - audio_offset.offset;
		DEB(DPRT("obsd pcm read data: read data = %d (ioctl st=%d)\n", *readdata, status));
		if (status < 0) *readdata = -1;
	}

	return 0;
}


int api_pcm_write (GSIPCM_HW, void *buf, int bytes_to_write) {
	int status = write (hw_status->fd, buf, bytes_to_write);

	if (status > 0) outspace -= status;
//DEB(DPRT("api_pcm_write : outspace from %d to %d\n", outspace+status, outspace));
	return status;
}
#endif /* USE_OPENBSD_AUDIO */

/*      EOF pcm_obsd.c */
