/* GSI - general sound interface
 *
 * pcm_oss.c - OSS pcm mode setting
 *
 * Lower level API dependencies:
 *   - OSS include locations are Linux/FreeBSD specific.
 *     Locations for other OS's?
 *   - OSS PCM mode setting
 *
 *
 * Copyright (C) 1996 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#if  USE_OSS

#ifdef OSS_INCLUDES

#  include OSS_INCLUDES"/soundcard.h" /* Commercial OSS */

#else

#if defined(__OpenBSD__) || defined(__NetBSD__)
#  include <soundcard.h>
#else /* linux, FreeBSD: OSS-free */
#  include <sys/soundcard.h>
#endif

#endif /* OSS_INCLUDES */

#include <errno.h>


/*---------------------------------------------------------------------------*
 *      setup_pcm (oss):
 *        opens device (hw->device; usually /dev/dsp)
 *        sets fragsize (2 fragments of size 'fragsize', or 3 of fragsize/2).
 *        sets samplerate ('samplerate')
 *        sets mono/stereo ('channels')
 *        sets sample type; 8 bit mulaw/8 bit unsigned/16 bit ('type')
 *      returns 0 if OK, < 0 if an error occurred.
 *---------------------------------------------------------------------------*/
int api_setup_pcm (GSIPCM_HW, int samplerate, int channels, int type, int log2_fragsize) {
	int  status;
	int  ioctl_value;
	int  n_bytes, n_bytes_kernel, j, my_log2_frag, my_num_frags;

	int  snd_formats[] = { AFMT_MU_LAW, AFMT_U8, AFMT_S16_LE };

#ifdef DEBUG
	char *snd_format_strings[] = { "mulaw", "unsigned 8 bit", "signed 16 bit"};
#endif
	int  first_pcm_type; /* Modesetting, first one to try. */
	//int  need_fragsize_change = True;

	DEB1(DPRT("("__FILE__"): *setup_pcm* (oss)\n"));

	/* A few sanity checks, put this later in the code? */
	samplerate = min(samplerate, 48000);
	samplerate = max(samplerate, 4000);
	type  = min(type, GSI_16BIT);
	type  = max(type, GSI_8BIT_MULAW);
	log2_fragsize = min(log2_fragsize, 13); /* Max 8K buffer at the moment. */
	DEB1(DPRT("request samplerate=%d, ch=%d, type=%s\n", samplerate, channels, snd_format_strings[ type -1 ]));

	/* MULAW -> 16 BIT (OSS on e.g. SB16/GUS PNP doesn't handle mulaw output
	 * contrary to the documentation). */
	if (type == GSI_8BIT_MULAW) type = GSI_16BIT;

	/* N.B. For non-blocking writes & SNDCTL_DSP_GETOSPACE we should set the
	 * fragsize quite small (each frag transfer generates an interrupt, so not
	 * too small I guess) that call only gives multiples of the fragsize as
	 * available space. */

	/* No longer necessary as we always close the device first: */
	/* NB. The following is necessary for OSS as the fragsize cannot be
	 * adjusted after a read/write to the pcm has been done (I assume a read
	 * or write has been done when pcm > 0). */
	//if (gsipcm_fd > 0 && log2_fragsize != gsipcm_log2_fragsize) { close(gsipcm_fd); gsipcm_fd = -1; }

  setup_pcm_retry:
	if (hw_status->fd > 0) {
		/* Always closing the device seems to solve another OSS bug where
		 * changing settings can cause rubbish sound output: */

		//if (log2_fragsize != gsipcm_log2_fragsize+1) {
		close(hw_status->fd); hw_status->fd = -1;
		//}
		//else need_fragsize_change = False;
	}

	if (hw_status->fd <= 0) { /* Not yet opened. */
		if (hw->device == NULL) return -1;
		hw_status->fd = open(hw->device, O_WRONLY);
		if (hw_status->fd < 0) {
			if (errno == EBUSY) {
				return GSIPCM_BUSY;
			}
			else {
				return GSIPCM_NOT_PRESENT;
			}
		}
	}

#if 1
	/* For some reason, we need to set the fragsize before the channels/rate/
	 * type (this sucks, as channels/rate/type may be changed so we'd need to
	 * go through this procedure again if any of these change from requested
	 * values) otherwise we get garbage sound output).
	 */
#  define SYS_TIME 10000
	/* Set fragment no/size. */

	/* Software_fragsize:
	 */
	n_bytes = (SYS_TIME * samplerate * channels * gsipcm_bytes_per_sample[ type ] + 1000000) / 1000000;
	DEB1(DPRT("nbytes %d\n", n_bytes));
	for (j=0; n_bytes != 0; j++) {
		n_bytes >>= 1;
	}
	DEB1(DPRT("nbytes 2n-rounded up: %d (log2=%d)\n", 1 << j, j));
	n_bytes = 1 << j;

	/* Minimal kernel buffer size = 8 * SYS_TIME, approximated with 8*n_bytes
	 * (for now).
	 */
	n_bytes_kernel = 8 * n_bytes;

	/* We try to set a hardware fragsize 1/2 (1/4?) of this software fragsize.
	 */
	for (;;) {
		my_log2_frag = j - 1;
		my_num_frags = n_bytes_kernel / (1 << my_log2_frag);

		ioctl_value = (my_num_frags << 16) | my_log2_frag;
		DEB1(DPRT("pcm oss: frag setting 0x%x\n", ioctl_value));
		status = ioctl(hw_status->fd, SNDCTL_DSP_SETFRAGMENT, &ioctl_value);
		if (status < 0) {
			if (errno == EINVAL) {
				/* Out of range. Assume log2_frag is too small and wrap around:
				 */
				my_log2_frag++;
				if (my_log2_frag > 13) my_log2_frag = 4;
				continue;
			}
			log_message(GSI_MESSAGE_ERROR, "%s: (OSS) SET FRAGMENT SIZE: %s", hw->device, strerror(errno));
			return -1;
		}
		break;
	}

	/* Set the software fragsize. */
	hw_status->log2_fragsize = j;
	hw_status->fragsize = 1 << j;
#endif


	/* Set channels. */
	//if (ioctl(pcm, SNDCTL_DSP_CHANNELS, &hw_status->channels)==-1) {
	ioctl_value = channels - 1; /* 0 = mono, 1 = stereo */
	status = ioctl(hw_status->fd, SNDCTL_DSP_STEREO, &ioctl_value);
	hw_status->channels = ioctl_value + 1;

	DEB1(if (status == -1) perror("PCM: SNDCTL_DSP_STEREO"));
	DEB1(if (channels == 2 && hw_status->channels != 2) DPRT("  pcm_device doesn't support stereo, using mono.\n"));
	DEB1(DPRT("  pcm_channels set to %d\n", hw_status->channels));


	/* Set format. */
	hw_status->type = type;
	first_pcm_type = hw_status->type;
	for (;;) {
		/* Sample size: 8 bit mulaw, 8 bit unsigned or 16 bit. */
		int format = snd_formats[ hw_status->type -1 ];
		if (format != GSI_8BIT_MULAW) {
			/* OSS (only <3.8?) seems to $#!* mulaw playing even when the ioctl
			 * succeeds, this is on GUS (OSS 3.5) and on SB 8 (OSS 3.5). */
			status = ioctl(hw_status->fd, SNDCTL_DSP_SETFMT, &format);
			if (status != -1) {
				if (format == snd_formats[ hw_status->type -1 ])
					break; /* It's supported */
			}
		}
		DEB1(DPRT(" audio format %s not supported\n", snd_format_strings[ hw_status->type -1 ]));
		hw_status->type++;
		if (hw_status->type > GSI_16BIT) hw_status->type = GSI_8BIT_MULAW;
		if (hw_status->type == first_pcm_type) {
			close(hw_status->fd);
			hw_status->fd = -1;
			return GSIPCM_NOT_PRESENT;
		}
	}
	DEB1(DPRT("  pcm_format set to %s\n", snd_format_strings[ hw_status->type -1 ]));


	/* Set samplerate. */
	ioctl_value = samplerate;
	status = ioctl(hw_status->fd, SNDCTL_DSP_SPEED, &ioctl_value);
	//if (status != 0) {
	//	log_message(GSI_MESSAGE_ERROR, "%s: (OSS) SNDCTL_DSP_SPEED: %s", hw->device, strerror(errno));
	//}
	DEB1(DPRT("  pcm_samplerate set to %d\n", ioctl_value));
	hw_status->samplerate = ioctl_value;

	/* Calculate some constants used to play fragments: */
	hw_status->bps = gsipcm_bytes_per_sample[ hw_status->type ];

	hw_status->samples = hw_status->fragsize / (hw_status->bps * hw_status->channels);

	/* Setting fragsize after rate/channels/type doesn't work, so: */

	/* To make sure that hw_status->samples / hw_status->samplerate >.01 s
	 * for most unices (.02s = minimal repeated usleep time for .01 s system
	 * resolution, .05s gives a good cushion against disk access etc). */

	if ( ((double) hw_status->samplerate /(double) samplerate) > 1.05
		 || ((double) hw_status->samplerate /(double) samplerate) < 0.95
		 || hw_status->bps != gsipcm_bytes_per_sample[ type ]
		 || hw_status->channels != channels) {
		DEB1(DPRT("Retrying pcm setup\n"));
		samplerate = hw_status->samplerate;
		type = hw_status->type;
		channels = hw_status->channels;
		goto setup_pcm_retry;
	}
	DEB1(DPRT("  pcm_fragsize (software) set to %d\n", hw_status->fragsize));


	hw_status->byte_order = GSI_LITTLE_ENDIAN;
	/* XXX Should change this as OSS can set little/big on some devices (should
	 * try to get native endianness first). */

	hw_status->max_write_space = api_pcm_write_space(hw, hw_status);

	api_pcm_set_mode (hw, hw_status, hw_status->modus);

	DEB1(DPRT("  pcm settings: bps=%d, samples=%d [ samplerate=%d, channels=%d, type=%d, log2_frag=%d, frag=%d ]\n", hw_status->bps, hw_status->samples, hw_status->samplerate, hw_status->channels, hw_status->type, hw_status->log2_fragsize, hw_status->fragsize));

	DEB1(DPRT("#setup_pcm#\n"));

	return 0;
}


int api_pcm_set_mode (GSIPCM_HW, int modus) {
	int status = -1;

	switch(modus) {
	case GSIPCM_MODE_NON_BLOCKING:
		status = fcntl(hw_status->fd, F_GETFL);
		if (status != -1)
			status = fcntl(hw_status->fd, F_SETFL, status | O_NONBLOCK);
		break;
	case GSIPCM_MODE_BLOCKING:
//2args doesn't compile on OpenBSD...
//		status = ioctl(hw_status->fd, SNDCTL_DSP_NONBLOCK);
// so, just use:
		status = fcntl(hw_status->fd, F_GETFL);
		if (status != -1)
			status = fcntl(hw_status->fd, F_SETFL, status & ~O_NONBLOCK);
		break;
	case GSIPCM_MODE_DMA:
		break;
	}

	return status;
}


int api_pcm_write_space (GSIPCM_HW) {
	int status = -1;
	struct audio_buf_info audio_buf;

	if (hw_status->fd > 0)
		status = ioctl(hw_status->fd, SNDCTL_DSP_GETOSPACE, &audio_buf);
	DEB(DPRT("oss SNDCTL_DSP_GETOSPACE = %d (ioctl st=%d)\n", audio_buf.bytes, status));
	if (status < 0) audio_buf.bytes = -1;

	return audio_buf.bytes;
}


int api_pcm_buffer_status (GSIPCM_HW, int *writespace, int *readdata) {
	int  status = -1;
	struct audio_buf_info audio_buf;

	if (writespace) {
		if (hw_status->fd > 0) {
			status = ioctl(hw_status->fd, SNDCTL_DSP_GETOSPACE, &audio_buf);
			DEB(DPRT("oss SNDCTL_DSP_GETOSPACE = %d (ioctl st=%d)\n", audio_buf.bytes, status));
		}
		if (status < 0) *writespace = -1;
		else *writespace = audio_buf.bytes;
	}
	if (readdata) {
		if (hw_status->fd > 0) {
			status = ioctl(hw_status->fd, SNDCTL_DSP_GETISPACE, &audio_buf);
			DEB(DPRT("oss SNDCTL_DSP_GETISPACE = %d (ioctl st=%d)\n", audio_buf.bytes, status));
		}
		if (status < 0) *readdata = -1;
		else *readdata = audio_buf.bytes;
	}

	return 0;
}


int api_pcm_write (GSIPCM_HW, void *buf, int bytes_to_write) {
	int  status;

	// MUCH TEST STUFF IN HERE..
	//usleep(10000);
	DEB(DPRT("before pcm write: "));
	api_pcm_write_space(hw, hw_status);

	status = write (hw_status->fd, buf, bytes_to_write);
	DEB(DPRT("after pcm write: "));
	api_pcm_write_space(hw, hw_status);

	return status;
}


# if 0
// NB. THIS CODE IS NOT YET FINISHED.
int   gsipcm_dma_enabled = False;
byte *gsipcm_dma_buf = NULL;

/* returns 0 if OK, -1 if not. */
int enable_pcm_dma(char *dev) {
	int  info;

	if (ioctl(fd, SNDCTL_DSP_GETCAPS, &caps) == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: (OSS) SNDCTL_DSP_GETCAPS: %s", dev, strerror(errno));
		return -1;
	}

	if (!(caps & DSP_CAP_TRIGGER) || !(caps & DSP_CAP_MMAP)) {
		return -1;
	}

	if (ioctl(fd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: (OSS) SNDCTL_DSP_GETOSPACE: %s", dev, strerror(errno));
		return -1;
	}

	if ((gsipcm_dma_buf = mmap(NULL, sz, PROT_WRITE, MAP_FILE|MAP_SHARED, fd, 0)) == (caddr_t)-1) {
		log_message(GSI_MESSAGE_ERROR, "%s: (OSS) mmap (write): %s", dev, strerror(errno));
		return -1;
	}
	gsipcm_dma_enabled = True;

	return 0;
}


int disable_pcm_dma() {
	gsipcm_dma_enabled = False;
	return 0;
}


int write_dma(int bytes, int ptr) {
	return -1;
}


int pcm_write(int bytes) {
	if (gsipcm_dma_enabled)
		return write_dma(bytes, gsipcm_dma_ptr);
	else {
		return write (hw_status->fd, fragment, fragsize);
	}
}
#endif /* Incomplete DMA + generalized code. */

#endif /* USE_OSS */


/*      EOF pcm_oss.c */
