/* GSI - general sound interface
 *
 * pcm_functions.c - gsipcm library; pcm functions
 *
 * Lower level API dependencies: 
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
//#include <gsi/gsipcm_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* These should go into the api*.c files where needed: */
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <errno.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIPCM_"
#include <gsi/gsi_debug.h>

#include "gsipcm.h"

#include "api.h"

#include "log.h"


/*__________________DEFINES___________________*/
#if  USE_ULTRASOUND
/* Can't get the native code to work properly at the moment (after settting
 * the fragment size in pcm/pcm.c, gus_write..() in gsi_server.c hangs) so use
 * the OSS emulation. */
#  undef  USE_ULTRASOUND
#  define USE_ULTRASOUND 0
#  undef  USE_OSS
#  define USE_OSS    1
#endif


/*---------------------------------------------------------------------------*
 *	The PCM abstraction library
 *---------------------------------------------------------------------------*/
gsipcm_device_t **gsipcm_devices = NULL;
int  gsipcm_num_pcms = 0;
gsipcm_devicedata_t **gsipcm_status;


#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
static int _gsipcm_native_byte_order = GSI_LITTLE_ENDIAN;
#else
static int _gsipcm_native_byte_order = GSI_BIG_ENDIAN;
#endif

/* This table is also available in gsif (sample_format.c: _gsif_Bps[]): */
int gsipcm_bytes_per_sample [ GSI_NUM_SAMPLE_TYPES ] = {
	1, 1, 1, 2, 3, 1, 2, 3, -4, -3, -5 /* negative = bits, not bytes. */
};


int gsipcm_init(GSIPCM_HW, int modus) {
	int  status;

	hw_status->modus = modus;
	hw_status->fd = -1;

	status = gsipcm_set_pcm(hw, hw_status, 8000, 1, GSI_8BIT, 8);

	if (status == GSIPCM_OK) {
		gsipcm_set_mode(hw, hw_status, modus);
	}

	return status;
}


int gsipcm_close(GSIPCM_HW) {
DEB(DPRT("gsipcm_close: fd=%d\n", hw_status->fd));
	if (hw_status->fd > 0) close(hw_status->fd);
	hw_status->fd = -1;
DEB(DPRT("gsipcm_close done\n"));

	return 0;
}


int gsipcm_set_pcm(GSIPCM_HW, int pcm_samplerate, int pcm_channels, int pcm_type, int pcm_log2_fragsize) {
//	if (hw_status->fd == -1) return GSIPCM_NOT_OPENED;

	return api_setup_pcm(hw, hw_status, /*0,*/ pcm_samplerate, pcm_channels, pcm_type, pcm_log2_fragsize);
	// for multi device change to:
	//return pcm[ dev ].setup(dev, pcm_samplerate, pcm_channels, pcm_type, pcm_log2_fragsize);
}


int gsipcm_write (GSIPCM_HW, char *buf, int n) {
	int  status;
	int  bytes_to_write;
	int  bytes_written;

	DEB1(DPRT("write (address = 0x%x, #bytes = %d)\n", hw_status->fragment, hw_status->fragsize));

	if (hw_status->fd == -1) return GSIPCM_NOT_OPENED;

	bytes_written = 0;

	bytes_to_write = n;

  fd_write:
DEB1(DPRT("before write: pcm free = %d\n", gsipcm_write_space(hw, hw_status)));

	status = api_pcm_write (hw, hw_status, buf+bytes_written, bytes_to_write);

DEB1(DPRT("pcm write (%d bytes): status = %d\n", bytes_to_write, status));
DEB1(DPRT("after write: pcm free = %d\n", gsipcm_write_space(hw, hw_status)));
	if (status < 0) {
		if (errno == EINTR) goto fd_write;
		else return status;
	}
	else {
		bytes_written += status;
		if ((status < bytes_to_write) && (hw_status->modus == GSIPCM_MODE_BLOCKING)) {
			bytes_to_write -= status;
			goto fd_write;
		}
	}

	return bytes_written;
}


int gsipcm_write_fragment (GSIPCM_HW) {
	return gsipcm_write(hw, hw_status, hw_status->fragment, hw_status->fragsize);
}


/*---------------------------------------------------------------------------*
 *	Set PCM characteristics
 *---------------------------------------------------------------------------*/
int gsipcm_set_fragment(GSIPCM_HW, int pcm_log2_fragsize) {
	return gsipcm_set_pcm(hw, hw_status, hw_status->samplerate, hw_status->channels, hw_status->type, pcm_log2_fragsize);
}

int gsipcm_set_samplerate(GSIPCM_HW, int pcm_samplerate) {
	return gsipcm_set_pcm(hw, hw_status, pcm_samplerate, hw_status->channels, hw_status->type, hw_status->log2_fragsize);
}

int gsipcm_set_channels(GSIPCM_HW, int pcm_channels) {
	return gsipcm_set_pcm(hw, hw_status, hw_status->samplerate, pcm_channels, hw_status->type, hw_status->log2_fragsize);
}

int gsipcm_set_type(GSIPCM_HW, int pcm_type) {
	return gsipcm_set_pcm(hw, hw_status, hw_status->samplerate, hw_status->channels, pcm_type, hw_status->log2_fragsize);
}


int gsipcm_set_mode(GSIPCM_HW, int modus) {
	if (hw_status->fd == -1) return GSIPCM_NOT_OPENED;
	return api_pcm_set_mode (hw, hw_status, modus);
}


/* NEW: Finish this code */
int gsipcm_convert_int32_to_output(GSIPCM_HW, int buf_in_len, int32 *buf_in, int *buf_out_len, byte *buf_out) {
	*buf_out_len = 0;
	return -1;
}


/*---------------------------------------------------------------------------*
 *	Read device status                                                       *
 */
int gsipcm_write_space(GSIPCM_HW) {
	if (hw_status->fd == -1) return GSIPCM_NOT_OPENED;
	return api_pcm_write_space(hw, hw_status);
}

int gsipcm_buffer_status(GSIPCM_HW, int *writespace, int *readdata) {
	if (hw_status->fd == -1) return GSIPCM_NOT_OPENED;
	return api_pcm_buffer_status(hw, hw_status, writespace, readdata);
}


/*---------------------------------------------------------------------------*
 *      API code.
 *---------------------------------------------------------------------------*/
#include "api/pcm_oss.c"
#include "api/pcm_ultra.c"
#include "api/pcm_devaudio.c"
#include "api/pcm_obsd.c"
#include "api/pcm_fbsd.c"


/*      EOF pcm_functions.c */
