/* GSI - general sound interface
 *
 * connect.c: connection logic etc. (to connect to the GSI server).
 *
 * ** See gsi.h for a description of the functions **
 *
 *
 * Copyright (C) W.H.Scholten 1996 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <errno.h>

#  include <time.h>
#  include <sys/types.h>
#  include <sys/socket.h>
#  include <netinet/in.h>
#  include <netdb.h>
#  ifndef AF_LOCAL
#    define AF_LOCAL AF_UNIX
#  endif

//#include <gsi/p_signal.h>
#include <gsi/p_signal.c> /* temporary */

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_"
#include <gsi/gsi_debug.h>

#include "gsi.h"

#include "log.h"


/*__________________VARIABELEN___________________*/
char gsi_thishost_name[ 64 ];
#warning "Need a lock (threads) for gsi_thishost_name[] (initialize once)"

char *gsi_version = GSI_INTERFACE_VERSION; 
char *gsi_device_type_name [ GSIDEV_NUM_DEVICE_TYPES ] = { "pcm", "synth", "cd", "line in", "microphone" };

void _gsi_close_now(gsi_server_t *c);


/*__________________DEFINITIES___________________*/
#define GSI_SERVER "gsi_server"


/*__________________PROTOTYPES___________________*/
static int  init_server (gsi_server_t *c, char *remote_host);
static int  init_connection (gsi_server_t *c, char *remote_host, int af_connection);

static void start_server (gsi_server_t *c);


/*__________________PROGRAMMA___________________*/
/*
 * This code is only useful for the execvp'ed server (AND NOT THREAD SAFE WITHOUT A LOCK)
 */
//static void (*main_sigchld_handler)(int); /* handler before we installed our own */

//struct connection_item {
//	struct gsi_connection *this;
//	struct gsi_connection *next;
//	struct gsi_connection *prev;
//};

static void child_killed(int i) {
//	int status;

#if 0 /* disable until it's fixed correctly */
//YYY	// list of connections: (or not needed as we can have only one server runnning locally?)
DEB(DPRT("Child killed, server_id =%d\n", server_id));
	if (server_id > 0 && waitpid(server_id, &status, WNOHANG) == server_id) {
		c->server_ok = False;
		close(c->send_fd);
		if (c->receive_fd != c->send_fd)
			close(c->receive_fd);
		c->send_fd = c->receive_fd = -1;

		log_message(GSI_MESSAGE_ERROR, "gsi_server killed");

		//For the socket server:
//XXX		c->server_execed = False; // useful to try starting it again
#endif
}


/* Initialize error handlers etc. */
static void init_handlers(void) {
	void *old_handler;

	old_handler = p_signal(SIGPIPE, SIG_IGN);
	if (old_handler != SIG_DFL) {
		/* already set to something (by the program using this library), so
		 * change back. */
		p_signal(SIGPIPE, old_handler);
	}

	old_handler = p_signal_restart(SIGCHLD, &child_killed, False);
	if (old_handler != &child_killed) {
//NOT THREAD SAFE WITHOUT A LOCK:		main_sigchld_handler = old_handler;
		/* already set to something (by the program using this library), so
		 * change back. */
		p_signal(SIGCHLD, old_handler);
	}
}


/*---------------------------------------------------------------------------*
 *      GSI initialization (make a new connection to a server).
 *
 *      Here read the hardware master volume, pcm volume, synth volume cd
 *      volume. Or do this with a different piece of code above? (init
 *      volumes at -1).
 *
 *      Returns a gsi_server_t struct if ok, NULL if unable to
 *      connect to the server.
 *---------------------------------------------------------------------------*/
gsi_server_t *gsi_init(char *remote_host) {
	gsi_server_t c;
	gsi_server_t *new_connection;

	/* initialize connection struct: */
	memset(&c, 0, sizeof(gsi_server_t));
	c.server_id = -1;

	init_handlers(); /* sigpipe */

	if (gethostname(gsi_thishost_name, 63)) {
		log_message(GSI_MESSAGE_ERROR, "gethostname overflow");
		return NULL;
	}

	if (remote_host == NULL || remote_host[0] == 0)
		remote_host = gsi_thishost_name;
	if (getenv("GSI_HOST") != NULL)
		remote_host = getenv("GSI_HOST");
	c.remote_host = strdup(remote_host);
	if (c.remote_host == NULL) {
		log_message(GSI_MESSAGE_ERROR, "Out of memory");
		return NULL;
	}

	DEB(DPRT("gsi_init remote host = %s\n", c.remote_host ));

	/* Here connect to the gsi_server (fork/rpc if connection fails?) */
	init_server(&c, remote_host);
	/*
	connect_status = init_server(c, remote_host, AF_INET);
	if (connect_status == -1) {
		connect_status = init_server(c, remote_host, AF_LOCAL);
	}
	*/

	/* after connection, the server will (should) automatically send the
	 * current hardware volumes so we don't need to do
	 *   gsi_send_commands(2, GSI_CMD_GET_MIXER_VOLUME, GSI_MIXER_SYNTH)
	 * etc, here.
	 */

	if (!c.server_ok)
		return NULL;

	c.num_commands = 0;
	DEB(DPRT("gsi_init: now sync\n"));

	gsi_sync(&c); /* we wait for the server version, mixer status etc. that is sent after initializing a new client (add GSI_EV_INIT_DONE ?) */
	DEB(DPRT("gsi_init: sync done\n"));

	/* initialize some stuff: */
	c.gsi_on = True;
	c.gsi_share_permissions = 0x600;

	new_connection = malloc(sizeof(gsi_server_t));
	if (new_connection == NULL) {
		log_message(GSI_MESSAGE_ERROR, "Out of memory");
		return NULL;
	}
	memcpy(new_connection, &c, sizeof(gsi_server_t));

	return new_connection;
}


/*---------------------------------------------------------------------------*
 *      Miscellaneous stuff.
 *---------------------------------------------------------------------------*/
void gsi_close (gsi_server_t *c) {
	int  i;

	if (c->server_ok) {
		gsi_release_pcm(c);

		for (i=0; i<c->_gsi_num_devices[ GSI_SYNTH ]; i++)
			gsi_release_synth(c, i);

		for (i=0; i<c->_gsi_num_devices[ GSI_CD ]; i++)
			gsi_release_cd(c, i);
		gsi_flush_all(c);
		/* should unload all sounds too. */

		//gsi_send_byte(c, GSI_CMD_CLOSE_SERVER);
		/* Should do a gsi_sync(c) here to make sure all above is sent. */
	}

	_gsi_close_now(c);
}

void _gsi_close_now(gsi_server_t *c) {
	int  i;

	// not needed when we do 'close(c->send_fd)':
	// send EOF:
	//if (shutdown(c->send_fd, /*1*/ 2) == -1) {
	//	log_message(GSI_MESSSAGE_ERROR, "%s: unable to shutdown socket", strerror(errno));
	//}

	close(c->send_fd);
	c->send_fd = -1;
	c->receive_fd = -1;

	if (c->gsi_cd_track_list != NULL) {
		free(c->gsi_cd_track_list);
		c->gsi_cd_track_list = NULL;
	}
	c->gsi_cd_track_no = 0;

	for (i=0; i< GSIDEV_NUM_DEVICE_TYPES; i++) {
		if (c->_gsi_mix_volume[i])
			free(c->_gsi_mix_volume[i]);
	}

	c->server_ok = False; /* aha, we can call gsi_init again! */
}


/* Low level --------------------------------------------------------------- */
/* put startup code from init_connection in 2 functions? */
#if 0
static int init_unix(void)
{
 	return 0;
}

static int init_inet(struct servent *sp)
{
	return 0;
}
#endif


/*
 * inet connection status: (so we can move the fprintf(..) stuff from
 * init_connection() into init_server())
 */
#define GSI_ERROR_SOCKET_NOT_CREATED     -1
#define GSI_ERROR_CONNECTION_FAILED      -2
#define GSI_ERROR_INET_NO_HOST_ENTRY     -3
#define GSI_ERROR_INET_NO_SERVICES_ENTRY -4

static int init_server(gsi_server_t *c, char *remote_host) {
	int  af_inet_status, af_local_status;

//ONLY FOR SINGLE THREAD:
//	if (c->server_ok) return True; /* already initialized */

	if (!(c->gsi_server = getenv("GSI_SERVER")))
		c->gsi_server = GSI_SERVER;
DEB(DPRT("Xinit_server: af_inet now\n"));

	af_inet_status = init_connection(c, remote_host, AF_INET);
DEB(DPRT("Xinit_server: af_inet done\n"));

	if (!c->server_ok) {
		DEB1(DPRT("%s: trying AF_LOCAL\n", gsi_version));
		af_local_status = init_connection(c, remote_host, AF_LOCAL);

		/* In case of AF_LOCAL we exec the server and try to connnect again.*/
		if (!c->server_ok && !c->server_execed && af_local_status == 1) {
			start_server(c);
			DEB1(DPRT("%s: sleep 1 s to let the server initialize\n", gsi_version));
			usleep(1000000L); /* wait 1 s for the server to initialize */

			af_local_status = init_connection(c, remote_host, AF_LOCAL);
		}

		/* The test if GSI_LOCAL_SOCKET_NAME exists could be placed
		 * earlier, and if it does, there's no need to exec the server, and
		 * if connection failed with enough access permissions to use the
		 * socket, it is probably dead and we could remove it. */
		if (af_local_status == 1) {
			int  status;
			struct stat my_stat;

			status = stat (GSI_LOCAL_SOCKET_NAME, &my_stat);
			if (status == -1 /*&& errno == ENOENT*/) {
				/* doesn't exist, we need to exec the server
				 * currently this never happens as this code is after the exec.
				 */
			}
			else {
				log_message(0, "Local connection failed:");
				if (S_ISSOCK(my_stat.st_mode)) {
					int  test_fd;

					/* Do we have the right access perms ? */
					if ((test_fd = open(GSI_LOCAL_SOCKET_NAME, O_RDWR)) < 0) {
						log_message(0, "You don't seem to have access permissions to "GSI_LOCAL_SOCKET_NAME);
					}
					else {
						close(test_fd);
						log_message(0, "As you seem to have access permissions to "GSI_LOCAL_SOCKET_NAME", remove the (apparantly stale) socket "GSI_LOCAL_SOCKET_NAME" and try again.\n");
					}
				}
				else {
					log_message(GSI_MESSAGE_ERROR, GSI_LOCAL_SOCKET_NAME" exists, but is not a socket.\nRemove it if not used by another program and try again.\n");
				}
			}
		}

	}

	if (!c->server_ok) {
		log_message(GSI_MESSAGE_ERROR, "Failed to connect to '%s' with either AF_INET or AF_LOCAL (see previous log messages for details)\n", c->gsi_server);
	}

	return c->server_ok;
}


/*
 * returns -1 for fatal (internal) error, 0 if OK, 1 if retry could be useful
 */
static int init_connection (gsi_server_t *c, char *remote_host, int af_connection) {
	int  oldflags;

	/*
	 * Connect to the server via a port
	 * ====================================================================
	 *  We request a service called "gsi_server".
	 *  An entry for it needs to exist in the /etc/services file.
	 *  The port address for this service can be any port number that is
	 *  likely to be unused, such as 22375, for example.  The host on which
	 *  the server runs must also have the same entry (same port number) in
	 *  its /etc/services file.
	 */

	int  addrlen;
	int  curstate = 1;


	/* clear address structures. */
	memset ((char *)&c->myaddr_in, 0, sizeof(struct sockaddr_in));
	memset ((char *)&c->peeraddr_in, 0, sizeof(struct sockaddr_in));

	if (af_connection == AF_INET) {
		/* Set up the peer address to which we will connect. */
		c->peeraddr_in.sin_family = AF_INET;
		/* Get the host information for the hostname that the user passed in.*/
		/*
		 * NOTE: must replace gethostbyname() by getnameinfo() for IPv6
		 */
		c->hp = gethostbyname (remote_host);

		if (c->hp == NULL) {
			DEB1(DPRT("%s: %s not found in /etc/hosts\n",
						 gsi_version, remote_host));
			return -1;
		}

		c->peeraddr_in.sin_addr.s_addr = ((struct in_addr *)(c->hp->h_addr))->s_addr;
		/*
		 * Find the information for the gsi_server in order to get the needed
		 * port number.
		 */
		c->sp = getservbyname (c->gsi_server, "tcp");
		if (c->sp == NULL) {
			DEB1(DPRT("%s: gsi_server not found in /etc/services\n", gsi_version));
			return -1;
		}
		c->peeraddr_in.sin_port = c->sp->s_port;
		DEB1(DPRT("%s: sp->s_port=%d\n", gsi_version, c->sp->s_port));
		/* <- AF_INET */
	}
	else {
		/* -> AF_LOCAL */
		c->peeraddr_in.sin_family = AF_LOCAL;
		 ((struct sockaddr *)(&c->peeraddr_in))->sa_family = AF_LOCAL;
		 strcpy(((struct sockaddr *)(&c->peeraddr_in))->sa_data, GSI_LOCAL_SOCKET_NAME);
	}


	/* shared part for AF_INET & AF_LOCAL */
	/* Create the socket. */
	c->socket_fd = socket (af_connection, SOCK_STREAM, 0);
	if (c->socket_fd == -1) {
		log_message(0, strerror(errno));
		DEB1(DPRT("%s: unable to create socket\n", gsi_version));
		return -1;
	}
	DEB1(DPRT("%s: socket created [socket fd=%d]\n", gsi_version, c->socket_fd));

	/* close the socket when an exec* function is executed. */
	oldflags = fcntl(c->socket_fd, F_GETFD);
	if (oldflags == -1) oldflags = 0;
	if (fcntl(c->socket_fd, F_SETFD, oldflags | FD_CLOEXEC) == -1) {
		log_message(GSI_MESSAGE_ERROR, "Unable to set socket to 'close on exec'");
		close(c->socket_fd);
		return -1;
	}
	/* enable local address reuse. */
	if (setsockopt(c->socket_fd, SOL_SOCKET, SO_REUSEADDR,
				   &curstate, sizeof(curstate) ) == -1) {
		log_message(GSI_MESSAGE_ERROR, "Unable to set for a fresh socket");
		close(c->socket_fd);
		return -1;
	}

	/*
	 * Try to connect to the remote server at the address which was just built
	 * into peeraddr.
	 */
	if (connect(c->socket_fd, (struct sockaddr *) &c->peeraddr_in, sizeof(struct sockaddr_in)) == -1) {
		DEB1(DPRT("%s: unable to connect to remote [errno=%d: %s]\n", gsi_version, errno, strerror(errno)));
		close(c->socket_fd);

		if (af_connection == AF_LOCAL) {
			return 1; /* might be useful to try to reconnect */
		}
		else
			return -1;
	}

	/*
	 * The connect call assigns a random address to the local end of the
	 * connection, so we use getsockname to see what it assigned.
	 */
	addrlen = sizeof(struct sockaddr_in);
	if (getsockname(c->socket_fd, (struct sockaddr *) &c->myaddr_in, &addrlen) == -1) {
		log_message(GSI_MESSAGE_ERROR, "%s: Unable to read socket address", strerror(errno));
		close(c->socket_fd);
		if (af_connection == AF_LOCAL) {
			unlink(GSI_LOCAL_SOCKET_NAME);
		}
		return -1;
	}

	time(&c->timevar);
	DEB1(DPRT("%s: Connected to %s on port %u at %s", gsi_version,
		   remote_host, ntohs(c->myaddr_in.sin_port), ctime(&c->timevar)));

	oldflags = fcntl(c->socket_fd, F_GETFL);
	if (oldflags == -1) /* error, so what now? assume all flags are clear */
		oldflags = 0;
	fcntl(c->socket_fd, F_SETFL, oldflags | O_NONBLOCK);

	c->send_fd  = c->socket_fd;
	c->receive_fd = c->socket_fd;

	c->server_ok = True;

	return 0;
	/* ==================================================================== */
}


/* start gsi_server if we couldn't connect to it (use rpc for remote?) */
static void start_server(gsi_server_t *c) {
	int  child;
	char *argv[5];

	DEB1(DPRT("%s: starting the server..\n", gsi_version));
	c->server_execed = True;
	if (strcmp(c->remote_host, gsi_thishost_name)) {
		log_message(GSI_MESSAGE_ERROR, "Connect to remote server failed and I can't start a server on a remote host");
		return;
	}

	setpgid(0,0);
	child = fork();
	if (child == 0) { /* child's thread */
		/* close the file descriptors inherited from the parent: */
		int  i, fd, max_fd = min(65535, getdtablesize());

		/* only for linux (indicating there's no system wide maximum? hence
		 * also on other OS'es? see OpenBSD's manpage too):
		 * if (max_allowed_fd<0) { max_allowed_fd=65535; } */
		for (i=0; i< max_fd; i++) close(i);

		/* when the client starts the server, discard server stdio
		 */
		fd = open ("/dev/null", O_RDWR);
		dup2 (fd, 1);   /* STDOUT */
		dup2 (fd, 2);   /* STDERR */

		argv[0] = c->gsi_server;
		argv[1] = "-q";
		argv[2] = "-p";
		argv[3] = "AF_LOCAL";
		argv[4] = NULL;
		execvp(c->gsi_server, argv);
		/* We only arrive here if execvp failed to replace this process
		 * with the soundserver code. */
		log_message(GSI_MESSAGE_ERROR, "Couldn't execute %s!", c->gsi_server);
		exit (0); /* child quits */
	}
	c->server_id = child;
}


/*      EOF connect.c */
