/* GSI - general sound interface
 *
 * event.c: event handling (for interface to the GSI server)
 *
 * ** See protocol/commands.spec for a description of the commands **
 * ** and protocol/events.spec for the events.                     **
 *
 *
 * Copyright (C) W.H.Scholten 1996 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include <errno.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_"
#include <gsi/gsi_debug.h>

#include "gsi.h"
#include "gsi_event_lengths.h"
#ifdef DEBUG
#  include "gsi_event_names.h"
#endif

#include "io.h"
#include "log.h"


/*__________________VARIABELEN___________________*/


/*__________________DEFINITIES___________________*/


/*__________________PROTOTYPES___________________*/
int  _gsi_get_channel_status (gsi_server_t *c, int ch);
void _gsi_set_channel_status (gsi_server_t *c, int ch, int bit);
int  gsi_time (void);

extern void _gsi_close_now(gsi_server_t *c);


/*---------------------------------------------------------------------------*
 *      Events & callbacks.
 *---------------------------------------------------------------------------*/
void gsi_enable_events(gsi_server_t *c, int bits) {
	gsi_send_commands(c, 5, GSI_CMD_ENABLE_EVENTS, bits >>24, bits >>16, bits >>8, bits);
}
void gsi_disable_events(gsi_server_t *c, int bits) {
	gsi_send_commands(c, 5, GSI_CMD_DISABLE_EVENTS, bits >>24, bits >>16, bits >>8, bits);
}


static void copy_event_data(gsi_server_t *c, int n) {
	int  i;

	for (i=0; i<n; i++)
		c->gsi_event_data[i] = c->status_queue[i];
}


/*---------------------------------------------------------------------------*
 *      Read the next events, returns 0 if an event is available, -1 if not.
 *---------------------------------------------------------------------------*/
enum { _ERROR_EVENT, _INFO_EVENT, _CHANNEL_EVENT, _SONG_EVENT, _CD_EVENT, _NO_ACTION };

//NEW: int gsi_next_event (gsi_server_t *c, gsi_event_t * ev) {
// pass ev=NULL to disable copying, i.e. just process internally only.
int gsi_next_event (gsi_server_t *c) {
DEB(DPRT("-- gsi_next_event --\n"));
	gsi_events_pending(c);
	if (c->_gsi_num_pending_events > 0) {
		/* Callbacks: */
		if (c->_gsi_event_copy_decoded_data)
			copy_event_data(c, c->_gsi_event_decoded_length);

		switch(c->_gsi_event_type) {
		case _ERROR_EVENT:
			if (c->error_callback_function != NULL)
				c->error_callback_function(c->status_queue[0]);
			break;
		case _INFO_EVENT:
			if (c->info_callback_function != NULL)
				c->info_callback_function(c->status_queue[0]);
			break;
		case _CHANNEL_EVENT:
		{
			int  j = c->status_queue[1];
			if (c->cb_function[j+2] != NULL) (c->cb_function[j+2])(j);
			break;
		}
		case _SONG_EVENT:
			if (c->cb_function[1] != NULL) (c->cb_function[1])(-1);
			break;
		case _CD_EVENT:
			if (c->cb_function[0] != NULL) (c->cb_function[0])(-2);
			break;
		default:
			break;
		}
		c->status_queue_len -= c->_gsi_event_decoded_length;
		if (c->status_queue_len > 0)
			memmove(c->status_queue, c->status_queue + c->_gsi_event_decoded_length, c->status_queue_len);
		c->_gsi_num_pending_events--;
DEB(DPRT("** gsi_next_event: got one **\n"));

		return 0;
	}

DEB(DPRT("** gsi_next_event: NONE **\n"));
	return -1;
}


/*---------------------------------------------------------------------------*
 *      Returns the number of (known to be) internally buffered commands.
 *---------------------------------------------------------------------------*/
int  gsi_commands_pending(gsi_server_t *c) {
	if (c->num_commands > 0) return 1;

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Returns the number of (known to be) available events.
 *---------------------------------------------------------------------------*/
int gsi_events_pending(gsi_server_t *c) {
	int  i;
	int  decoded_length = 0;
	int  event_type;
	byte b;
	byte *status_queue = &c->status_queue[0];

	DEB1(DPRT("--gsi_events_pending--\n"));

	if (c->_gsi_num_pending_events > 0) return c->_gsi_num_pending_events;

	c->_gsi_num_pending_events = 0;

	DEB2(DPRT("  checking for server events\n"));
	if (!c->server_ok)
		return -1;

	for (;;) {
		i = read(c->receive_fd, &b, 1);
		DEB2(DPRT("     -> read status=%d\n", i));
		if (i == 0) {
			/* EOF on pipe, i.e. gsi_server died. */
			close(c->socket_fd);
			c->socket_fd = -1;
			c->server_ok = False;
			c->status_queue_len = 0;
			i = 1;
			b = GSI_EV_SERVER_KILLED;
		}
		DEB2(if (i < 0) DPRT("  read errno=%d (%s)\n", errno, strerror(errno)));
		if (i < 0) {
			if (errno == EPIPE)
				c->server_ok = False;
			return 0 /*-1*/; /* no complete event available yet */
			//break;
		}
		if (i > 0) {
			/* there was something: */
			int  copy_decoded_data = True; /* copy all decoded data to gsi_event_data */
			c->status_queue[ c->status_queue_len++ ] = b;
			//int command = b; //=> if (command_size == 0) command = b;
			DEB2(DPRT("  server event byte received %d (for event %s (%d))\n", (int) b, gsi_event_name[c->status_queue[0]], c->status_queue[0]));

			if (c->status_queue_len < abs(gsi_event_length[ c->status_queue[0] ]))
				return 0 /*-1*/; /* no complete event available yet */

			if (gsi_event_length[ c->status_queue[0] ] == 0) {
				/* bad event */
				log_message(GSI_MESSAGE_ERROR, "unknown event %d", c->status_queue[0]);
				c->status_queue_len--;
				continue;
			}
			DEB1(DPRT(" event %s\n", gsi_event_name[ c->status_queue[0] ]));
			if (c->status_queue[0] < 64) {
				event_type = _ERROR_EVENT;
			}
			else if (c->status_queue[0] < 192) {
				event_type = _INFO_EVENT;
			}
			else {
				event_type = _NO_ACTION;
				/* play status => must split in pcm/cd/synth etc. */
			}

			decoded_length = abs(gsi_event_length[ c->status_queue[0] ]);
			/* only needs changing for variable length events, i.e. for those
			 * with gsi_event_length < 0. */
			DEB(DPRT("Decoding event %s\n", gsi_event_name[c->status_queue[0]]));
			DEB(DPRT(" decoded_length = %d\n", decoded_length));

			switch(c->status_queue[0]) {
				/* Errors: */
//			case GSI_EV_UNKNOWN_SAMPLE_FILEFORMAT:
//			case GSI_EV_NO_SUCH_SAMPLE_FILE:
//				DEB1(DPRT(" slot %d\n", c->status_queue[1]));
//				break;
			case GSI_EV_NO_SUCH_FILE:
				DEB1(DPRT(" slot %d\n", c->status_queue[1]));
				break;
			case GSI_EV_NO_ROOM:
				break;

//			case GSI_EV_UNKNOWN_SONG_FILEFORMAT:
//				break;

				/* Warnings: */
			case GSI_EV_PCM_BUSY:
			case GSI_EV_SYNTH_BUSY:
			case GSI_EV_CD_BUSY:
				/* change from error to info (why?) */
				//event_type = _INFO_EVENT;
				break;

			case GSI_EV_SERVER_PROTOCOL_VERSION:
			{
				int serv_version, serv_compat_version;

				event_type = _NO_ACTION;
				serv_version = ((status_queue[1]<<24) | (status_queue[2]<<16) | (status_queue[3]<<8) | (status_queue[4]));
				serv_compat_version = ((status_queue[5]<<24) | (status_queue[6]<<16) | (status_queue[7]<<8) | (status_queue[8]));

				if (serv_version < GSI_PROTOCOL_COMPATIBLE_VERSION_BCD) {
					log_message(0, "Server/client protocol version mismatch: client uses newer incompatible protocol");
					log_message(0, "Disabling GSI interface\n");
					_gsi_close_now(c);
				}
				if (GSI_PROTOCOL_VERSION_BCD < serv_compat_version) {
					log_message(0, "server/client protocol version mismatch: client uses older incompatible protocol");
					log_message(0, "Disabling GSI interface\n");
					_gsi_close_now(c);
				}
				c->server_version = serv_version;
				c->server_compat_version = serv_compat_version;
				break;
			}

			case GSI_EV_CHANNEL_END:
			{
				int  j;

				DEB1(DPRT(" -> ch=%d\n", status_queue[1]));
				j = status_queue[1];
				c->channel_status[j>>5] &= ~(1<<(j &31));
				event_type = _CHANNEL_EVENT;
				break;
			}
/*
 * Why   is  this  needed?   Suppose a   sound    is killed  or   ends, a
 * play_channel_sound is given  and gsi_set_channel_status(ch,  1) given,
 * then a kill EV can come  after this and  channel status will be 0 even
 * though  the  sound has just started.    This  means we   need either a
 * counter   (increase with gsi_adjust_channel_status(ch) which increases
 * it by 1, EV decreases it by 1, or  callbacks (which will arrive in the
 * same order  as things change in  the server) or ???
 * GSI_CMD_KILL_ALL  must give a callback to reset all counters to 0.
 */
			case GSI_EV_CHANNEL_STARTED:
			{
				int  j;

				DEB1(DPRT(" -> (ch=%d)\n", status_queue[1]));
				j = status_queue[1];
				c->channel_status[j>>5] |= (1<<(j&31));
				gsi_unlock_channel(c, j);
				event_type = _CHANNEL_EVENT;
				break;
			}
			case GSI_EV_CHANNEL_ENDALL:
			{
				int  j;

				copy_event_data(c, 1);
				for (j=0; j<256; j++) {
					if (_gsi_get_channel_status(c, j) && c->cb_function[j+2] != NULL)
						(c->cb_function[j+2])(j);
					_gsi_set_channel_status(c, j, 0);
				}
				//for (j=0; j<8; j++) c->channel_status[j] = 0;
			}
			break;
			case GSI_EV_STREAM_UPDATE:
			{
				DEB1(DPRT(" ->slot=%d\n", status_queue[1]));
				event_type = _INFO_EVENT;
				break;
			}
			case GSI_EV_STREAM_UNDERFLOW:
			{
				DEB1(DPRT(" -> slot=%d\n", status_queue[1]));
				event_type = _INFO_EVENT;
				break;
			}

			case GSI_EV_CD_TRACK_STARTED:
			case GSI_EV_CD_TRACK_END:
				event_type = _CD_EVENT;
				break;

			case GSI_EV_CD_TRACK_LIST:
			{
				if (c->status_queue_len < 3) continue;

				c->gsi_cd_track_no = status_queue[2];
				DEB1(DPRT(" -> cd %d: tracks = %d, queue_len=%d\n", status_queue[1], c->gsi_cd_track_no, c->status_queue_len));
				if (c->status_queue_len < (c->gsi_cd_track_no+2)*4+3) continue;
				decoded_length = (c->gsi_cd_track_no+2)*4+3;
				copy_decoded_data = False;
				copy_event_data(c, 3);

				c->gsi_cd_track_list = realloc(c->gsi_cd_track_list, 4*(c->gsi_cd_track_no+3));
				if (c->gsi_cd_track_list == NULL) {
					log_message(0, "Out of memory");
					break;
				}

				for (i=0; i<=c->gsi_cd_track_no+1; i++) {
					c->gsi_cd_track_list[i].min = status_queue[3+i*4];
					c->gsi_cd_track_list[i].sec = status_queue[3+i*4+1];
					c->gsi_cd_track_list[i].frame = status_queue[3+i*4+2];
					c->gsi_cd_track_list[i].type = status_queue[3+i*4+3];
				}
				event_type = _CD_EVENT;
				break;
			}
			case GSI_EV_CD_STATUS:
				DEB1(
					DPRT(" ->cd %d, status=%d\n", status_queue[1], status_queue[2]);
					DPRT("   track=%d, msf=%d,%d,%d\n", status_queue[3], status_queue[4], status_queue[5], status_queue[6]);
					);
				event_type = _CD_EVENT;
				break;

			case GSI_EV_MIXER_STATUS:
			{
				int  devtype, devno;

				devtype = status_queue[1];
				devno = status_queue[2];
				DEB1(DPRT("mix status dev [%d, %d]\n", devtype, devno));
				if (c->_gsi_num_devices [ devtype ] < devno+1) {
					int  *tmp;

					c->_gsi_num_devices [ devtype ] = devno+1;
					tmp = realloc(c->_gsi_mix_volume [ devtype ], sizeof(int) * (devno+1));
					if (tmp == NULL) {
						log_message(0, "Out of memory");
						break;
					}
					c->_gsi_mix_volume [ devtype ] = tmp;
				}
				DEB1(DPRT("mix vol: %d\n", status_queue[3]));
				c->_gsi_mix_volume [ devtype ] [ devno ] = (status_queue[3]>0 ? status_queue[3]+1 : 0);
				DEB1(DPRT("now _gsi_num_devices[%d] = %d\n", devtype, c->_gsi_num_devices [ devtype ]));
				break;
			}

			case GSI_EV_DEVICE_ADD:
			{
				int  i;
				int  *tmp;
				int  flags;

				i = status_queue[1];
				flags = status_queue[2];

				tmp = realloc( c->_gsi_mix_volume [i], (c->_gsi_num_devices[i] + 1) * sizeof(int) );
				if (tmp == NULL) {
					log_message(0, "Out of memory");
				}
				else {
					c->_gsi_mix_volume [i] = tmp;
					c->_gsi_mix_volume [i][ c->_gsi_num_devices[i] ] = 0;
					c->_gsi_num_devices [i]++;
					/* don't use flags (virtual etc) at the moment */
				}
			}

			case GSI_EV_SHMID:
			{
				int  slot, shmid;

				slot = status_queue[1];
				shmid = (status_queue[2]<<24) | (status_queue[3]<<16) | (status_queue[4]<<8) | (status_queue[5]);
				break;
			}

			case GSI_EV_SHARE_PERMISSIONS:
				c->gsi_share_permissions = (status_queue[1]<<24) | (status_queue[2]<<16) | (status_queue[3]<<8) | (status_queue[4]);
				break;

			case GSI_EV_SYNC:
				c->sync_flag = True;
				break;

				//case GSI_EV_PCM_SHARED:
			case GSI_EV_PCM_GRABBED:
			case GSI_EV_PCM_RELEASED:
				//case GSI_EV_SYNTH_SHARED:
			case GSI_EV_SYNTH_GRABBED:
			case GSI_EV_SYNTH_RELEASED:
			case GSI_EV_CD_SHARED:
			case GSI_EV_CD_GRABBED:
			case GSI_EV_CD_RELEASED:
			case GSI_EV_SERVER_KILLED:
				break;

			case GSI_EV_SLOT_LOADED:
				/* set slot_loaded to true as system used with channels */
				break;

			case GSI_EV_NO_PCM:
			case GSI_EV_NO_SYNTH:
				break;

			case GSI_EV_CHANNEL_NO_CURRENT:
				log_message(0, "GSI_EV_CHANNEL_NO_CURRENT for cmd...");
				exit(1);
				break;

			default:
				//DEB(DPRT("Bad event %d\n", status_queue[0]));
				log_message(0, "Bad event %d\n", status_queue[0]);
				DEB(DPRT("Fix this in libgsi!\nexiting..\n"));
				exit(1);
				break;
			}


			if (decoded_length > 0) {
				c->_gsi_num_pending_events = 1;
				c->_gsi_event_decoded_length = decoded_length;
				c->_gsi_event_copy_decoded_data = copy_decoded_data;
				c->_gsi_event_type = event_type;
				return 1;
			}
		}
	}
	DEB1(DPRT("++gsi_events_pending++\n"));

	return 0; /* no complete event available yet */
}


/*---------------------------------------------------------------------------*
 *      Callbacks.
 *---------------------------------------------------------------------------*/
void *gsi_set_channel_callback (gsi_server_t *c, int ch, void (*cbf)(int ch)) {
	void *old_func = c->cb_function[ch+2];

	NANA(I(ch >= -2 && ch < 256));
	c->cb_function[ch+2] = cbf;

	return old_func;
}
void gsi_clear_channel_callback (gsi_server_t *c, int ch) {
	NANA(I(ch >= -2 && ch < 256));
	c->cb_function[ch+2] = NULL;
}


void *gsi_set_error_callback (gsi_server_t *c, void (*cbf)(int error)) {
	void *old_func = c->error_callback_function;
	c->error_callback_function = cbf;

	return old_func;
}
void gsi_clear_error_callback (gsi_server_t *c) {
	c->error_callback_function = NULL;
}


void *gsi_set_info_callback (gsi_server_t *c, void (*cbf)(int error)) {
	void *old_func = c->info_callback_function;
	c->info_callback_function = cbf;

	return old_func;
}
void gsi_clear_info_callback (gsi_server_t *c) {
	c->info_callback_function = NULL;
}


int gsi_event_fd (gsi_server_t *c) {
	return c->receive_fd;
}


/*---------------------------------------------------------------------------*
 *      Give the current time in seconds *10.
 *---------------------------------------------------------------------------*/
int gsi_time(void) {
	struct timeval tv;
	int  current_time;

	gettimeofday(&tv, NULL);
	current_time = tv.tv_sec *10 /*+tv.tv_usec /100000*/ ;

	return current_time;
}


/*      EOF event.c */
