/* GSI - general sound interface
 *
 * functions.c: functions used to control the server
 *
 * ** See gsi.h for a description of the functions **
 *
 *
 * Copyright (C) W.H.Scholten 1996 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>

#if HAVE_SYSV_SHARED_MEMORY
#  include <sys/types.h>
#  include <sys/ipc.h>
#  include <sys/shm.h>
#endif

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_"
#include <gsi/gsi_debug.h>

#include "gsi.h"

#include "io.h"
#include "log.h"


/*__________________VARIABELEN___________________*/
extern char gsi_thishost_name[ 64 ];


/*__________________DEFINITIES___________________*/


/*__________________PROTOTYPES___________________*/
int  _gsi_get_channel_status (gsi_server_t *c, int ch);
void _gsi_set_channel_status (gsi_server_t *c, int ch, int bit);
int  gsi_time (void);

void _gsi_close_now(gsi_server_t *c);


/*---------------------------------------------------------------------------*
 *      Set the dir where the server can load files.
 *---------------------------------------------------------------------------*/
void gsi_chdir (gsi_server_t *c, char *dir)
{
	gsi_send_byte(c, GSI_CMD_SET_WORKING_DIRECTORY);
	gsi_send_int32(c, strlen(dir));
	gsi_send_string(c, dir);
	DEB1(DPRT("%s: gsi_chdir  %s\n", gsi_version, dir));
}


/*---------------------------------------------------------------------------*
 *      Init PCM system.
 *      If fragsize = 0 the server will find a suitable value for fast
 *      response 1 will give double that, 2/3 give 4/8 times the minimum.
 *      => (do the same for pcm samplerate? bytes, channels?) <=
 *      Returns 0 if OK, -1 if failed.
 *
 *      (Note that this function may be called multiple times; it's just a
 *      convenient way to set a lot of stuff at once)
 *---------------------------------------------------------------------------*/
int  gsi_init_pcm(gsi_server_t *c, int pcm_samplerate, int pcm_channels, int pcm_type, int pcm_fragsize)
{
	DEB1(DPRT("init_pcm\n"));
	if (c->server_ok) {
		/* server starts with PCM released. */
		gsi_send_byte(c, GSI_CMD_SET_PCM_SAMPLERATE);
        gsi_send_int32(c, pcm_samplerate);
		DEB1(DPRT(" set samplerate %d\n", pcm_samplerate));
		gsi_send_byte(c, GSI_CMD_SET_PCM_CHANNELS);
		gsi_send_byte(c, pcm_channels);
		DEB1(DPRT(" set chan  %d\n", pcm_channels));
		gsi_send_byte(c, GSI_CMD_SET_PCM_TYPE);
		gsi_send_byte(c, pcm_type);
		DEB1(DPRT(" set bytes %d\n", pcm_type));
		gsi_send_byte(c, GSI_CMD_SET_PCM_FRAGSIZE);
		gsi_send_byte(c, pcm_fragsize);
		DEB1(DPRT(" set frag  %d\n", pcm_fragsize));
		gsi_send_byte(c, GSI_CMD_GRAB_PCM); /* Sets the new parameters. */
		DEB1(DPRT(" GSI_CMD_GRAB_PCM\n"));
		DEB1(DPRT(" %s initialized.\n", c->gsi_server));
		DEB1(DPRT("#init_pcm#\n"));
		return 0;
	}
	else return -1;
}


/*---------------------------------------------------------------------------*
 *      Set sound volumes, balance, mute etc.
 *---------------------------------------------------------------------------*/
void gsi_set_balance(gsi_server_t *c, int balance) {
	gsi_send_commands(c, 3, GSI_CMD_SET_MIXER_BALANCE, balance>>8, balance &255);
// or maybe do it like this?:
//	gsi_send_commands(c, 3, GSI_CMD_MIXER_CONTROL, GSIMIXCMD_MIXER_BALANCE, balance);
}


/*---------------------------------------------------------------------------*
 *      Set sound volumes.
 *      Mixer volume settings:
 *        GSI_MASTER : master volume, i.e. mixed signal to line out.
 *        GSI_PCM    : pcm volume
 *        GSI_SYNTH  : onboard song playing device
 *                     (midi and other supported song formats)
 *        GSI_CD     : cd output via the soundcard
 *                     (headphone is unaffected)
 *
 *      Saving volumes is done to remove any rounding problems with repeated
 *      reading/writing (device values read may not be the same as what has
 *      been set).
 *---------------------------------------------------------------------------*/
/* ____ volumes (device multipliers) */
void gsi_set_master_volume (gsi_server_t *c, int vol) {
	c->_gsi_master_volume = vol;
	gsi_send_commands(c, 2, GSI_CMD_SET_MASTER_VOLUME, vol>2 ? vol-1: vol);
}

void gsi_set_mixer_volume (gsi_server_t *c, int mix_ch, int dev, int vol) {
	NANA(I(mix_ch > 0 && mix_ch < GSIDEV_NUM_DEVICE_TYPES));

	if (mix_ch < 0 || dev < 1 || dev >= c->_gsi_num_devices [ mix_ch ]) {
		log_message(0, "gsi_set_mixer_volume: bad ch/dev");
		return;
	}

	if (vol < 0) vol = 0; else if (vol > 256) vol = 256;
	c->_gsi_mix_volume [ mix_ch ] [ dev ] = vol;
	gsi_send_commands(c, 4, GSI_CMD_SET_MIXER_VOLUME, mix_ch, dev, vol>2 ? vol-1: vol);
}

int  gsi_get_mixer_volume (gsi_server_t *c, int mix_ch, int dev) {
	NANA(I(mix_ch > 0 && mix_ch < GSIDEV_NUM_DEVICE_TYPES));
DEB(DPRT("_gsi_mix_volume [%d] [%d]\n", mix_ch, dev));
DEB(DPRT("num_devices [%d]=%d\n", mix_ch, c->_gsi_num_devices [ mix_ch ]));

	if (mix_ch < 0 || dev < 0 || dev >= c->_gsi_num_devices [ mix_ch ]) {
		log_message(0, "gsi_get_mixer_volume: bad ch/dev");
		return -1;
	}

	return c->_gsi_mix_volume [ mix_ch ] [ dev ];
}
int  gsi_get_master_volume (gsi_server_t *c) {
//	return c->_gsi_master_volume;
	return c->_gsi_mix_volume [0] [0];
}


void gsi_set_channel_volume(gsi_server_t *c, int ch, int vol) {
	gsi_send_commands(c, 3, GSI_CMD_CHANNEL_SET_VOLUME, ch, vol);
}
void gsi_set_cd_volume(gsi_server_t *c, int cd, int vol) {
	if (cd < 0 || cd >= c->_gsi_num_devices[ GSI_CD ]) return;

	gsi_set_mixer_volume (c, GSI_CD, cd, vol);
}

/* The application's overall (master) volume: */
void gsi_set_volume(gsi_server_t *c, int volume) {
	int  i;

	for (i=0; i<256; i++) {
		gsi_set_channel_volume(c, i, volume);
	}
	for (i = 0; i< c->_gsi_num_devices [ GSI_CD ]; i++) {
		gsi_set_cd_volume(c, i, volume);
	}

	c->_gsi_default_volume = volume;
}


/*---------------------------------------------------------------------------*
 *      Play sounds 'n stuff.
 *---------------------------------------------------------------------------*/
void gsi_stop_channels(gsi_server_t *c) {
	gsi_send_byte(c, GSI_CMD_CHANNEL_STOP_ALL);
	gsi_flush(c);
}


int gsi_play_sound (gsi_server_t *c, int ch, int slot, int loop) {
	if (c->gsi_on /*&& c->server_ok*/) {
		if (ch == -1) {
			ch = gsi_find_free_channel(c);
		}
		if (ch != -1) {
			gsi_lock_channel(c, ch);

			gsi_send_commands(c, 7,
							  GSI_CMD_CHANNEL_SET_SLOT, ch, slot, 
							  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loop,
							  GSI_CMD_CHANNEL_START, ch);
		}
		ENSURE( c->num_commands<=_MAX_QUEUE_LENGTH );
	}

	return ch;
}


int gsi_play_sound_3d (gsi_server_t *c, int ch, int slot, int loop, int x, int y, int z, int dx, int dy, int dz, int doppler_on) {
	NANA(I(ch>=-1 && ch <256));

	if (c->gsi_on /*&& c->server_ok*/) {
		if (ch == -1) {
			ch = gsi_find_free_channel(c);
		}
		if (ch != -1) {
			gsi_lock_channel(c, ch);
			gsi_send_commands(c, 25,
							  GSI_CMD_CHANNEL_SET_SLOT, ch, slot,
							  GSI_CMD_CHANNEL_SET_X  | GSI_CMD_CURRENT, x >>8, x &255,
							  GSI_CMD_CHANNEL_SET_Y  | GSI_CMD_CURRENT, y >>8, y &255,
							  GSI_CMD_CHANNEL_SET_Z  | GSI_CMD_CURRENT, z >>8, z &255,
							  GSI_CMD_CHANNEL_SET_DX | GSI_CMD_CURRENT, dx >>8, dx &255,
							  GSI_CMD_CHANNEL_SET_DY | GSI_CMD_CURRENT, dy >>8, dy &255,
							  GSI_CMD_CHANNEL_SET_DZ | GSI_CMD_CURRENT, dz >>8, dz &255,
							  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loop,
							  GSI_CMD_CHANNEL_START, ch
				);
			if (doppler_on) gsi_send_commands(c, 1, GSI_CMD_CHANNEL_DOPPLER_ON | GSI_CMD_CURRENT);
		}
	}
	ENSURE( c->num_commands <= _MAX_QUEUE_LENGTH );

	return ch;
}


int gsi_play_sound_stereo (gsi_server_t *c, int ch, int slot, int loop, int x, int dx, int doppler_on) {
	NANA(I(ch >= -1 && ch < 256));

	if (c->gsi_on /*&& c->server_ok*/) {
		if (ch == -1) {
			ch = gsi_find_free_channel(c);
		}
		if (ch != -1) {
			gsi_lock_channel(c, ch);
			gsi_send_commands(c, 13, GSI_CMD_CHANNEL_SET_SLOT, ch, slot,
							  GSI_CMD_CHANNEL_SET_X  | GSI_CMD_CURRENT, x >>8, x &255,
							  GSI_CMD_CHANNEL_SET_DX | GSI_CMD_CURRENT, dx >>8, dx &255,
							  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loop,
							  GSI_CMD_CHANNEL_START, ch
				);
			if (doppler_on) gsi_send_commands(c, 1, GSI_CMD_CHANNEL_DOPPLER_ON | GSI_CMD_CURRENT);
		}
		ENSURE( c->num_commands <= _MAX_QUEUE_LENGTH );
	}

	return ch;
}


/* Implement this using gsi_play_sound_at with depth=0? */
int gsi_play_sound_pan (gsi_server_t *c, int ch, int slot, int loop, int pan) {
	NANA(I(ch >= -1 && ch < 256));
	NANA(I(pan <= 256 && pan >= -256));

	if (c->gsi_on /*&& c->server_ok*/) {
		if (ch == -1) {
			ch = gsi_find_free_channel(c);
		}
		if (ch != -1) {
			gsi_lock_channel(c, ch);
			gsi_send_commands(c, 11, GSI_CMD_CHANNEL_SET_SLOT, ch, slot,
							  GSI_CMD_CHANNEL_SET_PAN  | GSI_CMD_CURRENT, pan >>8, pan &255,
							  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loop,
							  GSI_CMD_CHANNEL_START, ch
				);
		}
		ENSURE( c->num_commands <= _MAX_QUEUE_LENGTH );
	}

	return ch;
}


int gsi_play_sound_at(gsi_server_t *c, int ch, int slot, int loop, int pan, int depth, int height) {
	NANA(I(ch >= -1 && ch < 256));
	NANA(I(pan <= 256 && pan >= -256));

	if (c->gsi_on /*&& c->server_ok*/) {
		if (ch == -1) {
			ch = gsi_find_free_channel(c);
		}
		if (ch != -1) {
			gsi_lock_channel(c, ch);
			gsi_send_commands(c, 13, GSI_CMD_CHANNEL_SET_SLOT, ch, slot,
							  GSI_CMD_CHANNEL_SET_PAN  | GSI_CMD_CURRENT, pan >>8, pan &255,
							  GSI_CMD_CHANNEL_SET_DEPTH  | GSI_CMD_CURRENT, depth >>8, depth &255,
							  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loop,
							  GSI_CMD_CHANNEL_START, ch
				);
		}
		ENSURE( c->num_commands <= _MAX_QUEUE_LENGTH );
	}

	return ch;
}


#if 0 /* unfinished, using the sound struct */
void gsi_play (gsi_server_t *c, struct gsi_sound_item *s, int effect_bits) {
	int  command = GSI_CMD_PLAY;
 
	if (c->gsi_on /*&& c->server_ok*/) {
		if (s->channel == -1) {
			s->channel = gsi_find_free_channel(c);
		}
		if (s->channel != -1) {
			gsi_lock_channel(c, s->channel);
			gsi_send_commands(c, 3, GSI_CMD_CHANNEL_SET_SLOT, ch, slot);
			if (s->loop != 1) gsi_send_commands(c, 3, GSI_CMD_CHANNEL_SET_LOOP, ch, s->loop);

			gsi_update(s, effect_bits);
			if (effect_bits & GSI_SOUNDMASK_PAN)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_PAN | GSI_CMD_CURRENT |
								  channel_flag, s->pan >>8, s->pan &255
					);
			if (effect_bits & GSI_SOUNDMASK_X)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_X | GSI_CMD_CURRENT |
								  channel_flag, s->x >>8, s->x &255
					);
			if (effect_bits & GSI_SOUNDMASK_ADJUST_X)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_ADJUST_X | GSI_CMD_CURRENT |
								  channel_flag, s->x >>8, s->x &255
					);
			if (effect_bits & GSI_SOUNDMASK_Z)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_Z | GSI_CMD_CURRENT |
								  channel_flag, s->z >>8, s->z &255
					);
			if (effect_bits & GSI_SOUNDMASK_ADJUST_Z)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_ADJUST_Z | GSI_CMD_CURRENT |
								  channel_flag, s->z >>8, s->z &255
					);
			if (effect_bits & GSI_SOUNDMASK_Y)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_Y | GSI_CMD_CURRENT |
								  channel_flag, s->y >>8, s->y &255
					);
			if (effect_bits & GSI_SOUNDMASK_ADJUST_Y)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_ADJUST_Y | GSI_CMD_CURRENT |
								  channel_flag, s->y >>8, s->y &255
					);

			if (effect_bits & GSI_SOUNDMASK_DX)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_DX | GSI_CMD_CURRENT |
								  channel_flag, s->dx >>8, s->dx &255
					);
			if (effect_bits & GSI_SOUNDMASK_DZ)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_DZ | GSI_CMD_CURRENT |
								  channel_flag, s->dz >>8, s->dz &255
					);
			if (effect_bits & GSI_SOUNDMASK_DY)
				gsi_send_commands(c, 4, slot, GSI_CMD_CHANNEL_SET_DY | GSI_CMD_CURRENT |
								  channel_flag, s->dy >>8, s->dy &255
					);
			gsi_send_commands(c, 2, GSI_CMD_CHANNEL_START, ch);
		}
		ENSURE( c->num_commands <= _MAX_QUEUE_LENGTH );
	}
}
#endif


void gsi_channel_set_slot(gsi_server_t *c, int ch, int slot) {
	gsi_send_commands(c, 3, GSI_CMD_CHANNEL_SET_SLOT, ch, slot);
}


void gsi_stop_channel(gsi_server_t *c, int ch) {
	gsi_send_commands(c, 2, GSI_CMD_CHANNEL_STOP, ch);
}


void gsi_pause_channel(gsi_server_t *c, int ch) {
	gsi_send_commands(c, 2, GSI_CMD_CHANNEL_PAUSE, ch);
}


void gsi_resume_channel(gsi_server_t *c, int ch) {
	gsi_send_commands(c, 2, GSI_CMD_CHANNEL_RESUME, ch);
}


/*---------------------------------------------------------------------------*
 *      Play music.
 *---------------------------------------------------------------------------*/
int gsi_play_song(gsi_server_t *c, int synth, int songslot, int ch, int loopno) {
	NANA(I(loopno >= -1 && loopno < 256));

	if (ch == -1) {
		ch = gsi_find_free_channel(c);
	}
	if (ch != -1) {
		//if (gsi_channel_busy(ch)) return GSI_ERROR_CHANNEL_BUSY;
		if (synth < 0 || synth >= c->_gsi_num_devices[ GSI_SYNTH ]) return -1;
		gsi_send_commands(c, 10,
						  GSI_CMD_CHANNEL_SET_SLOT, ch, songslot,
						  GSI_CMD_CHANNEL_SET_DEVICE, ch, songslot,
						  GSI_CMD_CHANNEL_SET_LOOP | GSI_CMD_CURRENT, loopno,
						  GSI_CMD_CHANNEL_START, ch);
	}

	return ch;
}


int  gsi_load_instrument (gsi_server_t *c, int pgm, char *filename, int offset) {
	return -1;
}


int  gsi_send_instrument_data (gsi_server_t *c, int pgm, int instrtype, int instrlen, unsigned char *instrdata) {
	gsi_send_commands(c, 3, GSI_CMD_LOAD_INSTRUMENT_DATA, pgm, instrtype);
	gsi_send_int32(c, instrlen);
	gsi_send_block(c, instrdata, instrlen);

	return 0;
}


/*---------------------------------------------------------------------------*
 *      CD stuff.
 *---------------------------------------------------------------------------*/
void gsi_play_cd (gsi_server_t *c, int cd, int t, int m, int s, int f, int t2, int m2, int s2, int f2) {
	if (t < 0)   t = 0; else if (t > 255)   t = 255;
	if (m < 0)   m = 0; else if (m > 255)   m = 255;
	if (s < 0)   s = 0; else if (s > 255)   s = 255;
	if (f < 0)   f = 0; else if (f > 255)   f = 255;

	if (t2 < 0) t2 = 0; else if (t2 > 255) t2 = 255;
	if (m2 < 0) m2 = 0; else if (m2 > 255) m2 = 255;
	if (s2 < 0) s2 = 0; else if (s2 > 255) s2 = 255;
	if (f2 < 0) f2 = 0; else if (f2 > 255) f2 = 255;

	gsi_send_commands(c, 11, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_PLAY, t,m,s,f, t2,m2,s2,f2);
}
void gsi_pause_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_PAUSE);
}
void gsi_resume_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_RESUME);
}
void gsi_stop_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_STOP);
}
void gsi_eject_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_EJECT);
}
void gsi_insert_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_INSERT);
}
void gsi_select_cd_slot (gsi_server_t *c, int cd, int cd_slot) {
	NANA(I(cd_slot >= 0 && cd_slot < 256));

	gsi_send_commands(c, 4, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CHANGE_CD, cd_slot);
}
void gsi_get_cd_status (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_GET_STATUS);
}
void gsi_get_cd_tracks (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 3, GSI_CMD_CD_CONTROL, cd, GSI_CMD_CD_GET_TRACKS);
}


/*---------------------------------------------------------------------------*
 *      Dynamic channel use, finding out which are free.
 *---------------------------------------------------------------------------*/
/* Two level lock: record time at which a lock is requested; if no event
 * claiming the channel has come within say 5 s then the lock is released.
 * If an event does arrive, clear the lock and set the channel_status flag. */


/*   returns -1 when all slots are used, 0-255 otherwise. */
int  gsi_find_free_slot(gsi_server_t *c) {
	int  i;
#if 0
	int  current_time = gsi_time();
#endif

	for (i=0; i<256; i++) {
		if ((c->slot_lock[ i >> 5 ] >> (i & 31)) & 1) return i;
#if 0
		/* simple and nice slot lock release code: */
		if (c->last_locked_slot == False) {
			c->lls_no = i; c->lls_time = current_time;
			c->last_locked_slot = True;
		}
		else {
			/* If too old (slot flag not set fast enough) remove the lock */
			if (current_time - c->lls_time > 50) { /* 50 =5s */
				gsi_unlock_slot(c, c->lls_no);
				/* Now place slot i in the test position at current time */
				c->lls_no = i; /* Already true: last_locked_slot=True; */
				c->lls_time = current_time;
			}
		}
#endif
	}
	return -1;
}


/*   returns -1 when all channels are playing sounds, 0-255 otherwise. */
int  gsi_find_free_channel(gsi_server_t *c) {
	int  i;
	int  current_time = gsi_time();

	for (i=0; i<256; i++) {
		if ((c->channel_lock[ i >> 5 ] >> (i & 31)) & 1) {

			/* simple and nice channel lock release: */
			if (c->last_locked_channel == False) {
				c->llc_no = i; c->llc_time = current_time;
				c->last_locked_channel = True;
			}
			else {
				/* If too old (channel flag not set fast enough) remove the lock */
				if (current_time - c->llc_time > 50) { /* 50 = 5s */
					gsi_unlock_slot(c, c->llc_no);
					/* Now place channel i in the test position at current time */
					c->llc_no = i;
					c->llc_time = current_time;
					/* Note that last_locked_channel is already 'True' */
				}
			}
		}
		else if (((c->channel_status[ i >> 5 ] >> (i & 31)) & 1) == 0) {
			return i;
		}
	}
	return -1;
}


/* locks: */
int  gsi_lock_channel(gsi_server_t *c, int ch) {
	c->channel_lock[ ch >> 5 ] |= 1 << (ch & 31);
	return 0;
}
int  gsi_unlock_channel(gsi_server_t *c, int ch) {
	c->channel_lock[ ch >> 5 ] &= ~(1 << (ch & 31));
	return 0;
}


int  gsi_lock_slot(gsi_server_t *c, int slot) {
	c->slot_lock[ slot >> 5 ] |= 1 << (slot & 31);
	return 0;
}
int  gsi_unlock_slot(gsi_server_t *c, int slot) {
	c->slot_lock[ slot >> 5 ] &= ~(1 << (slot & 31));
	return 0;
}


int  _gsi_get_channel_status (gsi_server_t *c, int ch) {
	NANA(I(ch >= -1 && ch < 256));
	return c->channel_status[ ch >> 5 ] & (1 << (ch & 31));
}
void _gsi_set_channel_status (gsi_server_t *c, int ch, int bit) {
	NANA(I(ch >= -1 && ch < 256));
	c->channel_status[ ch >> 5 ] = (c->channel_status[ ch >> 5 ] & ~(1 << (ch & 31))) | ( (bit & 1) << (ch & 31) );
}

/* Returns 0 if free, 1 if locked, 2 if playing: */
int  gsi_get_channel_status (gsi_server_t *c, int ch) {
	return ((c->channel_status[ ch >> 5 ] & (1 << (ch & 31))) << 1) | (c->channel_lock[ ch >> 5 ] & (1 << (ch & 31)));
}



/* ==== separate pause_pcm? */
/* Pause all (pcm/synth/cd) */
void gsi_pause (gsi_server_t *c) {
//	if (c->server_ok)
		gsi_send_byte(c, GSI_CMD_PAUSE);
	gsi_flush(c);
}


/* continue all (pcm/synth/cd) */
void gsi_continue (gsi_server_t *c) {
//	if (c->server_ok)
		gsi_send_byte(c, GSI_CMD_RESUME);
	gsi_flush(c);
}


void gsi_grab_mixer (gsi_server_t *c) {
	gsi_send_byte(c, GSI_CMD_GRAB_MIXER);
	gsi_flush(c);
}
void gsi_release_mixer (gsi_server_t *c) {
	gsi_send_byte(c, GSI_CMD_RELEASE_MIXER);
	gsi_flush(c);
}

// add these ?
//   gsi_grab_device (int device)
//   gsi_share_device (int device)
//   gsi_release_device (int device)
// gsi_grab_device(GSI_PCM) == gsi_grab_pcm(c)
// gsi_grab_device(GSI_MASTER) == gsi_grab_mixer(c) ??
// or leave the mixer as a meta device ?

void gsi_grab_pcm (gsi_server_t *c) {
//	if (c->server_ok)
		gsi_send_byte(c, GSI_CMD_GRAB_PCM);
	gsi_flush(c);
}


void gsi_release_pcm (gsi_server_t *c) {
//	if (c->server_ok)
		gsi_send_byte(c, GSI_CMD_RELEASE_PCM);
	gsi_flush(c);
}

void gsi_grab_synth (gsi_server_t *c, int dev) {
	gsi_send_commands(c, 2, GSI_CMD_GRAB_SYNTH, dev);
	gsi_flush(c);
}
void gsi_release_synth (gsi_server_t *c, int dev) {
	gsi_send_commands(c, 2, GSI_CMD_RELEASE_SYNTH, dev);
	gsi_flush(c);
}

void gsi_grab_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 2, GSI_CMD_GRAB_CD, cd);
	gsi_flush(c);
}
void gsi_share_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 2, GSI_CMD_SHARE_CD, cd);

	gsi_flush(c);
}
void gsi_release_cd (gsi_server_t *c, int cd) {
	gsi_send_commands(c, 2, GSI_CMD_RELEASE_CD, cd);
	gsi_flush(c);
}

/* Data & file stuff: */
//int  gsi_load_sample (gsi_server_t *c, int slot, char *filename) {
int  gsi_load_sound (gsi_server_t *c, int slot, char *filename) {
//	return gsi_load_sample_from(c, slot, filename, 0);
	return gsi_load_sound_from(c, slot, filename, 0);
}


//int  gsi_load_sample_from (gsi_server_t *c, int slot, char *filename, int offset) {
int  gsi_load_sound_from (gsi_server_t *c, int slot, char *filename, int offset) {
	if (filename == NULL) return -1;

	gsi_lock_slot(c, slot);

	gsi_send_byte(c, GSI_CMD_LOAD_SOUND_FROM_FILE);
	gsi_send_byte(c, slot);
	gsi_send_int32(c, strlen(filename));
	gsi_send_int32(c, offset);
	// block:
	gsi_send_string(c, filename);

	return 0;
}


int  gsi_load_raw_sample (gsi_server_t *c, int slot, char *filename, int offset, int length, int samplerate, int channels, int type) {
	if (filename == NULL) return -1;

	gsi_lock_slot(c, slot);

	gsi_send_byte(c, GSI_CMD_LOAD_SOUND_DATA_FROM_FILE);
	gsi_send_byte(c, slot);
	gsi_send_int32(c, strlen(filename));
	gsi_send_int32(c, offset);
	gsi_send_int32(c, length);
	gsi_send_int32(c, samplerate);
	gsi_send_byte(c, channels);
	gsi_send_byte(c, type);
	// block:
	gsi_send_string(c, filename);

	return 0;
}


#if 0
/* Use -1 for any of these parameters you don't want to change. */
void gsi_convert_slot (gsi_server_t *c, int slot, int samplerate, int channels, int type/*, int size*/) {
	gsi_send_commands(c, 8, GSI_CMD_CONVERT_SLOT, slot, SPLIT_MSB_INT32(samplerate), channels, type);
}
#endif


int gsi_load_song (gsi_server_t *c, int slot, char *filename) {
	return gsi_load_song_from(c, slot, filename, 0);
}
int gsi_load_song_from (gsi_server_t *c, int slot, char *filename, int offset) {
	if (filename == NULL) return -1;

	gsi_send_byte(c, GSI_CMD_LOAD_SONG_FROM_FILE);
	gsi_send_byte(c, slot);

	gsi_send_int32(c, strlen(filename));
	gsi_send_int32(c, offset);
	// block:
	gsi_send_string(c, filename);

	return 0;
}


int gsi_send_raw_sample (gsi_server_t *c, int slot, int samplerate, int channels, int type, int size, unsigned char *data) {
	gsi_lock_slot(c, slot);

	gsi_send_byte(c, GSI_CMD_LOAD_SOUND_DATA);
	gsi_send_byte(c, slot);
	gsi_send_int32(c, samplerate);
	gsi_send_byte(c, channels);
	gsi_send_byte(c, type);

	gsi_send_int32(c, size);
	// block:
	gsi_send_block(c, data, size);

	return 0;
}


int gsi_send_song (gsi_server_t *c, int slot, int data_len, unsigned char *data) {
	gsi_send_byte(c, GSI_CMD_LOAD_SONG_DATA);
	gsi_send_byte(c, slot);

	gsi_send_int32(c, data_len);
	// block:
	gsi_send_block(c, data, data_len);

	return 0;
}


int gsi_send_song_track (gsi_server_t *c, int slot, int trackno, int data_len, unsigned char *data) {
#if 0
	gsi_send_byte(c, GSI_CMD_LOAD_SONG_TRACK);
	gsi_send_byte(c, slot);
	gsi_send_byte(c, trackno);

	gsi_send_int32(c, data_len);
	gsi_send_block(c, data, data_len);
#endif
	return 0;
}


/*---------------------------------------------------------------------------*
 *      SYSV shared memory (or anything else that's fast if SYSV shm is not
 *      available).
 *---------------------------------------------------------------------------*/
void gsi_send_shm_stream(gsi_server_t *c, int bytes, char *buffer) {
	/* useful ? */
}


/*   Returns 0 if OK, < 0 if failed (error code from shmget, or the server is
 *   not a local one...). */
int gsi_make_shm (gsi_server_t *c, int bytes, char **shm) {
#if USE_SHARED_MEMORY
#  if HAVE_SYSV_SHARED_MEMORY
//	int   status;
	key_t ipc_key;
	int   shmid;
	char *segptr;
	int   ipc_permissions;

	/* check that the server is a local one. */
	if (strcmp(c->remote_host, gsi_thishost_name) != 0) {
		return -1;
	}

	ipc_key = rand();
	ipc_permissions = 0600;

//	if (share_server) {
//		ipc_permissions = af_local_permissions;
//	}
	ipc_permissions = c->gsi_share_permissions;

	/* Open the shared memory segment - create if necessary */
	while ((shmid = shmget(ipc_key, bytes + GSI_SHM_HEADER_SIZE, IPC_CREAT | IPC_EXCL | ipc_permissions)) == -1) {
		if (errno == EEXIST) {
			DEB1(DPRT("Shared memory segment exists - next\n"));
		}
		else {
			log_message(GSI_MESSAGE_ERROR, "shget: %s", strerror(errno));
			return errno;
		}
		ipc_key = rand();
	}

	/* Attach (map) the shared memory segment into the current process */
	if ((segptr = shmat(shmid, 0, 0)) == (void *) (-1)) {
		log_message(GSI_MESSAGE_ERROR, "shmat: %s", strerror(errno));
		return errno;
	}

	*shm = (char *) segptr;

	strncpy(*shm + GSI_SHM_HEADER_SIZE, GSI_SHM_ID, 16);
	((struct gsi_shm *) *shm)->header_size = sizeof(struct gsi_shm);

	/* Now initialize semaphore */

	return shmid;
#  endif

#  if HAVE_POSIX_SHARED_MEMORY
	status = sem_init(&((struct gsi_shm *) *shm)->write_lock, 1, 1);
	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "sem_init on shared memory: %s", strerror(errno));
		status = shmctl(shmid, IPC_RMID, 0);
		if (status < 0) {
			log_message(GSI_MESSAGE_ERROR, "shmctl: %s", strerror(errno));
		}
		shmid = -1;
	}
#  endif

#else
	/* No fast shared memory available. */
	return -1;
#endif
}


int gsi_update_shm_write (gsi_server_t *c, int bytes, char *shmdata) {
#if USE_SHARED_MEMORY
	struct gsi_shm *shm = ((struct gsi_shm *) shmdata);

	/* is locking code OK now? */
	shm->write_end_ptr += bytes;
#  if HAVE_POSIX_SHARED_MEMORY
	if (sem_trywait(&shm->write_lock) == 0) {
		shm->write_start_ptr = shm->read_start_ptr;
	};
#  endif
#  if HAVE_SYSV_SHARED_MEMORY

#  endif

//	shm->write_lock = 1;
//	if (shm->read_lock == 0) {
//		/* no access to read_end_ptr/read_start_ptr is done. */
//		shm->read_end_ptr = shm->write_end_ptr;
//		//shm->write_start_ptr = shm->read_start_ptr; // not allowed, as shm->read_start_ptr must be updateble by server at any time without locking. So server must do this.
//	}
//	shm->write_lock = 0;
	return 0;
#else
	return -1;
#endif
}


int gsi_update_shm_read (gsi_server_t *c, int bytes, char *shmdata) {
#if USE_SHARED_MEMORY
	struct gsi_shm *shm = ((struct gsi_shm *) shmdata);

	/* is locking code OK now? */
	shm->read_start_ptr += bytes;

#  if HAVE_POSIX_SHARED_MEMORY
	if (sem_trywait(&shm->write_lock) == 0) {
		shm->write_start_ptr = shm->read_start_ptr;
	};
#  endif

//	shm->read_lock = 1;
//	if (shm->write_lock == 0) {
//		/* no access to write_start_ptr is done. */
//		//shm->read_end_ptr = shm->write_end_ptr; // server must do this!
//		shm->write_start_ptr = shm->read_start_ptr;
//	}
//	shm->read_lock = 0;
	return 0;
#else
	return -1;
#endif
}


int gsi_remove_shm (gsi_server_t *c, char *shm) {
#if HAVE_SYSV_SHARED_MEMORY
	int  status, shmid;

	shmid = ((struct gsi_shm *) shm)->shmid;
	status = shmdt(shm);
	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "shmdt: %s", strerror(errno));
		return status;
	}

	gsi_send_commands(c, 2, GSI_CMD_SHMDETACH, shmid);
	/* or let the SHM release be handled by FREE_SLOT ? */

	status = shmctl(shmid, IPC_RMID, 0);
	if (status < 0) {
		log_message(GSI_MESSAGE_ERROR, "shmctl: %s", strerror(errno));
	}

	return status;
#else
	return -1;
#endif
}


/*      EOF functions.c */
