/* GSI - general sound interface
 *
 * io.c: io handling (for interface to the GSI server)
 *
 *
 * Copyright (C) W.H.Scholten 1996 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include <errno.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_"
#include <gsi/gsi_debug.h>

#include "gsi.h"

#include "io.h"
#include "log.h"

void _gsi_server_died(gsi_server_t *c) {
	c->server_ok = False;
	close(c->send_fd);
	c->send_fd = -1;
	/* Put a server died message on the queue */
	c->status_queue[0] = GSI_EV_SERVER_KILLED;
	c->status_queue_len = 1;
	/* NB. This doesn't get read at the moment as gsi_events_pending (and
	 * gsi_next_event through it) check c->server_ok, so we must change that
	 * or use the return value of -1 for gsi_events_pending.
	 */

	c->num_commands = 0;
}

/* reading/writing */
void gsi_flush (gsi_server_t *c) {
	int  j;

	if (c->num_commands > 0) {
		j = write(c->send_fd, &c->gsi_command_buffer[0], c->num_commands);
		if (j > 0)
			c->num_commands -= j;
		else if (j < 0 && errno == EPIPE) {
			DEB1(DPRT("%s (gsi_flush) server dead\n", gsi_version));
			_gsi_server_died(c);
		}
	}
}


void gsi_flush_all (gsi_server_t *c) {
	if (c->num_commands > 0)
		_gsi_send_block(c, &c->gsi_command_buffer[0], c->num_commands);
	c->num_commands = 0;
}


//static struct timeval _gsi_max_wait_tv = { 10, 0 }; /* 10 s timeout value */

void gsi_flush_until_free (gsi_server_t *c, int nfree) {
	fd_set write_fd_sel;
	int  j = 0;
	int  status;

	nfree = min(nfree, _MAX_QUEUE_LENGTH);
	if (_MAX_QUEUE_LENGTH - c->num_commands >= nfree) return;
	for  (;;) {
		j = write(c->send_fd, &c->gsi_command_buffer[0], c->num_commands);
		if (j > 0) {
			c->num_commands -= j;
			memmove(&c->gsi_command_buffer[0], &c->gsi_command_buffer[j], c->num_commands);
		}
		else if (j < 0 && errno == EPIPE) {
			DEB1(DPRT("%s (gsi_flush_until_free) server dead\n", gsi_version));
			_gsi_server_died(c);
			break;
		}
		if (_MAX_QUEUE_LENGTH - c->num_commands >= nfree) break;
		FD_ZERO(&write_fd_sel);
		FD_SET(c->send_fd, &write_fd_sel);
		status = select(c->send_fd+1, NULL, &write_fd_sel, NULL, NULL);
		/*
		status = select(c->send_fd+1, NULL, &write_fd_sel, NULL, &_gsi_max_wait_tv);
		if (status == 0) {
			//GSI_EV_COMMAND_TIMEOUT
			return -1;
		}
		*/
	}
}

int gsi_sync (gsi_server_t *c) {
	c->sync_flag = False;
DEB(DPRT("--gsi_sync--\n"));

	gsi_send_commands(c, 1, GSI_CMD_SYNC);
	gsi_flush_all(c);
	while (c->sync_flag == False) {
		int  status;

		status = gsi_wait_for_event(c, 20, 0);
		if (status < 0) {
DEB(DPRT("sync FAILS\n"));
			return -1;
		}

		if (status == 0) {
			/* 10s time out */
			log_message(GSI_MESSAGE_ERROR, "SYNC timeout");
			/* send a GSI_EV_SYNC_TIMEOUT event ? */
DEB(DPRT("sync TIMEOUT, gsi_wait_for_event failed\n"));
			return -1;
		}

		while (gsi_next_event(c) == 0) {
			if (c->sync_flag == True) {
				break;
			}
		}
	}
DEB(DPRT("sync OK\n"));
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Send commands/data (bytes).
 *---------------------------------------------------------------------------*/
void gsi_send_commands(gsi_server_t *c, int argc, ...) {
	va_list argp;

	va_start(argp, argc);

	while (argc) {
		int  i, free_now = _MAX_QUEUE_LENGTH - c->num_commands;
		if (free_now <= 0) {
			gsi_flush_until_free(c, 4);
			free_now = _MAX_QUEUE_LENGTH - c->num_commands;
		}
		free_now = argc < free_now ? argc : free_now; /* amount of the available space we will use */

		for (i=0; i< free_now; i++) {
			c->gsi_command_buffer[ c->num_commands++ ] = va_arg(argp, byte);
		}
		argc -= free_now;
	}

	va_end(argp);
}


/*---------------------------------------------------------------------------*
 *      low level.
 *---------------------------------------------------------------------------*/
void gsi_send_byte(gsi_server_t *c, int cmd) {
	gsi_flush_until_free(c, 1);

	c->gsi_command_buffer[ c->num_commands++ ] = cmd;
}


void gsi_send_int32 (gsi_server_t *c, int data) {
	byte *buf;

	gsi_flush_until_free(c, 4);
	buf = &c->gsi_command_buffer[ c->num_commands ];
	*buf++ = data >> 24;
	*buf++ = data >> 16;
	*buf++ = data >> 8;
	*buf++ = data;
	c->num_commands += 4;
}


void gsi_send_block(gsi_server_t *c, byte *data, int length) {
	gsi_flush_all(c);
	_gsi_send_block(c, data, length);
}


void gsi_send_string(gsi_server_t *c,char *string) {
	int  len;

//	if (string == NULL) {
//		gsi_send_int32(c, 0);
//		//_gsi_send_byte(c, 0);
//	}
//	else {
		/* small strings are not sent directly to the pipe, but buffered */
	len = strlen(string) /*+1*/;
//		gsi_send_int32(c, len);
		if (len < _MAX_QUEUE_LENGTH - c->num_commands) {
			memcpy(&c->gsi_command_buffer[ c->num_commands ], string, len);
			c->num_commands += len;
		}
		else {
			gsi_send_block(c, (byte *) string, len);
		}
//	}
}

/*---------------------------------------------------------------------------*
 *      Pipe/socket data transfer functions (internal use only, unless you
 *      really know what you are doing).
 *---------------------------------------------------------------------------*/
void _gsi_send_block (gsi_server_t *c, byte *data, int len) {
	fd_set write_fd_sel;
	int  i = 0, j = 0;
	int  status;

	DEB2(DPRT(" _gsi_send_block: sending a block of %d bytes\n", len));
	if (c->server_ok) {
		while((j = write (c->send_fd, data+i, len-i)) <len-i) {
			if (j > 0)
				i += j;
			else if (j < 0 && errno == EPIPE) {
				DEB1(DPRT("%s (_gsi_send_block) server dead\n", gsi_version));
				_gsi_server_died(c);
				break;
			}
			FD_ZERO(&write_fd_sel);
			FD_SET(c->send_fd, &write_fd_sel);
			status = select(c->send_fd+1, NULL, &write_fd_sel, NULL, NULL);
//			status = select(c->send_fd+1, NULL, &write_fd_sel, NULL, &_gsi_max_wait_tv);
//			if (status == 0) {
//				//GSI_EV_COMMAND_TIMEOUT
//				return -1;
//			}
			if (status < 0) { /* write error => server died ? */

			}
		}
	}
	DEB1(DPRT(" _gsi_send_block done\n"));
}


/*---------------------------------------------------------------------------*
 *      Wait for data on receive_fd, if timeout_sec < 0, wait forever. Calls
 *      gsi_flush if commands are in the buffer.
 *      Returns -1 on error, 0 if no events within timeout, 1 if an event
 *      arrived (or was already available..).
 *---------------------------------------------------------------------------*/
int gsi_wait_for_event (gsi_server_t *c, int timeout_sec, int timeout_musec) {
	int    status;
	fd_set read_fds;
	fd_set write_fds;
	fd_set *select_write_fds;
	struct timeval tv;
	struct timeval *select_tv;
	//fd_set exception_fds;
	int  max_fd = 0;

	DEB1(DPRT("--gsi_wait_for_event--\n"));
	if (timeout_sec < 0) {
		select_tv = NULL;
	}
	else {
		select_tv = &tv;
		tv.tv_sec  = timeout_sec;
		tv.tv_usec = timeout_musec;
	}

	for(;;) {
		gsi_flush(c);

		if (gsi_events_pending(c) > 0) return 1;

		if (c->num_commands > 0) {
			select_write_fds = &write_fds;
			FD_ZERO(&write_fds);
			if (c->send_fd > 0) {
				FD_SET(c->send_fd, &write_fds);
				max_fd = max(max_fd, c->send_fd);
			}
		}
		else {
			select_write_fds = NULL;
		}

		//FD_ZERO(&exception_fds);
		//FD_SET(c->receive_fd, &exception_fds);

		FD_ZERO(&read_fds);
		if (c->receive_fd > 0) {
			FD_SET(c->receive_fd, &read_fds);
			max_fd = max(max_fd, c->receive_fd);
		}

		DEB3(DPRT("voor select (num_commands=%d)\n", c->num_commands));
		// if receive_fd /= send_fd => max(send_fd, receive_fd)+1
		status = select(max_fd+1, &read_fds, select_write_fds, /*&exception_fds*/ NULL, select_tv);
		DEB3(DPRT("na select (status=%d)\n", status));
		if (status == 0) break; /* timeout */
	}
	DEB1(DPRT("++gsi_wait_for_event++\n"));

	return status;
}

/*      EOF io.c */
