/* GSI - general sound interface
 *
 * utils.c: loading files, as described in a list (moved out of the server
 *   and changed to give the server load commands for each file).
 *
 * ** See gsi.h for a description of the functions **
 *
 *
 * Copyright (C) W.H.Scholten 2000
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_"
#include <gsi/gsi_debug.h>

#include "gsi.h"

#include "io.h"
#include "log.h"

static int  _tmp_gsif_types[] = {
        GSI_8BIT_ALAW, GSI_8BIT_MULAW, GSI_8BIT, GSI_16BIT, GSI_24BIT, 
        GSI_8BIT_SIGNED, GSI_16BIT_LE, GSI_24BIT_LE, /* add ADPCM, mp3 -> dynamic */
};
static char *_tmp_gsif_type_strings[] = {
        "8BIT_ALAW", "8BIT_MULAW", "8BIT", "16BIT", "8BIT_24BIT", 
        "8BIT_SIGNED", "16BIT_LE", "24BIT_LE", /* add ADPCM, mp3 -> dynamic */
};
int  _tmp_gsif_type(char *type) {
        int  i;
        for (i=0; i<8; i++) {
                if (!strncmp(type, _tmp_gsif_type_strings[i], 10)) return _tmp_gsif_types[i]
;
        }
        return -1;
}


/*---------------------------------------------------------------------------*
 *      Get information from a list to load a file.
 *
 *      Format:
 *        <name> <file_name> [<offset> [<length> <samplerate> <ch> <type>]]
 *        [autofree] volume=<v>
 *
 *      => change to
 *        <name> <file_name> [<offset> [ RAW_SAMPLE <length> <samplerate> <ch> <type>] [ RAW_MUSIC <lenght> <type>] ]
 *        [autofree] volume=<v>
 *
 *
 *      Examples:
 *        MAN_DOWN  man_down.au volume=256
 *        DOOM_SND1  doom.wad 10003 1000 11025 2 GSI_8BIT autofree volume=128
 *        DESCENT_EXPLOSION1  descent.pig 24000 autofree volume=128
 *
 *      Returns: -1 if file list cannot be opened, -2 if named sound is not
 *      listed in the list file, 0 if all went OK.
 *---------------------------------------------------------------------------*/
#define _GSI_SOUND_TYPE_SAMPLE  1
#define _GSI_SOUND_TYPE_SONG    2

#warning "Finish gsi_load_sound_from_list"
int gsi_load_sound_from_list (gsi_server_t *c, char *list_file, char *sound, int slot) {
	int  i;
	char list_line[ 1024 ]; /* line read from the list */
	FILE *fp;

//	int  found_sound = False;
	int  read_header, data_offset, data_length;

	int  data_samplerate, data_channels, data_type;
	int  slot_volume = 256;

	int  token_no = 0; /* 0=name, 1=filename, 2=offset, 3=length, 4=samplerate, 5=channels, 6=type 7=autofree, 8=volume */

	int sound_type; /* sample or song, from the list file? */

	if (list_file == NULL) return -1;
	if (strlen(list_file) == 0) return -1;
	DEB1(DPRT("LOAD ("__FILE__"): *load sound list*\n"));

	fp = fopen(list_file, "r");
	if (fp == NULL) {
		log_message(GSI_MESSAGE_ERROR, "sound list %s could not be opened", list_file);
		return -1;
	}

	for (i=0; ;i++) {
		int  slot;
		char *file_name;
		char *token;

		fgets(list_line, 127, fp); list_line[ 127 ] = 0;
		if (feof(fp)) break;
		if (strlen(list_line) <= 1) continue; /* empty entry */

		list_line[ (int) strlen(list_line) -1 ] = 0; /* get rid of the linefeed */

		token = strtok(list_line, " \t");
		if (token == NULL) continue;
		if (strcmp(token, sound)) continue;
		token_no++;

		/* These 4 should be retrieved from the new list file format: */
		read_header = True;
		data_offset = 0;
		data_length = 0;
		sound_type = _GSI_SOUND_TYPE_SAMPLE;

		for (;;) {
			char *next = strtok(NULL, " \t,=");
			if (next == NULL) {
				/* Done getting the info, now load the file */
				if (sound_type == _GSI_SOUND_TYPE_SAMPLE) {
					if (read_header)
						gsi_load_sound_from(c, slot, file_name, data_offset);
					else
						gsi_load_raw_sample (c, slot, file_name, data_offset, data_length, data_samplerate, data_channels, data_type);
/*                      => Should be: gsi_load_sample_block(); */
				}
				else {
					if (read_header)
						gsi_load_song_from(c, slot, file_name, data_offset);
					else
						/*gsi_load_raw_song_from(c, slot, file_name, data_offset)*/;
/*                      => Should be: gsi_load_song_track(); */
				}
				free(file_name);
				return 0;
			}
			else {
				if (token_no == 0) {
					file_name = strdup(token);
					if (file_name == NULL) {
						log_message(GSI_MESSAGE_ERROR, "Out of memory");
						return -1;
					}
				}
				else if (!strcmp(next, "autofree")) {
					token_no = 7;
					//slot_flags |= SF_AUTO_FREE;
				}
				else if (!strcmp(next, "volume")) {
					slot_volume = atoi(strtok(NULL, " \t,="));
					/* OK, use this as the default volume of the slot: */
					token_no = 8;
				}
				else {
					/* numbers: */
					int  num;

					num = atoi(token);
					if (token_no == 2) { data_offset = num; }
					else if (token_no == 3) { data_length = num; read_header = False; }
					else if (token_no == 4) { data_samplerate = num; }
					else if (token_no == 5) { data_channels = num; }
					else if (token_no == 6) {
						/* type => use the string */
						data_type = _tmp_gsif_type(token);
					}
					token_no++;
				}
			}

		}
	}
	fclose(fp);
	DEB1(DPRT("LOAD #load_sound_list#\n"));

	return -2;
}


/*      EOF utils.c */
