/* GSI - general sound interface
 *
 * io.c - gsif library; byte order stuff + reading multiple bytes etc
 *
 * Lower level API dependencies: 
 *
 *
 * Copyright (C) 1998 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <unistd.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
//#include <gsi/gsi_debug.h>

#include "io.h"


/*---------------------------------------------------------------------------*
 *      Write a 32 or 16 bit msb-integer to a file.
 *---------------------------------------------------------------------------*/
void gsif_w_msb_int32(int fd, int val, int *s)
{
	int  status;
	int  file_int;

#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
	file_int = gsif_swap_int32(val);
	status = write (fd, &file_int, 4);
#else
	file_int = val;
	status = write (fd, &file_int, 4);
#endif
	if (status != 4) (*s) = status;
}

void gsif_w_msb_int24(int fd, int32 i, int *s)
{
	gsif_w_byte(fd, i >> 16, s);
	gsif_w_byte(fd, i >> 8, s);
	gsif_w_byte(fd, i, s);
}

void gsif_w_msb_int16(int fd, int16 i, int *s)
{
	int  status;
	int16 fint;

#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
	fint = gsif_swap_int16(i);
	status = write (fd, &fint, 2);
#else
	fint = val;
	status = write (fd, &fint, 2);
#endif
	if (status != 2) (*s) = status;
}

void gsif_w_lsb_int24(int fd, int32 i, int *s)
{
	gsif_w_byte(fd, i, s);
	gsif_w_byte(fd, i >> 8, s);
	gsif_w_byte(fd, i >> 16, s);
}

void gsif_w_byte(int fd, byte fbyte, int *s)
{
	int  status;

	status = write(fd, &fbyte, 1);
	if (status != 1) (*s) = status;
}

void gsif_w_block(int fd, void *data, int length, int *s) {
	int  status;

	status = write(fd, data, length);
	if (status != length) (*s) = status;
}


/*---------------------------------------------------------------------------*
 *      Read a 32, 24 or 16 bit msb- or lsb-integer from a file.
 *---------------------------------------------------------------------------*/
/* function value = value read */
int32 gsif_r_msb_int32(int fd, int *s)
{
	int  status;
	int32 val;

	status = read(fd, &val, 4);
	if (status != 4) *s = -1;

#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
	val = gsif_swap_int32(val);
#endif
	return val;
}

int32 gsif_r_msb_int24(int fd, int *s)
{
	int32 val;

	val = ((int) gsif_r_byte(fd, s) << 16) | ((int) gsif_r_byte(fd, s) << 8) | ((int) gsif_r_byte(fd, s));

	return val;
}

int16 gsif_r_msb_int16(int fd, int *s)
{
	int  status;
	int16 val;

	status = read(fd, &val, 2);
	if (status != 2) *s = -1;
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
	val = gsif_swap_int16(val);
#endif
	return val;
}

int32 gsif_r_lsb_int32(int fd, int *s)
{
	int  status;
	int32 val;

	status = read(fd, &val, 4);
	if (status != 4) *s = -1;
#ifdef SYSCONFIG_BIG_ENDIAN_MACHINE
	val = gsif_swap_int32(val);
#endif
	return val;
}

int32 gsif_r_lsb_int24(int fd, int *s)
{
	int32 val;

	val = ((int) gsif_r_byte(fd, s)) | ((int) gsif_r_byte(fd, s) << 8) | ((int) gsif_r_byte(fd, s) << 16);

	return val;
}

int16 gsif_r_lsb_int16(int fd, int *s)
{
	int  status;
	int16 val;

	status = read(fd, &val, 2);
	if (status != 2) *s = -1;
#ifdef SYSCONFIG_BIG_ENDIAN_MACHINE
	val = gsif_swap_int16(val);
#endif
	return val;
}

byte gsif_r_byte(int fd, int *s)
{
	int  status;
	byte val;

	status = read(fd, &val, 1);
	if (status != 1) *s = -1;

	return val;
}

void gsif_r_block(int fd, void *data, int length, int *s) {
	int  status;

	status = read(fd, data, length);
	if (status != length) (*s) = status;
}


/*---------------------------------------------------------------------------*
 *      Read 16/32 bit integers from a buffer.
 *---------------------------------------------------------------------------*/
int32 gsif_msb_int32_at(byte *buf) {
	return (((int32) buf[0]) << 24) | (((int32) buf[1]) << 16) | (((int32) buf[2]) << 8) | buf[3];
}

int16 gsif_msb_int16_at(byte *buf) {
	return (((int16) buf[0]) << 8) | buf[1];
}

int32 gsif_lsb_int32_at(byte *buf) {
	return (((int32) buf[0])) | (((int32) buf[1]) << 8) | (((int32) buf[2]) << 16) | (((int32) buf[3]) << 24);
}

int16 gsif_lsb_int16_at(byte *buf) {
	return (((int16) buf[1]) << 8) | buf[0];
}


/*---------------------------------------------------------------------------*
 *      Convert endian-ness.
 *---------------------------------------------------------------------------*/
int32 gsif_swap_int32(int32 a) {
	return ((a & 255) << 24) | ((a & 0xff00) << 8) | ((a >> 8) & 0xff00) | ((a >> 24) &0xff);
}

int16 gsif_swap_int16(int16 a) {
	return ((a & 0xff) << 8) | ((a >> 8) & 0xff);
}


/*      EOF io.c */
