/* GSI - general sound interface
 *
 * block.c - gsif library; sound block manipulations.
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */



/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "sample_format.h"
#include "../io.h"


/*---------------------------------------------------------------------------*
 *      Utilities: gsif_free_sound, gsif_free_block.
 *---------------------------------------------------------------------------*/
int gsif_free_sound (struct gsi_sound *snd)
{
	int  i;

	DEB1(DPRT("("__FILE__") gsif_free_sound:\n"));
	if (snd != NULL) {
		if (snd->filename) {
			/* remove this test and make NULL before freeing in higher level
			 * code? */
			if (strncmp(snd->filename, "/DATA", 6) != 0) {
				free(snd->filename);
			}
		}
		if (snd->comment) free(snd->comment);

		if (snd->block == NULL) gsif_make_linear_block_list(snd);
		if (snd->block != NULL) {
			for (i=0; i< snd->blocks; i++) {

				if (snd->block[i] != NULL) {
					if (snd->block[i]->data != NULL)
						free(snd->block[i]->data);

					free(snd->block[i]);
				}
			}
			free(snd->block);
		}
		free(snd);
	}
	DEB1(DPRT("-> gsif_free_sound done\n"));

	return 0;
}


int gsif_free_block (struct gsi_sound *snd, int i)
{
	DEB1(DPRT("("__FILE__") gsif_free_block (%d)\n", i));

	if (i>snd->blocks) return -1;

	if (snd->block[i] != NULL) {
		if (snd->block[i]->data !=NULL)
			free(snd->block[i]->data);

		free(snd->block[i]);
		snd->block[i] = NULL;
	}
	DEB1(DPRT("-> gsif_free_block done\n"));

	return 0;
}


/*---------------------------------------------------------------------------*
 *      gsi_block manipulations:
 *      Add a block to the linked list.
 *      Returns 0 if OK, < 0 for error.
 *---------------------------------------------------------------------------*/
int gsif_add_block (struct gsi_sound *snd, int samplerate, int channels, int type, int data_start, int data_length)
{
	struct gsi_block *new_block = (struct gsi_block *) calloc(1, sizeof (struct gsi_block));

	if (new_block == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
	DEB1(DPRT("("__FILE__") gsif_add_block (samplerate=%d, ch=%d, type=%d, off=%d, size=%d)\n", samplerate, channels, type, data_start, data_length));
	if (snd->blocks >0) {
		snd->last->next = new_block;
		new_block->prev = snd->last;
		snd->last = snd->last->next;
	}
	else {
		snd->first = new_block;
		snd->last = snd->first;
	}
	new_block->next = snd->first;

	new_block->samplerate = samplerate;
	new_block->channels = channels;
	new_block->type = type;

	new_block->data_offset = data_start;
	new_block->data_length = data_length;
	//new_block->data = (byte *) malloc(data_length);
	snd->blocks++;
	DEB1(DPRT("-> gsif_add_block done\n"));

	return GSIF_OK;
}


int  gsif_make_linked_block_list (struct gsi_sound *snd)
{
	int  i;

	DEB1(DPRT("("__FILE__") make_linked_block_list\n"));
	snd->first = snd->block[0];
	snd->last  = snd->block[ snd->blocks-1 ];

	for (i=0; i<snd->blocks; i++) {
		/* circular lists */
		snd->block[i]->next = snd->block[ (i+1) % snd->blocks ];
		snd->block[i]->prev = snd->block[ (i-1+snd->blocks) % snd->blocks ];
	}
	DEB1(DPRT("-> make_linked_block_list done\n"));

	return GSIF_OK;
}


/* Returns 0 if OK, < 0 if an error occurred. */
int  gsif_make_linear_block_list (struct gsi_sound *snd)
{
	int  i;
	struct gsi_block *next_block;

	DEB1(DPRT("("__FILE__") make_linear_block_list\n"));
	if (snd->block) free(snd->block);
	snd->block = NULL;
	if (snd->blocks > 0) {
		if (snd->first != NULL) {
			snd->block = (struct gsi_block **) malloc (snd->blocks *sizeof(struct gsi_block *));
			if (snd->block == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
			next_block = snd->first;
			for (i=0; i < snd->blocks; i++) {
				snd->block[i] = next_block;
				next_block = next_block->next;
			}
		}
	}
	DEB1(DPRT("-> make_linear_block_list done\n"));

	return GSIF_OK;
}


/*---------------------------------------------------------------------------*
 *      Load sample data in a block.
 *---------------------------------------------------------------------------*/
int  gsif_load_block_data (int fd, struct gsi_sound *snd, int block)
{
	int  data_start;
	int  size;

	DEB1(DPRT("("__FILE__") gsif_load_block_data (block %d, off=%ld, len=%ld)\n", block, snd->block[ block ]->data_offset, snd->block[ block ]->data_length));

	data_start = snd->offset+snd->block[ block ]->data_offset;
	size = snd->block[ block ]->data_length;

//	padded_size = size;
//	/* padding: add silence for missing few bytes */
//	pad = 1;
//	if (snd->block[ block ]->type == GSI_16BIT) pad *= 2;
//	pad *= snd->block[ block ]->channels;
//
//	padded_size = (size + (pad-1)) & ~(pad-1);
//
//	snd->data = (sbyte *) malloc(padded_size);

	if (size > 0) {
		if (snd->block[ block ]->data == NULL)
			snd->block[ block ]->data = (byte *) malloc(size);
		if (snd->block[ block ]->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

		lseek(fd, data_start, SEEK_SET);
		read(fd, snd->block[ block ]->data, size);
	}

//	if (padded_size > size) {
//		if (current->sound_type[slot] == GSI_8BIT) {
//			// not needed when using callloc
//		}
//	}

	DEB1(DPRT("-> gsif_load_block_data: done\n"));

	return GSIF_OK;
}


/*      EOF sample_block.c */
