/* GSI - general sound interface
 *
 * convert_block.c - gsif library; convert a samples to another format
 *
 * Lower level API dependencies:
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */



/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "mulaw.h"


extern int  _gsif_Bps[];
extern int  _gsif_type_value[];
int gsif_sizeof_sample(int format);


/*---------------------------------------------------------------------------*
 *      Convert a sample to the given type. Use -1 for any of the parameters
 *      you don't want to change. If new_blk has data != NULL the given
 *      pointer will be used to write the converted data to, otherwise space
 *      is malloc'ed.
 *---------------------------------------------------------------------------*/
#warning "implement gsif_convert_data"
int  gsif_convert_data(void *conversion_state, int input_type, unsigned char *input_data, int output_type, int output_endian, int output_length);
/* 'conversion_state' is for compression/decompression data,
 * 'output_length' is in bytes (necessary for compression/decompression, meta
 * samples won't work).
 */

int gsif_convert_block (struct gsi_block *blk, struct gsi_block *new_blk, int samplerate, int channels, int type)
{
	int  i;
	int  bps_in, bps_out;
	int  samples;
	byte *input, *output;
	int  input_val;
	int  input_type;

	DEB1(DPRT("("__FILE__") gsif_convert_block:\n"));

	new_blk->samplerate = blk->samplerate;
	new_blk->channels = blk->channels;
	new_blk->type = blk->type;

	if (samplerate > 0)
		new_blk->samplerate = samplerate;
	if (channels > 0)
		new_blk->channels = channels;
	if (type > 0)
		new_blk->type = type;

	bps_in  = _gsif_Bps [ blk->type ];
	bps_out = _gsif_Bps [ type ];
	samples = blk->data_length / (blk->channels* bps_in);
	input_type= blk->type;

	if (new_blk->data == NULL) {
		int  new_size = 0.5 + ((double) (new_blk->samplerate) * (samples * bps_out)) / (double) blk->samplerate;

		new_blk->data = malloc(new_size);
		if (new_blk->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
	}

	input = blk->data;
	output= new_blk->data;

	/* Assuming samplerate_in = samplerate_out for now: */
	for (i=0; i<samples; i++) {
		switch (input_type) {
		case GSI_8BIT_ALAW:
			/* NOT YET SUPPORTED */
			return GSIF_ERROR_UNHANDLED_FORMAT_CONVERSION;
			// input_val = alaw_to_linear16_inline(input_data[i]);
			break;
		case GSI_8BIT_MULAW:
			input_val = mulaw_to_linear16_inline(input[i]);
			break;
		case GSI_8BIT:
			input_val = ((sbyte) (*input++ ^ 0x80)) << 8;
			break;
		case GSI_16BIT:
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
			input_val = (int) (int16) ( ((*input++) << 8) | (*input++) );
#else
			input_val = *((int16 *)input)++;
#endif
			break;
		case GSI_8BIT_SIGNED:
			input_val = ((sbyte) *input++) << 8;
			break;
		case GSI_16BIT_LE:
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
			input_val = *((int16 *)input)++;
#else
			input_val = (int) (int16) (*input++ | ((*input++) << 8) );
#endif
			break;
		case GSI_24BIT:
			/* NOT YET SUPPORTED */
		case GSI_24BIT_LE:
			/* NOT YET SUPPORTED */
			return GSIF_ERROR_UNHANDLED_FORMAT_CONVERSION;
		default:
			free(new_blk->data);
			free(new_blk);
			return GSIF_ERROR_UNKNOWN_TYPE;
			break;
		}
		/* all data is now 16 bit signed  (change for 24 bit support) */



		switch(type) {
		case GSI_8BIT:
			*output++ = ((input_val >> 8) & 255) ^ 0x80;
			break;
		case GSI_8BIT_ALAW:
			/* NOT YET SUPPORTED */
			return GSIF_ERROR_UNHANDLED_FORMAT_CONVERSION;
			// *output++ = linear16_to_alaw(input_val);
			break;
		case GSI_8BIT_MULAW:
			*output++ = linear16_to_mulaw(input_val);
			break;
		case GSI_16BIT: /* MSB */
			/* native for MSB machines: */
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
			*output++ = (input_val >> 8) & 255;
			*output++ = (input_val & 255);
#else
			*((int16 *) output)++ = input_val;
#endif
			break;
		case GSI_8BIT_SIGNED:
			*output++ = (input_val >> 8);
			break;
		case GSI_16BIT_LE:
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
			*((int16 *) output)++ = input_val;
#else
			*output++ = (input_val >> 8) & 255;
			*output++ = (input_val & 255);
#endif
			break;
		case GSI_24BIT:
			/* NOT YET SUPPORTED */
		case GSI_24BIT_LE:
			/* NOT YET SUPPORTED */
			return GSIF_ERROR_UNHANDLED_FORMAT_CONVERSION;
		default:
			free(new_blk->data);
			free(new_blk);
			return -GSIF_ERROR_UNKNOWN_TYPE;
		}
	}
	DEB1(DPRT("-> gsif_convert_block: done\n"));

	return GSIF_OK;
}


/*---------------------------------------------------------------------------*
 *      EOF convert_block.c
 *---------------------------------------------------------------------------*/
