/* GSI - general sound interface
 *
 * load.c - gsif library; load samples
 *
 * Lower level API dependencies: 
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */



/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <fcntl.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "../io.h"
#include "../log.h"


/*__________________EXTERN___________________*/
extern int  _gsif_Bps[];
extern int  _gsif_type_value[];

/*---------------------------------------------------------------------------*
 *      Load a sample from a file using information from the header.
 *---------------------------------------------------------------------------*/
struct gsi_sound *gsif_load_sample (char *filename, int offset, int *status)
{
	struct gsi_sound *new_snd = NULL;

	new_snd = gsif_get_structure(filename, 1 /* read_header*/, offset, 0, 0, 0, 0, status);
	gsif_load_sound_data(new_snd);

	return new_snd;
}


/*---------------------------------------------------------------------------*
 *      Load a raw sample from a file using the given information.
 *---------------------------------------------------------------------------*/
struct gsi_sound *gsif_load_raw_sample (char *filename, int offset, int length, int snd_samplerate, int snd_channels, int snd_type, int *status)
{
	struct gsi_sound *new_snd = NULL;

	new_snd = gsif_get_structure(filename, 0 /* read_header*/, offset, length/*data_len*/, snd_samplerate, snd_channels, snd_type, status);
	gsif_load_sound_data(new_snd);

	return new_snd;
}


/*---------------------------------------------------------------------------*
 *      Get file/data information.
 *---------------------------------------------------------------------------*/
struct gsi_sound *gsif_get_sample_structure (char *filename, int offset, int *status)
{
	struct gsi_sound *new_snd = NULL;

	new_snd = gsif_get_structure(filename, 1, offset, 0, 0, 0, 0, status);
	//status = gsif_get_structure(&new_snd, filename, 1, offset, 0, 0, 0, 0);

	return new_snd;
}


/*---------------------------------------------------------------------------*
 *      Load sample data in the blocks.
 *---------------------------------------------------------------------------*/
int gsif_load_sound_data (struct gsi_sound *snd)
{
	int  i, fd;

	if (snd != NULL) {
		fd = open(snd->filename, O_RDONLY);
		if (fd > 0) {
			for (i=0; i<snd->blocks; i++) {
				gsif_load_block_data(fd, snd, i);	
			}
			close(fd);
			return 0;
		}
	}

	return -1;
}




/*---------------------------------------------------------------------------*
 *      Load sample data from all blocks, convert to one format (the best
 *      format, based on the highest type, highest bits, highest channels
 *      from all blocks) and store it in one block. If samplerate>=0 then
 *      this overrides the best samplerate and ditto for channels, type.
 *---------------------------------------------------------------------------*/
int gsif_load_sound_in_block (struct gsi_sound *snd, struct gsi_block *blk, int samplerate, int channels, int type)
{
	int  i, fd;

	DEB1(DPRT("("__FILE__") gsif_load_sound_in_block:\n"));

	if (snd != NULL) {
		if (snd->blocks < 1) return -1;

		fd = open(snd->filename, O_RDONLY);
		if (fd > 0) {
			int  best_samplerate = 0;
			int  best_channels = 0;
			int  best_type;

			int  total_bytes = 0;

			int  max_old_block_size = 0;
			int  max_new_block_size = 0;
			int  need_conversion = False; //could be replaced: use convert_block = NULL for no conversion.
			byte *load_address;
			byte *convert_block;

			// get the best samplerate, channels, type:
			best_type = snd->block[0]->type;
			for (i=0; i<snd->blocks; i++) {
				if (snd->block[i]->samplerate > best_samplerate) {
					best_samplerate = snd->block[i]->samplerate;
				}
				if (snd->block[i]->channels > best_channels) {
					best_channels = snd->block[i]->channels;
				}
				if (_gsif_type_value[ snd->block[i]->type ] > _gsif_type_value[ best_type ]) {
					best_type = snd->block[i]->type;
				}


				//gsif_load_block_data(fd, snd, i);
			}

			/* check for overrides: */
			if (samplerate > 0)
				best_samplerate = samplerate;
			if (channels > 0)
				best_channels = channels;
			if (type >= 0)
				best_type = type;

			blk->samplerate = best_samplerate;
			blk->channels = best_channels;
			blk->type = best_type;
			DEB2(DPRT("    -> best values { samplerate=%d, chan=%d, type=%d }\n", best_samplerate, best_channels, best_type));
			// find out size of the new single block:
			for (i=0; i<snd->blocks; i++) {
				/* Due to overflow problems, use floating point: */
				int  new_block_size = (best_channels/snd->block[i]->channels) * (_gsif_Bps[ best_type ]/_gsif_Bps[ snd->block[i]->type ]) * (int) ( ((double) best_samplerate * snd->block[i]->data_length) / ((double) snd->block[i]->samplerate) );

				total_bytes += new_block_size;
				//max_new_block_size = max(max_new_block_size, new_block_size);
				max_old_block_size = max(max_old_block_size, snd->block[i]->data_length);
				//if (new_block_size != snd->block[i]->data_length) need_conversion = True;
			}
			DEB2(DPRT("  total_bytes=%d\n", total_bytes));

			/* malloc it: */
			blk->data = malloc(total_bytes);
			if (blk->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
			blk->data_length = total_bytes;

//			need_conversion = (snd->block[i]->samplerate == best_samplerate &&
//							   snd->block[i]->channels == best_channels &&
//							   snd->block[i]->type == best_type
//				);

			if (need_conversion) {
				convert_block = malloc(max_new_block_size);
				if (convert_block == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
			}

			load_address = blk->data;
			/* load & convert: */
			for (i=0; i<snd->blocks; i++) {
				DEB2(DPRT("    block %d: rate=%d, chan=%d, type=%d\n", i, snd->block[i]->samplerate, snd->block[i]->channels, snd->block[i]->type));
				if (snd->block[i]->samplerate == best_samplerate &&
					snd->block[i]->channels == best_channels &&
					snd->block[i]->type == best_type
					) {
					/* load: */
					DEB2(DPRT("    loading block %d\n", i));
					lseek(fd, snd->offset + snd->block[i]->data_offset, SEEK_SET);
					read(fd, load_address, snd->block[i]->data_length);
					load_address += snd->block[i]->data_length;
				}
				else {
					/* load & convert: */
					struct gsi_block old_block;
					DEB2(DPRT("    loading & converting block %d\n", i));
					lseek(fd, snd->offset + snd->block[i]->data_offset, SEEK_SET);
					read(fd, convert_block, snd->block[i]->data_length);

					/* now convert .. */
					old_block.data = convert_block;
					old_block.data_length = snd->block[i]->data_length;
					old_block.samplerate = snd->block[i]->samplerate;
					old_block.channels = snd->block[i]->channels;
					old_block.type = snd->block[i]->type;
// OR:
//					snd->block[i]->data = convert_block;
//					old_block = snd->block[i];

					gsif_convert_block (&old_block, blk, best_samplerate, best_channels, best_type);

					load_address += /* new_block_size = */ (best_channels/snd->block[i]->channels) * (_gsif_Bps[ best_type ]/_gsif_Bps[ snd->block[i]->type ]) * (int) ( ((double) best_samplerate * snd->block[i]->data_length) / ((double) snd->block[i]->samplerate) );

				}
			}

			if (need_conversion) {
				free(convert_block);
			}

			close(fd);
			DEB1(DPRT("-> gsif_load_sound_in_block: done\n"));

			return GSIF_OK;
		}
	}

	DEB1(DPRT("-> gsif_load_sound_in_block: failed\n"));
	return -1;
}


/*      EOF load.c */
