/* GSI - general sound interface
 *
 * sample_format.c - gsif library, sample structure loading
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
//#include <errno.h>


#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "sample_format.h"
#include "../io.h"
//#include "mulaw.h"


/*---------------------------------------------------------------------------*
 *      Get file structure + comments (data segments are NOT loaded).
 *---------------------------------------------------------------------------*/
struct gsi_sound *gsif_get_structure (char *file_name, int read_header, int data_offset, int data_len, int samplerate, int channels, int type, int *status)
{
	int  fd, size, data_start;
	struct stat fd_stat;
	struct gsi_sound *new_snd;
	int  format_status = GSIF_ERROR_STATUS_UNKNOWN;
	int  file_status;
	int  sys_status;

	DEB1(DPRT("("__FILE__") gsif_get_structure: %s\n", file_name));
	if (file_name == NULL) {
		*status = GSIF_ERROR_NO_SUCH_FILE;
		return NULL;
	}

	if (data_len < 0 || data_offset < 0) return NULL; /* GSIF_ERROR_BAD_ARGUMENTS */
	/* also check channels etc.? */

	fd = open(file_name, O_RDONLY);
	if (fd <= 0) {
		*status = GSIF_ERROR_NO_SUCH_FILE;
		return NULL;
	}

	/* check that the file is a regular file (not a directory) */
	fstat(fd, &fd_stat);
	if ( ! S_ISREG(fd_stat.st_mode) ) {
		DEB1(DPRT("->  %s not a regular file\n", file_name));
		close(fd);
		*status = GSIF_ERROR_NO_SUCH_FILE;
		return NULL;
	}

	new_snd = calloc(1, sizeof(struct gsi_sound));
	if (new_snd == NULL) {
		*status = GSIF_ERROR_OUT_OF_MEMORY;
		return NULL;
	}

	new_snd->filename = strdup(file_name);
	if (new_snd->filename == NULL) {
		free(new_snd);
		*status = GSIF_ERROR_OUT_OF_MEMORY;
		return NULL;
	}

	/* comment out next line and uncomment warning to make code foolproof
	 * (gives 2nd time 0 blocks aiff) */
	file_status = 0;

	{
		int  name_len = strlen(file_name);
		char *name_end = file_name + name_len;
		int  file_size = lseek(fd, 0, SEEK_END);

		/* To make sure stereo files are multiple of 2 bytes etc: */
		//int  padded_size, pad;


		new_snd->offset = data_offset;

		sys_status = lseek(fd, 0+data_offset, SEEK_SET);
		if (sys_status == -1) {
			gsif_free_sound(new_snd);
			*status = GSIF_ERROR_BAD_ARGUMENTS;
			return NULL;
		}

		if (!read_header) {
			data_start = data_offset;
			size = min(file_size - data_start, file_size);
			if (size < 0) size = 0;
			if (data_len > 0) size = min(data_len, size);

			gsif_add_block(new_snd, samplerate, channels, type, data_start, size);
			gsif_make_linear_block_list(new_snd);

			new_snd->file_format = GSI_FILE_RAW;
			format_status = GSIF_OK;
		}
		else {
			/* Now find out what type of file we have. raw files (8000 Hz mono
			 * 8 bit unsigned) are recognized through the extension.
			 */
			if ((name_len > 4 && !strcmp(name_end-4, ".raw"))) {
				samplerate = 8000;
				channels = 1;
				type = GSI_8BIT;
				data_start = data_offset;
				size = file_size - data_start;

				gsif_add_block(new_snd, samplerate, channels, type, data_start, size);
				gsif_make_linear_block_list(new_snd);

				new_snd->file_format = GSI_FILE_RAW;
				format_status = GSIF_OK;
			}
			else {
				/* Check for file types by signature */
				int  file_sig;

				file_sig = gsif_r_msb_int32(fd, &file_status);
				if (file_status) {
					gsif_free_sound(new_snd);
					*status = GSIF_ERROR_FILE_CORRUPT;
					return NULL;
				}

DEB(DPRT("filesig=0x%8x (%c%c%c%c))\n", file_sig, file_sig>>24, (file_sig>>16) &255, (file_sig>>8) &255, (file_sig>>0) &255));

				switch (file_sig) {
				case MAGIC_AU:
					format_status = gsif_get_structure_au(fd, new_snd);
					break;

				case SIG('F','O','R','M'): /* AIFF/AIFC */
					format_status = gsif_get_structure_aif(fd, new_snd);
					break;

				case MAGIC_VOC:
					format_status = gsif_get_structure_voc(fd, new_snd);
					break;

				case SIG('R','I','F','F'): /* yuck!*/
					format_status = gsif_get_structure_wav(fd, new_snd);
					break;

				default:
					format_status = GSIF_ERROR_UNHANDLED_FILE_TYPE;
					break;
				}
			} /* NOT .raw */
		} /* read header etc. */

		close(fd);
		if (format_status != GSIF_OK) {
			DEB1(DPRT("-> get file format of %s failed (err %d)\n", file_name, format_status));
			gsif_free_sound(new_snd);
			*status = format_status;
			return NULL;
		}
	}
	DEB1(DPRT("-> gsif_get_structure: done\n"));

	return new_snd;
}


/*      EOF sample_format.c */
