/* GSI - general sound interface; libgsif
 *
 * sample_get_structure_aif.c - gsif library, sample structure loading
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "sample_format.h"
#include "../io.h"
#include "../log.h"


/*---------------------------------------------------------------------------*
 *      Returns True if recognized as a valid AIF[FC] (Apple's audio
 *      interchange format)
 *      return codes: GSIF_OK/GSIF_ERROR_* from gsif.h
 *---------------------------------------------------------------------------*/
int gsif_get_structure_aif(int fd, struct gsi_sound *new_snd) {
	int  format_status = GSIF_ERROR_STATUS_UNKNOWN;
	int  size, data_start;
	int  channels = 0, type = -1, samplerate = 0;

	int  sig;
	int  tmp;
	int  chunk_len;
	int  found_data = False;
	int  file_status = 0;

	/* Position just after the 4 byte signature we checked to get here: */
	file_status |= (lseek(fd, new_snd->offset + 4, SEEK_SET) == -1);

	size = gsif_r_msb_int32(fd, &file_status);
	sig  = gsif_r_msb_int32(fd, &file_status);
	if (file_status) {
		return GSIF_ERROR_FILE_CORRUPT;
	}

	if (sig != SIG('A','I','F','F') && sig != SIG('A','I','F','C')) {
		return GSIF_ERROR_FILE_CORRUPT;
	}

	for (;;) {
		DEB2(DPRT("   -> aiffc: file pos = %d\n", (int) lseek(fd, 0, SEEK_CUR)));
		sig = gsif_r_msb_int32(fd, &file_status);
		chunk_len = gsif_r_msb_int32(fd, &file_status);
		if (file_status) {
			return GSIF_ERROR_FILE_CORRUPT;
		}

		DEB2(DPRT("   -> new chunk: sig=%c%c%c%c len=%d\n", (char) (sig>>24), (char) (sig>>16), (char) (sig>>8), (char) sig, chunk_len));

		switch(sig) {
		case SIG('C','S','N','D'): /* compressed sound */
			DEB2(DPRT("   -> aiffc: CSND ignored\n"));
			file_status |= (lseek(fd, chunk_len, SEEK_CUR) == -1);
			break;

		case SIG('F','V','E','R'): /* format version */
			DEB2(DPRT("   -> aiffc: FVER ignored\n"));
			file_status |= (lseek(fd, chunk_len, SEEK_CUR) == -1);
			break;

		case SIG('I','N','S','T'): /* instrument chunk */
			DEB2(DPRT("   -> aiffc: INST ignored\n"));
			file_status |= (lseek(fd, chunk_len, SEEK_CUR) == -1);
			break;

		case SIG('M','A','R','K'): /* marker: name + index into sample data */
			DEB2(DPRT("   -> aiffc: MARK ignored\n"));
			file_status |= (lseek(fd, chunk_len, SEEK_CUR) == -1);
			break;

		case SIG('C','O','M','T'): /* comment chunk */
			DEB2(DPRT("   -> aiffc: COMT chunk\n"));
			tmp = gsif_r_msb_int32(fd, &file_status); /* sample frames */
			DEB1(DPRT("   -> aiffc: -> num comments=%d\n", tmp));
			//file_status |= (lseek(fd, new_snd->offset + 24, SEEK_SET) == -1);
			new_snd->comment = calloc(1, chunk_len+1);
			if (new_snd->comment == NULL) {
				return GSIF_ERROR_OUT_OF_MEMORY;
			}

			file_status |= (read(fd, new_snd->comment, chunk_len) != chunk_len);
			if (file_status) {
				return GSIF_ERROR_FILE_CORRUPT;
			}
			new_snd->comment[ chunk_len ] = 0;
			break;

		case SIG('A','P','P','L'): /* application specific chunk */
			DEB2(DPRT("   -> aiffc: APPL chunk ignored\n"));
			file_status |= (lseek(fd, chunk_len, SEEK_CUR) == -1);
			break;

		case SIG('C','O','M','M'):
			channels = gsif_r_msb_int16(fd, &file_status);
			tmp = gsif_r_msb_int32(fd, &file_status); /* sample frames */
			tmp = gsif_r_msb_int16(fd, &file_status); /* bits per sample */
			DEB2(DPRT("   -> aiffc: COMM: ch=%d, bits=%d\n", channels, tmp));
			if (file_status) {
				return GSIF_ERROR_FILE_CORRUPT;
			}

			if (tmp == 8) {
				type = GSI_8BIT_SIGNED;
				//type = GSI_8BIT;
			}
			else if (tmp == 16) type = GSI_16BIT;

			{ /* Apple dipsticks! using a long double, ridiculous */
				long double a;
				file_status |= (read(fd, &a, 10) != 10);
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
				/* Convert MSB to LSB for LSB machines */
				{
					int  i;
					char tmp;
					char *aa = (char *) &a;

					for (i=0; i<5; i++) {
						tmp = aa[ i ]; aa[ i ] = aa[ 9-i ]; aa[ 9-i ] = tmp;
					}
				}
#endif
				DEB2(DPRT("   -> aiffc: COMM: samplerate (long double)=%f\n", (float) a));
				samplerate = (int) a;
			}
			if (samplerate < 1 || channels < 1 || file_status) {
				format_status = GSIF_ERROR_FILE_CORRUPT; break;
			}

			file_status |= (lseek(fd, chunk_len-10-8, SEEK_CUR) == -1);
			break;

		case SIG('S','S','N','D'):
			tmp = gsif_r_msb_int32(fd, &file_status); /* offset */
			tmp = gsif_r_msb_int32(fd, &file_status); /* blocksize */
			if (file_status) {
				return GSIF_ERROR_FILE_CORRUPT;
			}

			size = chunk_len-8;
			data_start = lseek(fd, 0, SEEK_CUR);

			gsif_add_block(new_snd, samplerate, channels, type, data_start, size);
			gsif_make_linear_block_list(new_snd);
			//gsif_load_block_data(fd, new_snd, /*0*/new_snd->blocks-1);
			file_status |= (lseek(fd, chunk_len-8, SEEK_CUR) == -1);

			found_data = True;
			break;

		default:

			log_message (0, "unknown AIFF/C chunk type x%4x", sig);
		}
		if (found_data || format_status != GSIF_ERROR_STATUS_UNKNOWN) break;
	}

	new_snd->file_format = GSI_FILE_AIFF;
	if (format_status == GSIF_ERROR_STATUS_UNKNOWN)
		format_status = GSIF_OK;

	return format_status;
}


/*      EOF sample_get_structure_aif.c */
