/* GSI - general sound interface
 *
 * sample_get_structure_au.c - gsif library, sample structure loading
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "sample_format.h"
#include "../io.h"
//#include "mulaw.h"


/*---------------------------------------------------------------------------*
 *  au/snd
 *---------------------------------------------------------------------------*
 *   .au/.snd file format:
 *----------------------------------------------
 *  ".snd"  START  LENGTH  TYPE  FREQ  CHANNELS
 *  <DESCRIPTION>    (text)
 *  <DATA>           (starts at START, length=LENGTH)
 *----------------------------------------------
 *   START, LENGTH, TYPE, FREQ and CHANNELS are 32bit MSB-ints
 *   TYPE is:
 *      1 = mu-law compressed 8 bit, (GSI_8BIT_MULAW)
 *      2 = unsigned 8 bit,          (GSI_8BIT)
 *      3 = signed 16 bit MSB        (GSI_16BIT)
 *   etc. (see sample/sample_format.h for a complete list)
 *
 *   CHANNELS is:
 *      1 = mono
 *      2 = stereo
 *
 *---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*
 *      Returns 0 if recognized as a valid AU (sun/next)
 *      return codes: GSIF_OK/GSIF_ERROR_* from gsif.h
 *---------------------------------------------------------------------------*/
int gsif_get_structure_au(int fd, struct gsi_sound *new_snd) {
	int  file_size;
	int  data_size, data_start;
	int  channels = 0, type = -1, samplerate = 0;
	int  real_data_size;
	int  read_status = 0;

	file_size = lseek(fd, 0, SEEK_END);

	/* Position just after the 4 byte signature we checked to get here: */
	lseek(fd, new_snd->offset + 0+4, SEEK_SET);
	data_start = gsif_r_msb_int32(fd, &read_status);
	if (data_start < 24) return GSIF_ERROR_FILE_CORRUPT;

DEB(DPRT("libgsif: AU data start = %d, read_status=%d\n", data_start, read_status));
	/* In some files 'size' is junk:
	 * marvin.au (hhgttg): size=0xFFFFFFFF (=-1, see SNDSoundStruct)
	 * badkharma.au (the young ones): only 30K is used, but filesize=82K
	 * Hence the following approach:
	 */
	real_data_size = file_size - data_start;
	if (real_data_size < 0) {
		return GSIF_ERROR_FILE_CORRUPT;
	}

	data_size = gsif_r_msb_int32(fd, &read_status);
	if (data_size <= 0 || data_size > real_data_size) {
		data_size = real_data_size;
	}

	type = gsif_r_msb_int32(fd, &read_status);
	if (read_status) {
		return GSIF_ERROR_FILE_CORRUPT;
	}

	if (type < 1 || type > AU_FORMAT_ALAW_8) type = -1;
	else
	if (type == AU_FORMAT_4BIT_G721) type = GSI_4BIT_ADPCM;
	else
	if (type == AU_FORMAT_3BIT_G723) type = GSI_3BIT_ADPCM;
	else
	if (type == AU_FORMAT_5BIT_G723) type = GSI_5BIT_ADPCM;
	else
	if (type == AU_FORMAT_ALAW_8) type = GSI_8BIT_ALAW;

	if (type == -1) {
		return GSIF_ERROR_UNHANDLED_FILE_TYPE;
	}

	samplerate = gsif_r_msb_int32(fd, &read_status);
	channels = gsif_r_msb_int32(fd, &read_status);
	if (read_status || samplerate < 1 || channels < 1) {
		return GSIF_ERROR_FILE_CORRUPT;
	}

	gsif_add_block(new_snd, samplerate, channels, type, data_start, data_size);
	gsif_make_linear_block_list(new_snd);

	lseek(fd, new_snd->offset + 24, SEEK_SET);
	new_snd->comment = calloc(1, data_start-24+1);
	if (new_snd->comment == NULL) {
		return GSIF_ERROR_OUT_OF_MEMORY;
	}

	read_status |= (read(fd, new_snd->comment, data_start-24) != data_start-24);
	new_snd->comment[ data_start-24 ] = 0;
	new_snd->file_format = GSI_FILE_AU;

	return GSIF_OK;
}


/*      EOF sample_get_structure_au.c */
