/* GSI - general sound interface
 *
 * sample_get_structure_voc.c - gsif library, sample structure loading
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "sample_format.h"
#include "../io.h"
#include "../log.h"


/*---------------------------------------------------------------------------*
 *      Return GSIF_OK if recognized as a valid VOC (and one that the lib
 *      knows how to handle!)
 *      (VOC = creative labs format, used in esp. duke nukem 3d)
 *      return codes: GSIF_OK/GSIF_ERROR_* from gsif.h
 *---------------------------------------------------------------------------*/
int gsif_get_structure_voc(int fd, struct gsi_sound *new_snd) {
	int  format_status = GSIF_ERROR_STATUS_UNKNOWN;
	int  size, data_start;
	int  channels = 0, type = -1, samplerate = 0;

	int  sig;
	char buf[VOC_ID_SIZE+1];
	int  voc_extended = False;
	int  voc_compression = -1, voc_sample_rate = -1;
	int  voc_tracks = 1;

	char *voc_comment = NULL;
	int  found_data = False;

	int  n; /* chunk_size */
	int  voc_start; /* i.e. header size */
	int  voc_version, voc_format_2, voc_bps, gsi_format;
	int  file_status = 0;


	DEB1(DPRT("   -> sig = MAGIC_VOC\n"));
	// Position at the start of the 4 byte signature we checked to get here:
	lseek(fd, new_snd->offset /*data_offset*/, SEEK_SET);
	if (read(fd, buf, VOC_ID_SIZE) < VOC_ID_SIZE) {
		return format_status;
	}
	buf[VOC_ID_SIZE] = 0;

	DEB1(DPRT("   -> VOC ID = %s\n", buf));
	if (strncmp(buf, VOC_ID, VOC_ID_SIZE))
		return format_status;

	DEB1(DPRT("   -> VOC ID OK\n"));
	voc_start = gsif_r_lsb_int16(fd, &file_status);
	voc_version = gsif_r_lsb_int16(fd, &file_status);
	if (gsif_r_lsb_int16(fd, &file_status) != ((~voc_version + 0x1234) & 0xffff) ) {
		DEB1(DPRT(" VOC check failed\n"));
		return format_status;
	}

//fprintf(stderr, "seek: %ld\n", lseek(fd, 0, SEEK_CUR));
	for (;;) {
		int  tmp;

		DEB2(DPRT("   -> next voc block; file pos = %ld\n", (long) lseek(fd, 0, SEEK_CUR)));
		file_status = 0;
		sig = gsif_r_byte(fd, &file_status);

		if (file_status) {
			log_message(0, "file error");
			return GSIF_ERROR_FILE_CORRUPT;
		}
		DEB2(DPRT("   -> voc: sig %d\n", sig));
		n = gsif_r_lsb_int24(fd, &file_status); /* chunk len */
//fprintf(stderr, "seek: %ld\n", lseek(fd, 0, SEEK_CUR));

		switch(sig) {
		case VOC_TERMINATOR:
			DEB2(DPRT("   -> VOC_TERMINATOR\n"));
			found_data = True;
			break;

		case VOC_DATA:
			DEB2(DPRT("   -> VOC_DATA\n"));
			//size = gsif_r_lsb_int24(fd, &file_status);
			size = n;
			size -= 2;

			if (!voc_extended) {
				voc_sample_rate = 1000000 / (256 - gsif_r_byte(fd, &file_status));
				voc_compression = gsif_r_byte(fd, &file_status);
			}
			else {
				lseek(fd, 2, SEEK_CUR);
			}

			data_start = lseek(fd, 0, SEEK_CUR);
			lseek(fd, size, SEEK_CUR);

			DEB2(DPRT("   -> voc compression: %d\n", voc_compression));
			if (voc_compression == 0) {
				type = GSI_8BIT;
			}
			else {
				log_message(0, "voc compression %d not yet handled", voc_compression);
				gsif_free_sound(new_snd);
				//return NULL;
				//break; //should free allocated blocks here.
				return GSIF_ERROR_STATUS_UNKNOWN;
			}

			samplerate = voc_sample_rate;
			channels = voc_tracks;
			if (samplerate < 1 || channels < 1) {
				format_status = GSIF_ERROR_FILE_CORRUPT; break;
			}

			gsif_add_block(new_snd, samplerate, channels, type, data_start, size);
			gsif_make_linear_block_list(new_snd);

			DEB2(DPRT("   -> VOC_DATA: format=%d, channels=%d, rate=%d\n", voc_compression, voc_tracks, voc_sample_rate));
			break;

		case VOC_TEXT:
			DEB2(DPRT("   -> VOC_TEXT\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			if (!(voc_comment = (char *) malloc(n+1))) {
				return GSIF_ERROR_OUT_OF_MEMORY;
			}
			if (read(fd, voc_comment, n) < n) {
				log_message(0, "EOF in voc comment");
				return GSIF_ERROR_FILE_CORRUPT;
			}
			voc_comment[n] = 0;
			new_snd->comment = voc_comment;
			break;

		case VOC_EXTENDED:
			DEB2(DPRT("   -> VOC_EXTENDED\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);

			voc_extended = True; /* flag */
			tmp = gsif_r_byte(fd, &file_status);
			tmp |= gsif_r_byte(fd, &file_status) << 8;
			voc_sample_rate = 256000000 / (65536 - tmp);
			voc_compression = gsif_r_byte(fd, &file_status);

			voc_tracks = gsif_r_byte(fd, &file_status) +1;
			if (voc_tracks > 2) {
				log_message(0, "VOC tracks > 2");
				return GSIF_ERROR_FILE_CORRUPT;
//X				gsif_free_sound(new_snd);
			}

			voc_sample_rate /= voc_tracks;
			break;

		case VOC_SILENCE:
			DEB2(DPRT("   -> VOC_SILENCE ignored\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			file_status |= (lseek(fd, n, SEEK_CUR) == -1);
			break;

			/* following are all ignored for now */
		case VOC_CONTINUE:
			DEB2(DPRT("   -> VOC_CONTINUE ignored\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			file_status |= (lseek(fd, n, SEEK_CUR) == -1);
			break;

		case VOC_MARKER:
			DEB2(DPRT("   -> VOC_MARKER ignored\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			file_status |= (lseek(fd, n, SEEK_CUR) == -1);
			break;

		case VOC_REPEAT:
			DEB2(DPRT("   -> VOC_REPEAT ignored\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			file_status |= (lseek(fd, n, SEEK_CUR) == -1);
			break;

		case VOC_REPEAT_END:
			DEB2(DPRT("   -> VOC_REPEAT_END ignored\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			file_status |= (lseek(fd, n, SEEK_CUR) == -1);
			break;

		case VOC_DATA_2:
			DEB2(DPRT("   -> VOC_DATA_2\n"));
			//n = gsif_r_lsb_int24(fd, &file_status);
			size = n-12;
//fprintf(stderr, "seek: %ld\n", lseek(fd, 0, SEEK_CUR));

			voc_sample_rate = gsif_r_lsb_int32(fd, &file_status);
			voc_bps = gsif_r_byte(fd, &file_status);
			voc_tracks = gsif_r_byte(fd, &file_status);
			voc_format_2 = gsif_r_lsb_int16(fd, &file_status);
			switch(voc_format_2) {
			case 0: gsi_format = GSI_8BIT; break;
			case 4: gsi_format = GSI_16BIT_LE; break;
			case 6: gsi_format = GSI_8BIT_ALAW; break;
			case 7: gsi_format = GSI_8BIT_MULAW; break;
			default:
				DEB2(DPRT("   -> VOC format %d in block type 9 not handled\n", voc_format_2));
				gsi_format = GSI_8BIT; /* tmp */
			}
			file_status |= (lseek(fd, 4, SEEK_CUR) == -1);

			data_start = lseek(fd, 0, SEEK_CUR);
			file_status |= (lseek(fd, size, SEEK_CUR) == -1);

			type = gsi_format;

			samplerate = voc_sample_rate;
			channels = voc_tracks;
			if (file_status || samplerate < 1 || channels < 1) {
				format_status = GSIF_ERROR_FILE_CORRUPT;
				break;
			}

			gsif_add_block(new_snd, samplerate, channels, type, data_start, size);
			gsif_make_linear_block_list(new_snd);
			//gsif_load_block_data(fd, new_snd, /*0*/new_snd->blocks-1);

			DEB2(DPRT("   -> VOC_DATA_2: format=%d, channels=%d, rate=%d\n", voc_format_2, voc_tracks, voc_sample_rate));
			break;

		default:
			log_message(0, "unknown VOC block type %d", sig);
			file_status = 1; /* should not abuse file status for this ? */
			lseek(fd, n, SEEK_CUR);
			break;
		}
		if (found_data || file_status) break;
	}

	new_snd->file_format = GSI_FILE_VOC;
	format_status = GSIF_OK;

	return format_status;
}


/*      EOF sample_get_structure_voc.c */
