/* GSI - general sound interface
 *
 * convert_mus_to_midi.c: gsif library; convert a MUS file for internal use
 * (to MIDI).
 *
 * Lower level API dependencies: standard file stuff
 *
 *
 * Copyright (C) W.H.Scholten 1998 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"
#include "../include/gsif_midi.h"

#include "../io.h"
#include "midi_utils.h"


/*---------------------------------------------------------------------------*
 *      MUS -> MIDI conversion.
 *
 *      Returns 0 if OK, -1 if failed (e.g. out of memory).
 *---------------------------------------------------------------------------*/

/* MUS controller -> MIDI controller */
static byte mus_to_midi_ctrl[] = {
	/*  0 instrument--event (patch change) => NONE, not a controller but
		an event in MIDI terms  */ 0xFF,
	/*  1 bank select           */ MIDICTL_BANK_SELECT       /*   0 */,
	/*  2 modulation            */ MIDICTL_MODWHEEL          /*   1 */,
	/*  3 volume                */ MIDICTL_MAIN_VOLUME       /*   7 */,
	/*  4 pan                   */ MIDICTL_PAN               /*  10 */,
	/*  5 expression            */ MIDICTL_EXPRESSION        /*  11 */,
	/*  6 reverb depth          */ MIDICTL_EXT_EFF_DEPTH     /*  91 */,
	/*  7 chorus depth          */ MIDICTL_CHORUS_DEPTH      /*  93 */,
	/*  8 sustain pedal         */ MIDICTL_SUSTAIN           /*  64 */,
	/*  9 soft pedal            */ MIDICTL_SOFT_PEDAL        /*  67 */,
	/* 10 all sounds off        */ MIDICTL_SOUNDS_OFF        /* 120 */,
	/* 11 all notes off         */ MIDICTL_ALL_NOTES_OFF     /* 123 */,
	/* 12 mono                  */ MIDICTL_MONOPHONIC_MODE   /* 126 */,
	/* 13 poly                  */ MIDICTL_POLYPHONIC_MODE   /* 127 */,
	/* 14 reset all controllers */ MIDICTL_RESET_CONTROLLERS /* 121 */
};


#ifdef DEBUG
static char *mus_ctrl[] = {
	/*  0 */ "instrument change",
	/*  1 */ "bank select",
	/*  2 */ "modulation",
	/*  3 */ "volume",
	/*  4 */ "pan",
	/*  5 */ "expression",
	/*  6 */ "reverb",
	/*  7 */ "chorus",
	/*  8 */ "sustain",
	/*  9 */ "soft pedal",
	/* 10 */ "all sounds off",
	/* 11 */ "all notes off",
	/* 12 */ "mono",
	/* 13 */ "poly",
	/* 14 */ "reset all controllers",
};
#endif


int mus_to_midi(struct gsi_song *sng)
{
	int   i;
	byte *midi_data;
	int   midi_data_size;
	byte *midi_ptr;
	int   chn_velocity[16];

	int   total_time = 0;

	byte *new_midi_data; /* for realloc */
	int   new_midi_data_size;

	unsigned int  mus_time = 0;


	DEB2(DPRT("MUS->MIDI   sng->data_length=%d, malloc = %d \n", sng->data_length, sng->data_length *2));

	midi_data_size = sng->data_length *2+128; // estimate of the midi size.
	midi_data = (byte *) malloc(midi_data_size);
	if (midi_data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
	midi_ptr = midi_data;

	/* some tempo initializations:
	 * time sig e.g. 00 FF 58 04 04 02 60 08
	 */
	midi_data[0] = 0x00; /* dt = 0 ticks (vl) */
	midi_data[1] = MIDI_META_EVENT;
	midi_data[2] = MIDIMETA_TIME_SIGNATURE;
	midi_data[3] = 0x04; /* data = 4 bytes (vl) */
	midi_data[4] = 4;
	midi_data[5] = 2;
	midi_data[6] = 96;
	midi_data[7] = 8;

	/* tempo = 60 beats / minute => 1.0 s/beat
	 *  => 1E6 s/midi quarter note
	 */
	midi_data[8+0] = 0x00; /* dt = 0 ticks (vl) */
	midi_data[8+1] = MIDI_META_EVENT;
	midi_data[8+2] = MIDIMETA_SET_TEMPO;
	midi_data[8+3] = 0x03; /* data = 3 bytes (vl) */
	midi_data[8+4] = (1000000 >> 16);
	midi_data[8+5] = (1000000 >>  8) & 255;
	midi_data[8+6] = (1000000      ) & 255;

	midi_ptr += 15;

	/* XXX  I THINK 127 IS A BETTER INITIAL VELOCITY THAN 64 (bunny.mus) */
	for (i=0; i<16; i++) chn_velocity[i] = 127;


	for (i=0; i<sng->data_length; ) {
		int  done = False;
		int  event;
		int  last, channel, cmd;
		int  note, velocity, pitch_wheel, sys_event;


		if (midi_ptr - midi_data > midi_data_size - 128) {
			/* assuming a realloc that preserves the original data: */
			new_midi_data = realloc(midi_data, midi_data_size + 16384);
			if (new_midi_data == NULL) { /* realloc failed */
				DEB2(DPRT("MUS->MIDI   realloc failed\n"));
				free(midi_data);
				return GSIF_ERROR_OUT_OF_MEMORY;
			}
			midi_data_size = midi_data_size + 16384;
			midi_ptr += (new_midi_data-midi_data);
			midi_data = new_midi_data;
		}
		DEB3(DPRT("MUS->MIDI   position=%d\n", i));
		event = sng->data[i++];
		last = event & 0x80;
		channel = event & 0x0F;
		cmd = (event >> 4) & 0x07;
		DEB3(DPRT("MUS->MIDI   event=0x%02x, last=%d, channel=%d, cmd=%d\n", event, last, channel, cmd));

		/* change the channel numbers so that 9<->15 are swapped (other
		 * changes are irrelevant): */
		//channel ^= (9^15);

		/* to compare with mus2midi output: */
		if (channel == 15) channel=9; else if (channel == 9) channel=15;

		/* first write variable-length timing number: */
		midi_ptr += gsif_midi_write_vl(midi_ptr, mus_time);
		mus_time = 0;

		switch(cmd) {
		case 0: /* release note */
			note = sng->data[i++];

			// write MIDI stop note event:
			*midi_ptr++ = MIDI_NOTE_OFF | channel;
			*midi_ptr++ = note;

			/* what should the velocity be ?
			 * mus2midi:  0x7f
			 * musserver: last_volume
			 * qmus2mid:  0
			 */
			*midi_ptr++ = 0;
			break;

		case 1: /* play note */
			note = sng->data[i++];
			if (note & 0x80) {
				note &= 0x7F;
				velocity = sng->data[i++];
				chn_velocity[channel] = velocity;
			}
			else velocity = chn_velocity[channel];
			//chn_instrument[channel] = note;

			/* write MIDI start note event: */
			*midi_ptr++ = MIDI_NOTE_ON | channel;
			*midi_ptr++ = note;

			/* what should the velocity be ?
			 * mus2midi:  0x7F (?)
			 * musserver: last_volume
			 * qmus2mid:  velocity ('volume' in the MUS document)
			 */
			*midi_ptr++ = velocity;
			break;

		case 2: /* pitch wheel */
			pitch_wheel = sng->data[i++];

			/* write MIDI event: */
			*midi_ptr++ = MIDI_PITCH_BEND | channel;
			/* the MUS format uses all 8 bits for pitchwheel, hence: */
			*midi_ptr++ = (pitch_wheel & 1) << 6;
			*midi_ptr++ = pitch_wheel >> 1;
			break;

		case 3: /* system event */
		{
			sys_event = sng->data[i++];
			if (sys_event >= 10 && sys_event <= 14) {
				DEB3(DPRT("%s\n", mus_ctrl[sys_event]));
				/* write MIDI controller */
				*midi_ptr++ = MIDI_CTL_CHANGE | channel;
				*midi_ptr++ = mus_to_midi_ctrl[sys_event];
				if (sys_event == 12) {
					*midi_ptr++ = 16;
					/* should be no. of mus channels used */
				}
				else {
					*midi_ptr++ = 0;
				}
			}
			else {
				DEB3(DPRT("MUS->MIDI   unknown system event %d\n", sys_event));
				//return -1; ??
			}
			break;
		}

		case 4: /* change controller */
		{
			int  ctrl, ctrl_val;

			ctrl = sng->data[i++];
			ctrl_val= sng->data[i++];

			switch (ctrl) {
			case 0: /* change instrument/program */
				//chn_instrument[channel] = ctrl_val;
				/* write MIDI event, not controller! */
				*midi_ptr++ = MIDI_PGM_CHANGE | channel;
				*midi_ptr++ = ctrl_val;
				break;

			case 4: /* change pan */
				//chn_pan[channel] = ctrl_val - 0x40; // giving -64 .. 63
				DEB3(DPRT("%s (%+d)\n", mus_ctrl[ctrl], ctrl_val - 0x40));
				*midi_ptr++ = MIDI_CTL_CHANGE | channel;
				*midi_ptr++ = mus_to_midi_ctrl[ctrl];
				*midi_ptr++ = ctrl_val;
				break;

			default: /* other controllers */
				if (ctrl < 10) {
					DEB3(DPRT("%s (%d)\n", mus_ctrl[ctrl], ctrl_val));
					// write MIDI controller
					*midi_ptr++ = MIDI_CTL_CHANGE | channel;
					*midi_ptr++ = mus_to_midi_ctrl[ctrl];
					*midi_ptr++ = ctrl_val;
				}
				else {
					DEB3(DPRT("MUS->MIDI   Controller change for unknown controller (set %d (to %d)) at position %d\n", ctrl, ctrl_val, i));
				}
				break;
			}
			break;
		}

		case 6: /* score end */
			done = True;
			break;

		case 5: /* ?? */
		case 7: /* ?? */
			DEB3(DPRT("MUS->MIDI   unknown MUS event type %d at position %d\n", 7, i));
			//return -1;
			break;
		} /* switch */

		if (done) break;
		if (last) {
			byte  b;

			DEB3(DPRT("MUS->MIDI   reading time at position %d\n", i));
			mus_time = 0;
			do {
				b = sng->data[i++];
				if (i >= sng->data_length) break;
				mus_time <<= 7;
				mus_time += b & 0x7F;
			} while (b & 0x80);
			DEB3(DPRT("MUS->MIDI   mus_time=%d\n", mus_time));

			total_time += mus_time;
		} /* last */
	}

	/* write end of track event (dtime was already written above!) */
	*midi_ptr++ = 0xFF;
	*midi_ptr++ = 0x2F;
	*midi_ptr++ = 0x00;

DEB(DPRT("AA midi_data_size=%d\n", midi_data_size));
	/* free unused space:
	 * assuming a realloc that preserves the original data: */
	new_midi_data_size = midi_ptr - midi_data;
DEB(DPRT("AA new_midi_data_size=%d\n", new_midi_data_size));
	new_midi_data = realloc(midi_data, new_midi_data_size);
	if (new_midi_data != NULL) {
		/* Realloc OK (otherwise we just keep the old data). */
		midi_ptr += (new_midi_data-midi_data);
		midi_data = new_midi_data;
		midi_data_size = new_midi_data_size;
	}
DEB(DPRT("BB\n"));
DEB(DPRT("BB midi_data_size=%d\n", midi_data_size));

	sng->track = malloc(sizeof(void *) * 1);
	if (sng->track == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

	sng->tracks = 1;
	sng->track[0] = calloc(1, sizeof(struct gsi_track));
	if (sng->track[0] == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

	sng->track[0]->data = midi_data;
	sng->track[0]->data_length = midi_ptr - midi_data;
	// offset =0

	DEB2(DPRT("MUS->MIDI   sng->track[0]->data_length = %d\n", midi_ptr - midi_data));
	DEB2(DPRT("MUS->MIDI   mus total_time = %d ticks (%5.2f s)\n", total_time, total_time / 140.0));
	sng->midi_format = 1;
	sng->midi_division = 140;
	/* According to mus_format.doc: doom/doom2/heretic: 140, raptor: 70
	 * Other values will be selectable by setting the playing speed. */

	return GSIF_OK;
}


/*      EOF convert_mus_to_midi.c */
