/* GSI - general sound interface
 *
 * load_hmp.c: gsif library; load a hmp file and convert for internal use.
 *
 * Lower level API dependencies: standard file functions
 *
 *
 * Copyright (C) W.H.Scholten 1998-2000
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"
#include "../include/gsif_midi.h"

#include "../io.h"
#include "../file.h" /* For SIG() */



/*---------------------------------------------------------------------------*
 *      Load a midi in the file at the given offset.
 *
 *      Returns number of tracks (0 if invalid midi file).
 *---------------------------------------------------------------------------*/
//int gsif_load_song(char *filename, int offset, struct gsi_song *new_sng) {

int load_song_hmp(int fd, struct gsi_song *sng)
{
	/* hmp/hmq format (descent, abuse, slipstream 5000...), see
	 * '_refdoc/Hmp_format' for a description of the format.
	 */
	int  i;
	int  file_status = 0;

	char id_buffer[32];
	int  hmp_format = -1;
	enum { _HMP_ORIGINAL, _HMP_013195, _HMP_061595 };
	int  midi_division, num_tracks, track_len;


	lseek(fd, sng->offset, SEEK_SET);
	read(fd, id_buffer, 0x20); id_buffer[ 0x1F ] = 0;

	if (!strcmp(id_buffer, "HMIMIDIP")) hmp_format = _HMP_ORIGINAL;
	else if (!strcmp(id_buffer, "HMIMIDIP013195")) hmp_format = _HMP_013195;
	else if (!strcmp(id_buffer, "HMI-MIDISONG061595")) hmp_format = _HMP_061595;

	if (hmp_format >= 0) {
		int  track, chunk, chunks;
		int  hmp_header_size = 0;
		int  chunkno_offset;

		if (hmp_format == _HMP_061595) {
			DEB1(DPRT("L   Found 16-5-95 HMI\n"));
			/* hmp_header_size will be determined later */
			chunkno_offset = 0xe4;
		}
		else if (hmp_format == _HMP_013195) {
			hmp_header_size = 0x388;
			/* To crash-proof midi_decode use a bogus header size: */
			// hmp_header_size = 0x308;
			chunkno_offset = 0x30;
			DEB1(DPRT("L   Found 31-1-95 HMP\n"));
		}
		else {
			hmp_header_size = 0x308;
			chunkno_offset = 0x30;
			DEB1(DPRT("L   Found ORIGINAL HMP\n"));
		}
		sng->file_format = GSI_FILE_HMP;

		lseek(fd, sng->offset + chunkno_offset /*0x30*/, SEEK_SET);
		chunks = gsif_r_lsb_int32(fd, &file_status);

		if (chunks <= 0 || chunks > 255) {
			/* Must be an error. */
			DEB2(DPRT("HMP  num chunks <=0!\n"));
//XXX			gsif_free_song(sng);
//XXX			close(fd);
			return GSIF_ERROR_FILE_CORRUPT;
		}
		DEB2(DPRT("L   hmp chunks=%d\n", chunks));

		if (hmp_format == _HMP_061595) {
			lseek(fd, sng->offset + 0xD2, SEEK_SET);
			midi_division = gsif_r_lsb_int16(fd, &file_status);
		}
		else {
			//unknown = gsif_r_lsb_int32(fd, &file_status);
			lseek(fd, sng->offset + 0x38, SEEK_SET);
			midi_division = gsif_r_lsb_int32(fd, &file_status);
		}

		num_tracks = chunks;
		sng->track = malloc(sizeof(void *) * num_tracks);
		if (sng->track == NULL) return GSIF_ERROR_OUT_OF_MEMORY;


		/* with mmap this could be much cleaner. */
		for (i=0; i<chunks; i++) {
			sng->track[i] = calloc(1, sizeof(struct gsi_track));
			if (sng->track[i] == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
		}

		/* Determine header size & track sizes for the _HMP_061595 fornat: */
		if (hmp_format == _HMP_061595) {
			int  prev_offset;
			int  file_length = lseek(fd, 0, SEEK_END) - sng->offset;

			lseek(fd, sng->offset + 0x172, SEEK_SET);
			hmp_header_size = prev_offset = gsif_r_lsb_int32(fd, &file_status);

			for (i=0; i<chunks; i++) {
				int  next_offset, track_internal_offset;

				lseek(fd, sng->offset + 0x172 + 4*(i+1), SEEK_SET);
				next_offset = gsif_r_lsb_int32(fd, &file_status);

				if (i == chunks-1) next_offset = file_length;

				lseek(fd, sng->offset + prev_offset + 0x57, SEEK_SET);
				track_internal_offset = gsif_r_lsb_int32(fd, &file_status);

				sng->track[i]->data_length = next_offset - prev_offset - track_internal_offset;
				//DEB(DPRT("trk off=0x%x, internal offset=%x, len=0x%x\n", prev_offset, track_internal_offset, sng->track[i]->data_length));
				prev_offset = next_offset;
			}
		}

		lseek(fd, sng->offset + hmp_header_size, SEEK_SET);

		for (i=0; i<chunks; i++) {
			if (hmp_format < _HMP_061595) {
				chunk = gsif_r_lsb_int32(fd, &file_status);

				track_len = gsif_r_lsb_int32(fd, &file_status) - 12;

#define MAX_SONG_MALLOC 1*1024*1024 /* 1 MB per track */

				if (track_len < 0 || track_len > MAX_SONG_MALLOC) {
					/* Must be an error. */
					DEB1(DPRT("HMP  too big or negative malloc\n"));
					return GSIF_ERROR_FILE_CORRUPT;
				}
				track = gsif_r_lsb_int32(fd, &file_status);

				DEB2(DPRT("\nL   hmp chunk=%d\n", chunk));
				DEB2(DPRT("L   hmp track_len=%d\n", track_len));
				DEB2(DPRT("L   hmp track=%d\n", track));

				sng->track[i]->data_length = track_len;
			}
			else {
				int  track_internal_offset;

				lseek(fd, 0x57, SEEK_CUR);
				track_internal_offset = gsif_r_lsb_int32(fd, &file_status);

				lseek(fd, track_internal_offset - 0x57 -4, SEEK_CUR);
			}

			/* For track 0 add tempo setting to 60 beats/s: */
			if (i == 0) {
				int  tmp_offset = 8+7;

				sng->track[i]->data = malloc(sng->track[i]->data_length + tmp_offset);
				if (sng->track[i]->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

				/* time sig e.g. 00 FF 58 04 04 02 60 08 but with hmp (little
				 * endian) vl numbers!
				 */
				sng->track[i]->data[0] = 0x80; /* dt = 0 ticks (hmp vl) */
				sng->track[i]->data[1] = MIDI_META_EVENT;
				sng->track[i]->data[2] = MIDIMETA_TIME_SIGNATURE;
				sng->track[i]->data[3] = 0x84; /* data = 4 bytes (hmp vl) */
				sng->track[i]->data[4] = 4;
				sng->track[i]->data[5] = 2;
				sng->track[i]->data[6] = 96;
				sng->track[i]->data[7] = 8;

				/* tempo = 60 beats / minute => 1.0 s/beat
				 *  => 1E6 s/midi quarter note
				 */
				sng->track[i]->data[8+0] = 0x80; /* dt = 0 ticks (hmp vl) */
				sng->track[i]->data[8+1] = MIDI_META_EVENT;
				sng->track[i]->data[8+2] = MIDIMETA_SET_TEMPO;
				sng->track[i]->data[8+3] = 0x83; /* data = 3 bytes (hmp vl) */
				sng->track[i]->data[8+4] = (1000000 >> 16);
				sng->track[i]->data[8+5] = (1000000 >>  8) & 255;
				sng->track[i]->data[8+6] = (1000000      ) & 255;

				read(fd, sng->track[i]->data + tmp_offset, sng->track[i]->data_length);
				sng->track[i]->data_length += tmp_offset;
			}
			else {
				sng->track[i]->data = malloc(sng->track[i]->data_length);
				if (sng->track[i]->data == NULL)
					return GSIF_ERROR_OUT_OF_MEMORY;

				read(fd, sng->track[i]->data, sng->track[i]->data_length);
			}
		}

		sng->tracks = chunks;
		if (hmp_format == _HMP_061595)
			sng->midi_format = +1;
		else
			sng->midi_format = -1; /* negative = LSB vl numbers */
		sng->midi_division = midi_division;
	}

	return GSIF_OK;
}


/*      EOF load_hmp.c */
