/* GSI - general sound interface
 *
 * load_midi.c: gsif library; load a midi file and convert for internal use.
 *
 * Lower level API dependencies: standard file functions
 *
 *
 * Copyright (C) W.H.Scholten 1998 - 2000
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"

#include "../io.h"
#include "../file.h" /* For SIG() */



/*---------------------------------------------------------------------------*
 *      Load a midi in the file at the given offset.
 *
 *      Returns number of tracks (0 if invalid midi file).
 *---------------------------------------------------------------------------*/
//int gsif_load_song(char *filename, int offset, struct gsi_song *new_sng) {

int load_song_midi(int fd, struct gsi_song *sng)
{
	int  header_len;
	int  sig;

	int  num_tracks, track_offset, track_len;
	int  midi_format;
	int  midi_division;
	int  data_length;

	int  i;
	int  file_status = 0;

	DEB1(DPRT("L    Found MIDI\n"));
	sng->file_format = GSI_FILE_MIDI;

	header_len = gsif_r_msb_int32(fd, &file_status); // the header track
	midi_format = gsif_r_msb_int16(fd, &file_status);
	num_tracks = gsif_r_msb_int16(fd, &file_status);
	midi_division = gsif_r_msb_int16(fd, &file_status);
	DEB2(DPRT("L    -> header_len=%d\n", header_len));
	DEB2(DPRT("L    -> midi_format=%d\n", midi_format));
	DEB2(DPRT("L    -> num_tracks=%d\n", num_tracks));
	DEB2(DPRT("L    -> midi_division=%d\n", midi_division));
	lseek(fd, header_len-6, SEEK_CUR); // to next chunk

	if (sng->track != NULL)
		free(sng->track);
	sng->track = calloc(1, sizeof(void *) * num_tracks);
	if (sng->track == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

	track_offset = sng->offset; // temp
	for (i = 0; i < num_tracks; i++) {
		for (;;) {
			sig = gsif_r_msb_int32(fd, &file_status);
DEB(DPRT("track %d: sig file_status=%d, sig = 0x%x\n", i, file_status, sig));
			track_len = gsif_r_msb_int32(fd, &file_status);
DEB(DPRT("track %d: track_len file_status=%d, track_len = %d\n", i, file_status, track_len));

			if (file_status || track_len < 0) return GSIF_ERROR_FILE_CORRUPT;
			if (sig == SIG('M','T','r','k')) break;
			DEB1(DPRT("Skipping unknown MIDI chunk type %d (%x), track_len = %d\n", sig, sig, track_len));
			/* skip unknown chunck type */
			lseek(fd, track_len, SEEK_CUR);
		}

		/* sanity check */
		//track_len = min(track_len, sng->data +sng->data_length-midi_ptr);

		/* add a new midi track struct: */
		sng->track[i] = calloc(1, sizeof(struct gsi_track));
		if (sng->track[i] == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

		sng->track[i]->data_length = track_len;
		sng->track[i]->data = malloc(track_len);
		if (sng->track[i]->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;
		read(fd, sng->track[i]->data, track_len);
	}
	num_tracks = i;
	data_length= lseek(fd, 0, SEEK_CUR);

	/* This section is really obsolete (only wastes memory but still): */
	sng->data = malloc(data_length);
	if (sng->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

	lseek(fd, sng->offset, SEEK_SET);
	sng->data_length = data_length;
	read(fd, sng->data, sng->data_length);

	sng->tracks = num_tracks;
	sng->midi_format = midi_format;
	sng->midi_division = midi_division;

	return GSIF_OK;
}


/*      EOF load_midi.c */
