/* GSI - general sound interface
 *
 * load_song.c: gsif library; load a midi/hmp/mus file and convert for internal
 * use.
 *
 * Lower level API dependencies: standard file functions
 *
 *
 * Copyright (C) W.H.Scholten 1998 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________INCLUDES___________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"
#include <gsi/gsi_midi_names.h>

#include "../io.h"
#include "../file.h" /* For SIG() & _FILE* defs */


/*__________________PROTOTYPES___________________*/
extern int load_song_hmp (int fd, struct gsi_song *sng);
extern int load_song_midi (int fd, struct gsi_song *sng);
extern int load_song_mus (int fd, struct gsi_song *sng);


/*---------------------------------------------------------------------------*
 *      Load a midi in the file at the given offset.
 *
 *      Returns number of tracks (0 if invalid midi file).
 *---------------------------------------------------------------------------*/
//int gsif_load_song(char *filename, int offset, struct gsi_song *new_sng) {

struct gsi_song *gsif_load_song(char *filename, int offset, int *status)
{
	struct stat file_info;
	struct gsi_song *sng;
	int  fd;
	int  sig;
	int  format_status = GSIF_ERROR_STATUS_UNKNOWN;
	int  file_status = 0;

	DEB1(DPRT("GSIF ("__FILE__"): gsif_load_song (%s)\n", filename));
	if (filename == NULL) {
		*status = GSIF_ERROR_NO_SUCH_FILE;
		return NULL;
	}

	if (stat(filename, &file_info) == -1 || (file_info.st_mode & S_IFMT ) != S_IFREG ) {
		DEB1(DPRT("GSIF ->  %s not a regular file\n", filename));
		*status = GSIF_ERROR_CANT_OPEN_FILE;
		return NULL;
	}

	fd = open(filename, O_RDONLY);
	if (fd <= 0) {
		*status = GSIF_ERROR_CANT_OPEN_FILE;
		return NULL;
	}

	sng = (struct gsi_song *) calloc(1, sizeof(struct gsi_song));
	if (sng == NULL) {
		*status = GSIF_ERROR_OUT_OF_MEMORY;
		return NULL;
	}

	sng->filename = strdup(filename);
	if (sng->filename == NULL) {
		free(sng);
		*status = GSIF_ERROR_OUT_OF_MEMORY;
		return NULL;
	}

	sng->offset = offset;

	lseek(fd, offset, SEEK_SET);
	sig = gsif_r_msb_int32(fd, &file_status);

	if (sig == SIG('M','T','h','d')) {
		format_status = load_song_midi(fd, sng);
	}
	if (sig == SIG('M','U','S',0x1a)) {
		format_status = load_song_mus(fd, sng);
		/* put mus -> midi subr call here? */
	}
	if ((sig &0xffffff00) == SIG('H','M','I', 0)) {
		format_status = load_song_hmp(fd, sng);
	}

	close(fd);

	if (format_status != GSIF_OK) {
		DEB1(DPRT("GSIF -> get file format of %s failed (err %d)\n", filename, format_status));
		gsif_free_song(sng);
		if (format_status == GSIF_ERROR_STATUS_UNKNOWN)
			format_status = GSIF_ERROR_UNHANDLED_FILE_TYPE;

		*status = format_status;
		return NULL;
	}
	DEB1(DPRT("GSIF -> #gsif_load_song#\n"));

	return sng;
}


// RENAME: gsif_song_to_tracks ? (only handles MIDI btw).
int gsif_convert_song_to_tracks(struct gsi_song *sng)
{
	int  sig;
	int  track, track_len;
	int  midi_format;
	int  num_tracks;
	int  midi_division;
	byte *midi_ptr;         /* pointer to current position in midi file */

	DEB1(DPRT("GSIF ("__FILE__"): gsif_convert_song_to_tracks\n"));
	midi_ptr = sng->data;
	if (midi_ptr == NULL) return GSIF_ERROR_NULL_DEREFERENCE;

	sig = msb_int32_at(midi_ptr);
	midi_ptr += 4;

	if (sig == SIG('M','T','h','d')) {
		DEB1(DPRT("GSIF    -> sig = MThd (MIDI)\n"));
		sng->file_format = GSI_FILE_MIDI;

		track_len = msb_int32_at(midi_ptr);
		midi_ptr += 4;
		DEB2(DPRT("GSIF    -> track_len=%d\n", track_len));

		midi_format = msb_int16_at(midi_ptr);
		midi_ptr += 2;
		DEB2(DPRT("GSIF    -> song_format=%d\n", midi_format));

		num_tracks = msb_int16_at(midi_ptr);
		midi_ptr += 2;
		DEB2(DPRT("GSIF    -> num_tracks=%d\n", num_tracks));

		midi_division = msb_int16_at(midi_ptr);
		midi_ptr += 2;
		DEB2(DPRT("GSIF    -> midi_division=%d\n", midi_division));
	}
	else {
		DEB1(DPRT("GSIF: unsupported file type\n"));
		return GSIF_ERROR_UNHANDLED_FILE_TYPE;
	}

	if (sng->track != NULL) free(sng->track);

	/* calloc in case we return without initializing all pointers (needed to
	 * be able to free allocated space). */
	sng->track = calloc(sizeof(void *) * num_tracks, 1);
	if (sng->track == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

	for (track = 0; track < num_tracks; track++) {
		/* we need to be prepared to encounter new block types according to
		 * the MIDI specification, so skip unknown block types. */
		for (;;) {
			sig = msb_int32_at(midi_ptr);
			midi_ptr += 4;
			track_len = msb_int32_at(midi_ptr);
			midi_ptr += 4;

			if (sig == SIG('M','T','r','k')) break;
			/* skip unknown chunck type */
			midi_ptr += track_len;
		}

		/* sanity check: */
		track_len = min(track_len, sng->data + sng->data_length - midi_ptr);

		/* add a new midi track struct: */
		sng->track[track] = calloc(1, sizeof(struct gsi_track));
		if (sng->track[track] == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

		/* alloc space for each track (could just point to the data (midiptr)
		 * but ... */
		sng->track[track]->data = malloc(track_len);
		if (sng->track[track]->data == NULL) return GSIF_ERROR_OUT_OF_MEMORY;

		memcpy(sng->track[track]->data, midi_ptr, track_len);
		sng->track[track]->data_length = track_len;
		midi_ptr += track_len;
	}
	num_tracks = track;

	sng->tracks = num_tracks;
	sng->midi_format = midi_format;
	sng->midi_division = midi_division;

	DEB1(DPRT("GSIF -> #gsif_convert_song_to_tracks#\n"));

	return num_tracks;
}


int  gsif_free_song (struct gsi_song *sng)
{
	int  i;

	if (sng != NULL) {
		if (sng->filename != NULL) free (sng->filename);
		if (sng->comment != NULL) free (sng->comment);
		if (sng->data != NULL) free (sng->data);
		if (sng->track != NULL) {
			for (i=0; i< sng->tracks; i++) {
				if (sng->track[i] != NULL) {
					if (sng->track[i]->data != NULL)
						free(sng->track[i]->data);
					free(sng->track[i]);
				}
			}
			free(sng->track);
		}
		free(sng);
		return 0;
	}
	return -1;
}


int  gsif_free_track (struct gsi_song *sng, int i)
{
	if (sng != NULL && sng->track != NULL) {
		if (sng->track[i] != NULL) {
			if (sng->track[i]->data != NULL) {
				free(sng->track[i]->data);
			}
			free(sng->track[i]);
			sng->track[i] = NULL;
		}
		return 0;
	}
	return -1;
}


void gsif_add_track (struct gsi_sound *sng, int data, int data_length) {
}

void gsif_make_linked_track_list (struct gsi_sound *sng) {
}

void gsif_make_track_list (struct gsi_sound *sng) {
}


/*      EOF song_load.c */
