/* GSI - general sound interface
 *
 * midi_decode.c: get MIDI events from MIDI data.
 *
 * Lower level API dependencies:
 *   - only the MIDI specifications are used.
 *
 *
 * Copyright (C) W.H.Scholten 1998 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


/*__________________  INCLUDES  __________________*/
//#include <gsi/gsif_config.h>
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>

#include <limits.h>

#include <gsi/whs_defines.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIF_"
#include <gsi/gsi_debug.h>

#include "../include/gsif.h"
#include "../include/gsif_midi.h"



/*---------------------------------------------------------------------------*
 *      Make song (playing) status to read events from it.
 *---------------------------------------------------------------------------*/
struct gsi_song_status *gsif_make_song_status(struct gsi_song *song) {
	struct gsi_song_status *ss;
	int  i;

	if (song == NULL) return NULL;

	ss = (struct gsi_song_status *) calloc(1, sizeof(struct gsi_song_status));
	if (ss == NULL) return NULL; /*GSIF_ERROR_OUT_OF_MEMORY;*/
	ss->tracks = song->tracks;
	ss->track = calloc(1, sizeof(struct gsi_track_status) * song->tracks);
	if (ss->track == NULL) {
		free(ss);
		return NULL; /*GSIF_ERROR_OUT_OF_MEMORY;*/
	}

	ss->midi_format = song->midi_format;

	//ss->reverb=?;
	//ss->chorus=?;
	for (i=0; i<song->tracks; i++) {
		/* copy data from the song: (remove?) */
		ss->track[i].data = song->track[i]->data;
		ss->track[i].data_length = song->track[i]->data_length;
	}
	ss->midi_division = song->midi_division;
	gsif_reset_song_status(ss);

	return ss;
}


/* free the structures, not the song data */
void gsif_free_song_status(struct gsi_song_status *ss) {
	if (ss != NULL) {
		if (ss->track != NULL) {
			free(ss->track);
		}
		free(ss);
	}
}


/*---------------------------------------------------------------------------*
 *      Reset file pointer etc to play/restart a song.
 *---------------------------------------------------------------------------*/
void gsif_reset_song_status(struct gsi_song_status *status) {
	int  i;

	/* reset track playing status: (with a lot of sanity checking ..) */
	if (status != NULL && status->track != NULL) {
		//status->current_ticks = 0;
		status->ticks = 0;
		status->last_ticks = 0;
		status->time = 0.0;
		//status->last_time = 0.0;
		status->last_synth_ticks = 0;
		status->song_end_time = -1; /* not yet known. */

		/* midi_division was copied from song struct */
		status->midi_tempo = 500000; /* default_tempo 120 beats/s = 0.5 s per beat */
		status->percussion_chn = 0x200;

		status->reverb = 0; // ??
		status->chorus = 0; // ??

		for (i=0; i<status->tracks; i++) {
			status->track[i].index = 0;
			status->track[i].running = -1;
			//status->track[i].ticks = 0;
			status->track[i].ticks = gsif_midi_read_vl( (status->midi_format >=0 ? True : False), &status->track[i]);
		}
	}
}


/*---------------------------------------------------------------------------*
 *      Read variable-length number (7 bits per byte, MSB first, LSB for HMP).
 *      Updates the track's index accordingly.
 *---------------------------------------------------------------------------*/
unsigned int gsif_midi_read_vl(int msb_midi_format, struct gsi_track_status *s)
{
	unsigned int  len = 0;
	byte c;

	if (msb_midi_format) {
		for (;;) {
			if (s->index >= s->data_length) break; /* sanity check */
			c = s->data[ (s->index)++ ];
			len = (len << 7) | (c & 0x7f);
			if (!(c & 0x80)) break;
		}
	}
	else { /* hmp LSB variant */
		int  i;
		for (i=0; ;i++) {
			if (s->index >= s->data_length) break; /* sanity check */
			c = s->data[ (s->index)++ ];
			len = len | (((int) (c & 0x7f)) << (7*i) );
			if ( (c & 0x80)) break;
		}
	}

	return len;
}


/*---------------------------------------------------------------------------*
 *      Get the next midi event. Use preferred_track>=0 to force a track,
 *      start_ticks>=0 to get the first event from the given ticks.
 *
 *      Returns 0 if OK, -1/-2 if an error occurred (NULL status, track data
 *      past end of track etc),  1 if the next event is later than max_ticks,
 *      and 2 if there are no more events.
 *---------------------------------------------------------------------------*/

/* length of MIDI commands (0 denotes a length given by a variable-length
 * number in the file):
 */
static int gsif_midi_cmd_length [ 16 ] = {
	0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 1, 1, 2, 0
};

int gsif_midi_next_event(int max_ticks, struct gsi_song_status *status, int preferred_track, int start_ticks, struct gsi_song_event *event)
{
	struct gsi_track_status *track_status;
	byte *trk_data;
	unsigned int  trk_length;
	unsigned int  index;
	unsigned int  trk_ticks;
	int  midi_division;

	unsigned int  earliest;
	int  earliest_track;

	byte cmd;
	int  cmd_length;
	byte *data;
	int  i;


	DEB2(DPRT("MIDI ("__FILE__"): *gsif_midi_next_event*\n"));
	if (status == NULL) {
		return GSIF_ERROR_NULL_DEREFERENCE;
	}

	midi_division = status->midi_division;
	earliest = UINT_MAX;

	for (i = 0; i < status->tracks; i++) {
		if (status->track[i].ticks < earliest) {
			earliest_track = i;
			earliest = status->track[i].ticks;
		}
	}

	event->ticks = earliest;
	if (earliest == UINT_MAX)
		return GSIFMIDI_STATUS_NO_MORE_EVENTS; /* no more events. */

	event->track = earliest_track;
	if (event->ticks > max_ticks) {
		/* no new event within time limit (the event has the ticks and track
		 * of the earliest that would be used). */
		return GSIFMIDI_STATUS_NO_EVENTS_WITHIN_LIMIT;
	}

	track_status = &status->track[ earliest_track ];

	/* track data: */
	trk_data   = track_status->data;
	trk_length = track_status->data_length;

	/* track status: */
	index      = track_status->index;
	cmd        = track_status->running;
	trk_ticks  = track_status->ticks;


	DEB2(DPRT("MIDI   -> trk_ticks = %d\n", (int) trk_ticks));

	/* Look at the midi data for instructions */
	if (index >= trk_length) return GSIF_ERROR_FILE_CORRUPT; /* track error */

	/* new command? (otherwise use running status) */
	if (trk_data[ index ] & 0x80) {
		cmd = trk_data[ index++ ];
	}

	/* MIDI_META_EVENT is predecoded a bit */
	if (cmd == MIDI_META_EVENT) {
		if (index >= trk_length) return GSIF_ERROR_FILE_CORRUPT; /* track error */
		cmd = trk_data[ index++ ];
	}
	/* meta commands are 0-7f so this works nicely */

	if (cmd > 0xf7)	{
		/* midi real-time message (currently ignored) */
		cmd_length = 0;
	}
	else {
		cmd_length = gsif_midi_cmd_length[ cmd >> 4 ];
		if (cmd_length == 0) { /* length=0 denotes vl-number */
			track_status->index = index;
			cmd_length = gsif_midi_read_vl( (status->midi_format >= 0 ? True : False), track_status);
			index = track_status->index;
		}
	}

	data = &(trk_data[ index ]);
	if (cmd == MIDIMETA_SET_TEMPO) {
		status->midi_tempo = ((data[0] << 16) | (data[1] << 8) | data[2]);
	}

	if (trk_ticks > status->ticks) {
		if (midi_division > 0) {
			/* this works if tempo is changed during the song. */
			status->time += ((trk_ticks - status->ticks) * (1.0e-6 * (double) status->midi_tempo)) /(double) (midi_division);
			status->last_ticks = status->ticks;
			status->ticks = trk_ticks;
		}
		else if (midi_division < 0) {
			status->time = ( (double) trk_ticks / ( (double) ((midi_division & 0xff00 >> 8) * (midi_division & 0xff)) * 1.0e6) );
		}
	}

	track_status->running = cmd;
	track_status->ticks = trk_ticks;
	track_status->index = index;

	event->ticks  = track_status->ticks;

	/* adjust MIDI time and track index */
	track_status->index += cmd_length;
	if (track_status->index >= track_status->data_length) {
		track_status->ticks = UINT_MAX; /* track done */
	}
	else {
		track_status->ticks += gsif_midi_read_vl( (status->midi_format >= 0 ? True : False), track_status);
	}

	event->track  = earliest_track;
	event->event  = cmd;
	event->data   = data;
	event->length = cmd_length;

	DEB2(DPRT(" event->ticks=%d\n", event->ticks);
		 DPRT(" event->track=%d\n", event->track);
		 DPRT(" event->event=0x%x\n", event->event);
		 DPRT(" event->length=%d\n", event->length);
		 for (i=0; i<event->length; i++) {
			 DPRT(" event->data[%d]=%d\n", i, event->data[i]);
		 }
		 DPRT(" track_status->index = %d\n", (int) track_status->index);

		);
	DEB2(DPRT("MIDI #gsif_midi_next_event#\n"));

	return GSIF_OK;
}



#if 0
int gsif_midi_decode_sysex(struct gsi_song_status *status, struct gsi_song_event *event) {
	byte *data = event->data;
	int  cmd = event->event;

	/* GS & GM:
	 * gm_on  = { 0x7e, 0x7f, 0x09, 0x01 }
	 * gm_off = { 0x7e, 0x7f, 0x09, 0x02 }
	 * gs_on  = { 0x41, 0x10, 0x42, 0x12, 0x40, 0x00, 0x7f, 0x00, 0x41 }
	 * gs_off = { 0x41, 0x10, 0x42, 0x12, 0x40, 0x00, 0x7f, 0x7f, 0x41 }

	 * gs_drum_channel = { 0x41, ?, 0x42, 0x12, 0x40, 0x1?, 0x15 }
	 * gs_reset = { 0x41, ?, 0x42, 0x12, 0x40/0x00, 0x00, 0x7f }

	 * gs_text (??) = { 0x41, 0x10, 0x45, 0x12, 0x10, 0x00, 0x00, "..." }
	 */

	/* sysex decoding, remove and put in separate function? */
	if ((cmd & 0xf0) == MIDI_SYSTEM_PREFIX) {
		/* The sysex decoding is very incomplete at the moment */
		if (length > 1) {
			unsigned int  i;

			DEB1(DPRT("MIDI   -> decode_sysex\n"));
			if (length > 7 && data[0] == 0x41 && data[2] == 0x42 && data[3] == 0x12) {
				/* GS data set messages: */

				if (data[4] == 0x40 && (data[5] & 0xf0) == 0x10 && data[6] == 0x15) {
					/* Use rythm part (rythm channels used for drums) */
					i = data[5] & 0xf;
					if (i == 0)
						i = 0x09;
					else if (i < 10)
						i--;

					i = 1<<i;
					if (data[7])
						status->percussion_chn |= i;
					else
						status->percussion_chn &= ~i;
				}

				if ((data[4] == 0x40 || data[4] == 0) && data[5] == 0x00 && data[6] == 0x7f) {
					/* GS reset: */
					//status->percussion_chn = 0x0200;	// percussion only in channel 10 (no more rythm channels for drums)

					//for (i = 0; i < 16; i++) {	// set status info
					//	for (j = 0; j < 128; j++)
					//		note_vel[i][j] = 0;
					//	channel[i].bender = 8192;
					//	channel[i].bender_range = 2;
					//	channel[i].controller[ MIDICTL_PAN ] = 64;
					//	channel[i].controller[ MIDICTL_SUSTAIN ] = 0;
					//}
				}
				/* All this handles the first of alienz' sysex messages,
				 * but the second, hmm. */
			}
		}
	}
}
#endif


/*      EOF midi_decode.c */
