/* GSI - general sound interface
 *
 * gsimidi_pcm.c - wrappers for Keith Packard's software midi playing library
 * (wave table card emulation on a pcm device)
 *
 * Lower level API dependencies:
 *   - Keith Packard's midi playing library (see kpmidi.h)
 *
 *
 * Copyright (C) W.H.Scholten 1999 - 2001
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <gsi/gsimidi_pcm_config.h>

#include <stdio.h>
#include <stdlib.h>
//#include <string.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSISYNTH_"
//#include <gsi/gsi_debug.h>

#include <gsi/whs_defines.h>

#include "gsimidi_pcm.h"
#include "kpmidi.h"
#include "volume.h"
#include "utils/log.h"


/*__________________VARIABLES___________________*/
/* For (MidiData *) use on (void *d): */
#define D ((MidiData *) d)

/* For debugging & information: */
void gsimidipcm_print_settings(void *d) {
	printf("\ngsimidipcm settings: [\n");
	printf("  bufSamples = %d\n", D->bufSamples);
	printf("  chorus = %d\n", D->chorus);
	printf("  patchFile = %s\n", D->patchFile);
	//printf("  tuning = %s\n", D->tuning);
	printf("  panPhase = %f\n", D->panPhase);
	printf("  reverbName = %s\n", D->reverbName);
	//printf("  volume = %d\n", D->volume);
	printf("  stereo = %d\n", D->stereo);
	printf("  auto_pan = %d\n", D->auto_pan);
	printf("  enveloprate = %d\n", D->envelopeRate);

	printf("  scale = %f\n", D->scale);
	printf("]\n");
}


int  gsimidipcm_get_status (void *d, struct gsipcmmodule *m) {
	m->done = D->done;
	m->queued = D->queued;
	m->put = D->put;
	m->get = D->get;
	m->bufSamples = D->bufSamples;
	m->buf = D->buf;
	return 0;
}


int  gsimidipcm_calculate (void *d, int samples) {
//D->scale = 8192.0;
	MidiQueue (D, samples);
//fprintf(stderr, "D->scale=%f\n", D->scale);
	return 0;
}


//double gsimidipcm_seek (void *d, double t) {
//	return MidiSeek (D, t*1.0e6) / 1.0e6;
//}
int gsimidipcm_seek (void *d, double t) {
	MidiSeek (D, t*1.0e6);
	return 0;
}


int  gsimidipcm_set_output_format (void *d, int format, int rate, int channels) {
	/* This must be changed: */
	MidiSetFormat (D, True /* sixteenBit */, False /* unsign */, False /* msbFirst */, True /* stereo */, True /* leftFirst*/);

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Open the pseudo synth device.
 *---------------------------------------------------------------------------*/
int gsimidipcm_open (midi_softsynth_t *softsynth, void **p)
{
	int  status;
	MidiData *d;

	printf("*gsimidipcm_open* [\n");

	*p = malloc(sizeof(MidiData));
	if ((*p) == NULL) {
		gsimidipcm_log_message(GSI_MESSAGE_ERROR, "Out of memory creating MidiData");
		return -1;
	}
	d = (MidiData *) *p;

	MidiSetDefaults (d);

	printf("(na MidiSetDefaults):\n");
	printf("  d->dataRate = %d, d->divisions = %d\n", d->dataRate, d->divisions);

#if 0
	/* ==== Softsynth defaults (will be removed here and go to the app: ==== */
	softsynth->Bps = 16;
	softsynth->sampleRate = 44100;
	softsynth->stereo = True; /* 2 channels */
	softsynth->volume = 70;
	softsynth->bufSamples = 4096;
	//softsynth->maxVoices = 64;
	//softsynth->envelopeRate = 0;
	softsynth->reverb = NULL;
	softsynth->auto_pan = True;
	softsynth->chorus = True;
	softsynth->patch = NULL;
	softsynth->panPhase = 100.0;
	softsynth->tuning = "equal";
	/* ==== end of app code ================================================ */


	/* Set values from default sofsynth values. */
	/* == Not necessary as MidiSetDefaults(d) does this ? == */
	d->dataRate = softsynth->sampleRate;

#   define VOLUME_RANGE           10.0 /* from ui.c */
#   define PosToValue(pos, max)   pow(max, (double) (pos * 2.0f - 1.0f))

	d->scale = PosToValue ((float) softsynth->volume / 100.0f, VOLUME_RANGE) * GSIMIDI_PCM__VOLUME_SCALE;
	/* scale is thus in the range GSIMIDI_PCM__VOLUME_SCALE * VOLUME_RANGE to
	 * GSIMIDI_PCM__VOLUME_SCALE * 1/VOLUME_RANGE
	 */

	if (softsynth->Bps == 8) {
		d->dataBytesPerSample = 1;
	}
	else if (softsynth->Bps == 16) {
		d->dataBytesPerSample = 2;
	}
	else {
		gsimidipcm_log_message(GSI_MESSAGE_ERROR, "Bad Bps %d", softsynth->Bps);
		return -1;
	}

	d->bufSamples = softsynth->bufSamples;
	//d->maxVoices = softsynth->maxVoices;
	//d->envelopeRate = softsynth->envelopeRate;
	d->stereo = softsynth->stereo;
	d->reverbName = softsynth->reverb;
	d->auto_pan = softsynth->auto_pan;
	d->chorus = softsynth->chorus;
	d->patchFile = softsynth->patch;
	d->panPhase *= softsynth->panPhase / 100.0f;

	MidiSetTemperment (d, softsynth->tuning);
#endif

	gsimidipcm_print_settings(d);

	/* Default scale = 0.0 which means no output !! */
	d->scale = 8192.0;
	//d->scale = 16384.0;
	//d->scale = 32768.0;

d->patchFile = NULL;

	MidiSetTemperment (d, "equal");
	MidiSetReverbType (d, "room");
	//d->panPhase = 50.0;
	d->auto_pan = True;

	status = MidiInit (d);

	printf("(na MidiInit):\n");
	printf("  d->tempo = %ld, d->midiSpeedFactor = %f\n", d->tempo, d->midiSpeedFactor);
	printf("  d->dataRate = %d, d->divisions = %d\n", d->dataRate, d->divisions);
	printf("] #gsimidipcm_open#\n");

	return !status;
}


/*---------------------------------------------------------------------------*
 *      Close the pseudo synth device (release all used memory).
 *---------------------------------------------------------------------------*/
int gsimidipcm_close (void *d)
{
	MidiDispose (D);
	MidiClose (D);
	free(d);
	return 0;
}


int gsimidipcm_input_file (void *d, char *name) {
	int  status;

	status = MidiSetFile (D, name);

	return status;
}


int gsimidipcm_clear_samples (void *d) {
	/* Not yet implemented. */
	fprintf(stderr, "ERROR: gsimidipcm_clear_samples not implemented\n");
	return -1;
}


int gsimidipcm_set_temperment (void *d, char *temperment) {
	return ! MidiSetTemperment (D, temperment);
}


int  gsimidipcm_set_reverb (void *d, char *reverb) {
	return ! MidiSetReverbType (D, reverb);
}


/* Instrument volume: 1.0 = default */
/* instrument 0 - 127 = patch, 128 - 255 = drums */
int gsimidipcm_set_instrument_volume (void *d, int instrument, float volume) {
	if (instrument < 128)
		patchVolume[ instrument ] = volume;
	else
		drumVolume[ instrument - 128 - MIN_DRUM ] = volume;

	return 0;
}


int gsimidipcm_set_volume_scale (void *d, int scale_max) {
	/* e.g. 16384 to avoid clipping with multiple voices */
	if (scale_max < 0) scale_max = 0;
	D->scale = (float) scale_max;

	return 0;
}

/*---------------------------------------------------------------------------*
 *      Some sequencer buffer management functions.
 *---------------------------------------------------------------------------*/
/* Buffer status: 0=empty, 1=used, 2=full. */
int gsimidipcm_buffer_status (void *d) {
	if (D->queued) {
		if (D->queued == D->bufSamples)
			return 2;

		return 1;
	}

	return 0;
}


/* Is the sequencer (or buffer) ready to accept data or not? */
int gsimidipcm_buffer_full (void *d) {
	return (D->queued == D->bufSamples);
}


int gsimidipcm_buflen (void *d) {

	return D->queued;
}


/* gsimidipcm_remove returns 0 if all went well, < 0 on error, which is if
 * n > available data, or n <= 0. */
int gsimidipcm_remove (void *d, int n) {
	if (n > 0 && D->queued > 0 && n <= D->queued) {
		D->queued -= n;
		D->get += n;

		if (D->get >= D->bufSamples)
			D->get -= D->bufSamples;

		return 0;
	}

	return -1;
}


/*      EOF init_lib.c */
