/*
 * $Id: midifile.c,v 1.9 1998/06/01 06:03:15 keithp Exp $
 * => partially upgraded to 1.11 1998/08/21
 *
 * Modified by W.H.Scholten 2000 - 2001
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
//#include <gsi/gsimidi_pcm_config.h> /* endianness */
#include <gsi/config.h> /* endianness */
#include "../kpmidi.h"
#include "../utils/log.h"
#include "../utils/debug.h"

//NEWtypedef union {
//NEW	uint8   data[4];
//NEW	uint32  value;
//NEW} ufile32;
//NEW
//NEWtypedef union {
//NEW	uint8   data[2];
//NEW	uint16  value;
//NEW} ufile16;


typedef struct _MidiFileHeader {
    uint8	magic[4];
    uint32	len;
    uint16	format;
    uint16	tracks;
    uint16	divisions;
//NEW	ufile32  len;
//NEW	ufile16  format;
//NEW	ufile16  tracks;
//NEW	ufile16  divisions;
} MidiFileHeader;

#define MIDI_FILE_HEADER_SIZE	14

typedef struct _MidiTrackHeader {
    uint8	magic[4];
    uint32	len;
//NEW	ufile32 len;
} MidiTrackHeader;

#define MIDI_TRACK_HEADER_SIZE	8

#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
#  define BE_INT32(l) ((((l) & 0xff) << 24) | \
					   (((l) & 0xff00) << 8) | \
					   (((l) & 0xff0000) >> 8) | \
					   (((l) & 0xff000000) >> 24))

#  define BE_INT16(s) ((((s) & 0xff) << 8) | \
					   (((s) & 0xff00) >> 8))

#else
#  define BE_INT32(l) (l)
#  define BE_INT16(s) (s)
#endif


//NEW#define SWAP_INT32(l)   l.value = ((l.data[0] << 24) | \
//NEW			(l.data[1] << 16) | \
//NEW			(l.data[2] << 8) | \
//NEW			(l.data[3]))
//NEW
//NEW#define SWAP_INT16(l)  l.value = ((l.data[0] << 8) | \
//NEW			(l.data[1]))


typedef struct _midiTrack {
    unsigned char   *buf;
    int		    len;
    int		    pos;
    int		    running;
    double	    eventTime;
    MidiParseState  state;
} MidiTrack;

typedef struct _midiFile {
    MidiTrack	    *tracks;
    MidiFileHeader  fileHead;
    int		    track;
} MidiFile;

void
MidiFileDispose (MidiData *d)
{
	int		track;
	MidiFile	*f = d->source;
	MidiTrack	*t;

	for (track = 0; track < f->fileHead.tracks; track++)
//NEW	for (track = 0; track < f->fileHead.tracks.value; track++)
	{
		t = &f->tracks[track];
		free (t->buf);
		t->buf = 0;
	}
	free (f->tracks);
	free (d->source);
	d->source = 0;
}

int
MidiFileRead (MidiData *d, char *name)
{
    int		    t;
    MidiFile	    *f;
    MidiTrackHeader thead;
    FILE	    *mf;
    
    mf = fopen (name, "r");
    if (!mf)
		return 0;

    f = calloc (1, sizeof (MidiFile));
    if (!f)
    {
		fclose (mf);
		return 0;
    }

    if (fread (&f->fileHead, MIDI_FILE_HEADER_SIZE, 1, mf) != 1)
    {
		free (f);
		fclose (mf);
		return 0;
    }

    f->fileHead.len = BE_INT32 (f->fileHead.len);
    f->fileHead.format = BE_INT16 (f->fileHead.format);
    f->fileHead.tracks = BE_INT16 (f->fileHead.tracks);
    f->fileHead.divisions = BE_INT16 (f->fileHead.divisions);
    d->divisions = f->fileHead.divisions;
DEB(DPRT("midifile: d->divisions = %d\n", d->divisions));
    if (memcmp (f->fileHead.magic, "MThd", 4))
    {
		free (f);
		fclose (mf);
		return 0;
    }

    if (f->fileHead.format > 2)
    {
		gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Unknown MIDI format %f: %s", f->fileHead.format, name);
		free (f);
		fclose (mf);
		return 0;
    }
    f->tracks = calloc (f->fileHead.tracks, sizeof (MidiTrack));
    if (!f->tracks)
    {
		gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Can't allocate %d tracks", f->fileHead.tracks);
		free (f);
		fclose (mf);
		return 0;
    }
    for (t = 0; t < f->fileHead.tracks; t++)
    {
		if (fread (&thead, MIDI_TRACK_HEADER_SIZE, 1, mf) != 1)
		{
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Can't read track %d header: %s", t, name);
			free (f);
			fclose (mf);
			return 0;
		}
		if (memcmp (thead.magic, "MTrk", 4))
		{
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Bad track magic %d: %s", t, name);
			free (f);
			fclose (mf);
			return 0;
		}
		f->tracks[t].len = BE_INT32 (thead.len);
		f->tracks[t].buf = malloc (f->tracks[t].len);
		if (f->tracks[t].buf == NULL) {
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Out of memory");
			free (f);
			fclose (mf);
			return 0;
		}

		f->tracks[t].pos = 0;
		MidiInitState (&f->tracks[t].state);
		if (fread (f->tracks[t].buf, 1, f->tracks[t].len, mf) !=
			f->tracks[t].len)
		{
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Short track read %d: %s", t, name);
			free (f);
			fclose (mf);
			return 0;
		}
    }
    fclose (mf);
    d->source = f;

    return 1;
}

int
MidiFileEvent (MidiData *d, Boolean seek)
{
    MidiFile	    *f = d->source;
    int		    ntrack;
    int		    ftrack, track, ltrack;
    MidiTrack	    *t;
    int		    n;
    int		    ready;
    unsigned long   delta;

    if (f->fileHead.format == 1)
		ntrack = f->fileHead.tracks;
    else
		ntrack = 1;

    ftrack = f->track;
    ltrack = ftrack + ntrack;
    ready = 0;
    d->eventTime = 1e38;
    for (track = ftrack; track < ltrack; track++)
    {
		t = &f->tracks[track];
		while (t->eventTime <= d->time)
		{
			n = MidiParseOne (d, &t->state, t->buf + t->pos, t->len - t->pos, seek);
			if (n < 0)
				t->pos = t->len;
			else
				t->pos += n;

			if (t->pos < t->len)
			{
				n = MidiParseInt (t->buf + t->pos, &delta);
				t->pos += n;
				t->eventTime += delta;
			}
			if (t->pos >= t->len)
			{
				t->running = False;
				t->eventTime = 1e38;
				if (f->fileHead.format == 2)
				{
					f->track++;
					if (f->track < f->fileHead.tracks)
						ltrack++;
				}
				break;
			}
		}
		if (t->eventTime < d->eventTime)
			d->eventTime = t->eventTime;
		if (t->running)
			ready++;
    }
    if (!ready)
	{
		MidiAllNotesOff (d);
		if (!seek && d->repeat)
		{
			MidiReset (d);
			ready = True;
		}
	}

    return ready;
}

void
MidiFileReset (MidiData *d)
{
    MidiFile	    *f = d->source;
    MidiTrack	    *t;
    unsigned long   delta;
    int		    track;

	if (!f)
		return;

    /*
     * Read initial delta time and set eventTime
     */
    for (track = 0; track < f->fileHead.tracks; track++)
    {
		t = &f->tracks[track];
		MidiInitState (&t->state);
		t->running = True;
		t->pos = MidiParseInt (t->buf, &delta);

//REMOVED FOR UPGRADE		t->eventTime = delta;
		t->eventTime = d->time + delta;

		if (t->eventTime < d->eventTime)
			d->eventTime = t->eventTime;
    }
    f->track = 0;
}

SourceFuncs MidiFileSource = {
    MidiFileRead,
    MidiFileEvent,
    MidiFileDispose,
    MidiFileReset,
    MidiSample,
};
