/*
 * $Id: patchgram.y,v 1.5 1999/08/26 18:58:48 keithp Exp $
 *
 * Modified by W.H.Scholten
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

%{
#include <stdio.h>
#include "../kpmidi.h"
#include "../utils/log.h"
#include "../midi_to_pcm.h"

static MidiPatch	*patch;
static Instrument	*instrument;
static SampleElement	*sample;
static float		*data;
static float		scale;
static int		data_size, data_used;
static EnvelopeElement	*envelope;
static int		envelope_used;
void			MidiPatch_error (char *format, ...);
static void		setRanges (MidiPatch *);
static void		MidiPatch_free (const MidiPatch *);
static int  MidiPatch_parse (void);
extern int yylex (void);
%}

%union {
    int32    	    ival;
    float	    fval;
    char	    *sval;
    MidiFix	    xval;
}

%token		    OC CC COMMA SEMI EQUAL LESS RIGHT
%token		    RELEASE RATE LOW HIGH ROOT SCALE LOOP
%token		    TREMELO VIBRATO CHORUS ALWAYS
%token		    FIXED_FREQUENCY FIXED_VELOCITY RELEASE_LOOP AUTO_PAN
%token <ival>	    INTEGER
%token <fval>	    FLOAT
%token <sval>	    STRING

%type <ival>	    optint
%type <xval>	    fix

%%
patch	    :	STRING { 
		    instrument = calloc (1, sizeof (Instrument));
		    patch = calloc (1, sizeof (MidiPatch));
		    instrument->tremelo_scale = 0;
		    instrument->tremelo_frequency = 0;
		    instrument->vibrato_scale = 0;
		    instrument->vibrato_frequency = 0;
		    instrument->chorus = INSTRUMENT_CHORUS_OFF;
		    instrument->flags = 0;
		    patch->instrument = instrument;
		    patch->free = MidiPatch_free;
		    if (data)
			free (data);
		    data = 0;
		    data_used = 0;
		    data_size = 0;
		    strcpy (instrument->magic, _GSIMIDIPCM_INSTRUMENT_NATIVE_BINARY__MAGIC);
		    envelope_used = 0;
		    envelope = 0;
		    }
		instrument
		{
		    patch->name = $1;
		    patch->data = realloc (data, 
					   data_used * sizeof (float));
		    patch->envelopes = envelope;
		    instrument->dataTotal = data_used;
		    instrument->envelopeTotal = envelope_used;
		    setRanges (patch);
		    instrument = 0;
		    data = 0;
		    envelope = 0;
		}
	    ;
instrument  :	options samples
	    ;
options	    :	options option
	    |
	    ;
option	    :	flag
	    |	vibrato
	    |	tremelo
	    |	chorus
	    ;
flag	    :	FIXED_FREQUENCY
		    { instrument->flags |= INSTRUMENT_FIXED_FREQUENCY; }
	    |	FIXED_VELOCITY
		    { instrument->flags |= INSTRUMENT_FIXED_VELOCITY; }
	    |	RELEASE_LOOP
		    { instrument->flags |= INSTRUMENT_RELEASE_LOOP; }
	    |	AUTO_PAN
		    { instrument->flags |= INSTRUMENT_AUTO_PAN; }
	    ;
tremelo	    :	TREMELO EQUAL FLOAT COMMA INTEGER SEMI
		    {	instrument->tremelo_scale = $3;
			instrument->tremelo_frequency = $5; }
	    ;
vibrato	    :	VIBRATO EQUAL FLOAT COMMA INTEGER SEMI
		    {	instrument->vibrato_scale = $3;
			instrument->vibrato_frequency = $5; }
	    ;
chorus	    :	CHORUS
		    {	instrument->chorus = INSTRUMENT_CHORUS_ON; }
	    |	CHORUS ALWAYS
		    {	instrument->chorus = INSTRUMENT_CHORUS_ALWAYS; }
	    ;
samples	    :	samples sample
	    |
	    ;
sample	    :	OC 
		    { 
			SampleElement *samples;

			samples = calloc (instrument->nsamples + 1, sizeof (SampleElement));
			if (patch->samples)
			{
			    memcpy (samples, patch->samples,
				    instrument->nsamples * sizeof (SampleElement));
			    free ((void *) patch->samples);
			}
			sample = &samples[instrument->nsamples];
			sample->loop_start = FIX_INVALID;
			sample->loop_end = FIX_INVALID;
			sample->low_freq = -1;
			sample->high_freq = -1;
			sample->release = 0;
			sample->data = data_used;
			sample->envelope = envelope_used;
			patch->samples = samples;
			instrument->nsamples++;
			scale = 1.0f;
		    }
		    params envelope data CC
			{
			    int32   len;
			    
			    if (!sample->sample_rate)
			    {
				yyerror ("Instrument must have sample rate");
				YYERROR;
			    }
			    if (!sample->root_freq)
			    {
				yyerror ("Instrument must have root frequency");
				YYERROR;
			    }
			    if (!sample->release)
			    {
				yyerror ("Instrument must have release specified");
				YYERROR;
			    }
			    len = data_used - sample->data;
			    if (len >= MAX_MIDI_SAMPLES)
			    {
				yyerror ("Too many samples %d (max %d)",
					 len, MAX_MIDI_SAMPLES);
				YYERROR;
			    }
			    sample->data_length = FixFromInt (len);
			}
	    ;
params	    :	params param
	    |
	    ;
param	    :	RATE EQUAL INTEGER SEMI
		{ sample->sample_rate = $3; }
	    |	LOW EQUAL INTEGER SEMI
		{ sample->low_freq = $3; }
	    |	HIGH EQUAL INTEGER SEMI
		{ sample->high_freq = $3; }
	    |	ROOT EQUAL INTEGER SEMI
		{ sample->root_freq = $3; }
	    |	SCALE EQUAL FLOAT SEMI
		{ scale = $3; }
	    |	LOOP EQUAL fix COMMA fix SEMI
		{ sample->loop_start = $3; sample->loop_end = $5; }
	    ;
fix	    :	INTEGER
		{ $$ = FixFromInt ($1); }
	    |	FLOAT
		{ $$ = FixFromFloat ($1); }
	    ;
envelope    :	OC elements CC
	    ;
elements    :	element elements
	    |
	    ;
element	    :	INTEGER optint COMMA INTEGER SEMI
		{
		    EnvelopeElement *ee;

		    ee = calloc (envelope_used + 1, sizeof (EnvelopeElement));
		    if (envelope)
		    {
			memcpy (ee, envelope,
				envelope_used * sizeof (EnvelopeElement));
			free (envelope);
		    }
		    envelope = ee;
		    ee = envelope + envelope_used;
		    envelope_used++;
		    ee->rate = $1;
		    ee->noteRate = $2;
		    ee->target = $4;
		    sample->nenvelope++;
		}
	    |	RELEASE
		{   sample->release = sample->nenvelope; }
	    ;
optint	    :	INTEGER
	    |
		{ $$ = 0; }
	    ;
data	    :	data datum
	    |
	    ;
datum	    :	FLOAT
		{
		    if (data_used == data_size)
		    {
			if (data)
			    data = realloc (data, 
					    (data_size *= 2) * sizeof (float));
			else
			    data = malloc ((data_size = 128) * sizeof (float));
		    }
		    data[data_used++] = $1 * scale;
		}
	    |	LESS
		{
		    if (sample->loop_start != FIX_INVALID)
		    {
			yyerror ("Instrument can have only one loop");
			YYERROR;
		    }
		    sample->loop_start = FixFromInt (data_used);
		}
	    |	RIGHT
		{
		    if (sample->loop_end != FIX_INVALID)
		    {
			yyerror ("Extra loop end");
			YYERROR;
		    }
		    if (sample->loop_start == FIX_INVALID)
		    {
			yyerror ("Loop end must be after loop start");
			YYERROR;
		    }
		    sample->loop_end = FixFromInt (data_used);
		}
	    ;
%%

#include <stdarg.h>
int		MidiPatch_line;
const char	*MidiPatch_file;

int
MidiPatch_wrap (void)
{
    return 1;
}

void
MidiPatch_error (char *format, ...)
{
    va_list	args;
    char	buf[4096];

    va_start (args, format);
    sprintf (buf, "%s: %d: ", MidiPatch_file, MidiPatch_line);
    vsprintf (buf + strlen (buf), format, args);
    gsimidipcm_log_message (GSI_MESSAGE_ERROR, "%s", buf);
    va_end (args);
}

struct range {
    int32	    root, low, high;
    int	    id;
};

int
compareRange (const void *a, const void *b)
{
    const struct range    *ra = a, *rb = b;

    return ra->root - rb->root;
}

void
setRanges (MidiPatch *p)
{
    const Instrument    *ip = p->instrument;
    int			    i;
    struct range	    ranges[256];
    SampleElement	    *s;

    for (i = 0; i < ip->nsamples; i++)
    {
	s = (SampleElement *) &p->samples[i];
	ranges[i].root = s->root_freq;
	ranges[i].low = s->low_freq;
	ranges[i].high = s->high_freq;
	ranges[i].id = i;
    }
    qsort (ranges, ip->nsamples, sizeof (struct range), 
	   compareRange);
    for (i = 0; i < ip->nsamples; i++)
    {
	if (i == 0)
	    ranges[i].low = 0;
	else
	    ranges[i].low = ranges[i-1].high+1;
	if (i == ip->nsamples - 1)
	    ranges[i].high = 0x7fffffff;
	else
	{
	    if (ranges[i].high < 0)
	    {
		if (ranges[i+1].low > ranges[i].root)
		    ranges[i].high = ranges[i+1].low - 1;
		else
		    ranges[i].high = (ranges[i].root + ranges[i+1].root) / 2;
	    }
	}
    }
    for (i = 0; i < ip->nsamples; i++)
    {
	s = (SampleElement *) &p->samples[ranges[i].id];
	s->low_freq = ranges[i].low;
	s->high_freq = ranges[i].high;
    }
}

const MidiPatch *
MidiPatchParse (const char *file)
{
    FILE	*input;
    extern FILE	*MidiPatch_in;

    input = fopen (file, "r");
    if (!input)
    {
	gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Can't open patch file \"%s\"\n", file);
	return 0;
    }
    MidiPatch_in = input;
    MidiPatch_file = file;
    MidiPatch_line = 1;
    patch = 0;
    if (MidiPatch_parse () != 0)
    {
	if (patch)
	{
	    if (patch->samples)
		free ((void *) patch->samples);
	    free (patch);
	    patch = 0;
	}
	if (instrument)
	{
	    free (instrument);
	    instrument = 0;
	}
	if (data)
	{
	    free (data);
	    data = 0;
	}
	if (envelope)
	{
	    free (envelope);
	    envelope = 0;
	}
    }
    fclose (input);
    MidiPatch_in = 0;
    return patch;
}

static void
MidiPatch_free (const MidiPatch *p)
{
    free ((void *) p->name);
    free ((void *) p->instrument);
    free ((void *) p->samples);
    free ((void *) p->envelopes);
    free ((void *) p->data);
    free ((void *) p);
}
