/*
 * $Id: gus.c,v 1.4 1998/09/14 00:48:43 keithp Exp $
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <toivonen@clinet.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   instrum.c 
   
   Code to load and unload GUS-compatible instrument patches.

*/

#include "../kpmidi.h"
#include "../midi_to_pcm.h"
#include <stdlib.h>
#include <stdio.h>

#ifdef USE_GUS_PATCH_CODE

#define MODES_16BIT	(1<<0)
#define MODES_UNSIGNED	(1<<1)
#define MODES_LOOPING	(1<<2)
#define MODES_PINGPONG	(1<<3)
#define MODES_REVERSE	(1<<4)
#define MODES_SUSTAIN	(1<<5)
#define MODES_ENVELOPE	(1<<6)

static int32 convert_envelope_rate(int32 sample_rate, uint8 rate)
{
    int32 exp, mant;

    exp= (rate>>6) * 3;
    mant = (int32) (rate & 0x3f);

    return mant << (20 - exp);
}


static int32 convert_envelope_offset(uint8 offset)
{
  /* This is not too good... Can anyone tell me what these values mean?
     Are they GUS-style "exponential" volumes? And what does that mean? */

  return offset << 23;
}

static void reverse_data(int16 *sp, int32 ls, int32 le)
{
  int16 s, *ep=sp+le;
  sp+=ls;
  le-=ls;
  le/=2;
  while (le--)
    {
      s=*sp;
      *sp++=*ep;
      *ep--=s;
    }
}

void *
safe_malloc (int len)
{
    void    *r;

    r = malloc (len);
    if (!r)
    {
	fprintf (stderr, "malloc %d failed\n", len);
	exit (1);
    }
    return r;
}

static void
FreeGUSPatch (const MidiPatch *p)
{
    if (p->name)
	free ((void *) p->name);
    free ((void *) p->data);
    free ((void *) p->samples);
    free ((void *) p->envelopes);
    free ((void *) p->instrument);
    free ((void *) p);
}

/* 
   If panning or note_to_use != -1, it will be used for all samples,
   instead of the sample-specific values in the instrument file. 

   For note_to_use, any value <0 or >127 will be forced to 0.
 
   For other parameters, 1 means yes, 0 means no, other values are
   undefined.

   TODO: do reverse loops right */

const MidiPatch *
ReadGUSPatch(const char *name)
{
    MidiPatch	    *p;
    Instrument	    *ip;
    SampleElement   *sp, *samples;
    EnvelopeElement *ep;
    FILE	    *fp;
    uint8	    tmp[1024];
    int		    i,j/*,noluck=0*/;
    float	    *data;
    int16	    *cdata, *cd;
    int16	    average;
    double	    total;
    int		    data_size = 0;

    if (!name) goto bail0;

    p = calloc (1, sizeof(MidiPatch));
    if (!p)
	goto bail0;
    
    /* Open patch file */
    if (!(fp=fopen(name, "r")))
	goto bail0a;

    /* Read some headers and do cursory sanity checks. There are loads
     of magic offsets. This could be rewritten... */

    if ((239 != fread(tmp, 1, 239, fp)) ||
	(memcmp(tmp, "GF1PATCH110\0ID#000002", 22) &&
	 memcmp(tmp, "GF1PATCH100\0ID#000002", 22))) /* don't know what the
						      differences are */
    {
	goto bail1;
    }

    if (tmp[82] != 1 && tmp[82] != 0) /* instruments. To some patch makers, 
				       0 means 1 */
	goto bail1;

    if (tmp[151] != 1 && tmp[151] != 0) /* layers. What's a layer? */
	goto bail1;

    ip = calloc(1, sizeof(Instrument));
    if (!ip)
	goto bail1;
    
    ip->nsamples = tmp[198];
    samples = calloc(ip->nsamples, sizeof(SampleElement));
    p->instrument = ip;
    p->samples = samples;
    p->envelopes = calloc (6 * ip->nsamples, sizeof (EnvelopeElement));
    if (!samples)
	goto bail2;
    
    for (i=0; i<ip->nsamples; i++)
    {
	uint8	tmp[4];
	uint8	modes;
	uint8	fractions;
	int32	data_length;
	int32	loop_start;
	int32	loop_end;

#define READ_CHAR(thing) \
	if (1 != fread(tmp, 1, 1, fp)) goto bail4; \
	    thing = tmp[0];
#define READ_SHORT(thing) \
	if (1 != fread(tmp, 2, 1, fp)) goto bail4; \
	    thing = tmp[0] | (tmp[1] << 8);
#define READ_LONG(thing) \
	if (1 != fread(tmp, 4, 1, fp)) goto bail4; \
	    thing = tmp[0] | (tmp[1] << 8) | (tmp[2] << 16) | (tmp[3] << 24);

#define skip(fp,i)  fseek (fp,i,SEEK_CUR)

	skip (fp, 7); /* Skip the wave name */

	if (1 != fread(&fractions, 1, 1, fp))
	    goto bail4;

	sp=&(samples[i]);

	READ_LONG(data_length);
	READ_LONG(loop_start);
	READ_LONG(loop_end);
	READ_SHORT(sp->sample_rate);
	READ_LONG(sp->low_freq);
	READ_LONG(sp->high_freq);
	READ_LONG(sp->root_freq);
	/* READ_SHORT(sp->tuning); */
	skip(fp, 2); /* Why have a "root frequency" and then "tuning"?? */

	READ_CHAR(tmp[0]);
	/*      fprintf (stderr, " %dhz loop %d -> %d end %d\n",
			 sp->root_freq / 1000,
			 loop_start / 2,
			 loop_end / 2,
			 data_length / 2); */

#if 0
	if (panning==-1)
	    sp->panning = (tmp[0] * 8 + 4) & 0x7f;
	else
	    sp->panning=(uint8)(panning & 0x7F);
#endif

	/* envelope, tremolo, and vibrato */
	if (18 != fread(tmp, 1, 18, fp)) goto bail4; 

#if 0
	if (!tmp[13] || !tmp[14])
	{
	    sp->tremolo_sweep_increment=
	    sp->tremolo_phase_increment=sp->tremolo_depth=0;
	    /*	  fprintf (stderr, " * no tremolo\n");*/
	}
	else
	{
	    sp->tremolo_sweep_increment=tmp[12];
	    sp->tremolo_phase_increment=tmp[13];
	    sp->tremolo_depth=tmp[14];
	    /*	  fprintf (stderr,
			   " * tremolo: sweep %d, phase %d, depth %d\n",
			   sp->tremolo_sweep_increment, sp->tremolo_phase_increment,
			   sp->tremolo_depth); */
	}

	if (!tmp[16] || !tmp[17])
	{
	    sp->vibrato_sweep_increment=
	    sp->vibrato_control_ratio=sp->vibrato_depth=0;
	    /*	  fprintf (stderr, " * no vibrato\n"); */
	}
	else
	{
	    sp->vibrato_control_ratio=tmp[16];
	    sp->vibrato_sweep_increment=tmp[15];
	    sp->vibrato_depth=tmp[17];
	    /*	  fprintf (stderr,
			   " * vibrato: sweep %d, ctl %d, depth %d\n",
			   sp->vibrato_sweep_increment, sp->vibrato_control_ratio,
			   sp->vibrato_depth); */

	}
#endif

	READ_CHAR(modes);

	skip(fp, 40); /* skip the useless scale frequency, scale factor
		       (what's it mean?), and reserved space */

#if 0
	/* Mark this as a fixed-pitch instrument if such a deed is desired. */
	if (note_to_use!=-1)
	    ip->flags |= INSTRUMENT_FIXED_FREQUENCY;
#endif

	/* seashore.pat in the Midia patch set has no Sustain. I don't
			understand why, and fixing it by adding the Sustain flag to
			all looped patches probably breaks something else. We do it
			anyway. */

	if (modes & MODES_LOOPING) 
	    modes |= MODES_SUSTAIN;

#if 0
	/* Strip any loops and envelopes we're permitted to */
	if ((strip_loop==1) && 
	    (modes & (MODES_SUSTAIN | MODES_LOOPING | 
			  MODES_PINGPONG | MODES_REVERSE)))
	{
	    /*	  fprintf (stderr, " - Removing loop and/or sustain"); */
	    modes &=~(MODES_SUSTAIN | MODES_LOOPING | 
			  MODES_PINGPONG | MODES_REVERSE);
	}
#endif

#if 0
	if (strip_envelope==1)
	{
	    /*	  if (modes & MODES_ENVELOPE)
	     fprintf (stderr, " - Removing envelope"); */
	    modes &= ~MODES_ENVELOPE;
	}
	else if (strip_envelope != 0)
	{
	    /* Have to make a guess. */
	    if (!(modes & (MODES_LOOPING | MODES_PINGPONG | MODES_REVERSE)))
	    {
		/* No loop? Then what's there to sustain? No envelope needed
		 either... */
		modes &= ~(MODES_SUSTAIN|MODES_ENVELOPE);
		/*	      fprintf (stderr, 
				       " - No loop, removing sustain and envelope"); */
	    }
	    else if (!memcmp(tmp, "??????", 6) || tmp[11] >= 100) 
	    {
		/* Envelope rates all maxed out? Envelope end at a high "offset"?
		 That's a weird envelope. Take it out. */
		modes &= ~MODES_ENVELOPE;
		/*	      fprintf (stderr, 
				       " - Weirdness, removing envelope"); */
	    }
	    else if (!(modes & MODES_SUSTAIN))
	    {
		/* No sustain? Then no envelope.  I don't know if this is
		 justified, but patches without sustain usually don't need the
		 envelope either... at least the Gravis ones. They're mostly
		 drums.  I think. */
		modes &= ~MODES_ENVELOPE;
		/*	      fprintf (stderr, 
				       " - No sustain, removing envelope"); */
	    }
	}
#endif

	sp->envelope = i * 6;
	sp->nenvelope = 6;
	sp->release = 3;
	ep = (EnvelopeElement *) MSEnvelope(p,sp);
	
	for (j=0; j<6; j++)
	{
	    ep[j].rate = convert_envelope_rate(sp->sample_rate, tmp[j]);
	    ep[j].target = convert_envelope_offset(tmp[6+j]);
	}

	/* Then read the sample data */
	cdata = malloc (data_length);
	if (!cdata)
	    goto bail4;
	if (1 != fread (cdata, data_length, 1, fp))
	    goto bail4;

	if (!(modes & MODES_16BIT)) /* convert to 16-bit data */
	{
	    uint8   *cp= (uint8 *)cdata;
	    uint16  *tmp,*new;
	    
	    tmp = new = malloc(data_length*2);
		//WHS: added this malloc check:
		if (!tmp)
			goto bail4;

	    j = data_length;
	    while (j--)
	    {
		*tmp++ = ((uint16)*cp << 8) | (*cp);
		cp++;
	    }
	    free (cdata);
	    cdata = new;
	}
	else
	{
	    union {
		uint8	b[2];
		uint16	s;
	    } t;
	    
	    data_length /= 2;
	    loop_start /= 2;
	    loop_end /= 2;
	    /* convert to machine byte order */

	    t.b[0] = 1;
	    t.b[1] = 0;
	    if (t.s != 1)
	    {
		uint16	*tmp = cdata, s;
		j = data_length;
		while (j--)
		{ 
		    s = *tmp;
		    s = (s << 8) | (s >> 8);
		    *tmp++ = s;
		}
	    }
	}

	if (modes & MODES_UNSIGNED) /* convert to signed data */
	{
	    uint16  *tmp = cdata;
	    
	    j = data_length;
	    while (j--)
		*tmp++ ^= 0x8000;
	}

	/* Reverse reverse loops and pass them off as normal loops */
	if (modes & MODES_REVERSE)
	{
	    int32 t;
	    /* The GUS apparently plays reverse loops by reversing the
	     whole sample. We do the same because the GUS does not SUCK. */

	    /*	  fprintf (stderr, "Reverse loop in %s\n", name); */
	    reverse_data((int16 *)cdata, 0, data_length);

	    t = loop_start;
	    loop_start = data_length - loop_end;
	    loop_end = data_length - t;

	    modes &= ~MODES_REVERSE;
	    modes |= MODES_LOOPING; /* just in case */
	}

#if 0
	if (strip_tail==1)
	{
	    /* Let's not really, just say we did. */
	    fprintf (stderr, " - Stripping tail");
	    data_length = loop_end;
	}
#endif
	/*
	 * DC offset adjustment
	 */
	cd = cdata;
	j = data_length;
	total = 0.0;
	while (j--)
	{
	    total += (double) *cd++;
	}
	average = (int16) (total / data_length);

	/*
	 * convert to floating point
	 */
// XXX WHS: add malloc checks:
	if (p->data)
	    p->data = realloc ((float *) p->data, (data_size + data_length) * sizeof (float));
	else
	    p->data = malloc (data_length * sizeof (float));

	sp->data = data_size;
	data_size += data_length;
	data = (float *) MSData(p,sp);
	cd = cdata;
	j = data_length;
	while (j--)
	    *data++ = (float) (*cd++ - average) / 32768.0f;
	free (cdata);
	sp->data_length = FixFromInt (data_length);
	sp->loop_start = FixFromInt (loop_start);
	sp->loop_end = FixFromInt (loop_end);

// WHS: To be able to to write a .mb directly after reading a .pat I added:
	ip->dataTotal += data_length;
	ip->envelopeTotal += 6;
// WHS: This also seems to fix weird play problems when playing using .pat's
// directly.


// NOTE: going A {.pat->.ma->.mb} or B {.pat->.mb} will produce different
// .mb files.
//
// 1. sp->low_freq, sp->high_freq are not the same. In case A low_freq=0,
//    high_freq=-1, whereas in B these are the values as given by the .pat
//    file.
//
//    Does this mean low_freq/high_freq are not read from the .ma file?
//    (they ARE in the file).
//
// 2. The actual floating point data is different. This is caused by the
//    floating point conversion binary->decimal->binary in case A vs.
//    binary->binary in case B.
//

    }

// WHS: To be able to write a .mb file directly after reading a .pat I added:
	strcpy (ip->magic, _GSIMIDIPCM_INSTRUMENT_NATIVE_BINARY__MAGIC);

    fclose(fp);
    p->free = FreeGUSPatch;
    return p;

bail4:
//bail3:
    free (samples);
bail2:
    free (ip);
bail1:
    fclose (fp);
bail0a:
    if (p->data)
	free ((void *) p->data);
    if (p->samples)
	free ((void *) p->samples);
    if (p->envelopes)
	free ((void *) p->envelopes);
    free (p);
bail0:
    return 0;
}

#endif /* USE_GUS_PATCH_CODE */
