/*
 * $Id: instrum.c,v 1.6 1998/05/27 19:25:16 keithp Exp $
 * => upgraded to 1.8
 *
 * Modified by W.H.Scholten
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <string.h>
#include "../utils/string_match.h"
#include "../midi_data.h"
#include "../midi_to_pcm.h"
#include "../kpmidi.h"
#include "write_instrument.h"
#include "../volume.h"
#include "../utils/debug.h"

#define DATA_PER_LINE   5

// Rename: gsimidi_pcm_write_instrument_ascii()
int
InstrumWriteText (const MidiPatch   *p,
				  const float       scale, 
				  const char        *prefix, 
				  const char        *file)
{
	const Instrument        *ip = p->instrument; 
	FILE  *f;
	int   ret = 0;
	const SampleElement     *sp;
	const EnvelopeElement   *ep;
	int   s;
	int   e;
	int   d;
	int   l;

	f = fopen (file, "w");
	if (!f)
		return ret;


	fprintf (f, "\"%s\"\n", prefix);
	if (ip->flags & INSTRUMENT_FIXED_FREQUENCY)
		fprintf (f, "fixed_frequency\n");
	if (ip->flags & INSTRUMENT_FIXED_VELOCITY)
		fprintf (f, "fixed_velocity\n");
	if (ip->flags & INSTRUMENT_RELEASE_LOOP)
		fprintf (f, "release_loop\n");
	if (ip->tremelo_frequency)
	{
		fprintf (f, "tremelo = %10.7f , %d;\n",
				 ip->tremelo_scale, ip->tremelo_frequency);
	}
	if (ip->vibrato_frequency)
	{
		fprintf (f, "vibrato = %10.7f , %d;\n",
				 ip->vibrato_scale, ip->vibrato_frequency);
	}

	switch (ip->chorus) {
	case INSTRUMENT_CHORUS_OFF:
		break;
	case INSTRUMENT_CHORUS_ON:
		fprintf (f, "chorus\n");
		break;
	case INSTRUMENT_CHORUS_ALWAYS:
		fprintf (f, "chorus always\n");
		break;
	}

	for (s = 0; s < ip->nsamples; s++)
	{
		sp = &p->samples[s];
		fprintf (f, "{\n");
		fprintf (f, "\trate = %d;\n", sp->sample_rate);
		fprintf (f, "\tlow = %d;\n", sp->low_freq);
		fprintf (f, "\thigh = %d;\n", sp->high_freq);
		fprintf (f, "\troot = %d;\n", sp->root_freq);
		fprintf (f, "\tscale = %f;\n", scale);
		fprintf (f, "\tloop = %f, %f;\n",
				 FixToFloat (sp->loop_start), FixToFloat (sp->loop_end));

		fprintf (f, "\t{\n");
		for (e = 0; e < sp->nenvelope; e++)
		{
			ep = &MSEnvelope(p, sp)[e];
			if (e == sp->release)
				fprintf (f, "\t\trelease\n");
			fprintf (f, "\t\t%10d, %10d;\n", ep->rate, ep->target);
		}
		fprintf (f, "\t}\n");
		l = 0;

		for (d = 0; d < FixToInt (sp->data_length); d++)
		{
			fprintf (f, " %9.6f", MSData(p,sp)[d]);
			l++;
			if (l == DATA_PER_LINE)
			{
				fprintf (f, "\n");
				l = 0;
			}
		}
		if (l != 0)
			fprintf (f, "\n");
		fprintf (f, "}\n");
	}
	if (!ferror (f))
		ret = 1;
	fclose (f);

//DEBUG:
#if 0
 { int i;  for (i=0; i<100; i++) printf("f[%d]=%f\n", i, p->data[i]); }
#endif
//END DEBUG

	return ret;
}


// Rename: gsimidi_pcm_write_instrument()
int
InstrumWriteBinary (const MidiPatch	*p,
					const float  scale,
					const char   *prefix, 
					const char   *file)
{
	const Instrument  *ip = p->instrument;
	FILE  *f;
	int   ret = 0;

	f = fopen (file, "w");

	if (!f) 
		return ret;

	fwrite (ip, sizeof (Instrument), 1, f);
	fwrite (p->samples, sizeof (SampleElement), ip->nsamples, f);
DEB(DPRT("write: ip->nsamples=%d\n", ip->nsamples));
	fwrite (p->envelopes, sizeof (EnvelopeElement), ip->envelopeTotal, f);
DEB(DPRT("write: ip->envelopeTotal=%d\n", ip->envelopeTotal));
	fwrite (p->data, sizeof (float), ip->dataTotal, f);
DEB(DPRT("write: ip->dataTotal=%d\n", ip->dataTotal));

//DEBUG:
#if 0
 { int i;  for (i=0; i<100; i++) printf("f[%d]=%f\n", i, p->data[i]); }
#endif
//END DEBUG

	if (!ferror (f))
		ret = 1;

	fclose (f);

	return ret;
}


/*
 * Write all instruments out
 */
static void
PatchName (const char *names, char *name)
{
	*name = '\0';
	while ( (names = match_next (names, name)) );
	while (*name)
	{
		if (*name == '+')
			*name = '_';
		name++;
	}
}


float
PatchVolume (const char *patch)
{
	FILE   *f;
	char   name[1024];
	float  volume;
	float  ret = 1.0f;

	f = OpenMidiVolumes("r");
	if (f)
	{
		while (fscanf (f, "%s %f", name, &volume) == 2)
		{
			if (match_any (name, patch))
			{
				ret = volume;
				break;
			}
		}
		fclose (f);
	}

	return ret;
}


int
PatchWrite (const char *patch,
			const MidiMap *map,
			Boolean writeInstruments,
			Boolean writeBinary)
{
	const MidiPatch *mp;
	FILE  *p;
	char  name[256];
	char  file[256];
	int   i;

	p = fopen (patch, "w");
	if (!p)
		return 0;

	for (i = 0; i < map->nmap; i++)
	{
		PatchName (map->map[i].name, name);
		strcpy (file, name);
		if (writeBinary)
			strcat (file, _GSIMIDIPCM_INSTRUMENT_NATIVE_BINARY__SUFFIX);
		else
			strcat (file, _GSIMIDIPCM_INSTRUMENT_NATIVE_TEXT__SUFFIX);

		if (writeInstruments)
		{
			PatchName (map->map[i].name, name);
			mp = LoadPatch (map, name);
			if (mp)
			{
				if (writeBinary)
					InstrumWriteBinary (mp,
										PatchVolume (map->map[i].name),
										name, file);
				else
					InstrumWriteText (mp,
									  PatchVolume (map->map[i].name),
									  name, file);
			}
		}
		fprintf (p, "%s %s\n", map->map[i].name, file);
	}
	fclose (p);

	return 1;
}
