/*
 * $Id: init.c,v 1.22 1998/06/04 05:33:28 keithp Exp $
 *
 * Modified by W.H.Scholten (dd 2000 - 2001)
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <string.h>
#include <stdlib.h>

#include "gsimidi_pcm_config.h"
#include "kpmidi.h"
#include "ui.h"
#include "midi_data.h"
#include "input/input.h"
#include "utils/debug.h"
#include "utils/log.h"


int
MidiInit (MidiData *d)
{
	DEB(DPRT("MidiInit\n"));

	if (!d->patchFile)
		d->patchFile = DEFAULT_PATCH;
	DEB(DPRT("init: d->patchFile=%s\n", d->patchFile));

	InitMidiPatch (d);
	d->sourceFuncs = &MidiFileSource;

	/*
	 * Compute envelope sub-sample rate, must be a divisor
	 * of the data rate.  By default, set it to 1000 per second
	 */
	if (!d->envelopeRate)
		d->envelopeRate = 1000;
	d->envelopeDiv = d->dataRate / d->envelopeRate;
	if (d->envelopeDiv < 1)
		d->envelopeDiv = 1;
	d->envelopeRate = d->dataRate / d->envelopeDiv;
	d->divisions = 1;

	/*
	 * Allocate voices array
	 */
	if (!d->maxVoices)
		d->maxVoices = GSIMIDI_PCM__VOICES;
	d->voices = calloc (d->maxVoices, sizeof (MidiVoice));
	if (!d->voices) {
		gsimidipcm_log_message(GSI_MESSAGE_ERROR, "Out of memory");
		return 0;
	}

	/*
	 * Allocate output buffer
	 */
	if (!d->buf)
		d->buf = malloc (d->bufSamples);
	if (!d->buf) {
		gsimidipcm_log_message(GSI_MESSAGE_ERROR, "Out of memory");
		return 0;
	}
	DEB(DPRT("init: d->buf=%p\n", d->buf));

	/*
	 * Initialize note table if not already done
	 */
	if (!d->noteTable[0])
		MidiSetTuning (d, equal440Table);

	DEB(DPRT("MidiInit DONE\n"));

	return 1;
}


void
MidiSetDefaults (MidiData *d)
{
	memset (d, '\0', sizeof (MidiData));
	d->dataRate = 44100;
	d->stereo = True;
	d->dataBytesPerSample = 2;
	d->bufSamples = 16384 /* 4096 */;
	d->patchFile = DEFAULT_PATCH;
	d->envelopeRate = 1000;
	d->maxVoices = GSIMIDI_PCM__VOICES;
	d->panPhase = GSIMIDI_PCM__DEFAULT_PAN_PHASE;
}


int
MidiSetFile (MidiData *d, char *fname)
{
	DEB(DPRT("MidiSetFile: before, d->divisions=%d, d->dataRate=%d\n", d->divisions, d->dataRate));
	if (*fname == ':') {
		d->sourceFuncs = &MidiInSource;
		if (!(*d->sourceFuncs->Read) (d, fname+1)) {
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Can't open raw midiport: %s", fname);
			return 0;
		}
	}
	else {
		d->sourceFuncs = &MidiFileSource;
		DEB(DPRT("MidiSetFile: file\n"));
		if (!(*d->sourceFuncs->Read) (d, fname)) {
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Not a known song format: %s", fname);
			return 0;
		}
	}
	DEB(DPRT("MidiSetFile: ok, d->divisions=%d, d->dataRate=%d\n", d->divisions, d->dataRate));
	MidiOutInit (d);
//	gsimidipcm_UiStart (d);
	d->stopped = False;

	return 1;
}


void
MidiDispose (MidiData *d)
{
	if (d->source)
		(*d->sourceFuncs->Dispose) (d);
}


void
MidiClose (MidiData *d)
{
	if (d->voices) {
		free (d->voices);
		d->voices = NULL;
	}
	if (d->buf) {
		free (d->buf);
		d->buf = NULL;
	}
}


void
MidiFini (MidiData *d)
{
	MidiDispose (d);
	MidiClose (d);
	FreeMidiPatch (d);
}


/* temperment = NULL will set default ("equal") */
Boolean
MidiSetTemperment (MidiData *d, char *temperment)
{
	if (!temperment || !strcmp (temperment, "equal"))
		MidiSetTuning (d, equal440Table);
	else if (!strcmp (temperment, "meanTone"))
		MidiSetTuning (d, mean415Table);
	else if (!strcmp (temperment, "WerkmeisterIII"))
		MidiSetTuning (d, werk3Table);
	else if (!strcmp (temperment, "equal415"))
		MidiSetTuning (d, equal415Table);
	else if (!strcmp (temperment, "vallotti"))
		MidiSetTuning (d, vallottiTable);
	else
		return False;

	return True;
}
