/*
 * $Id: midiparse.c,v 1.1 1998/06/01 06:03:15 keithp Exp $
 *
 * Modified by W.H.Scholten (dd 2000 - 2001)
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>

#include "kpmidi.h"
#include "ui.h"
#include "midi_data.h"
#include "utils/debug.h"

int
MidiParseInt (unsigned char	*buf,
	      unsigned long	*i)
{
    unsigned char   *s = buf;
    unsigned char   c;
    unsigned long   l;

    l = 0;
    for (;;)
    {
	c = *s++;
	l = (l << 7) | (c & 0x7f);
	if (!(c & 0x80))
	    break;
    }
    *i = l;
    return s - buf;
}

void
MidiAfterTouch (MidiData *d, 
		unsigned char channel, 
		unsigned char note,
		unsigned char velocity)
{
    int		voice;
    MidiVoice	*v;
    
    DPRINT ((DEBUG_NOTE, "After touch: %d %d %d", channel, note, velocity));
    for (voice = d->nactive; --voice >= 0;)
    {
	v = &d->voices[voice];
	if (v->channel == channel && v->note == note)
	{
	    v->velocity = velocity;
	    voiceVolume (d, voice);
	}
    }
}
		    
void
MidiControlChange (MidiData	    *d,
		   unsigned char    channel,
		   unsigned char    controller,
		   unsigned char    value)
{
    MidiChannel	*c = &d->channels[channel];
    int		voice;
    MidiVoice	*v;
    
    switch (controller) {
    case 1:	/* Modulation */
	DPRINT ((DEBUG_MIDI, "Modulation channel %d value %d",
		channel, value));
	c->modulation = value;
	MidiComputeTremelo (d, channel);
	MidiComputeVibrato (d, channel);
	break;
    case 6:	/* Data Entry MSB */
	if (c->pn_registered)
	{
	    switch ((c->pn_r_msb << 8) | c->pn_r_lsb) {
	    case 0x0000:	    /* Pitch-Bend sensitivity */
		c->bend_semi = value;
		MidiSetBendRange (d, channel);
		break;
	    }
	}
	break;
    case 7:	/* Main Volume */
	DPRINT ((DEBUG_MIDI, "Volume channel %d value %d",
		channel, value));
	c->volume = value;
	MidiChangeVolume (d, channel);
	gsimidipcm_UiSetChannelVolume (d, channel);
	break;
    case 10:	/* Pan */
	DPRINT ((DEBUG_MIDI, "Pan channel %d value %d",
		channel, value));
	c->pan = value;
	MidiChangeVolume (d, channel);
	break;
    case 11:	/* Expression */
	DPRINT ((DEBUG_MIDI, "Expression channel %d value %d",
		channel, value));
	c->expression = value;
	MidiChangeVolume (d, channel);
	break;
    case 38:	/* Data Entry LSB */
	if (c->pn_registered)
	{
	    switch ((c->pn_r_msb << 8) | c->pn_r_lsb) {
	    case 0x0000:	    /* Pitch-Bend sensitivity */
		c->bend_cent = value;
		MidiSetBendRange (d, channel);
		break;
	    }
	}
	break;
    case 64:	/* Sustain */
	DPRINT ((DEBUG_MIDI, "Sustain channel %d value %d",
		channel, value));
	c->sustain = value;
	/*
	 * Release all notes not currently held when
	 * sustain is released
	 */
	if (c->sustain < 0x40)
	{
	    for (voice = d->nactive; --voice >= 0; )
	    {
		v = &d->voices[voice];
		if (v->channel == channel && !v->on)
		    voiceRelease (d, voice, 0);
	    }
	}
	break;
    case 93:	/* Effects 3 Depth */
	c->chorus = value;
	break;
    case 96:	/* Data Increment */
	break;
    case 97:	/* Data Decrement */
	break;
    case 98:	/* Unregistered Paramter number MSB */
	c->pn_u_msb = value;
	c->pn_registered = 0;
	break;
    case 99:	/* Unregistered Paramter number LSB */
	c->pn_u_lsb = value;
	c->pn_registered = 0;
	break;
    case 101:	/* Registered Parameter number MSB */
	c->pn_r_msb = value;
	c->pn_registered = 1;
	break;
    case 102:	/* Registered Parameter number LSB */
	c->pn_r_lsb = value;
	c->pn_registered = 1;
	break;
    case 120:	/* All Sound Off */
    case 123:	/* All notes off */
	DPRINT ((DEBUG_MIDI, "All off channel %d %d",
		channel, controller));
	for (voice = d->nactive; --voice >= 0; )
	{
	    v = &d->voices[voice];
	    if (v->channel == channel)
	    {
		v->on = False;
		voiceRelease (d, voice, 0);
	    }
	}
	break;
    case 121:	/* Reset all controllers */
	DPRINT ((DEBUG_MIDI, "Reset controllers channel %d",
		channel));
	MidiResetControllers (d, channel);
	break;
	/* case 122:    Local Control On/Off */
    case 124:	/* Omni Mode Off */
    case 125:	/* Omni Mode On */
    case 126:	/* Mono Mode On */
    case 127:	/* Poly Mode On */
	DPRINT  ((DEBUG_MIDI, "Mode change message %d", controller));
	break;
    default:
	DPRINT ((DEBUG_MIDI, "Unknown controller %d channel %d value %d",
		controller, channel, value));
	break;
    }
}

void
MidiProgramChange (MidiData	    *d,
		   unsigned char    channel,
		   unsigned char    program)
{
    MidiChannel	*c = &d->channels[channel];

    DPRINT ((DEBUG_PROGRAM, "Program change: %d %d (%s)",
	   channel, program, patchNames[program]));
    /*
     * Attempt to avoid changes to the drum channel -- assume
     * that setting the drum channel to 0 is not a request
     * to select the acoustic_grand
     */
    if (!c->drum || (channel != 9 && program))
    {
	c->program = program;
	c->drum = False;
	gsimidipcm_UiProgramChange (d, channel, c->program);
    }
}

void
MidiChannelPressure (MidiData	    *d,
		     unsigned char  channel,
		     unsigned char  pressure)
{
    int		voice;
    MidiVoice	*v;
    
    DPRINT ((DEBUG_NOTE, "Channel pressure: %d %d", channel, pressure));
    for (voice = d->nactive; --voice >= 0; )
    {
	v = &d->voices[voice];
	if (v->channel == channel)
	{
	    v->velocity = pressure;
	    voiceVolume (d, voice);
	}
    }
}

void
MidiPitchWheel (MidiData	*d,
		unsigned char	channel,
		unsigned short	value)
{
    MidiChannel	*c = &d->channels[channel];
    
    DPRINT ((DEBUG_NOTE, "Pitch wheel: %d %d", channel, value));
    c->bend = value;
    MidiComputeBend (d, channel);
}

static unsigned char *
saveCountedText (unsigned char *buf, int len)
{
    unsigned char   *s;

    s = malloc (len + 1);
	if (!s)
		return NULL; /* Out of memory */
    memcpy (s, buf, len);
    s[len] = '\0';
    return s;
}

void
MidiFileSysex (MidiData		*d,
	       MidiParseState	*s,
	       unsigned char	*buf,
	       unsigned char	type,
	       int		len)
{
    char    *text;
    
    switch (type)
    {
    case 0x00:
	/* Trace sequence */
	break;
    case 0x01:
    case 0x02:
    case 0x03:
    case 0x04:
    case 0x05:
    case 0x06:
    case 0x07:
	text = saveCountedText (buf, len);
	gsimidipcm_UiSetText (d, 0, text, (MidiText) type);
	free (text);
	break;
    case 0x20:	/* set data channel */
	break;
    case 0x2f:	/* EOF */
	s->state = midiEOF;
	break;
    case 0x51:	/* set tempo */
	/*
	 * Tempo is measured in usec/qnote
	 */
	d->tempo = (((int32) buf[0] << 16) | 
		    ((int32) buf[1] << 8) | 
		    ((int32) buf[2]));
	DPRINT ((DEBUG_MIDI, "Tempo %d\n", d->tempo));
	d->midiSpeed = 1.0 / (d->tempo * d->midiSpeedFactor);
	gsimidipcm_UiSetTempo (d);
	break;
    case 0x54:	/* SMPTE offset */
	break;
    case 0x58:	/* Time signature */
	break;
    case 0x59:	/* Key signature */
	break;
    default:	/* unknown */
	break;
    }
}

void
MidiParseTune (MidiData		*d,
	       unsigned char	*buf,
	       int		note)
{
    unsigned char   base, c0, c1;
    int		    cents;

    base = buf[0];
    c0 = buf[1];
    c1 = buf[2];
    cents = ((c0 & 0x7f) << 7) | (c1 & 0x7f);
    MidiChangeTuning (d, note, base, cents * 100 / 16384);
}

void
MidiRealTime (MidiData *d, unsigned char message)
{
    switch (message) {
    case 0xf8:	    /* Midi clock */
    case 0xfa:	    /* Midi start */
    case 0xfb:	    /* Midi continue */
    case 0xfc:	    /* Midi stop */
    case 0xfe:	    /* Active sense */
    case 0xff:	    /* Reset */
    }
}

void
MidiInitState (MidiParseState *s)
{
    s->state = midiStart;
    s->value = 0;
    s->count = 0;
    s->limit = 0;
    s->obeyRealtime = False;
    s->status = 0;
}

int
MidiParseOne (MidiData		*d,
	      MidiParseState	*s,
	      unsigned char	*buf,
	      int		avail,
	      Boolean		seek)
{
    unsigned char   *buf_start = buf, *buf_end = buf + avail;
    unsigned char   b;
    
    do {
	/*
	 * Check for real-time messages
	 */
	b = *buf;
	if (s->obeyRealtime && 0xf8 <= b && b <= 0xff)
	{
	    MidiRealTime (d, b);
	    buf++;
	} 
	else switch (s->state) 
	{
	case midiStart:
	    b = *buf;
	    s->state = midiRunning;
	    /*
	     * Check for running status
	     */
	    if ((b & 0x80) != 0)
	    {
		s->status = b;
		buf++;
		break;
	    }
	    /* fall through ... */
	case midiRunning:
	    /* quick check for out-of-sync */
	    if (!(s->status & 0x80))
	    {
		s->state = midiSkip;
		break;
	    }
	    switch ((s->status >> 4) & 0x07) 
	    {
	    case 0:	/* note off */
		s->buf[0] = *buf++;
		s->state = midiNoteOff;
		break;
	    case 1:	/* note on */
		s->buf[0] = *buf++;
		s->state = midiNoteOn;
		break;
	    case 2:	/* after touch */
		s->buf[0] = *buf++;
		s->state = midiAfterTouch;
		break;
	    case 3:	/* control change */
		s->buf[0] = *buf++;
		s->state = midiControlChange;
		break;
	    case 4:	/* program change */
		MidiProgramChange (d, s->status & 0xf, *buf++);
		s->state = midiStart;
		break;
	    case 5:	/* channel pressure */
		MidiChannelPressure (d, s->status & 0xf, *buf++);
		s->state = midiStart;
		break;
	    case 6:	/* pitch wheel */
		s->buf[0] = *buf++;
		s->state = midiPitchWheel;
		break;
	    case 7:	/* sysex */
		switch (s->status & 0xf)
		{
		case 0x0:
		    s->state = midiSysex;
		    break;
		case 0xf:	/* midi file extensions */
		    s->state = midiSysexFile;
		    break;
		}
		break;
	    }
	    break;
	case midiNoteOff:
	    b = *buf++;
	    MidiNoteOff (d, s->status & 0xf, s->buf[0], b);
	    s->state = midiStart;
	    break;
	case midiNoteOn:
	    b = *buf++;
	    MidiNoteOn (d, s->status & 0xf, s->buf[0], b, seek);
	    s->state = midiStart;
	    break;
	case midiAfterTouch:
	    b = *buf++;
	    MidiAfterTouch (d, s->status & 0xf, s->buf[0], b);
	    s->state = midiStart;
	    break;
	case midiControlChange:
	    b = *buf++;
	    MidiControlChange (d, s->status & 0xf, s->buf[0], b);
	    s->state = midiStart;
	    break;
	case midiPitchWheel:
	    b = *buf++;
	    MidiPitchWheel (d, s->status & 0xf, (b << 7) | s->buf[0]);
	    s->state = midiStart;
	    break;
	case midiSysex:
	    b = *buf++;
	    switch (b) {
	    case 0x7e:
		s->state = midiSysexUniversalNonRealTime;
		break;
	    case 0x7f:
		s->state = midiSysexUniversalRealTime;
		break;
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexUniversalNonRealTime:
	    b = *buf++;
	    switch (b) {
	    case 0x08:	    /* Midi tuning standard */
		s->state = midiSysexTuningNRT;
		break;
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexTuningNRT:
	    b = *buf++;
	    switch (b) {
	    case 0x00:	    /* bulk dump request */
		s->state = midiSkip;
		break;
	    case 0x01:	    /* bulk dump reply */
		s->value = 0;
		
		s->count = 0;
		s->limit = 17;
		s->state = midiSkipSome;
		
		s->returnState = midiSysexTuningBulkReply;
		break;
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexTuningBulkReplyMore:
	    MidiParseTune (d, s->buf, s->value);
	    if (++s->value == 128)
	    {
		s->state = midiStart;
		break;
	    }
	    /* fall through ... */
	case midiSysexTuningBulkReply:
	    s->count = 0;
	    s->limit = 3;
	    s->returnState = midiSysexTuningBulkReplyMore;
	    s->state = midiBufSome;
	    break;
	case midiSysexUniversalRealTime:
	    b = *buf++;
	    switch (b) 
	    {
	    case 0x04:	    /* Device control */
		s->state = midiSysexDeviceControl;
		break;
	    case 0x08:	    /* Midi tuning standard */
		s->state = midiSysexTuningRT;
		break;
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexDeviceControl:
	    b = *buf++;
	    switch (b)
	    {
	    case 0x00:	    /* Master volume */
	    case 0x01:	    /* Master balance */
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexTuningRT:
	    b = *buf++;
	    switch (b)
	    {
	    case 0x02:	    /* Note change */
		s->state = midiSysexTuningRTNoteChange;
		break;
	    default:
		s->state = midiSkip;
		break;
	    }
	    break;
	case midiSysexTuningRTNoteChange:
	    /*
	     * Real time note changes:
	     *
	     *	nnotes [ note tuning ]*
	     */
	    s->value = *buf++;	/* number of notes */
	    s->state = midiSysexTuningRTNoteChangeMore;
	    break;
	case midiSysexTuningRTNoteChangeMoreCont:
	    MidiParseTune (d, s->buf + 1, s->buf[0]);
	    /* fall through ... */
	case midiSysexTuningRTNoteChangeMore:
	    if (--s->value == 0)
	    {
		s->state = midiStart;
	    }
	    else
	    {
		s->limit = 4;
		s->count = 0;
		s->state = midiBufSome;
		s->returnState = midiSysexTuningRTNoteChangeMoreCont;
	    }
	    break;
	case midiSysexFile:
	    s->buf[0] = *buf++;
	    s->returnState = midiSysexFileMore;
	    s->state = midiInt;
	    break;
	case midiSysexFileMore:
	    MidiFileSysex (d, s, buf, s->buf[0], s->value);
	    buf += s->value;
	    s->state = midiStart;
	    break;
	case midiInt:
	    s->value = 0;
	    s->state = midiIntMore;
	    /* fall through */
	case midiIntMore:
	    b = *buf++;
	    s->value = s->value << 7 | (b & 0x7f);
	    if (!(b & 0x80))
		s->state = s->returnState;
	    break;
	case midiBufSome:
	    s->buf[s->count++] = *buf++;
	    if (s->count == s->limit)
		s->state = s->returnState;
	    break;
	case midiSkipSome:
	    ++buf;
	    if (++s->count == s->limit)
		s->state = s->returnState;
	    break;
	case midiSkip:
	    b = *buf++;
	    if (b & 0x80)
		s->state = midiStart;
	    break;
	}
	if (s->state == midiEOF)
	    return -1;
    } while (buf < buf_end && s->state != midiStart);
    return buf - buf_start;
}
