/*
 * $Id: midi.c,v 1.41 1998/06/01 06:03:14 keithp Exp $
 * => upgraded to 1.44 1999/08/26
 *
 * Modified by W.H.Scholten (dd 2000 - 2001)
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <string.h>
#include "kpmidi.h"
#include "ui.h"
#include "midi_to_pcm.h"
#include "utils/string_match.h"
#include "utils/debug.h"
#include "volume.h"

void	
voiceWaveInit (MidiData   *d,
		MidiVoice  *v,
		double      frequency,
		Boolean     loopRelease,
		float       pan)
{
	VoiceWave		*vw = &v->wave;
	const SampleElement	*se = v->se;

	vw->freqFactor = ((double) se->sample_rate * (double) FixFromInt (1) / 
					  ((double) se->root_freq * (double) d->dataRate));
	vw->samples = MSData (v->patch, se);

	vw->loop_len = se->loop_end - se->loop_start;

	vw->loopRelease = loopRelease;
	vw->sample = 0;

	vw->end = se->loop_end;
	vw->data_end = se->data_length;

	voiceWaveChange (vw, frequency);
}

#define PAN_RATIO   (1.0f / 2.0f)

/*
 * Given value in range 1.0->0.5, generate value
 * in range 1.0 -> PAN_RATIO
 */

#define PanM	(2.0f - 2.0f * PAN_RATIO)
#define PanB	(2.0f * PAN_RATIO - 1.0f)
#define PanF(p) (PanM * (p) + PanB)

static void
voiceEnvelopeSet (MidiVoice *v, float out_scale)
{
	VoiceEnvelope *ve = &v->envelope;
	ve->fValue = voiceEnvelopeValue (ve) * out_scale;
	if (ve->fPan <= 0.5) {
		ve->fLeft = ve->fValue;
		ve->fRight = ve->fValue * PanF(0.5f + ve->fPan);
	}
	else {
		ve->fLeft = ve->fValue * PanF(1.5f - ve->fPan);
		ve->fRight = ve->fValue;
	}
}

static void
voiceComputeEnvelope (MidiVoice *v, float out_scale)
{
	VoiceEnvelope           *ve = &v->envelope;
	const SampleElement     *se = v->se;
	const EnvelopeElement   *ee;
	double			rate;
	int32			target;

	ee = &MSEnvelope(v->patch, se)[ve->element];
	target = ee->target;
	/*
	 * negative rates -- sustain for that many "samples"
	 */
	if (ee->rate < 0) {
		ve->valueStep = 0;
		ve->target = target;
		ve->value = ve->target;
		ve->remain = -ee->rate / ve->envelopeRatio;
	}
	else {
		rate = (double) (ee->rate + ee->noteRate * v->note) * ve->envelopeRatio;
		if (rate < 0)
			rate = 1;
		if (rate > 2e9)
			rate = 2e9;
		ve->valueStep = (int32) rate;
		ve->target = target;
		if (ve->valueStep == 0) {
			ve->remain = -1;
		}
		else {
			if (ve->target < ve->value)
				ve->valueStep = -ve->valueStep;
			ve->remain = (ve->target - ve->value) / ve->valueStep;
			if (ve->remain == 0) {
				ve->valueStep = 0;
				ve->remain = 1;
			}
		}
	}
	voiceEnvelopeSet (v, out_scale);
	DPRINT ((DEBUG_ENVELOPE, "element %d %08x->%08x(%08x) at %08x remain %d\n",
			 ve->element,
			 ve->value,
			 ve->target,
			 ve->value + ve->valueStep * ve->remain,
			 ve->valueStep,
			 ve->remain));
}

static void
voiceChangePan (MidiVoice *v, float pan, float panPhase, unsigned long dataRate)
{
	v->envelope.fPan = pan;
	v->phase = (pan - 0.5f) * 2.0f * panPhase * dataRate;
	if (v->phase < -MAX_PHASE)
		v->phase = -MAX_PHASE;
	else if (v->phase > MAX_PHASE)
		v->phase = MAX_PHASE;
}

static void
voiceChangeVolume (MidiVoice *v, float volume)
{
	VoiceEnvelope		*ve = &v->envelope;
//	const SampleElement		*se = v->se;
	int				vind;

	vind = (int) (volume * ((1 << VOLUME_BITS) - 1));
	ve->volume = volumeTable[vind];
	DPRINT ((DEBUG_NOTE, "Change Volume %g %d %g\n", volume, vind, ve->volume));
	ve->mult = ve->volume * ve->in_scale;
}

static Boolean
voiceSetEnvelopeElement (MidiVoice *v, int element, float out_scale)
{
	VoiceEnvelope           *ve = &v->envelope;
	const SampleElement     *se = v->se;
	const EnvelopeElement   *ee = MSEnvelope(v->patch, se);

	if (element >= se->nenvelope) {
		ve->element = se->nenvelope;
		ve->value = 0;
		ve->valueStep = 0;
		ve->remain = -1;
		ve->fValue = 0.0f;
		ve->fLeft = 0.0f;
		ve->fRight = 0.0f;
		return False;
	}
	if (ve->on && element == se->release) {
		if (ee[se->release-1].target > 0x8000000) {
			ve->valueStep = 0;
			return True;
		}
	}

	ve->element = element;
	voiceComputeEnvelope (v, out_scale);

	return True;
}

Boolean
voiceEnvelopeInit (MidiData *d,
		 MidiVoice		*v,
		 float		volume,
		 float		in_scale,
		 float		pan)
{
	VoiceEnvelope	*ve = &v->envelope;

	ve->on = True;
	ve->element = 0;
	ve->in_scale = in_scale;
	ve->envelopeRatio = (double) ENVELOPE_RATE / (double) d->envelopeRate;
	ve->fTrem = 1.0f;
	ve->value = 0;
	voiceChangePan (v, pan, d->panPhase, d->dataRate);
	voiceChangeVolume (v, volume);
	DPRINT ((DEBUG_ENVELOPE, "Envelope volume %d\n", (int) (ve->volume * 1000.0)));

	return voiceSetEnvelopeElement (v, 0, d->scale);
}

void
voiceLFOChange (VoiceLFO *vl,
		int32		frequency,
		unsigned long	envelopeRate,
		float		scale)
{
	double	step;

	if (scale) {
		step = (double) frequency / (double) envelopeRate * (65536.0 * 65536.0 / 1000.0);
		vl->step = (MidiSine) step;
	}
	else
		vl->step = 0;

	vl->scale = scale;
}

void
voiceLFOInit (VoiceLFO  *vl,
		int32		frequency,
		unsigned long envelopeRate,
		double		scale)
{
	vl->value = 0;
	voiceLFOChange (vl, frequency, envelopeRate, scale);
}

void
voiceIdle (MidiData *d, int voice)
{
	MidiVoice *v = &d->voices [ voice ];

	DPRINT ((DEBUG_NOTE, "Idle %d %d\n",
		d->voices[voice].channel,
		d->voices[voice].note));

	gsimidipcm_UiNoteIdle (d, v->channel, v->note);

	d->nactive--;
	if (voice < d->nactive)
	{
	memmove (&d->voices[voice], &d->voices[voice+1],
		(d->nactive - voice) * sizeof (MidiVoice));
	}
}

void
voicePress (MidiData		*d,
		int			voice,
		const MidiPatch	*patch,
		float		volume,
		float		in_scale,
		double		frequency,
		float		pan)
{
	const Instrument	*instrument = patch->instrument;
	MidiVoice		*v = &d->voices[voice];
//	VoiceEnvelope   *ve = &v->envelope;
//	VoiceWave		*vw = &v->wave;
	const SampleElement   *se;
	int			s;
	Boolean		loopRelease = False;
	int			closest;
	int32			ifreq;
	int32		dist;
	int32		mindist;

	DPRINT ((DEBUG_NOTE, "Press %d %d\n",
		v->channel, v->note));
	v->on = True;
	v->instrument = instrument;
	v->patch = patch;
	ifreq = (int) frequency;
	mindist = 0x7fffffff;
	closest = 0;
	for (s = 0; s < instrument->nsamples; s++)
	{
	se = &patch->samples[s];
	if (se->low_freq <= ifreq && ifreq <= se->high_freq)
	{
		closest = s;
		break;
	}
	dist = se->root_freq - ifreq;
	if (dist < 0)
		dist = -dist;
	if (dist < mindist)
	{
		closest = s;
		mindist = dist;
	}
	}
	for (s = 0; s < MAX_PHASE; s++)
	v->delay[s] = 0.0f;
	se = &patch->samples[closest];
	v->se = se;
	if (instrument->flags & INSTRUMENT_FIXED_FREQUENCY)
	frequency = se->root_freq;
	if (instrument->flags & INSTRUMENT_RELEASE_LOOP)
	loopRelease = True;
	voiceEnvelopeInit (d, v, volume, in_scale, pan);
	voiceWaveInit (d, v, frequency, loopRelease, pan);
	gsimidipcm_UiNoteOn (d, v->channel, v->note);
}

void
voiceBend (MidiData		*d,
	 int			voice,
	 double		frequency,
	 unsigned long	sampleRate)
{
	MidiVoice		*v = &d->voices[voice];
	VoiceWave		*vw = &v->wave;

	voiceWaveChange (vw, frequency);
}

void
voiceVolume (MidiData		*d,
		int		voice)
{
	MidiVoice		*v = &d->voices[voice];
	MidiChannel		*c = &d->channels[v->channel];
	VoiceEnvelope	*ve = &v->envelope;
	float		old;
	float		volume;

	volume = Volume (v->velocity, c->volume, c->expression);
	old = ve->volume;
	voiceChangeVolume (v, volume);
	DPRINT ((DEBUG_NOTE, "Volume voice %d (vel %d vol %d exp %d) %g %g->%g\n",
		voice, 
		v->velocity, c->volume, c->expression,
		volume, old, ve->volume));
	voiceEnvelopeSet (v, d->scale);
}
		
void
voicePan (MidiData  *d,
	int		voice,
	float		pan)
{
	MidiVoice		*v = &d->voices[voice];

	voiceChangePan (v, pan, d->panPhase, d->dataRate);
	voiceEnvelopeSet (v, d->scale);
}

void
voiceRelease (MidiData		*d,
		int		voice,
		int		velocity)
{
	MidiVoice		*v = &d->voices[voice];
	VoiceWave		*vw = &v->wave;
	VoiceEnvelope	*ve = &v->envelope;
	const SampleElement	*se = v->se;

	DPRINT ((DEBUG_NOTE, "Release %d %d\n",
		v->channel, v->note));
	vw->end = vw->data_end;
	ve->on = False;
	gsimidipcm_UiNoteOff (d, v->channel, v->note);
	if (ve->element < se->release)
	{
	if (!voiceSetEnvelopeElement (v, se->release, d->scale))
		voiceIdle (d, voice);
	}
}

int
AllocVoices (MidiData	*d, int chorus)
{
	int		voice, minVoice;
	float	min;
	MidiVoice	*v;

	while (d->nactive + chorus > d->maxVoices)
	{
	minVoice = 0;
	min = d->voices[0].envelope.fValue;
	for (voice = 1; voice < d->nactive; voice++)
	{
		v = &d->voices[voice];
		if (v->envelope.fValue < min &&
		v->envelope.element >= v->se->release)
		{
		min = d->voices[voice].envelope.fValue;
		minVoice = voice;
		}
	}
	voiceIdle (d, minVoice);
	}
	voice = d->nactive;
	d->nactive += chorus;
	return voice;
}

void
MidiNoteOn (MidiData		*d,
		unsigned char   channel,
		unsigned char   note,
		unsigned char   velocity,
		Boolean		seek)
{
	int			voice;
	double		frequencyOffset;
	float		in_scale;
	float		volume;
	float		pan, panOffset;
	const MidiPatch	*patch;
	const Instrument	*instrument;
	MidiVoice		*v;
	int			chorus, odd;
	MidiChannel		*c = &d->channels[channel];
	Boolean		first;

	if (velocity == 0)
	{
		MidiNoteOff (d, channel, note, 0);
		return;
	}

	if (c->drum)
	{
		if (note < MIN_DRUM || MAX_DRUM < note)
			note = MIN_DRUM + 3;
		patch = d->drumMap[note-MIN_DRUM];
		if (!patch)
		{
			patch = MidiLoadDrum (d, note-MIN_DRUM);
			if (!patch)
				return;
			d->drumMap[note-MIN_DRUM] = patch;
		}
		in_scale = drumVolume[note-MIN_DRUM];
	}
	else
	{
		patch = d->patchMap[c->program];
		if (!patch)
		{
			patch = MidiLoadProgram (d, c->program);
			if (!patch)
				return;
			d->patchMap[c->program] = patch;
		}
		in_scale = patchVolume[c->program];
	}
	if (seek) return;

	instrument = patch->instrument;

	switch (d->chorus) {
	case 0:
		chorus = 1;
		break;
	case 1:
		chorus = 1;
		switch (instrument->chorus) {
		case INSTRUMENT_CHORUS_OFF:
			if (c->chorus > 96)
				chorus = 2;
			break;
		case INSTRUMENT_CHORUS_ON:
			if (c->chorus > 32)
				chorus = 2;
			break;
		case INSTRUMENT_CHORUS_ALWAYS:
			chorus = 2;
			break;
		}
		break;
	default:
		chorus = d->chorus;
		break;
	}


	if (instrument->flags & INSTRUMENT_FIXED_VELOCITY)
		velocity = 64;

	DPRINT ((DEBUG_NOTE, "Note on: ch %d no %d vel %d vol %d exp %d %g", 
			 channel, note, velocity, c->volume, c->expression,
			 Volume (velocity, c->volume, c->expression)));

	odd = chorus & 1;
	volume = Volume (velocity, c->volume, c->expression);
	in_scale = in_scale / chorus;

	first = True;
	voice = AllocVoices (d, chorus);
	while (chorus--)
	{
#if 0
	{
		static int maxVoice;

		if (voice > maxVoice)
		{
			maxVoice = voice;
			printf ("%d\n", maxVoice);
		}
	}
#endif
	v = &d->voices[voice];

	if (first && odd)
	{
		frequencyOffset = 1;
		panOffset = 0.0f;
	}
	else
	{
		if ((chorus - odd) & 1)
		{
#if 0
			/* Old v1.41 code: */
			frequencyOffset = centsTable[(random() & 7)+1];
			panOffset = (float) (chorus + 1) / 20.0f;
#else
			/* Upgraded to v1.44: */
			frequencyOffset = centsTable[chorus*2 + 2];
			panOffset = (float) (chorus + 1) / 10.0f;
#endif
		}
		else
		{
			frequencyOffset = 1/frequencyOffset;
			panOffset = -panOffset;
		}
	}

	pan = (float) c->pan / 128.0f;
	pan = pan + panOffset;
	if (d->auto_pan && (instrument->flags & INSTRUMENT_AUTO_PAN))
	{
		float   auto_dist = (note / 127.0f) - 0.5f;

		if (pan > 0.5f)
		{
		pan = pan + (1.0f - pan) * auto_dist;
		}
		else
		{
		pan = pan + pan * auto_dist;
		}
	}
	if (pan > 1.0f)
		pan = 1.0f;
	else if (pan < 0.0f)
		pan = 0.0f;
	v->channel = channel;
	v->velocity = velocity;
	v->note = note;
	voiceLFOInit (&v->vibrato,
				  instrument->vibrato_frequency,
				  d->envelopeRate,
				  instrument->vibrato_scale * c->modulation);

	voiceLFOInit (&v->tremelo,
				  instrument->tremelo_frequency,
				  d->envelopeRate,
				  instrument->tremelo_scale * c->modulation);

	voicePress (d, voice,
				patch,
				volume,
				in_scale,
				VoiceFrequency(d, voice) * frequencyOffset,
				pan);
	first = False;
	voice++;
	}
}

void
MidiNoteOff (MidiData		*d,
		unsigned char  channel,
		unsigned char  note,
		unsigned char  velocity)
{
	MidiVoice		*v;
	MidiChannel		*c = &d->channels[channel];
	int			voice;

	DPRINT ((DEBUG_NOTE, "Note off: %d %d %d", channel, note, velocity));
	for (voice = d->nactive; --voice >= 0;)
	{
	v = &d->voices[voice];
	if (v->channel == channel && v->note == note && v->on)
	{
		v->on = False;
		if (c->sustain < 0x40)
		voiceRelease (d, voice, velocity);
	}
	}
}

float	last_preloaded_value;

#define	PRELOAD_STEP	128

float
MidiPreloadSample (const volatile float *data, int32 data_length)
{
	float   t;

	t = 0.0f;
	while (data_length > 0)
	{
	t += *data;
	data += PRELOAD_STEP;
	data_length -= PRELOAD_STEP;
	}
	return t;
}

void
MidiPreloadInstrument (const MidiPatch *patch)
{
	int			n;
	float		t;
	const Instrument	*instrument = patch->instrument;

	t = 0.0f;
	for (n = 0; n < instrument->nsamples; n++)
	t += MidiPreloadSample (MSData(patch, &patch->samples[n]),
				FixToInt (patch->samples[n].data_length));

	last_preloaded_value = t;
}

const MidiPatch *
MidiLoadProgram (MidiData *d, unsigned char program)
{
	const MidiPatch *patch;

	patch = LoadPatch (d->map, patchNames[program]);
	if (patch)
	MidiPreloadInstrument (patch);
	return patch;
}

const MidiPatch *
MidiLoadDrum (MidiData *d, unsigned char program)
{
	const MidiPatch	*patch;

	patch = LoadPatch (d->map, drumNames[program]);
	if (patch)
	MidiPreloadInstrument (patch);
	return patch;
}

double
VoiceFrequency (MidiData *d, int voice)
{
	MidiVoice	*v = &d->voices[voice];
	MidiChannel	*c = &d->channels[v->channel];
	double	vib;

	vib = SineValue(v->vibrato.value) * v->vibrato.scale + 1.0;
	return d->noteTable[v->note] * (vib * c->bend_ratio);
}

void
MidiComputeBend (MidiData *d, unsigned char channel)
{
	MidiChannel	*c = &d->channels[channel];
	int		voice;
	int		cents;
	int		semis;
	int		s;
	MidiVoice	*v;

	s = ((int) c->bend) - (int) 0x2000;
	cents = (c->bend_range * s) / 0x1fff;
	if (s < 0)
	cents = -cents;
	semis = cents / 100;
	cents = cents % 100;
	c->bend_ratio = semiTable[semis] * centsTable[cents];
	if (s < 0)
	c->bend_ratio = 1/c->bend_ratio;
	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	if (v->channel == channel)
	{
		voiceWaveChange (&v->wave, VoiceFrequency (d, voice));
	}
	}
}

void
MidiComputeTremelo (MidiData *d, unsigned char channel)
{
	MidiChannel	*c = &d->channels[channel];
	int		voice;
	MidiVoice	*v;

	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	if (v->channel == channel)
	{
		voiceLFOChange (&v->tremelo,
				v->instrument->tremelo_frequency,
				d->envelopeRate,
				v->instrument->tremelo_scale * c->modulation);
	}
	}
}

void
MidiComputeVibrato (MidiData *d, unsigned char channel)
{
	MidiChannel	*c = &d->channels[channel];
	int		voice;
	MidiVoice	*v;

	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	if (v->channel == channel)
	{
		voiceLFOChange (&v->vibrato,
				v->instrument->vibrato_frequency,
				d->envelopeRate,
				v->instrument->vibrato_scale * c->modulation);
	}
	}
}

void
MidiChangePan (MidiData *d, unsigned char channel)
{
//	MidiChannel	*c = &d->channels[channel];
	int		voice;
	MidiVoice	*v;
	float	pan;

	pan = (float) d->channels[channel].pan / 128.0f;
	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	if (v->channel == channel)
	{
		voicePan (d, voice, pan);
	}
	}
}

void
MidiChangeVolume (MidiData *d, unsigned char channel)
{
//	MidiChannel	*c = &d->channels[channel];
	int		voice;
	MidiVoice	*v;

	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	if (v->channel == channel)
	{
		voiceVolume (d, voice);
	}
	}
}

void
MidiChangePanPhase (MidiData *d, float panPhase)
{
	MidiVoice	*v;
	int		voice;

	d->panPhase = panPhase;
	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	voiceChangePan (v, v->envelope.fPan, panPhase, d->dataRate);
	}
}

void
MidiSetBendRange (MidiData *d, unsigned char channel)
{
	MidiChannel	*c = &d->channels[channel];
//	int		cents, semi;

	c->bend_range = (int) c->bend_semi * 100 + (int) c->bend_cent;
}

Boolean
MidiAddReverbTap (MidiData *d, float amount, int us)
{
	MidiReverbTap   *taps;
	MidiReverb		*rev;
	int			t;
	unsigned long   offset;
	unsigned long   length;

	offset = (unsigned long) ((double) d->dataRate * (double) us / 1000000.0);
	for (t = 0; t < d->reverbCount; t++) {
		if (d->reverbTaps[t].offset == offset) {
			d->reverbTaps[t].amount = amount;
			return; // WHS: Return what?? False ??
		}
	}
	taps = calloc (sizeof (MidiReverbTap), d->reverbCount + 1);
	if (!taps)
		return False;

	for (t = 0; t < d->reverbCount; t++)
	taps[t] = d->reverbTaps[t];
	taps[t].amount = amount;
	taps[t].offset = (unsigned long) offset;
	if (d->reverbMask <= offset)
	{
	length = 1;
	while (length <= offset)
		length <<= 1;
	rev = calloc (sizeof (MidiReverb), length);
	if (!rev)
	{
		free (taps);
		return False;
	}
	if (d->reverb)
	{
		memcpy (rev, d->reverb, sizeof (MidiReverb) * (d->reverbMask + 1));
		free (d->reverb);
	}
	d->reverb = rev;
	d->reverbMask = length - 1;
	}
	if (d->reverbTaps)
	free (d->reverbTaps);
	d->reverbTaps = taps;
	d->reverbCount++;

	return True;
}

void
MidiResetReverb (MidiData *d)
{
	if (d->reverbTaps)
	{
	free (d->reverbTaps);
	d->reverbTaps = 0;
	}
	if (d->reverb)
	{
	free (d->reverb);
	d->reverb = 0;
	}
	d->reverbMask = 0;
	d->reverbCount = 0;
	d->reverbPos = 0;
}

void
MidiResetControllers (MidiData *d, int channel)
{
	MidiChannel		*c = &d->channels[channel];

	c->volume = 90;
	c->expression = 127;
	gsimidipcm_UiSetChannelVolume (d, channel);
	c->pan = 64;
	c->sustain = 0;
	c->bend = 0x2000;
	c->bend_semi = 2;
	c->bend_cent = 0;
	c->pn_r_msb = 0;
	c->pn_r_lsb = 0;
	c->pn_u_msb = 0;
	c->pn_u_lsb = 0;
	c->pn_registered = 0;
	c->modulation = 64;
	c->chorus = 64;
	MidiSetBendRange (d, channel);
	MidiComputeBend (d, channel);
	MidiComputeVibrato (d, channel);
	MidiComputeTremelo (d, channel);
}

void
MidiAllNotesOff (MidiData *d)
{
	int	voice;
	MidiVoice	*v;
	for (voice = d->nactive; --voice >= 0; )
	{
	v = &d->voices[voice];
	v->on = False;
	voiceRelease (d, voice, 0);
	}
}

/*
 * Compute length of midi song in microseconds
 */
double
MidiSeek (MidiData *d, double usec)
{
	int			ready;
	double		total;

	MidiReset (d);
	total = 0.0;
	d->eventTime = d->time;
	do
	{
		/*
		 *   ticks * usec/qnote / (ticks/qnote)
		 * = ticks * usec/qnote * qnote / ticks
		 * = usec
		 */
	total += (d->eventTime - d->time) * ((double) d->tempo / 
						(double) d->divisions);
	d->time = d->eventTime;
	ready = (*d->sourceFuncs->Event) (d, True);
	if (total >= usec)
		break;
	} while (ready);
	d->samples = total / 1.0e6 * d->dataRate;
	return total;
}

MidiReverbTap roomReverb[] = {
	{ 0.25f,	15000 },  /* ceiling */
	{ 0.166667,	10000 },	/* walls */
	{ 0.125f,	35000 },	/* back */
};

MidiReverbTap concertReverb[] = {
	{ 0.125f,	100000 },	/* ceiling */
	{ 0.25f,	10000 },	/* orchestra shell */
	{ 0.066667f,	75000 },	/* walls */
	{ 0.05f,	300000 },	/* back */
};

MidiReverbTap cathedralReverb[] = {
	{ 0.4f,	100000 },	/* ceiling */
	{ 0.2f,	75000 },	/* walls */
	{ 0.15f,	300000 },	/* back */
};

typedef struct _midiReverbType {
	char		*name;
	MidiReverbTap   *taps;
	int			count;
} MidiReverbType;

MidiReverbType reverbTypes[] = {
	{ "none",	0,		0 },
	{ "room",	roomReverb,	3 },
	{ "concert",	concertReverb,	4 },
	{ "cathedral",cathedralReverb,3 },
};

#define NREVERB	(sizeof (reverbTypes) / sizeof (reverbTypes[0]))

Boolean
MidiSetReverbType (MidiData *d, char *name)
{
	int		i, t;
	char	*next;
	float   amount;
	int		offset;

	MidiResetReverb (d);
	if (!name)
	name = "room";
	if (!*name)
	return True;
	for (i = 0; i < NREVERB; i++)
	if (!strcmp (name, reverbTypes[i].name))
	{
		for (t = 0; t < reverbTypes[i].count; t++)
		{
		MidiAddReverbTap (d, reverbTypes[i].taps[t].amount,
				reverbTypes[i].taps[t].offset);
		}
		return True;
	}
	if (*name < '0' || '9' < *name)
	return False;
	while (*name)
	{
	amount = (float) strtod (name, &next);
	if (next == name)
		return False;
	name = next;
	if (*name++ != ',')
		return False;
	offset = (unsigned long) strtod (name, &next);
	if (next == name)
		return False;
	MidiAddReverbTap (d, amount, offset);
	name = next;
	if (*name == ';')
		name++;
	else if (*name)
		return False;
	}

	// WHS: do we get here? If so, return True or False?
}

void
MidiReset (MidiData *d)
{
	unsigned char   channel;
	MidiChannel		*c;

	d->nactive = 0;
	d->done = False;
	d->midiDone = False;
	d->stopped = False;
	d->eventTime = 1e38;
	d->time = 0.0;
	d->tempo = 128;
	d->samples = 0;
	MidiSetReverbType (d, d->reverbName);
	/*
	*   1 / (usec/qnote * sample/usec * qnote / tick)
	* = 1 / (sample / tick)
	* = tick / sample
	*
	* Each output sample bumps event times by this
	* many ticks
	*/
	d->midiSpeed = 1.0 / (d->tempo * d->midiSpeedFactor);
	gsimidipcm_UiSetTempo (d);

	for (channel = 0, c = d->channels; 
		 channel < MIDI_CHANNELS; 
		 channel++, c++)
	{
		c->program = 0;
		/*
		 * By default, channel 9 and 15 are drums, a program
		 * change will cancel that
		 */
		if (channel == 9 || channel == 15)
			c->drum = True;
		else
			c->drum = False;
		MidiResetControllers (d, channel);
	}
	(*d->sourceFuncs->Reset) (d);
}

void
MidiOutInit (MidiData *d)
{
	/*
	 * divisions is measured in ticks/qnote
	 *
	 * samples/second / (usec/second * ticks/qnote)
	 *
	 * = samples/usec * qnote / ticks
	 */
	d->midiSpeedFactor = ((double) d->dataRate / 
			(1000000.0 * (double) d->divisions));
	/*
	 * Length in seconds
	 */
	d->lengthSeconds = MidiSeek (d, 1e38) / 1e6;

	MidiReset (d);
}

#warning "Rewrite MidiSample & MidiSetFormat to give (32 bit stereo <current pcm samplerate>) output?"
int
MidiSample (MidiData *d, char *buf)
{
	float		left, right, value;
	//float	 reverb;
	unsigned short  sample;
	MidiVoice		*v;
	int			i;
	MidiFix		fix;
	const float		*s;
	float		interp;

	if (!d->midiDone) {
		while (d->eventTime <= d->time) {
			if (!(*d->sourceFuncs->Event) (d, False)) {
				d->midiDone = True;
				break;
			}
		}
	}
	if (d->midiDone && d->nactive == 0)
		return 0;

	i = d->nactive;
	v = d->voices + i;

#define GetValue(result) \
	fix = v->wave.sample; \
	s = v->wave.samples + FixToInt(fix); \
	interp = midiInterp[FixToInterp(fix)]; \
	result (s[0] * (1.0f - interp) + s[1] * interp); \

	if (d->format.stereo) {
		left = 0.0f;
		right = 0.0f;
		while (--i >= 0) {
			v--;

			GetValue (value = );
			voiceDelay(v, ++v->delay_pos) = value;
			if (v->phase > 0) {
				right += v->envelope.fRight * value;
				left  += (v->envelope.fLeft * 
						  voiceDelay(v, v->delay_pos-v->phase));
			}
			else {
				left  += v->envelope.fLeft * value;
				right += (v->envelope.fRight * 
						  voiceDelay(v, v->delay_pos+v->phase));
			}

			if ((v->wave.sample = fix + v->wave.sampleStep) >= v->wave.end) {
				if (v->envelope.on || v->wave.loopRelease)
					v->wave.sample -= v->wave.loop_len;
				if (v->wave.sample >= v->wave.data_end) {
					voiceIdle (d, i);
					continue;
				}
			}
		}
		if (left >= 32768.0f)
			left = 32767.0f;
		else if (left <= -32768.0f)
			left = -32767.0f;
		if (right >= 32768.0f)
			right = 32767.0f;
		else if (right <= -32768.0f)
			right = -32767.0f;
		if (d->reverbCount) {
			int			r;
			unsigned long   pos;
			unsigned long   mask;
			MidiReverbTap   *tap;
			MidiReverb		*rev;

			pos = d->reverbPos;
			mask = d->reverbMask;
			for (r = d->reverbCount, tap = d->reverbTaps; 
				 --r >= 0; 
				 ++tap) {
				rev = &d->reverb[(pos - tap->offset) & mask];
				left += rev->left * tap->amount;
				right += rev->right * tap->amount;
			}
			rev = &d->reverb[pos & mask];
			rev->left = left;
			rev->right = right;
			++d->reverbPos;
		}

#if 1
		sample = (unsigned short) left;
		buf[0] = sample & 0xff;
		buf[1] = sample >> 8;
		sample = (unsigned short) right;
		buf[2] = sample & 0xff;
		buf[3] = sample >> 8;
#else
		/*
		 * To convert from signed to unsigned, simply xor the MSB -- try it
		 */
		sample = ((unsigned short) left) ^ d->format.xor;

		/*
		 * Store LSB first so byte format ends up with MSB
		 */
		buf[d->format.l_lsb] = sample & 0xff;
		buf[d->format.l_msb] = sample >> 8;

		/*
		 * Now the right channel
		 */
		sample = (unsigned short) right ^ d->format.xor;
		buf[d->format.r_lsb] = sample & 0xff;
		buf[d->format.r_msb] = sample >> 8;
#endif
	}
	else { /* Mono output */
		left = 0.0f;
		while (--i >= 0) {
			v--;

			GetValue (left += v->envelope.fValue * );

			if ((v->wave.sample = fix + v->wave.sampleStep) >= v->wave.end) {
				if (v->envelope.on || v->wave.loopRelease)
					v->wave.sample -= v->wave.loop_len;
				if (v->wave.sample >= v->wave.data_end)
					voiceIdle (d, i);
			}
		}
		if (left >= 32768.0f)
			left = 32767.0f;
		else if (left <= -32768.0f)
			left = -32767.0f;
		if (d->reverbCount) {
			int			r;
			unsigned long   pos;
			unsigned long   mask;
			MidiReverbTap   *tap;
			MidiReverb		*rev;

			pos = d->reverbPos;
			mask = d->reverbMask;
			for (r = d->reverbCount, tap = d->reverbTaps; 
				 --r >= 0; 
				 ++tap) {
				rev = &d->reverb[(pos - tap->offset) & mask];
				left += rev->left * tap->amount;
			}
			rev = &d->reverb[pos & mask];
			rev->left = left;
			++d->reverbPos;
		}

		/*
		 * To convert from signed to unsigned, simply xor the MSB -- try it
		 */
		sample = (unsigned short) left ^ d->format.xor;

		/*
		 * Store LSB first so byte format ends up with MSB
		 */
		buf[d->format.l_lsb] = sample & 0xff;
		buf[d->format.l_msb] = sample >> 8;
	}

	if (--d->envelopeStep < 0) {
		d->envelopeStep = d->envelopeDiv - 1;
		i = d->nactive;
		v = d->voices + i;
		while (--i >= 0) {
			v--;
			if (v->vibrato.step) {
				v->vibrato.value += v->vibrato.step;
				voiceWaveChange (&v->wave, VoiceFrequency (d, i));
			}
			if (v->tremelo.step) {
				v->tremelo.value += v->tremelo.step;
				v->envelope.fTrem = SineValue(v->tremelo.value) * v->tremelo.scale + 1.0;
			}
			v->envelope.value += v->envelope.valueStep;
			voiceEnvelopeSet (v, d->scale);
			if (--v->envelope.remain == 0) {
				v->envelope.value = v->envelope.target;
				if (!voiceSetEnvelopeElement (v,
											  v->envelope.element + 1,
											  d->scale)) {
					voiceIdle (d, i);
					continue;
				}
			}
		}
	}
	d->time = d->time + d->midiSpeed;
	d->samples++;

	return d->dataBytesPerUnit;
}

void
MidiQueue (MidiData *d, int amount)
{
	int			n;
	int			t;
	unsigned char   *b;

	/* WHS: Ringbuffer: put is where new samples are put, get is where they are
	 * removed, i.e. get is the start, put is end+1. t = samples we can write
	 * into the buffer.
	 */

	if (d->put >= d->get)
		t = d->bufSamples - d->put;
	else
		t = d->get - d->put;

	if (t > amount)
		t = amount;

	b = d->buf + d->put;

	while (t) {
		n = (*d->sourceFuncs->Sample) (d, b);
		if (n == 0) {
			//DEB(DPRT("MidiQueue:  Sample returned 0, quit\n"));
			d->done = True;
			break;
		}
		b += n;
		t -= n;
		d->queued += n;
	}

	d->put = b - d->buf;
	if (d->put == d->bufSamples)
		d->put = 0;
}


void
MidiSetFormat (MidiData *d,
		 Boolean	sixteenBit,
		 Boolean	unsign,
		 Boolean	msbFirst,
		 Boolean	stereo,
		 Boolean	leftFirst)
{
	int	off;

	DEB(DPRT("MidiSetFormat: sixteenBit=%d, unsign=%d, msbFirst=%d, stereo=%d, leftFirst=%d\n", sixteenBit, unsign, msbFirst, stereo, leftFirst));

	if (unsign)
		d->format.xor = 0x8000;
	else
		d->format.xor = 0x0;
	if (sixteenBit) {
		if (msbFirst) {
			d->format.l_msb = 0;
			d->format.l_lsb = 1;
		}
		else {
			d->format.l_msb = 1;
			d->format.l_lsb = 0;
		}
		d->dataBytesPerUnit = 2;
	}
	else {
		d->format.l_msb = 0;
		d->format.l_lsb = 0;
		d->dataBytesPerUnit = 1;
	}

	d->format.stereo = stereo;
	if (stereo) {
		if (sixteenBit)
			off = 2;
		else
			off = 1;
		d->dataBytesPerUnit *= 2;
	}
	else
		off = 0;

	d->format.r_msb = d->format.l_msb + off;
	d->format.r_lsb = d->format.l_lsb + off;

	DEB(DPRT("MidiSetFormat: dataBytesPerUnit=%d\n", d->dataBytesPerUnit));
}


static const MidiPatch *
FindInstrument (const MidiPatch **instruments, int ninstruments, const char *name)
{
	int		i;
	char	npart[256];
	const char	*nt;

	for (nt = name; (nt = match_next (nt, npart));) {
		for (i = 0; i < ninstruments; i++) {
			if (match_any (npart, instruments[i]->name)) {
				return instruments[i];
			}
		}
	}

	return 0;
}

Boolean
InitMidiPatch (MidiData *d)
{
	if (d->map)
		return True;

	d->map = GetPatchMap (d->patchFile);
	if (!d->map)
		return False;

	InitMidiVolumes ();

	return True;
}

void
FreeMidiPatch (MidiData *d)
{
	if (d->map) {
		FreePatchMap (d->map);
		d->map = 0;
	}
}


double
MidiNoteFrequency (int  note,
		const double table[12])
{
	double  factor;

	factor = 1.0;
	while (note < NOTE_TABLE_BASE) {
		note += 12;
		factor /= 2.0;
	}
	while (note >= NOTE_TABLE_BASE + 12) {
		note -= 12;
		factor *= 2.0;
	}
	return table [ note - NOTE_TABLE_BASE ] * factor;
}

void
MidiSetNote (MidiData *d, int note,
			 const double table[12],
			 int  base,
			 int  cents)
{
	d->noteTable [ note ] = (MidiNoteFrequency (base, table) * centsTable [ cents ] * 1000.0);
}

void
MidiChangeTuning (MidiData  *d,
		int  note,
		int  base,
		int  cents)
{
	int         voice;
	MidiVoice   *v;
	MidiChannel *c;

	MidiSetNote (d, note, equal440Table, base, cents);
	for (voice = d->nactive; --voice >= 0; ) {
		v = &d->voices[voice];
		if (v->note == note) {
			c = &d->channels[v->channel];
			voiceWaveChange (&v->wave, VoiceFrequency (d, voice));
		}
	}
}

void
MidiSetTuning (MidiData  *d, const double table[12])
{
	int  note;

	for (note = 0; note < 128; note++)
		MidiSetNote (d, note, table, note, 0);
}
