/*
 * $Id: patch.c,v 1.3 1998/05/27 19:25:16 keithp Exp $
 *
 * Modified by W.H.Scholten
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "kpmidi.h"
#include "patchmap.h"
#include "utils/string_match.h"
#include "utils/log.h"

#define PATCH_LOAD_FAILED   ((const MidiPatch *) 1)

const MidiMap *
GetPatchMap (char *mapFile)
{
	MidiMap         *map;
	MidiMapElement  elements[1024], *elt;
	int             n;
	FILE            *m;
	//char          dir_tmp[1024];
	char            *dir;
	char            line[1024];
	char            name_tmp[1024], file_tmp[1024];
	char            *name, *file;
	char            *slash;
	//char          *dot;
	int             lno;

	m = fopen (mapFile, "r");
	if (!m)
	{
		gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Can't open map file %s", mapFile);
		goto bail1;
	}
	lno = 0;
	n = 0;
	while (fgets (line, sizeof (line) - 1, m))
	{
		lno++;
		if (line[0] == '#')
			continue;
		if (sscanf (line, "%s %s", name_tmp, file_tmp) != 2)
		{
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "Map file %s: %d: syntax error", mapFile, lno);
			goto bail2;
		}
		name = malloc (strlen (name_tmp) + 1);
		if (!name)
			goto bail2;
		file = malloc (strlen (file_tmp) + 1);
		if (!file)
		{
			free (name);
			goto bail2;
		}
		strcpy (name, name_tmp);
		strcpy (file, file_tmp);
		elements[n].name = name;
		elements[n].file = file;
		elements[n].patch = 0;
		n++;
	}
	map = calloc (1, sizeof (MidiMap));
	if (!map)
		goto bail2;
	elt = calloc (n, sizeof (MidiMapElement));
	if (!elt)
		goto bail3;
	slash = strrchr (mapFile, '/');
	if (!slash)
	{
		dir = malloc (3);
		if (!dir)
			goto bail4;
		strcpy (dir, "./");
	}
	else {
		dir = malloc (slash - mapFile + 2);
		if (!dir)
			goto bail4;
		strncpy (dir, mapFile, slash-mapFile + 1);
	}
	memcpy (elt, elements, n * sizeof (MidiMapElement));
	map->nmap = n;
	map->dir = dir;
	map->map = elt;
	fclose (m);
	return map;
  bail4:
	free (elt);
  bail3:
	free (map);
  bail2:
	while (n--)
	{
		free ((void *) elements[n].name);
		free ((void *) elements[n].file);
	}
	fclose (m);
  bail1:
	return 0;
}

void
FreePatchMap (const MidiMap *map)
{
	int              i;
	const MidiPatch *patch;

	for (i = 0; i < map->nmap; i++)
	{
		patch = map->map[i].patch;
		if (patch && patch != PATCH_LOAD_FAILED) {
			(*patch->free) (patch);
		}
		free ((void *) map->map[i].name);
		free ((void *) map->map[i].file);
	}
	free ((void *) map);
}

static MidiMapElement *
FindPatch (const MidiMap *map, const char *name)
{
	int  i;
	char npart[256];
	const char *nt;

	for (nt = name; (nt = match_next (nt, npart));)
	{
		for (i = 0; i < map->nmap; i++)
		{
			if (match_any (npart, map->map[i].name))
			{
				return &map->map[i];
			}
		}
	}

	return 0;
}

const MidiPatch *
LoadPatchFile (const char *fname)
{
	char *dot, *slash;

	dot = strrchr (fname, '.');
	slash = strrchr (fname, '/');
	if (slash && dot && dot < slash)
		dot = 0;
	if (!dot || !strcmp (dot, _GSIMIDIPCM_INSTRUMENT_NATIVE_BINARY__SUFFIX))
		return MidiPatchMap (fname);
	else if (!strcmp (dot, _GSIMIDIPCM_INSTRUMENT_NATIVE_TEXT__SUFFIX))
		return MidiPatchParse (fname);
#ifdef USE_GUS_PATCH_CODE
	else if (!strcmp (dot, _GSIMIDIPCM_INSTRUMENT_GUS__SUFFIX))
		return ReadGUSPatch (fname);
#endif
	else {
		gsimidipcm_log_message (GSI_MESSAGE_ERROR, "%s: invalid patch file extension \"%s\"", fname, dot);
		return 0;
	}
}

const MidiPatch *
LoadPatch (const MidiMap *map, const char *name)
{
	//const MidiPatch  *patch;
	MidiMapElement *element;
	char   fname[1024];
    //char *dot;

	if (!map)
		return 0;

	element = FindPatch (map, name);
	if (!element)
		return 0;

	if (element->patch == PATCH_LOAD_FAILED)
		return 0;

	if (!element->patch)
	{
		if (element->file[0] != '/')
		{
			strcpy (fname, map->dir);
			strcat (fname, element->file);
			if (access (fname, F_OK) != 0)
			{
				strcpy (fname, DEFAULT_DIR);
				strcat (fname, element->file);
			}
		}
		else
			strcpy (fname, element->file);
		element->patch = LoadPatchFile (fname);
		if (!element->patch)
		{
			gsimidipcm_log_message (GSI_MESSAGE_ERROR, "%s: can't load patch file", element->file);
			element->patch = PATCH_LOAD_FAILED;
			return 0;
		}
	}

	return element->patch;
}
