/* GSI - general sound interface
 *
 * client.c - shared sound server, client management
 *
 * Lower level API dependencies: standard socket code
 *
 * CHANGES
 * ++eero, 8/98:
 * - Cleaned up unnecessary global variables and added support
 *   for AF_UNIX sockets.
 *
 *
 * Copyright (C) 1998 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <errno.h>

//#include <sys/types.h>
#include <sys/stat.h> /* for chmod(2) */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#ifndef AF_LOCAL
#  define AF_LOCAL AF_UNIX
#endif

#include <time.h>

#include <syslog.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_SERVER (CLIENT)"
#include <gsi/gsi_debug.h>
#include <gsi/gsi_commands.h>

#include "client.h"
#include "client_data.h"
#include "client_device_requests.h"
#include "gsi_server.h"
#include "log.h"


/*__________________DEFINITIES___________________*/


/*__________________EXTERN___________________*/


/*__________________VARIABELEN___________________*/
int  num_clients;
struct gsi_client  *client_list[ MAX_CLIENTS ];


static int  mysock;			/* listen socket descriptor */
int  connection_type = GSI_CONNECTION_ANY;

int  af_local_permissions = 0777;

/* Note: some of the comments in this file are (based on) comments from a
 * public domain example on using sockets; Just FYI. */

/*---------------------------------------------------------------------------*
 *      Socket initialization
 *---------------------------------------------------------------------------*/
static int init_unix(void)
{
	struct sockaddr addr;
	int sh;

	sh = socket(AF_LOCAL, SOCK_STREAM, 0);
	if (sh < 0) {
		log_message(GSI_MESSAGE_ERROR, "%s -> Unable to create AF_LOCAL socket "GSI_LOCAL_SOCKET_NAME, strerror(errno));

		return -1;
	}

	addr.sa_family = AF_LOCAL;
	strncpy(addr.sa_data, GSI_LOCAL_SOCKET_NAME, 14); /* char sa_data[14] */
	addr.sa_data[13] = 0;

	//umask(0700); // default = don't share anything.
	//Q: bind with AF_LOCAL: umask taken into account?

	if (bind(sh, &addr, sizeof(struct sockaddr)) < 0) {
		char *msg = "Note: This means the local socket for communication with the GSI server"
				" already exists. If there are no \""SERVER_NAME"\" processes running, you can"
				" just remove the file "GSI_LOCAL_SOCKET_NAME" and rerun the server.";

		log_message(GSI_MESSAGE_ERROR, "%s %s", strerror(errno), msg);

		close(sh);
		return -1;
	}
	af_local_socket_created = True; /* we may 'unlink' it (prevents another started server from removing another server's socket) */
	DEB1(DPRT("%s: bound AF_LOCAL socket to " GSI_LOCAL_SOCKET_NAME "\n", server_name));

	/* to allow others to connect (e.g. root starts AF_LOCAL server, but non
	 * root users want sound output):
	 */
	if (share_server)
		chmod(GSI_LOCAL_SOCKET_NAME, af_local_permissions);
	else
		chmod(GSI_LOCAL_SOCKET_NAME, 0700);
	// replace with: if (share_server) umask(af_local_permissions); before bind?

	return sh;
}


static int init_inet(struct servent *sp)
{
	struct sockaddr_in addr;
	int sh;

	sh = socket(AF_INET, SOCK_STREAM, 0);
	if (sh < 0) {
		log_message(GSI_MESSAGE_ERROR, "%s -> Unable to create AF_INET socket", strerror(errno));
		return -1;
	}

	/* The server should listen on the wildcard address, rather than its
	 * own internet address.  This is generally good practice for
	 * servers, because on systems which are connected to more than one
	 * network at once will be able to have one server listening on all
	 * networks at once.  Even when the host is connected to only one
	 * network, this is good practice, because it makes the server
	 * program more portable.
	 */
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = sp->s_port;
	addr.sin_family = AF_INET;

	if (bind(sh, (struct sockaddr *)&addr, sizeof(struct sockaddr)) < 0) {
		log_message(GSI_MESSAGE_ERROR, "%s -> AF_INET socket: unable to bind address", strerror(errno));

		close(sh);
		return -1;
	}
	DEB1(DPRT("%s: using AF_INET socket on port %d\n", server_name, sp->s_port));

	return sh;
}


/*---------------------------------------------------------------------------
 *      Set up the listening socket etc.
 *---------------------------------------------------------------------------
 */
int init_connection(void)
{
	int  oldflags; /* Old fd/file status flags. */
	struct servent *sp;	/* Pointer to service information. */

	DEB1(DPRT("("__FILE__"): *init_connection*\n"));

	sp = getservbyname (GSI_SERVICE, "tcp");
	if (sp && forced_connection != GSI_CONNECTION_AF_LOCAL /*&& allow_AF_INET*/) {
		syslog(LOG_NOTICE, GSI_SERVICE " found in /etc/services");
		mysock = init_inet(sp);
		connection_type = GSI_CONNECTION_AF_INET;
	}
	else {
		if (forced_connection == GSI_CONNECTION_AF_INET) {
			syslog(LOG_NOTICE, "Can't use AF_INET");
			exit(1);
		}
		mysock = init_unix();
		connection_type = GSI_CONNECTION_AF_LOCAL;
	}
	if (mysock < 0) {
		syslog(LOG_NOTICE, "init_connection failed..");
		exit(1);
	}

	/* Initiate the listen on the socket so remote users can connect.
	 * The listen backlog is set to 5, which is the largest currently
	 * supported (BSD at least).
	 */
	if (listen(mysock, 5) == -1) {
		close(mysock);

		log_message(GSI_MESSAGE_ERROR, "%s -> Unable to listen on socket, exiting", strerror(errno));

		exit(1);
	}
	DEB1(DPRT("listening , mysock=%d\n", mysock));

	/* Set the 'close on exec' flag, not really needed at the moment */
	oldflags = fcntl(mysock, F_GETFD);
	if (oldflags == -1 || fcntl(mysock, F_SETFD, oldflags | FD_CLOEXEC) == -1) {
		syslog(LOG_ERR, "Unable to set socket to 'close on exec'");
	}
	oldflags = fcntl(mysock, F_GETFL);
	if (oldflags == -1 || fcntl(mysock, F_SETFL, oldflags | O_NONBLOCK) == -1) {
		close(mysock);
		syslog(LOG_ERR, "Unable to set socket to non-blocking");

		return -1;
	}
	DEB1(DPRT("#init_connection#\n"));

	return mysock;
}


/*---------------------------------------------------------------------------
 *      Try to add a new client.  Returns -1 if failed (i.e.  no
 *      connection request) If succesful, initialize variables for the new
 *      client
 *---------------------------------------------------------------------------
 */
char thishost_name[ 64 ];
int add_client(void)
{
	int  socket_fd;	        /* connected socket descriptor */
	int  oldflags;          /* of socket_fd */


	int  addrlen;
	char *hostname;	        /* points to the remote host's name string */
	struct hostent *hp;     /* pointer to host info for remote host */
	time_t timevar;	        /* contains time returned by time() */
	struct linger linger;   /* allow a lingering, graceful close; */
	struct sockaddr_in peeraddr_in;	/* for peer socket address */

	if (num_clients >= MAX_CLIENTS) return -1;

	addrlen = sizeof(struct sockaddr_in);

	/* This call will block until a new connection arrives unless ls was
	 * set to non-blocking.
	 *
	 * Then, it will return the address of the connecting peer, and a
	 * new socket descriptor, socket_fd, for that connection.
	 */
	socket_fd = accept(mysock, (struct sockaddr *) &peeraddr_in, &addrlen);

	DEB1(DPRT("("__FILE__"): *add_client* socket_fd=%d\n", socket_fd));

	if (socket_fd == -1) return -1;

	/* Enable local address reuse */
	{
		int  curstate = 1;
		if (setsockopt(socket_fd, SOL_SOCKET, SO_REUSEADDR, &curstate, sizeof(curstate)) < 0) {
			syslog(LOG_ERR, "Unable to set for a fresh socket");
			return -1;
		}
	}

	/* Look up the host information for the remote host that we have
	 * connected with.  Its internet address was returned by the accept
	 * call, in the main daemon loop above.
	 */
	hp = gethostbyaddr ((char *) &peeraddr_in.sin_addr,
						sizeof (struct in_addr), peeraddr_in.sin_family);

	if (hp) {
		hostname = hp->h_name;	/* point to host's name */
	}
	else {
		/* The information is unavailable for the remote host.  Just
		 * format its internet address to be printed out in the
		 * logging information.  The address will be shown in
		 * "internet dot format".
		 */
		hostname = inet_ntoa(peeraddr_in.sin_addr);
	}

	/* Check host access & log a startup message */
	time (&timevar);
	if (connection_type == GSI_CONNECTION_AF_INET) {
		/* Now see if the host is allowed access .. */

#define host_allow(a,b) 0
		/* XXX Make the host_allow(,) function to make this work */
		if (host_allow("/etc/gsi.hosts", hostname) < 0) {
			syslog(LOG_ERR, "Access from host %s denied", hostname);
			close (socket_fd);
			return -1;
			/* should also add a delay to prevent too many attempts in a row
			 * and filling up the logs (or not send logs any more about that
			 * server for a while). */
		}

		/* For non local clients log startups */
		if (gethostname(thishost_name, 63)) {
			syslog(LOG_ERR, "gethostname overflow");
			return -1;
		}
		if (strcmp(hostname, thishost_name)) {
			syslog(LOG_NOTICE, "AF_INET: Connection with client started from %s port %u at %s", hostname, ntohs(peeraddr_in.sin_port), ctime(&timevar));
		}
	}

	/* Set the socket for a lingering, graceful close. This will cause
	 * a final close of this socket to wait until all of the data sent
	 * on it has been received by the remote host.
	 *
	 * but what if a client disconnects/is killed? Trouble? I got 100%
	 * CPU use in some cases when doom was killed .. (check this out)
	 */
	linger.l_onoff  = 1;
	linger.l_linger = 1;
	if (setsockopt(socket_fd, SOL_SOCKET, SO_LINGER, &linger, sizeof(linger)) == -1) {
		if (connection_type == GSI_CONNECTION_AF_INET)
			syslog(LOG_ERR, "Connection with %s aborted on error", hostname);
		else
			syslog(LOG_ERR, "AF_LOCAL: Connection aborted on error");
		exit(1);
	}


	/* Now we have a connection, so add a new client */
	{
		struct gsi_client  *new;

		new = calloc(1, sizeof(struct gsi_client));
		if (new == NULL) {
			close(socket_fd);
			log_message (GSI_MESSAGE_ERROR, "add client: malloc failed");
			return -1;
		}

		client_list[ num_clients ] = new;
		num_clients++;

		/* Initialize data */
		initialize_client_data(new, socket_fd);
	}

	/* Non-blocking reads from the socket */
	oldflags = fcntl(socket_fd, F_GETFL);
	if (oldflags != -1) fcntl(socket_fd, F_SETFL, oldflags | O_NONBLOCK);
	DEB1(DPRT("#add_client#\n"));

	return 0;
}


int remove_client(int client) {
	if ((client & 255) < num_clients) {
		remove_device_requests(client);
		remove_client_data(client);
		num_clients--;
		if (num_clients == 0) init_device_requests();
		return 0;
	}

	return -1;
}


/*---------------------------------------------------------------------------*
 *      Remove a client, the socket for communication, the list entry and all
 *      allocated space (samples etc).
 *---------------------------------------------------------------------------*/
int remove_client_data(int client)
{
	if ((client & 255) < num_clients) {
		close(client_list[ client ]->socket_fd);

		free_client_data (client);

		free(client_list[ client ]);
		if (client < num_clients-1)
			memmove(client_list+client, client_list+client+1, sizeof(void *) * (num_clients-1-client));
		return 0;
	}

	return -1;
}


/*      EOF client.c */
