/* GSI - general sound interface
 *
 * client_data.h
 *
 * by W.H.Scholten 1998 - 2001
 *
 * This file is released to the public domain.
 */


#ifndef _GSI__server__client_data_h
#define _GSI__server__client_data_h

#include <unistd.h>

#include <gsi/whs_types.h>
#include <gsi/whs_defines.h>

#include <gsi/mixer_data.h>

#define GSI_CONNECTION_ANY       -1
#define GSI_CONNECTION_AF_LOCAL   0
#define GSI_CONNECTION_AF_INET    1

/* Make these maxima dynamic? */
#define MAX_SOUNDS            256
/*   Maximum number of sounds loaded (slots), this can have any value you want
 *   (note that it uses n*4 bytes for the 'playing' table), actually the inter-
 *   face sends 1 byte, so could use byte playing[..] and restrict to 256)
 *
 *   perhaps use a 1-2 byte scheme to use 0-127, 128-255 +2nd byte for
 *   128+32768 possible samples (like midi's vl-numbers) */


#define MAX_CHANNELS          256
/*   Maximum number of simultaneous sounds on channels (this cannot be more
 *   than 256) */

#define MAX_STATUS_QUEUE_LEN 1024

#include "play/slot.h"
#include "play/channel.h"


/* -------------------------------------------------------------------------
 *     CLIENT structure:
 *     Defaults of the gsi_client members are in comments:
 * ------------------------------------------------------------------------- */
typedef struct gsi_client {
	char *client_name; /*   for identification. */

	uid_t   pw_uid;    /* user id */
	gid_t   pw_gid;    /* group id */


	/* ------------------------- INPUT -------------------------------- */
    int  socket_fd; /* File descriptor for the socket */

#define RING_BUFFER_SIZE      2048
#define SCRATCH_BUFFER_SIZE    128

	/* Ring buffer: */
	byte input_buffer [ RING_BUFFER_SIZE + SCRATCH_BUFFER_SIZE ];
	int  input_buffer_start; /* first byte */
	int  input_buffer_len;

	int  input_nuke;       /* = 0;  number of byte to ignore */

	/* Asynchronous command decoding: state system */
	int  input_cmd;        /* = -1; the command we're working on, -1 if none */
	int  input_cmd_len;    /* = 0; the number of bytes we still expect for the command */

	char input_read_block; /* = False; Do we let the server read directly into a malloced input block? */
	byte *input_block;     /* = NULL; */
	int  input_block_alloc_len; /* = 0; size of malloced space */
	int  input_block_size; /* = 0; bytes wanted from network (rest is scratch space) */
	uint input_block_len;  /* = 0; number of bytes received */

	uint input_block_len_left;  /* = 0; number of bytes to be received after the current block */
	uint input_block_no;  /* = 0; number of the block we've split the incoming data into */

	/* Asynchronous command decoding: saved state before block reading */
	int  input__slot;
	int  input__type;
	int  input__rate;
	int  input__channels;
	int  input__length;
	int  input__offset;

	int  input_stage;
	/*   0 = nothing yet,
	 *   1 = first byte done, command identified
	 *   2, 3, .. = stages for commands that read blocks */


	/* -------------------------- EVENTS ------------------------------ */
	int  event_mask;        /* = 0 */
    /*   the events we send back to the program. */
	byte status_queue [ MAX_STATUS_QUEUE_LEN ];
	int  status_queue_len;


	/* ---------------------- DEVICE GRABBING ------------------------- */
	int  mixer_request;     /* = GSI_CMD_RELEASE_MIXER */
	int  pcm_request;       /* = GSI_CMD_RELEASE_PCM */
	int  synth_request;     /* = GSI_CMD_RELEASE_SYNTH */
	int  cd_request[ 16 ];  /* = GSI_CMD_RELEASE_CD */


	/* ----------------------- PCM PLAY SYSTEM ------------------------ */
	int  max_channels;      /* = MAX_CHANNELS */
	/*   max number of simultaneous sounds playing, reduce with
	 *   GSI_CMD_SET_MAX_CHANNELS to limit the maximum CPU use */

	int  mach1;             /* = 256 << 8  (default speed of sound = 256.0) */
	int  base_width;        /* = 256, range is 0-256 (0=mono, 256=full stereo) */
	int  base_depth;
	int  base_height;

	int  paused;            /* = False */


	/* ------------------------- PCM STATE ---------------------------- */
	int  new_pcm_samplerate;/* = 8000 */
	int  new_pcm_channels;  /* = 1 */
	int  new_pcm_type;      /* = GSI_8BIT */
	int  new_pcm_log2_frag; /* = 8 */


	int  autorelease_pcm;   /* = False */
	/*   If true, the PCM is released when no sounds are playing, it will be
	 *   grabbed again when a 'play sound' command is received. */

	int  autoset_pcm;       /* = False */
	/*   If true, the PCM settings are changed whenever a sound is played after
	 *   a period of silence, to the rate/bytes/channels of that sound. */

	char sounds_dir[ 256 ];

	/* These two are set by sound start & modify commands: */
	int  last_channel;      /* = -1 */
	/*   channel of last command (for immediately following update commands),
	 *   -1 if none (done playing or a kill command issued on the channel). */


	/* --------------------- APPLICATION VOLUMES ---------------------- */
	int  *application_volume [ GSIMIXER_NUM_DEVICE_TYPES ];


	/* ------------------------- SLOTS -------------------------------- */
	/* This also has the info for streams and partial loading: */
	struct gsis_slot *slot [ MAX_SOUNDS ];

	int  max_slot_used;     /* = 0 */
	/* = 1+ highest number of sound slots used. */


	/* ------------------------- CHANNELS ----------------------------- */
	struct gsis_channel_info *channel [ MAX_CHANNELS ];

	int  channel_playnum;   /* = 0 */
	/*   Number of sounds being played on channels */
	byte channel_used [ MAX_CHANNELS ];
	/*   simple linear list of which channels are playing */
	int  current_time;      /* = 0 */


	/* ------------------------- CHANNELS: OBSERVER ------------------- */
	float observer_x;       /* = 0 */
	float observer_z;       /* = 0 */
	float observer_y;       /* = 0 */
	float observer_angle;   /* = 0 */ /* fi */
	float observer_theta;   /* = 0 */
	/*   For observer directioning: observer_look_fi/observer_look_theta */

	float observer_dx;      /* = 0 */
	float observer_dz;      /* = 0 */
	float observer_dy;      /* = 0 */
} gsi_client_t;

int free_client_data (int client);
int initialize_client_data (struct gsi_client *new, int socket_fd);

#endif /* _GSI__server__client_data_h */
