/* GSI - general sound interface
 *
 * commands.c - shared sound server, command decoding
 *
 * Lower level API dependencies:
 *   - libgsif, libgsicd, libgsisynth, SYSV shared memory
 *
 *
 * Copyright (C) 1996 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

#include <gsi/whs_types.h>
//#define DEBUG_PREFIX "GSI_SERVER (CMD)"
#define DEBUG_PREFIX "(CMD)"
#include <gsi/gsi_debug.h>

#include <gsi/gsi_commands.h>
#include <gsi/gsi_subcommands.h>
#include <gsi/gsi_events.h>
#include <gsi/gsi_eventmasks.h>

#include <gsi/gsif.h>
#include <gsi/gsif_midi.h> /* temp */
#include <gsi/gsicd.h>
#include <gsi/synth_data.h>
#include <gsi/gsi_command_lengths.h>
#ifdef DEBUG
#  include <gsi/gsi_command_names.h>
#endif

#include "gsi_server.h"
#include "client.h"
#include "client_device_requests.h"

#include "device/mixer/mixer.h"
#include "device/pcm/pcm.h"
#include "device/attach_device.h"

#include "play/cd/cd.h"
#include "load/slot.h"
#include "play/song/song_play.h"
#include "play/song/synth.h"

#include "conf.h"

#include "io.h"

#include "command_decode.h"

#include "utils/string_util.h"
#include "utils/shm.h"
#include "utils/tables.h"

#include "log.h"


/*__________________DEFINITIES___________________*/


/*__________________EXTERN___________________*/
/* gsi_server.c: */
extern int  current_time;      /* 1/32s fragment counter for doppler effects */


/*__________________FUNCTIE PROTOTYPES___________________*/
void  check_dxyz(struct gsis_channel_info *ch); /* Sets flags for 3d and doppler etc. if needed. */




/*---------------------------------------------------------------------------*
 *      Decode commands, async.
 *---------------------------------------------------------------------------*/
void decode_command (gsi_client_t *c) {
	int  i;
	int  status;
	int  cmd;
	int  cmd_len;

	byte *cmd_buf;

	DEB1(DPRT("("__FILE__"): *decode_command*\n"));
	if (c->input_cmd == -1) {
		if (c->input_buffer_len == 0) return;
		c->input_cmd = c->input_buffer[ c->input_buffer_start ];
		c->input_cmd_len = abs(gsi_command_length[ c->input_cmd ]);

		cmd_bytes_used = 1;
		if (c->input_buffer_len < c->input_cmd_len) {
			cmd_decode_finished = False;
			return;
		}
		adjust_for_cmd_bytes_used (c);
	}

	cmd = c->input_cmd;
	cmd_len = c->input_cmd_len;

	/* Changed in partially done, long (block reading) commands: */
	cmd_bytes_used = cmd_len;

	if (c->input_buffer_start + cmd_len > RING_BUFFER_SIZE) {
		int  copy_len1 = (RING_BUFFER_SIZE-c->input_buffer_start);
		int  copy_len2 = cmd_len - copy_len1;
		memcpy(c->input_buffer+RING_BUFFER_SIZE, c->input_buffer, copy_len2);
	}
	cmd_buf = &c->input_buffer[ c->input_buffer_start ];

	/* Nuked in partially done, long (block reading) commands: */
	cmd_decode_finished = True;



	DEB(DPRT("Decoding command %s\n", gsi_command_name[cmd]));


	/* Preparations done, now execute -------------------------------------- */
	if (cmd < 64) {


	switch(cmd) {
	/* --------------- CD commands ------------------------------------- */
	case GSI_CMD_CD_CONTROL:
	{
		int  cdno, cd_command;
#ifdef DEBUG
		char *cd_cmd_name[ 19 ] = {
			"GSI_CMD_CD_PLAY", "-invalid-", "-invalid-", "-invalid-", "-invalid-", "-invalid-", "GSI_CMD_CD_STOP", "GSI_CMD_CD_PAUSE", "GSI_CMD_CD_RESUME", "GSI_CMD_CD_MUTE", "GSI_CMD_CD_LEFT_CHANNEL", "GSI_CMD_CD_RIGHT_CHANNEL", "GSI_CMD_CD_MONO", "GSI_CMD_CD_STEREO", "GSI_CMD_CD_EJECT", "GSI_CMD_CD_INSERT", "GSI_CMD_CHANGE_CD", "GSI_CMD_CD_GET_TRACKS" , "GSI_CMD_CD_GET_STATUS"
		};
#endif

		cdno = next_byte(cmd_buf);
		cd_command = next_byte(cmd_buf);
		DEB1(DPRT("-> cd_command=%d for cd %d (%s)\n", cd_command, cdno, cd_cmd_name[ cd_command ]));
		/* Should only execute if cd_owner = current_client or shared */
		switch(cd_command) {

		case GSI_CMD_CD_PLAY:
		{
			int  t1,m1,s1,f1,t2,m2,s2,f2;

			t1 = next_byte(cmd_buf);
			m1 = next_byte(cmd_buf);
			s1 = next_byte(cmd_buf);
			f1 = next_byte(cmd_buf);
			t2 = next_byte(cmd_buf);
			m2 = next_byte(cmd_buf);
			s2 = next_byte(cmd_buf);
			f2 = next_byte(cmd_buf);

			if (is_a_cd_owner(cdno, current_client)) {
				int  st = gsicd_play(cdno, t1,m1,s1,f1, t2,m2,s2,f2);

				/* Notify all owners: */
				if (!st) {
					DEB(DPRT("GSICD_EVENT_TRACK_STARTED\n"));
					send_list_to_cd_owners (cdno, 3, GSI_EV_CD_TRACK_STARTED, cdno, gsicd_status[ cdno ]->first.track);
				}

			}
			break;
		}

		case GSI_CMD_CD_STOP:
			if (is_a_cd_owner(cdno, current_client)) {
				int  st = gsicd_stop(cdno);

				/* Notify all owners */
				if (!st) {
					DEB(DPRT("GSICD_EVENT_TRACK_END\n"));
					send_list_to_cd_owners (cdno, 3, GSI_EV_CD_TRACK_END, cdno, gsicd_status[ cdno ]->current_track);
				}

			}
			break;
		case GSI_CMD_CD_PAUSE:
			if (is_a_cd_owner(cdno, current_client)) {
				gsicd_pause(cdno);
			}
			break;
		case GSI_CMD_CD_RESUME:
			if (is_a_cd_owner(cdno, current_client)) {
				gsicd_resume(cdno);
			}
			break;
		case GSI_CMD_CD_MUTE:
		case GSI_CMD_CD_LEFT_CHANNEL:
		case GSI_CMD_CD_RIGHT_CHANNEL:
		case GSI_CMD_CD_MONO:
		case GSI_CMD_CD_STEREO:
			if (is_a_cd_owner(cdno, current_client)) {
				gsicd_set_channels(cdno, cd_command);

				/* Notify all owners */
				send_list_to_cd_owners (cdno, 3, GSI_EV_CD_OUTPUT, cdno, cd_command);
			}
			break;

		case GSI_CMD_CD_EJECT:
			if (is_a_cd_owner(cdno, current_client)) {
				int st = gsicd_eject(cdno);

				if (st == -1) {
				}
				else if (st == GSICD_ERROR_MOUNTED) {
					//perror("EJECT CD");
					send_list(c, 2, GSI_EV_CD_MOUNTED, cdno);
				}
				else {
					/* Notify all owners */
					send_list_to_cd_owners (cdno, 2, GSI_EV_CD_EJECTED, cdno);
				}

			}
			break;
		case GSI_CMD_CD_INSERT:
			if (is_a_cd_owner(cdno, current_client)) {
				int status = gsicd_insert(cdno);
			}
			break;
		case GSI_CMD_CHANGE_CD:
		{
			int  cd_slot = next_byte(cmd_buf);
			if (is_a_cd_owner(cdno, current_client)) {
				int status = gsicd_select_slot(cdno, cd_slot);
				if (status == GSICD_ERROR_BAD_SLOT) {
					send_list(c, 3, GSI_EV_CD_BAD_SLOT, cdno, cd_slot);
				}
			}
			break;
		}
		case GSI_CMD_CD_GET_TRACKS:
		{
			DEB1(DPRT("GSI_CMD_CD_GET_TRACKS\n"));
			if (is_a_cd_owner(cdno, current_client)) {
				send_list(c, 3, GSI_EV_CD_TRACK_LIST, cdno, gsicd_status[ cdno ]->num_tracks);
				for (i=0; i <= gsicd_status[ cdno ]->num_tracks+1; i++) {
					send_list(c, 4,
							  gsicd_status[ cdno ]->track[i].min,
							  gsicd_status[ cdno ]->track[i].sec,
							  gsicd_status[ cdno ]->track[i].frame,
							  gsicd_status[ cdno ]->track[i].type
						);
				}
			}
			else {
				DEB1(DPRT("CD: NOT OWNER! (client =%d, req=%d, owner=%d)\n", current_client, c->cd_request[ cdno ], cd_owner));
				// send event saying information is not available?
				//send_list(c, 2, GSI_EV_CD_NOT_OWNER, cdno);
			}
			break;
		}
		case GSI_CMD_CD_GET_STATUS:
		{
			if (is_a_cd_owner(cdno, current_client)) {
				gsicd_update_status(cdno);
				send_list(c, 7, GSI_EV_CD_STATUS, cdno,
						  gsicd_status[ cdno ]->status,
						  gsicd_status[ cdno ]->current_track,
						  gsicd_status[ cdno ]->current_m,
						  gsicd_status[ cdno ]->current_s,
						  gsicd_status[ cdno ]->current_f
					);
						
			}
			break;
		}
		default:
			log_message(GSI_MESSAGE_ERROR, "Bad CD command %d", cmd);
			kill(0, SIGSEGV);
			/* should send error message to the client */
			break;
		} /* cd_command */
	}



	/* --------------- Channel commands -------------------------------- */
		case GSI_CMD_CHANNEL_STOP:
		{
			int  chn;

			chn = next_byte(cmd_buf);
			DEB1(DPRT("-> terminating sound on channel %d\n", chn));

			for (i=c->channel_playnum-1; i >= 0; i--) {
				if (c->channel_used[i] == chn) {
					DEB1(DPRT("-> found sound, playlist_no=%d\n", i));
					if (c->last_channel == chn) c->last_channel = -1;
					if (i < c->channel_playnum-1) {
						memmove(c->channel_used+i, c->channel_used+i+1, (c->channel_playnum-i-1)*sizeof(int));
					}
					//if (c->event_mask & GSI_EVMASK_CHANNEL) {
					send_list(c, 2, GSI_EV_CHANNEL_END, chn);
					DEB1(DPRT("-> GSI_EV_CHANNEL_END event in queue\n"));
					//}
					c->channel_playnum--;
					total_playnum--;
					break;
				}
			}
			DEB1(if (i < 0) DPRT("-> not found !\n"));
			break;
		}

		case GSI_CMD_CHANNEL_START:
		{
			int  chn;

			chn = next_byte(cmd_buf);
			DEB(DPRT("-> start sound on channel %d\n", chn));
			c->last_channel = chn;

			/* Is the channel already playing a sound ? */
			if (c->channel[ chn ]->sample.flags & CHANNEL_FLAG_PLAYING) {
				return; /* yes */
			}
			else {
				/* no, start playing on this channel, but first: */
				DEB(DPRT("-> slot no = %d\n", c->channel[ chn ]->slot));
				if (c->channel[ chn ]->type == _GSI_SOUNDTYPE_SAMPLE) {
					gsis_slot_t *s;
					gsis_slot_sample_t *ch_sample;

					if (c->channel[ chn ]->slot < 0) {
						send_list(c, 2, GSI_EV_CHANNEL_END, chn);
						return;
					}

					s = c->slot[ c->channel[ chn ]->slot ];
					ch_sample = s->data.sample;
					/* autoset pcm after silence? */
					if (!total_playnum && autoset_pcm) {
						int  status;
						DEB(DPRT("start channel %d: autoset_pcm -> setup_pcm\n"));
						status = setup_pcm(server_pcm, server_pcm_status, /*dev,*/ ch_sample->samplerate, ch_sample->channels, ch_sample->type, 0);
					}

					c->channel[ chn ]->sample.flags |= CHANNEL_FLAG_PLAYING;
				}
				else {
					c->channel[ chn ]->song.flags |= CHANNEL_FLAG_PLAYING;
				}

				c->channel_used[ c->channel_playnum ] = chn;
				c->channel_playnum++;
				total_playnum++;
				DEB1(DPRT("-> new sound on channel\n"));
				send_byte(c, GSI_EV_CHANNEL_STARTED);
				send_byte(c, chn);
			}
			break;
		}

		case GSI_CMD_CHANNEL_PAUSE:
		{
			int  chn;
			gsis_channel_info_t *ch;

			chn = next_byte(cmd_buf);

			c->last_channel = chn;

			ch = c->channel[ chn ];
			//if (ch->sample.flags & CHANNEL_FLAG_PLAYING) {
			if (ch->slot >= 0 && !(ch->sample.flags & CHANNEL_FLAG_PAUSED)) {
				ch->sample.flags |= CHANNEL_FLAG_PAUSED;
				/* remove channel from the linear list: */
				if (chn < c->channel_playnum-1) {
					memmove(c->channel_used+chn, c->channel_used+chn+1, (c->channel_playnum-chn-1)*sizeof(int));
				}
				c->channel_playnum--;
				total_playnum--;
			}
			break;
		}

		case GSI_CMD_CHANNEL_RESUME:
		{
			int  chn;
			gsis_channel_info_t *ch;

			chn = next_byte(cmd_buf);

			c->last_channel = chn;

			ch = c->channel[ chn ];
			if (ch->sample.flags & CHANNEL_FLAG_PAUSED) {
				ch->sample.flags ^= CHANNEL_FLAG_PAUSED;
				/* add sound to the linear play list again. */
				c->channel_used[ c->channel_playnum ] = chn;
				c->channel_playnum++;
				total_playnum++;
			}
			break;
		}

		case GSI_CMD_CHANNEL_START_ALL:
			log_message(GSI_MESSAGE_NOTICE, "GSI_CMD_CHANNEL_START_ALL not done");
			kill(0, SIGSEGV);
			break;
		case GSI_CMD_CHANNEL_STOP_ALL:
		{
			int  pc;

			DEB1(DPRT("GSI_CMD_KILLALL\n"));
			total_playnum -= c->channel_playnum;
			for (pc=0; pc < c->channel_playnum; pc++) {
				c->channel[ c->channel_used[ pc ] ]->slot = -1;
			}
			c->channel_playnum = 0;
			//if (c->event_mask & GSI_EVMASK_CHANNEL) {
			send_byte(c, GSI_EV_CHANNEL_ENDALL);
			//}
			break;
		}

		case GSI_CMD_CHANNEL_PAUSE_ALL:
			break;
		case GSI_CMD_CHANNEL_RESUME_ALL:


	case GSI_CMD_CHANNEL_SET_SLOT:
	{
		int chn = next_byte(cmd_buf);
		int slot = next_byte(cmd_buf);
		gsis_channel_info_t *ch;

		c->last_channel = chn;

		ch = c->channel[ chn ];
		if (ch == NULL) {
			ch = calloc(1, sizeof(struct gsis_channel_info));
			if (ch == NULL) {
				/* out of memory */
				log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CHANNEL_SET_SLOT: out of memory");
				kill(0, SIGSEGV);
			}
			c->channel[ chn ] = ch;
			ch->slot = -1; /* not associated with a valid slot yet.. */
		}

		if (c->slot[ slot ] == NULL) {
			//log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CHANNEL_SET_SLOT: no such slot");
			
		}
		else {
			ch->play_ct = 1;

			if (c->slot[ slot ]->type == _GSI_SOUNDTYPE_SAMPLE) {
				ch->sample.flags = 0; /* repeat is set separately now */

				ch->sample.position = 0;
				ch->slot = slot;

				/* defaults: */
				ch->volume = 256;
				ch->mix_volume = c->application_volume[GSI_PCM] [0]; // same as: (ch->volume * c->application_volume[GSI_PCM])>>8;
				ch->sample.x  = 0.0;
				ch->sample.dx = 0;
				ch->sample.z  = 64.0;
				ch->sample.dz = 0;
				ch->sample.y  = 0.0;
				ch->sample.dy = 0;
				ch->sample.echo_dt  = 0;
				ch->sample.echo_fvol = 255;
				ch->sample.old_dist = 64.0;
				ch->sample.pitch_f = 65536;

				ch->type = _GSI_SOUNDTYPE_SAMPLE;
			}
			else {
				/* _GSI_SOUNDTYPE_SONG */
				log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CHANNEL_SET_SLOT song not done");

				ch->type = _GSI_SOUNDTYPE_SONG;
			}
		}
		log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CHANNEL_SET_SLOT not finished");
		break;
	}
		case GSI_CMD_CHANNEL_SET_DEVICE:
			log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CHANNEL_SET_SLOT not finished");
			kill(0, SIGSEGV);
			break;

			/* all the observer decoding must go here */
		default:
			log_message(GSI_MESSAGE_ERROR, "0-63: Bad command %d", cmd);
			kill(0, SIGSEGV);
			//remove_client(current_client);
			//kill(0, SIGSEGV);
			/* should send error message to the client */
			break;
		} /* switch(cmd) */
	} /* cmd < 64 */




	/* --------------------------------------------------------------------- */
	else if (cmd < 192) {
	/* Commands 64-191 */
	int  cmd_no_current;
	int  chn;
	struct gsis_channel_info *ch;

	if (cmd & GSI_CMD_CURRENT) {
		DEB1(DPRT("-> last_channel=%d\n", c->last_channel));
		chn = c->last_channel;
	}
	else {
		chn = next_byte(cmd_buf);
		c->last_channel = chn;
	}

	cmd_no_current = cmd & ~GSI_CMD_CURRENT;
	if (chn < 0) {
		/* happens if last_channel is not set yet, so skip the command. */
#warning "fix nuke length and adjust input buffer length."
		c->input_nuke = gsi_command_length[ cmd_no_current ] - 2;
		//DEB(DPRT("cmd %d (cmd2 %d) skip %d bytes\n", command, cmd_no_current, c->input_nuke));
		send_list(c, 2, GSI_EV_CHANNEL_NO_CURRENT, cmd);
		return;
	}

	ch = c->channel[ chn ];
	if (ch == NULL) {
		log_message(GSI_MESSAGE_NOTICE, "uninitialized channel");
		kill(0, SIGSEGV);
	}

	DEB1(DPRT("-> command=%d, cmd_no_current=%d, ch=%d\n", cmd, cmd_no_current, chn));
	switch (cmd_no_current) {
	case GSI_CMD_CHANNEL_SET_LOOP:
		ch->play_ct = (int) next_byte(cmd_buf);
fprintf(stderr, "Loop = %d\n", ch->play_ct);
		break;

	case GSI_CMD_CHANNEL_SET_SAMPLE_FACTOR:
		ch->sample.pitch_f = next_msb_int32(cmd_buf);
		DEB1(DPRT("-> f=%d\n", ch->sample.pitch_f));
		break;

	case GSI_CMD_CHANNEL_SET_SAMPLE_POSITION:
		ch->sample.position = next_msb_int32(cmd_buf);
		break;

	case GSI_CMD_CHANNEL_SET_X:
		ch->sample.x = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE | CHANNEL_FLAG_INIT_POSITION;
		ch->sample.last_position_time = current_time;
		break;

	case GSI_CMD_CHANNEL_SET_Z:
		ch->sample.z = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE | CHANNEL_FLAG_INIT_POSITION;
		ch->sample.last_position_time = current_time;
		break;

	case GSI_CMD_CHANNEL_SET_Y:
		ch->sample.y = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE | CHANNEL_FLAG_INIT_POSITION;
		ch->sample.last_position_time = current_time;
		break;

	case GSI_CMD_CHANNEL_ADJUST_X:
		ch->sample.x = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE;
		ch->sample.last_position_time = current_time;
		break;

	case GSI_CMD_CHANNEL_ADJUST_Z:
		ch->sample.z = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE;
		ch->sample.last_position_time = current_time;
		break;

	case GSI_CMD_CHANNEL_ADJUST_Y:
		ch->sample.y = (int) next_msb_int16(cmd_buf);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE;
		ch->sample.last_position_time = current_time;
		break;


	case GSI_CMD_CHANNEL_SET_DX:
		ch->sample.dx = (int) next_msb_int16(cmd_buf) /256.0;
		check_dxyz(ch); /* sets CHANNEL_FLAG_RECALC_3D if needed */
		break;

	case GSI_CMD_CHANNEL_SET_DZ:
		ch->sample.dz = (int) next_msb_int16(cmd_buf) /256.0;
		check_dxyz(ch);
		break;

	case GSI_CMD_CHANNEL_SET_DY:
		ch->sample.dy = (int) next_msb_int16(cmd_buf) /256.0;
		check_dxyz(ch);
		break;

	case GSI_CMD_CHANNEL_SET_ANGLE:
	{
		int   angle;
		float x = ch->sample.x;
		float z = ch->sample.z;
		float radius;
		int   cos_angle, sin_angle;

		angle = next_byte(cmd_buf);

		cos_angle = sinus_table[ ((64-angle) & 255) << 2 ];
		sin_angle = sinus_table[ angle << 2 ];
		/* no slow sqrt when the sound is in initial position: */
		if (x == 0) radius = z; else radius = sqrt(x*x+z*z);

		ch->sample.old_dist = sqrt(x*x+z*z+ ch->sample.y*ch->sample.y /* y*y */);
		ch->sample.x = (radius*cos_angle) /65536.0;
		ch->sample.z = (radius*sin_angle) /65536.0;
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE;
		ch->sample.last_position_time = current_time;
		break;
	}

	case GSI_CMD_CHANNEL_SET_ANGLE_R:
	{
		int   angle;
		float radius;
		int   cos_angle, sin_angle;

		angle = next_byte(cmd_buf);
		radius = next_msb_int16(cmd_buf);

		cos_angle = sinus_table[ ((64-angle) & 255) << 2 ];
		sin_angle = sinus_table[ angle << 2 ];

		ch->sample.x = (radius*cos_angle)/65536.0;
		ch->sample.z = (radius*sin_angle)/65536.0;
		//ch->old_dist = radius;
		ch->sample.old_dist = sqrt(radius*radius+ch->sample.y*ch->sample.y /* y*y */);
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D_ONCE;
		ch->sample.last_position_time = current_time;
		break;
	}

#if 0
	case GSI_CMD_CHANNEL_SET_ANGLE_DEFAULT_R:
	{
		int angle;
		int cos_angle, sin_angle;

		/* sets the angle and standard radius (i.e. no sqrt needed) */
		angle = next_byte(cmd_buf);

		cos_angle = sinus_table[ ((64-angle) & 255) << 2 ];
		sin_angle = sinus_table[ angle << 2 ];
		ch->sample.x = (64*cos_angle) /65536.0;
		ch->sample.z = (64*sin_angle) /65536.0;
		ch->sample.old_dist = sqrt(64*64+ch->sample.y*ch->sample.y /* y*y */);
		ch->sample.last_position_time = current_time;
		break;
	}
#endif
	case GSI_CMD_CHANNEL_SET_PAN:
	{
		ch->sample.pan = next_msb_int16(cmd_buf);
		DEB1(DPRT("-> pan = %d\n", (int) ch->sample.pan));
		break;
	}

	case GSI_CMD_CHANNEL_SET_DEPTH:

	case GSI_CMD_CHANNEL_SET_HEIGHT:
		/* Not used as only 2 channel cards are supported by the API's. */
		next_msb_int16(cmd_buf);
		break;

	case GSI_CMD_CHANNEL_SET_VOLUME:
	{
		ch->volume = next_byte(cmd_buf);
		if (ch->volume > 0) ch->volume++;
DEB(DPRT("ch->volume=%d\n", ch->volume));
		ch->mix_volume = (ch->volume * c->application_volume[GSI_PCM] [0])>>8;
DEB(DPRT("ch->mix_volume=%d\n", ch->mix_volume));
		break;
	}

	case GSI_CMD_CHANNEL_ECHO_ON:
		ch->sample.echo_dt = (int) (unsigned int) next_msb_int16(cmd_buf);
		ch->sample.echo_fvol = next_byte(cmd_buf);
		break;

	case GSI_CMD_CHANNEL_ECHO_OFF:
		ch->sample.echo_dt = 0;
		break;

	case GSI_CMD_CHANNEL_DOPPLER_ON:
		//ch->doppler = cmd_int16();
		// f'=f *Vg/(Vg+Vobj). the same factor is used for the volume change (number of J/(m^2 s) increases/decreases with the frequency)
		ch->sample.flags |= CHANNEL_FLAG_DOPPLER;
		break;

	case GSI_CMD_CHANNEL_DOPPLER_OFF:
		// For all channels or just one?
		ch->sample.flags &= ~CHANNEL_FLAG_DOPPLER;
		break;






/* Observer */
	case GSI_CMD_OBSERVER_SET_X:
		c->observer_x = ((int) next_msb_int16(cmd_buf));
		break;

	case GSI_CMD_OBSERVER_SET_Z:
		c->observer_z = ((int) next_msb_int16(cmd_buf));
		break;

	case GSI_CMD_OBSERVER_SET_Y:
		c->observer_y = ((int) next_msb_int16(cmd_buf));
		break;

	case GSI_CMD_OBSERVER_SET_ANGLE:
		c->observer_angle = (int) (int8) next_byte(cmd_buf);
		break;

	case GSI_CMD_OBSERVER_SET_DX:
		c->observer_dx = (int) next_msb_int16(cmd_buf);
		break;

	case GSI_CMD_OBSERVER_SET_DZ:
		c->observer_dz = (int) next_msb_int16(cmd_buf);
		break;

	case GSI_CMD_OBSERVER_SET_DY:
		c->observer_dy = (int) next_msb_int16(cmd_buf);
		break;

	case GSI_CMD_OBSERVER_DOPPLER_ON:
		break;

	case GSI_CMD_OBSERVER_DOPPLER_OFF:
		break;

	default:
		log_message(GSI_MESSAGE_NOTICE, "64-191: Bad command %d", cmd);
		/* Should close the fd to the client as we get out of sync now, the
		 * clien can then reconnect and try again.
		 */
		kill(0, SIGSEGV);
		//remove_client(current_client);
		//kill(0, SIGSEGV);
	} /* switch (cmd_no_current) */

	} /* 64-191 */














	/* ------ 192 - 255 ---------------------------------------------------- */
	else switch(cmd) {

		/* Events ---------------------------------------------------------- */
	case GSI_CMD_ENABLE_EVENTS:
	{
		int  bits;

		bits = next_msb_int32(cmd_buf);
		c->event_mask |= bits;
		break;
	}

	case GSI_CMD_DISABLE_EVENTS:
	{
		int  bits;

		bits = next_msb_int32(cmd_buf);
		c->event_mask &= ~bits;
		break;
	}

	case GSI_CMD_SYNC:
		DEB(DPRT("SYNC from client %d\n", current_client));
		send_byte(c, GSI_EV_SYNC);
		while (c->status_queue_len > 0) flush(c);
#warning "change sync"
		// Much better: disable reading from this client until the status
		// queue is empty.
		break;


		/* Misc. ----------------------------------------------------------- */
	case GSI_CMD_SET_WORKING_DIRECTORY:
	{
		/* XXX Block loading command. */
		char *name;

		if (c->input_stage == 0) {
			int  block_len;

			c->input_stage++;

			block_len = next_msb_int32(cmd_buf);
			if (!cmd_init_block(c, 4, block_len + 1, block_len, &status)) {
				cmd_decode_finished = False;
				return;
			}
		}

		/* Done loading command data, so execute: */
		c->input_block [ c->input_block_size ] = 0; /* C string terminator */
		name = c->input_block;

		DEB1(DPRT("-> working directory = %s\n", name));
		strncpy(c->sounds_dir, name, 255);
		c->sounds_dir[ 255 ] = 0;

		/* Reset flags */
		c->input_stage = 0;
		break;
	}

	case GSI_CMD_SET_SPEED_OF_SOUND:
		c->mach1 = next_msb_int16(cmd_buf);
		break;


		/* Grab/release PCM ------------------------------------------------ */
	case GSI_CMD_GRAB_PCM:
		if (c->pcm_request == GSI_CMD_RELEASE_PCM) { /* => new request */
			pcm_request[ pcm_requests ] = current_client;
			pcm_requests++;
			DEB1(DPRT("-> pcm_requests=%d, current_client=%d\n", pcm_requests, current_client));
		}
		c->pcm_request = GSI_CMD_GRAB_PCM;
		c->autorelease_pcm = False;
		c->autoset_pcm = False;

		if (pcm_request[ 0 ] == current_client) {
			int st = setup_pcm(server_pcm, server_pcm_status, /*dev,*/ c->new_pcm_samplerate, c->new_pcm_channels, c->new_pcm_type, c->new_pcm_log2_frag);
			if (!st) {
				autorelease_pcm = False;
				autoset_pcm = False; /* Correct? or with SET_PCM_SAMPLERATE etc? */
			}
		}
		break;

	case GSI_CMD_RELEASE_PCM:
		c->pcm_request = GSI_CMD_RELEASE_PCM;
		c->autorelease_pcm = False;
		remove_pcm_request(current_client);
		break;

	case GSI_CMD_AUTO_PCM:
		c->pcm_request = GSI_CMD_AUTO_PCM;
		c->autorelease_pcm = True;
		if (pcm_request[ 0 ] == current_client) autorelease_pcm = True;
		break;


		/* Grab/release/share synth ---------------------------------------- */
	case GSI_CMD_GRAB_SYNTH:
	{
		int  devno;

		devno = next_byte(cmd_buf);

		if (synth_owner < 0) {
#if 0 /* rewrite this */
			if (c->song_status == NULL) {
				c->song_status = calloc(1, sizeof(c->song_status));
			}

			if (gsisynth_open(c->song_status->hw, c->song_status->hw_status) > -1) {
				synth_owner = current_client;
				send_list(c, 2, GSI_EV_SYNTH_GRABBED, devno);
//				song_init(current_client, channel);
				c->song_status->song_play_status = _SONG_NONE;
				c->song_status->song_play_ct = 0; // not needed any more?

				mixer_set_application_volume(GSI_SYNTH, devno, c->application_volume[GSI_SYNTH] [devno]);
			}
			else {
				if (num_synths == 0) {
					send_list(c, 2, GSI_EV_NO_SYNTH, devno);
				}
				else {
					send_list(c, 2, GSI_EV_SYNTH_BUSY, devno);
				}
			}
#endif /* rewrite */
		}
		else if (synth_owner != current_client) {
			send_list(c, 2, GSI_EV_SYNTH_BUSY, devno);
		}
		break;
	}

	case GSI_CMD_RELEASE_SYNTH:
	{
		int  devno;

		devno = next_byte(cmd_buf);

		remove_synth_request(current_client);
		break;
	}


	/* Grab/release/share cd ----------------------------------------------- */
	case GSI_CMD_GRAB_CD:
	{
		int cdno = next_byte(cmd_buf);

		/* cd_owner: 0-255 is real client, 65536+n=shared by n clients. */
		if (cd_owner[ cdno ] < 0) {
			if (gsicd_init(cdno) == 0) {
				cd_owner[ cdno ] = current_client;
				send_list(c, 2, GSI_EV_CD_GRABBED, cdno);
				c->cd_request[ cdno ] = GSI_CMD_GRAB_CD;

				mixer_set_application_volume(GSI_CD, cdno, c->application_volume[GSI_CD] [ cdno ]);
			} else {
				/* check for errors: */
				if (gsicd_status[ cdno ]->status > GSICD_STATUS_PAUSED) {
					switch (gsicd_status[ cdno ]->status) {
					case GSICD_STATUS_NO_DISK:
						send_list(c, 2, GSI_EV_CD_NO_DISK, cdno);
						break;
					case GSICD_STATUS_NO_DRIVE:
						send_list(c, 2, GSI_EV_NO_CD, cdno);
						break;
					case GSICD_STATUS_UNAVAILABLE:
						send_list(c, 2, GSI_EV_CD_BUSY, cdno);
						DEB1(DPRT("-> CD device busy (%s).\n", server_name, gsicd_devices[ cdno ]->device));
						break;
					default:
						break;
					}
				}
			}
		}
		else {
			/* grabbed by other client also means BUSY */
			send_list(c, 2, GSI_EV_CD_BUSY, cdno);
		}
		break;
	}

	case GSI_CMD_RELEASE_CD:
	{
		int cdno = next_byte(cmd_buf);

		remove_cd_request(cdno, current_client);
		c->cd_request[ cdno ] = GSI_CMD_RELEASE_CD;
		break;
	}

	case GSI_CMD_SHARE_CD: // replaces GSI_CMD_CONTINUE_CD
	{
		int cdno = next_byte(cmd_buf);

		if (cd_owner[ cdno ] < 0 || cd_owner[ cdno ] > 65535) {
			if (cd_owner[ cdno ] < 0) {
				if (gsicd_init(cdno) == 0) {
					cd_owner[ cdno ] = 65536+1;
					send_list(c, 2, GSI_EV_CD_SHARED, cdno);
					c->cd_request[ cdno ] = GSI_CMD_SHARE_CD;
				}
				/* events for error conditions (no CD, no disk in CD etc)
				 * are sent in cd_init. NOT ANY MORE. */
			}
			else {
				cd_owner[ cdno ]++; /* increment the share-counter. */
				c->cd_request[ cdno ] = GSI_CMD_SHARE_CD;
			}
		}
		else {
			/* privately grabbed, so not shareable */
			send_list(c, 2, GSI_EV_CD_BUSY, cdno);
		}

		break;
	}


	/* Grab/release/share mixer ------------------------------------------------ */
	case GSI_CMD_GRAB_MIXER:
		if (mixer_owner < 0) {
			mixer_owner = current_client;
		}
		else {
			send_byte(c, GSI_EV_MIXER_BUSY);
		}
		break;

	case GSI_CMD_RELEASE_MIXER:
		remove_mixer_request(current_client);
		c->mixer_request = GSI_CMD_RELEASE_MIXER;
		break;


/* Add a device ------------------------------------------------------------ */
	case GSI_CMD_ADD_DEVICE:
	{
		/* XXX Block loading command. */
		char *device_name;

		if (c->input_stage == 0) {
			int  block_len;

			c->input_stage++;

			c->input__type = next_byte(cmd_buf);
			block_len = next_msb_int32(cmd_buf);
			if (!cmd_init_block(c, 5, block_len + 1, block_len, &status)) {
				cmd_decode_finished = False;
				return;
			}
		}

		/* Done loading command data, so execute: */
		c->input_block [ c->input_block_size ] = 0; /* C string terminator */
		device_name = c->input_block;
		DEB(DPRT("adding device %s (type %d)\n", device_name, c->input__type));

		switch(c->input__type) {
		case GSI_PCM:
			attach_cd_device(device_name, "PCM");
			break;
		case GSI_SYNTH:
			attach_synth_device(device_name, "SYNTH");
			break;
		case GSI_CD:
			attach_cd_device(device_name, "CD");
			check_cd_config(); /* For keep_open. */
			break;
		default:
			//_decode_error__bad_device_type(c, device_type);
			fprintf(stderr, "GSI_CMD_ADD_DEVICE: Bad device type %d\n", c->input__type);
			break;
		}

		/* Reset flags */
		c->input_stage = 0;
		break;
	}


/* Set PCM characteristics ------------------------------------------------- */
	case GSI_CMD_USE_BEST_PCM_SETTINGS:
		/* rename?: GSI_CMD_AUTO_SET_PCM / GSI_CMD_AUTO_CONFIG_PCM */
		c->autoset_pcm = True;
		if (pcm_request[ 0 ] == current_client) autoset_pcm = True;
		break;

	case GSI_CMD_SET_PCM_SAMPLERATE:
		c->new_pcm_samplerate = (int) next_msb_int32(cmd_buf);
		break;

	case GSI_CMD_SET_PCM_CHANNELS:
		c->new_pcm_channels = (int) next_byte(cmd_buf);
		break;

	case GSI_CMD_SET_PCM_TYPE:
		c->new_pcm_type = (int) next_byte(cmd_buf);
		break;

	case GSI_CMD_SET_PCM_FRAGSIZE:
		c->new_pcm_log2_frag = next_byte(cmd_buf);
		break;


/* Set/get MIXER volumes --------------------------------------------------- */
	case GSI_CMD_SET_BASE_WIDTH:
		c->base_width = next_msb_int16(cmd_buf);
		break;

	case GSI_CMD_SET_MIXER_VOLUME:
	{
		int  devtype, devno, vol;

		devtype = next_byte(cmd_buf);
		devno = next_byte(cmd_buf);
		vol = next_byte(cmd_buf);

		if (vol > 0) vol++; /* Range change 0-255 -> 0, 2-256 */
		DEB1(DPRT("set mixer dev [%d, %d ], vol=%d\n", devtype, devno, vol));
		DEB1(DPRT("mixer_owner=%d, current_client=%d\n", mixer_owner, current_client));
		if (mixer_owner == current_client) {
			mixer_set_device_volume(devtype, devno, vol);
		}
		break;
	}

	case GSI_CMD_GET_MIXER_VOLUME:
	{
		int  devtype, devno;

		devtype = next_byte(cmd_buf);
		devno = next_byte(cmd_buf);

		mixer_send_status_of(devtype, devno);
		flush_all();
		break;
	}

	case GSI_CMD_SET_MIXER_BALANCE:
	{
		int  balance;

		balance = next_msb_int16(cmd_buf);

		if (current_client == mixer_owner || mixer_owner == -1) {
			mixer_set_balance(balance);
		}
		break;
	}


/* Pause/continue all devices ---------------------------------------------- */
	case GSI_CMD_PAUSE:
		c->paused = True;
		break;

	case GSI_CMD_RESUME:
		c->paused = False;
		break;

	case GSI_CMD_SLOTS_LOAD_DATA:
	{
		load_all_slots(c);
		break;
	}
	case GSI_CMD_SLOTS_FREE_DATA:
	{
		unload_all_slots(c);
		break;
	}


	case GSI_CMD_LOAD_SOUND:
	{
		int   slot;
		struct gsis_slot *s;

		slot = next_byte(cmd_buf);

		s = c->slot[ slot ];
		/* For now, we ignore a load on an undefined file/slot: */
		if (s == NULL) return;

		if (slot_busy(c, slot)) {
			_decode_error__slot_busy(c, slot);
			return;
		}

		if (s->type == _GSI_SLOTTYPE__PCM_SAMPLE) {
			if (s->data.sample->input != _GSI_INPUT_FILE) return;

			if (s->data.sample->data.file == NULL) return;
			if (s->data.sample->data.file->filename == NULL) return;
		}
		else if (s->type == _GSI_SLOTTYPE__MIDI_SONG) {
//			if (s->data.input != _GSI_INPUT_FILE) return;
//			if (s->data.sample.file == NULL) return;
//			if (s->data.file->filename == NULL) return;
		}

		/* N.B. Only useful if we allow changing the name, when a
		 * sound is still used (which is currently the case). */
		status = load_slot(c, slot);
		if (status < 0)
			_decode_error__no_such_file(c, slot);
		else {
			send_list(c, 2, GSI_EV_SLOT_LOADED, slot);
		}
		break;
	}

	case GSI_CMD_LOAD_SOUND_FROM_FILE:
	{
		/* XXX Block loading command. */
		int slot, offset;
		char *name;

fprintf(stderr, "GSI_CMD_LOAD_SOUND_FROM_FILE ok\n");
		if (c->input_stage == 0) {
			int block_len;

			c->input_stage ++;

			slot = next_byte(cmd_buf);
			block_len = next_msb_int32(cmd_buf);
			c->input__offset = next_msb_int32(cmd_buf);
			if (slot >= MAX_SOUNDS) {
				_decode_error__bad_slot(c, slot);
				_decode_nuke_command(c, block_len);
				return;
			}
			if (slot_busy(c, slot)) {
				_decode_error__slot_busy(c, slot);
				_decode_nuke_command(c, block_len);
				return;
			}

			/* save data: */
			c->input__slot = slot;
			if (!cmd_init_block(c, 9, block_len + 1, block_len, &status)) {
				cmd_decode_finished = False;
				return;
			}
		}

		/* Done loading command data, so execute: */
		c->input_block [ c->input_block_size ] = 0; /* C string terminator */
		name = c->input_block;
		slot = c->input__slot;
		offset = c->input__offset;

		free_slot(c, slot);

		c->slot[ slot ] = calloc(1, sizeof(struct gsis_slot));
		if (c->slot[ slot ] == NULL) {
			_decode_error__no_room(c, slot, 0);
			break;
		}
fprintf(stderr, "GSI_CMD_LOAD_SOUND_FROM_FILE ok 2\n");
		status = init_slot_with_sample(c, slot, absolute_file_name(c->sounds_dir, name), True, offset, 0);
fprintf(stderr, "GSI_CMD_LOAD_SOUND_FROM_FILE st=%d\n", status);
		if (status < 0) {
			_decode_error__no_such_file(c, slot);
		}
        else {
			status = load_slot(c, slot);
		}

		/* Reset flags */
		c->input_stage = 0;
		break;
	}

	case GSI_CMD_LOAD_SOUND_DATA:
	{
		/* XXX Block loading command. */
		int slot;
		int  sound_size, sample_len;

		if (c->input_stage == 0) {
			c->input_stage ++;

			c->input__slot = next_byte(cmd_buf);
			c->input__rate = next_msb_int32(cmd_buf);
			c->input__channels = next_byte(cmd_buf);
			c->input__type = next_byte(cmd_buf);
			c->input__length = next_msb_int32(cmd_buf); /* Number of bytes (samples?) */

			DEB1(DPRT("LOAD_SOUND_DATA with samplerate=%d, ch=%d, type=%d\n", c->input__rate, c->input__channels, c->input__type));
			/* Works for uncompressed only: */
			sample_len = c->input__length;
			//sound_size = sample_len * sample_channels * _server_gsi_bytes_per_sample[ sample_type ];
			sound_size = sample_len;
			slot = c->input__slot;

			if (slot >= MAX_SOUNDS) {
				_decode_error__bad_slot(c, slot);
				_decode_nuke_command(c, sample_len);
				return;
			}

			if (slot_busy(c, slot)) {
				_decode_error__slot_busy(c, slot);
				_decode_nuke_command(c, sample_len);
				return;
			}

			free_slot(c, slot);
			c->slot[ slot ] = calloc(1, sizeof(struct gsis_slot));
			if (c->slot[ slot ] == NULL) {
				/* handle this properly */
			}
			c->slot[ slot ]->type = _GSI_SLOTTYPE__PCM_SAMPLE;

			c->slot[ slot ]->data.sample = calloc(1, sizeof(struct gsis_slot_sample));
			if (c->slot[ slot ]->data.sample == NULL) {
				/* handle this properly */
			}

			c->slot[ slot ]->data.sample->buffer = (byte *) malloc(sound_size);
			if (c->slot[ slot ]->data.sample->buffer == NULL) {
				/* handle this properly */
			}

			if (!cmd_init_block(c, 11, sample_len, sample_len, &status)) {
				cmd_decode_finished = False;
				return;
			}
		}

		/* Done loading command data, so execute: */
//		init_sample(c->input__slot, c->input__rate, c->input__channels, c->input__type, c->input__length); /* error status? */
		init_sample(c, c->input__slot, c->input__rate, c->input__channels, c->input__type, c->input__length); /* error status? */

		/* Reset flags */
		c->input_stage = 0;
		break;
	}

	case GSI_CMD_LOAD_SOUND_DATA_FROM_FILE:
	{
		int slot;

		char *name;
		struct gsi_sound *new_snd;

		if (c->input_stage == 0) {
			int block_len;

			c->input_stage ++; /* flag */

			c->input__slot = next_byte(cmd_buf);
			block_len = next_msb_int32(cmd_buf);
			c->input__offset = next_msb_int32(cmd_buf);
			c->input__length = next_msb_int32(cmd_buf);
			c->input__rate = next_msb_int32(cmd_buf);
			c->input__channels = next_byte(cmd_buf);
			c->input__type = next_byte(cmd_buf);

			if (!cmd_init_block(c, 19, block_len + 1, block_len, &status)) {
				cmd_decode_finished = False;
				return;
			}
		}

		/* Done loading command data, so execute: ========================== */
		c->input_block [ c->input_block_size ] = 0; /* C string terminator */
		name = c->input_block;
		slot = c->input__slot;

		if (slot >= MAX_SOUNDS) {
			_decode_error__bad_slot(c, slot);
			return;
		}

		if (slot_busy(c, slot)) {
			_decode_error__slot_busy(c, slot);
			return;
		}

		free_slot(c, slot);

		status = init_slot_with_sample(c, slot, absolute_file_name(c->sounds_dir, name), False, c->input__offset, c->input__length);

		if (status < 0) {
			_decode_error__no_such_file(c, slot);
			return;
		}

		c->max_slot_used = max(c->max_slot_used, slot+1);

		c->slot[ slot ] = calloc(1, sizeof(struct gsis_slot));
		if (c->slot[ slot ] == NULL) {
			// handle this properly
		}
		c->slot[ slot ]->type = _GSI_SLOTTYPE__PCM_SAMPLE;

		c->slot[ slot ]->data.sample = calloc(1, sizeof(struct gsis_slot_sample));
		if (c->slot[ slot ]->data.sample == NULL) {
			// handle this properly
		}

		c->slot[ slot ]->data.sample->input = _GSI_INPUT_FILE;
		new_snd = calloc(1, sizeof(struct gsi_sound));
		if (new_snd == NULL) {
			// handle this properly
		}
		new_snd->filename = strdup(absolute_file_name(c->sounds_dir, name));
		if (new_snd->filename == NULL) {
			// handle this properly
		}
		gsif_add_block(new_snd, c->input__rate, c->input__channels, c->input__type, c->input__offset, c->input__length); /* error status? */
		gsif_make_linear_block_list (new_snd); /* error status? */
		c->slot[ slot ]->data.sample->data.file = new_snd;

		status = load_slot(c, slot);
		/* error status? */

		/* Reset flags */
		c->input_stage = 0;
		break;
	}

	case GSI_CMD_CONFIGURE_SLOT:
	{
		int  slot;
		int  samplerate;
		int  chn;
		int  type;
		int  num_samples;

		struct gsis_slot *s;
		struct gsis_slot_sample *s_sample;

		int  slot_size;
		void *new_buf;

		slot = next_byte(cmd_buf);
		samplerate = next_msb_int32(cmd_buf);
		chn = next_byte(cmd_buf);
		type = next_byte(cmd_buf);
		num_samples = next_msb_int32(cmd_buf);

		DEB1(DPRT("configure slot %d to speed=%d, ch=%d, type=%d, num_samples=%d\n", slot, samplerate, chn, type, num_samples));

		slot_size = num_samples * _server_gsi_bytes_per_sample[ type ];

		/* Add: check if slot busy... */

		init_slot(&c->slot[ slot ], _GSI_INPUT_STREAM);

		s = c->slot[ slot ];
		s_sample = s->data.sample;

		new_buf = realloc(s_sample->buffer, slot_size);
		if (new_buf == NULL) {
			/* How do we handle this error? */
			//_decode_error__slot_out_of_memory(c, slot);
			log_message(GSI_MESSAGE_ERROR, "GSI_CMD_CONFIGURE_SLOT: out of memory\n");
			return;
		}
		else s_sample->buffer = new_buf;

		s_sample->data.stream->samplerate = samplerate;
		s_sample->data.stream->channels = chn;
		s_sample->data.stream->type = type;

		s_sample->size = slot_size;

		s_sample->samplerate = samplerate;
		s_sample->channels = chn;
		s_sample->type = type;

		sample_internal_type_of((struct gsi_sample *) s_sample);

		s_sample->meta_sample_size = s_sample->channels * _server_gsi_bytes_per_sample[ s_sample->type ];
		s_sample->meta_samples = s_sample->size / s_sample->meta_sample_size;
		break;
	}

	case GSI_CMD_STREAM:
	{
		int  chn;
		int  slot;

		struct gsis_channel_info *ch;

		chn = next_byte(cmd_buf);
		slot = next_byte(cmd_buf);

		DEB(DPRT("streaming ch %d, slot %d\n", ch, slot));
		if (c->channel[ chn ] == NULL) {
			c->channel[ chn ] = calloc(1, sizeof(struct gsis_channel_info));
			c->channel[ chn ]->slot = -1;
		}
		ch = c->channel[ chn ];

		if (ch->slot == -1) {
			ch->slot = slot;
			//OLD ch->sample.flags |= CHANNEL_FLAG_STREAM;
			c->slot[ slot ]->data.sample->input = _GSI_INPUT_STREAM;

			/* Other inits: */
			ch->sample.position = 0;

			/* defaults: */
			ch->volume = 256;
			ch->mix_volume = c->application_volume[ GSI_PCM ] [0]; /* Same as: (ch->volume * c->application_volume[GSI_PCM])>>8; */
			ch->sample.x  = 0.0;
			ch->sample.dx = 0;
			ch->sample.z  = 64.0;
			ch->sample.dz = 0;
			ch->sample.y  = 0.0;
			ch->sample.dy = 0;
			ch->sample.echo_dt  = 0;
			ch->sample.echo_fvol = 255;
			ch->sample.old_dist = 64.0;
			ch->sample.pitch_f = 65536;

			DEB1(DPRT("-> added STREAM; total_playnum=%d, autoset=%d\n", total_playnum, autoset_pcm));
			/* Autoset pcm after silence? */
			if (!total_playnum && autoset_pcm) {
				struct gsis_slot *s = c->slot [ slot ];

				int status = setup_pcm(server_pcm, server_pcm_status, /*dev,*/ s->data.sample->samplerate, s->data.sample->channels, s->data.sample->type, 0);
			}

			total_playnum++;
			c->channel_used [ c->channel_playnum++ ] = chn;
		}
		break;
	}

	case GSI_CMD_SEND_SLOT_DATA:
	{
		/* XXX Block loading command. */
		int  slot;
		int  length;
		int  wrap_size;
		struct gsis_slot *s;
		struct gsis_slot_sample *ss;

		if (c->input_stage == 0) {
			c->input_stage ++;

			c->input_block_no = 0;

			slot = next_byte(cmd_buf);
			length = next_msb_int32(cmd_buf);

			s = c->slot[ slot ];
			DEB(DPRT("SEND_SLOT_DATA: slot=%d, length=%d, slot_size = %d\n", slot, length, s->data.sample->size));

			if (!s || !s->data.sample) {
				_decode_error__bad_slot(c, slot);
				_decode_nuke_command(c, length);
				return;
			}

			ss = s->data.sample;

			if (length > s->data.sample->size) {
				/* XXX should just nuke partially: */
				_decode_error__no_room(c, slot, length);
				_decode_nuke_command(c, length);
				fprintf(stderr, "SEND_SLOT_DATA: No room (blk len=%d, slot size = %d\n", length, s->data.sample->size);
				return;
			}

			if (ss->data.stream == NULL) {
				/* GSI_CMD_SEND SLOT_DATA should be used after
				 * GSI_CMD_CONFIGURE_SLOT, but the following could be useful
				 * to reuse a loaded slot: */
				ss->data.stream = malloc(sizeof(struct gsis_stream));

				if (ss->data.stream == NULL) {
					/* How do we handle such a failure? Nuke the data ? */
					_decode_error__no_room(c, slot, length);
					_decode_nuke_command(c, length);
					fprintf(stderr, "SEND_SLOT_DATA: out of memory\n");
					return;
				}
				memcpy(ss->data.stream, ss, sizeof(struct gsis_stream));
			}
			DEB1(DPRT("loading %d bytes at %d\n", length, ss->load_ptr));

			wrap_size = min(ss->size - ss->load_ptr, length);
			c->input_block_len_left = length - wrap_size;

			/* Save data. */
			c->input__slot = slot;
			c->input__length = length;

			if (!cmd_init_direct_block(c, 5 /* offset */, ss->buffer + ss->load_ptr, wrap_size)) {
				cmd_decode_finished = False;
				return;
			}
fprintf(stderr, "cmd_init_direct_block returned FALSE\n");
		}

		/* Recall saved data: */
		s = c->slot[ c->input__slot ];
		ss = s->data.sample;

	  send_slot_data__next_block:
		convert_data_to_internal(ss->buffer + ss->load_ptr, c->input_block_len, (struct gsi_sample *) ss->data.stream);
		ss->load_ptr += c->input_block_len;
		ss->current_size += c->input_block_len;
		if (ss->load_ptr >= ss->size) ss->load_ptr -= ss->size;

		if (c->input_block_len_left) {
			if (!cmd_init_direct_block(c, 0 /* offset */, ss->buffer + ss->load_ptr, c->input_block_len_left)) {
				cmd_decode_finished = False;
				return;
			}
			goto send_slot_data__next_block;
		}

		/* Reset flags */
		c->input_stage = 0;
		break;
	}


	case GSI_CMD_GET_SLOT_INFO:
	{
		int  slot;
		struct gsis_slot *s;

		slot = next_byte(cmd_buf);

		s = c->slot[ slot ];
		if (s && s->data.sample) {
			struct gsis_slot_sample *ss = s->data.sample;

			send_byte(c, GSI_EV_SLOT_INFO);
			send_byte(c, slot);
			send_int32(c, ss->samplerate);
			send_byte(c, ss->channels);
			send_byte(c, ss->type);
			send_int32(c, ss->size);
			DEB1(DPRT("-> GSI_EV_SLOT_INFO in queue\n"));
		}
		else {
			_decode_error__bad_slot(c, slot);
			// nuke??
		}
		break;
	}


	case GSI_CMD_FREE_SOUND:
	{
		int  slot;

		slot = next_byte(cmd_buf);

		if (!slot_busy(c, slot))
			free_slot(c, slot);
		else {
			send_byte(c, GSI_EV_SLOT_BUSY);
			send_byte(c, slot);
		}
		break;
	}

	case GSI_CMD_SHMSLOT:
	{
#if USE_SHARED_MEMORY
#  if HAVE_SYSV_SHARED_MEMORY
		/* Get the shmid from the client */
		int  slot, size;
		int  shmid;
		//struct gsis_shm *shm;
		//char *segptr;

		slot = next_byte(cmd_buf);
		/* Free slot .. */

		shmid = next_msb_int32(cmd_buf);
		/* Attach (map) the shared memory segment into the current process */
		status = attach_shm(shmid);
		if (status) {
			log_message(GSI_MESSAGE_ERROR, "slot %d: attach_shm failed", slot);
			send_byte(c, GSI_EV_SHMATTACH_FAILED);
			send_byte(c, slot);
			/* nuke 4 bytes */
		}
		else {
			/* Now use segptr in the slot .. */
			// XXX unfinished SHM code:
			size = next_msb_int32(cmd_buf);

			//shmid = get_shmid(size, shm);
			/* Make 'slot' use the shm... */

			send_byte(c, GSI_EV_SHMID);
			send_byte(c, slot);
			send_int32(c, shmid);
			flush_all();
		}
#  endif
#else
		/* XXX No shm, so send an error */
#endif
		break;
	}

	default:
		log_message(GSI_MESSAGE_NOTICE, "Bad command %d", cmd);
		kill(0, SIGSEGV);
		//remove_client(current_client);
		break;
	} /* switch(cmd) for 192-255 */

	DEB1(DPRT("command decoded.\n"));
}


/*---------------------------------------------------------------------------*
 *      Set CHANNEL_FLAG_RECALC_3D flag after commands that set dx/dy/dz.
 *---------------------------------------------------------------------------*/
void check_dxyz(struct gsis_channel_info *ch) {
	DEB1(DPRT("-> check_dxyz\n"));
	if (ch->sample.dx | ch->sample.dz | ch->sample.dy) {
		ch->sample.flags |= CHANNEL_FLAG_RECALC_3D;
		DEB1(DPRT("-> 3d!\n"));
	}
	else
		ch->sample.flags &= ~CHANNEL_FLAG_RECALC_3D;
	DEB1(DPRT("-> #check_dxyz#\n"));
}

/*      EOF commands.c */
