/* GSI - general sound interface
 *
 * conf.c - shared sound server, reads the config file (/etc/gsi.conf,
 * ~/.gsi.conf), and sets defaults for options not set.
 *
 * Lower level API dependencies:
 *     - standard file system stuff
 *
 *
 * Copyright (C) 1998 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */



/*__________________INCLUDES___________________*/
#include <gsi/gsi_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/stat.h>
//#include <errno.h>

#include <syslog.h>

#include <gsi/whs_types.h>
//#define DEBUG_PREFIX "GSI_SERVER (CONF)"
#define DEBUG_PREFIX "(CONF)"
#include <gsi/gsi_debug.h>

#include <gsi/version.h>
#include "conf.h"
#include "client.h"
//#include "gsi_server.h"

#include "device/mixer/mixer.h"
#include <gsi/gsimixer.h>
#include <gsi/gsicd.h>
#include <gsi/device.h>

#include "device/attach_device.h"

#include "log.h"


/*__________________DEFINITIES___________________*/
#define _GSI_CONFIG_FILE           "gsi.conf"
#define _GSI_MASTER_CONFIG_FILE    "/etc/gsi.conf"


/*__________________EXTERN___________________*/


/*__________________VARIABELEN___________________*/
int  allow_AF_INET = False; /* Not used, looks useless. Better idea for more access control? (per user) */
int  allow_loading = False; /* Not yet used */
int  allow_saving = False; /* Not yet used */

static int _mixer_initialized = False;
static char my_configfile [ 512 ];


/*__________________FUNCTIE PROTOTYPES___________________*/
int  read_config(char *name);


/* ------------------------------------------------------------------------- *
 *      Read config file (/etc/gsi.conf)                                     *
 *      RENAME: read_startup_server_config ?                                 *
 */
int  read_server_config()
{
	int  status;

	snprintf(my_configfile, 512, "%s/.%s", getenv("HOME"), _GSI_CONFIG_FILE);

	/* First try the config file in the home directory. */
	status = read_config(my_configfile);
	if (status != 0) {
		strcpy(my_configfile, _GSI_MASTER_CONFIG_FILE);
		status = read_config(my_configfile);
	}

	if (status) {
		log_message(GSI_MESSAGE_ERROR, "Config files %s and %s could not be used!", _GSI_CONFIG_FILE, _GSI_MASTER_CONFIG_FILE);
	}

	return status;
}


/* Sections: */
enum {
	_MIXER, _PCM_INPUT, _PCM_OUTPUT, _SYNTH_INPUT, _SYNTH_OUTPUT, _CD_OUTPUT, _PSEUDO_SYNTH, _SECURITY, _OPTIONS, _NONE
};
static char *device_string[] = {
	"mixer", "pcm_input", "pcm_output", "synth_input", "synth_output",
	"cd_output", "pseudo_synth",
	"security", "options", "end"
};

/* Entries: */
static char *volume_string[] = {
	"master_volume", "pcm_volume", "synth_volume", "cd_volume",
	"line_in_volume", "mic_volume"
};

static char *af_local_permissions_str = "AF_LOCAL_permissions";
static char *allow_af_inet_str = "allow_AF_INET";
static char *allow_saving_str = "allow_saving";
static char *allow_loading_str = "allow_loading";


/* ------------------------------------------------------------------------- *
 *      Returns 0 if OK, -1 if the file could not be opened, -2 if the file  *
 *      could not be parsed.                                                 *
 */
int read_config(char *name)
{
	int  i;
	char config_line[ 256 ];
	int  line_number = 0;
	int  bad_config_file_version = 0;
	FILE *fp;

	DEB1(DPRT("("__FILE__"): *read_config*\n"));

	if (name == NULL) return -1;
	if (strlen(name) == 0) return -1;

	fp = fopen(name, "r");
	if (fp == NULL) {
		//syslog(LOG_NOTICE, "Config file %s could not be opened!", name);
		//log_message (GSI_MESSAGE_NOTICE, "Config file %s could not be opened!", name);
		return -1;
	}

	bad_config_file_version = 0;
	for(;;) {
		fgets(config_line, 127, fp); config_line[ 127 ] = 0;
		if (feof(fp)) {
			log_message (GSI_MESSAGE_ERROR, "Config file %s: server version not specified", name);
			//return -2;
			bad_config_file_version = -1;
			fseek(fp, 0L, SEEK_SET); /* return to start to get the actual config data */
		}

		line_number++;
		if (strlen(config_line) == 0) {
			log_message (GSI_MESSAGE_ERROR, "Config file %s: empty file", name);
			return -2; /* Empty file. */
		}

		/* Get rid of the linefeed. */
		if (config_line [ (int) strlen(config_line) -1 ] == 10)
			config_line [ (int) strlen(config_line) -1 ] = 0;

		if (!strncmp(config_line, "server_version ", 15)) {
			if (strcmp(config_line + 15, GSI_VERSION_STRING)) {
				log_message (GSI_MESSAGE_ERROR, "Config file %s: wrong server version", name);
				//return -1;
				bad_config_file_version = 1;
			}
			break;
		}
	}

	if (bad_config_file_version) {
		log_message (GSI_MESSAGE_ERROR, "Will try to use this config file anyway...");
	}


	/* go through rest of the file for all settings */
	for (;;) {
		int  bad_entry;
		char *token;
		static int  device = _NONE;
		int  done;

		fgets(config_line, 127, fp); config_line[ 127 ] = 0;
		if (feof(fp)) break;
		line_number++;
		if (strlen(config_line) == 0) break; /* Empty file. */

		/* Get rid of the linefeed. */
		if (config_line [ (int) strlen(config_line) -1 ] == 10)
			config_line [ (int) strlen(config_line) -1 ] = 0;

		token = strtok(config_line, " \t");
		done = False;
		if (token == NULL) continue;
		if (token[0] == '#') continue;
		for (i=0; i<_NONE; i++) {
			if (!strcmp(token, device_string[i])) {
				device = i; done = True;
				DEB2(DPRT("token = no %d (%s)\n", i, device_string[i]));
			}
		}
		if (done) continue;

		bad_entry = False;
		if (device != _NONE) {
			if (!strcmp(token, "device")) {
				token = strtok(NULL, " \t");
				if (token != NULL && token[0] != '#') {
					char *devstr = token;
					switch(device) {
					case _MIXER:
						//close_mixer();
						if (mixer_device) free(mixer_device);
						mixer_device = strdup(devstr);
						gsimixer_num_devices [ GSI_MASTER ] = 1;
						DEB1(DPRT("mixer device=%s\n", devstr));
						break;

					case _PCM_INPUT:
						DEB1(DPRT("pcm_in device=%s\n", devstr));
						//attach_pcmin_device(devstr, "PCM_RECORD");
						break;

					case _PCM_OUTPUT:
						//close_pcm();
						attach_pcm_device(devstr, "PCM");
						//NEW attach_pcm_device(devstr, "PCM", GSIDEV_VIRTUAL);
						//NEW2 attach_device(devstr, "PCM", GSIDEV_VIRTUAL, sizeof(gsipcm_device_t));
						DEB1(DPRT("pcm_out device=%s\n", devstr));
						break;

					case _SYNTH_INPUT:
						//attach_synthin_device(devstr, "MIDI_IN");
						DEB1(DPRT("synth_in device=%s\n", devstr));
						break;

					case _SYNTH_OUTPUT:
						//close_synth();
						DEB(DPRT("config: attach synth\n"));
						attach_synth_device(devstr, "SYNTH");
						//NEW attach_synth_device(devstr, "PCM", GSIDEV_REAL);
						DEB1(DPRT("synth_out device=%s\n", devstr));
						break;

					case _CD_OUTPUT:
						//close_cd();
						attach_cd_device(devstr, "CD");
						//NEW attach_cd_device(devstr, "PCM", GSIDEV_SHARED);

						DEB1(DPRT("cd_out device=%s\n", devstr));
						break;

					case _PSEUDO_SYNTH:
#warning "FINISH attach_pcm_module_device"
						//attach_pcm_module_device(devstr, gsimidipcm_functions, "MIDI_PCM (KP)");
						break;
					}
				}
				//else bad_entry = True; /* not giving a volume level is allowed for now */
			}

			/* ------------------------------------------------------------- */
			else if (!strcmp(token, "end")) {
				device = _NONE;
			}

			/* ------------------------------------------------------------- */
			else if (device == _MIXER) {
				int  device_offset = 0;
				int  device_no = -1;
				int  device_volume = -1;

				if (!strncmp(token, "rec_", 4)) {
					token += 4;
					device_offset = _RECORD_DEVICE_OFFSET;
				}

				for (i=0; i < GSIMIXER_NUM_DEVICE_TYPES; i++) {
					if (!strcmp(token, volume_string[i])) {
						if (!_mixer_initialized) {
							mixer_init(); _mixer_initialized = True;
							/* Note that this must be done after attaching
							 * devices. */
						}

						token = strtok(NULL, " \t");
						device_no = -1;
						device_volume = -1;
						if (token != NULL && token[0] != '#') {
							device_no = (int) atol(token);
							token = strtok(NULL, " \t");
							if (token != NULL && token[0] != '#') {
								device_volume = (int) atol(token);
#warning "Fix this multi device index no."
								// N.B. gsimixer_volume is not yet allocated so
								// this must be changed..
								mixer_set_device_volume(device_offset + i, device_no, device_volume);
								// This instead ?:
								//gsimixer_set_device_volume((int) atol(token));
								DEB2(DPRT("init mixer ch (device type) %d device_no %d to %d\n", device_offset + i, device_no, device_volume));
							}
						}
						if (device_volume == -1) bad_entry = True;
						break;
					}
				}
				if (i == GSIMIXER_NUM_DEVICE_TYPES) bad_entry = True;
			}

			/* ------------------------------------------------------------- */
			else if (device == _PSEUDO_SYNTH) {
				/* handle the following:
				 * max channels
				 * max rate
				 * max bits */
			}

			/* ------------------------------------------------------------- */
			else if (device == _SECURITY) {
				if (!strncmp(token, af_local_permissions_str, 20)) {
					token = strtok(NULL, " \t");
					if (token != NULL && token[0] != '#') {
						char *endptr;
						int  permissions = strtol(token, &endptr, 0);

						if (*endptr == 0) af_local_permissions = permissions;
						else bad_entry = True;
						DEB1(DPRT("af_local_permissions = %0x\n", af_local_permissions));
					}
					else bad_entry = True;
				}
				else if (!strncmp(token, allow_af_inet_str, 20)) {
					allow_AF_INET = True;
				}
				else if (!strncmp(token, allow_saving_str, 20)) {
					allow_saving = True;
				}
				else if (!strncmp(token, allow_loading_str, 20)) {
					allow_loading = True;
				}
				else bad_entry = True;
			}

			/* ------------------------------------------------------------- */
			else if (device == _CD_OUTPUT) {
				if (!strcmp(token, "keep_open")) { }
				else if (!strncmp(token, "slots", 6)) {
					token = strtok(NULL, " \t");
					if (token != NULL && token[0] != '#') {
						char *endptr;
						int  slots = strtol(token, &endptr, 0);

						if (*endptr == 0) gsicd_devices[ gsicd_num_cds-1 ]->num_slots = slots;
						else bad_entry = True;
						DEB1(DPRT("slots for cd %d = %d\n", gsicd_num_cds-1, slots));
					}
					else bad_entry = True;
				}
				else bad_entry = True;
			}
			/* ------------------------------------------------------------- */
			else bad_entry = True;
		}

		if (bad_entry) {
			syslog(LOG_ERR, "%s: bad entry in line %d (%s)", my_configfile, line_number, token != NULL ? token : "missing parameter");
		}

		for (;;) {
			token = strtok(NULL, " \t");
			if (token == NULL) break;
			if (token[0] == '#') break;
			if (done) continue;
		}
	}

	fclose(fp);

#warning "added default mixer init"
	if (!_mixer_initialized) {
		mixer_init(); _mixer_initialized = True;
		/* This has been done above iff there are mixer settings (not useful
		 * for devaudio? Or is pcm?). */
	}

	/* See if we want to keep the cd fd open for the cd device found. */
	check_cd_config();

	DEB1(DPRT("#read_config#\n"));
	return 0;
}


void check_cd_config()
{
	int  i, j;
	char *entry;

	i = 0;
	for(;;) {
		entry = get_config_entry(my_configfile, "cd_output", "keep_open", i);
		if (entry == NULL) break;
		for (j=0; j<gsicd_num_cds; j++) {
			if (!strcmp(gsicd_devices[j]->device, entry))
				gsicd_devices[j]->flags |= GSICD__FLAG_KEEP_OPEN;
			return;
		}
		i++;
	}
}


/* ------------------------------------------------------------------------- *
 * Get a config entry; e.g. for keep_open with CD (multiple devices listed   *
 * as 'keep_open'). Returns a string.                                        *
 */
char *get_config_entry(char *filename, char *device, char *entry, int entry_no)
{
	FILE *fp;
	char config_line[ 256 ];
	int  i = 0;

	fp = fopen(filename, "r");
	if (fp == NULL) return NULL;

	for (;;) {
		char *token;
		static int  dev = _NONE;

		fgets(config_line, 127, fp); config_line[127] = 0;
		if (feof(fp)) break;
		if (strlen(config_line) == 0) break; /* empty entry */
		config_line[(int) strlen(config_line) -1] = 0; /* get rid of the linefeed */

		token = strtok(config_line, " \t");
		if (token == NULL) continue;
		if (token[0] == '#') continue;
		if (dev == _NONE) {
			if (!strcmp(token, device)) {
				dev = _NONE+1;
			}
		}
		if (dev != _NONE) {
			if (!strcmp(token, "end")) {
				dev = _NONE;
				break;
			}
			if (!strcmp(token, entry)) {
				if (i == entry_no) {
					fclose(fp);
					return strtok(NULL, " \t");
				}
				else i++;
			}
		}
	}
	fclose(fp);

	return NULL;
}


/*      EOF config.c */
