/* GSI - general sound interface
 *
 * mixer.c - shared sound server, mixer setting/reading
 *
 * Lower level API dependencies:
 *
 * Copyright (C) 1996 - 2000 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_SERVER (device mixer)"
#include <gsi/gsi_debug.h>

#include <gsi/gsi_events.h>
#include <gsi/whs_defines.h>
#include <gsi/gsimixer.h>
#include "mixer.h"
#include "../../gsi_server.h"
#include "../../io.h"


/*__________________VARIABELEN___________________*/
char *mixer_device;


static int  *gsi_application_volume [ GSIMIXER_NUM_DEVICE_TYPES ];

#if 0
/* To linearize all devices:
 * pcm's + synths + cd's; first attach all pcm's then all synth's then all cd's
 */
int  device_index[ GSIMIXER_NUM_DEVICE_TYPES ] = { 0 };


void make_device_index(void);

/* after attaching devices call this: */
void make_device_index(void) {
	int  i;

	for (i=1; i < GSIMIXER_NUM_DEVICE_TYPES; i++) {
		device_index [ i ] = device_index [ i-1 ] + gsimixer_num_devices [ i-1 ];
	}
}
#endif


/*---------------------------------------------------------------------------*
 *      Initialize mixer, checks whether we have a master volume and reads
 *      the current volumes if there are no previous (saved between open/
 *      close mixer) values.
 *---------------------------------------------------------------------------*/
int mixer_init () {
	int  i;
	int  status;

	for (i=0; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
		int  j;
		gsi_application_volume [ i ] = malloc(sizeof(int) * gsimixer_num_devices[i]);
		for (j=0; j<gsimixer_num_devices[i]; j++) {
			gsi_application_volume [ i ][ j ] = 256;
		}
	}

	gsimixer_device = mixer_device;
	status = gsimixer_init();

	if (status != 0) {
		log_message(-1, "Couldn't open mixer %s", gsimixer_device);
	}

	return status;
}


/*---------------------------------------------------------------------------*
 *      Close the mixer
 *---------------------------------------------------------------------------*/
int mixer_close () {
	gsimixer_close();

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Mixer volume settings:
 *
 *      Saving volumes is done to remove any rounding problems with repeated
 *      reading/writing (device values read may not be the same as what has
 *      been set, due to range conversions and/or hardware limitations).
 *---------------------------------------------------------------------------*/
int mixer_mute () {
//	master_save = gsi_mixer_volume[ GSI_MASTER ];
	gsimixer_set_device_volume(GSI_MASTER, 0, 0);
//	gsimixer_muted = True; /* could use gsimixer_muted */

	return -1;
}


int mixer_unmute() {
	int  i, j;

	gsimixer_unmute();

	for (i=0; i< GSIMIXER_NUM_DEVICE_TYPES; i++) {
		for (j=0; j<gsimixer_num_devices[i]; j++) {
			mixer_set_application_volume(i, j, gsi_application_volume[i][j]);
		}
	}

	return -1;
}


int mixer_set_balance (int new_balance) {
	//balance = new_balance;
	//mixer_set_device_volume(GSI_MASTER, get_mixer(GSI_MASTER));
	gsimixer_set_balance(new_balance);

	return -1;
}


int mixer_set_device_volume (int devtype, int devno, int vol) {
	int  hw_volume = -1;

	DEB1(DPRT("set mixer dev[%d, %d] vol = %d\n", devtype, devno, vol));
	if (devtype < 0 || devtype >= GSIMIXER_NUM_DEVICE_TYPES) {
		log_message(0, "set mixer: devtype %d out of range (0...%d)", devtype, GSIMIXER_NUM_DEVICE_TYPES -1);

		return -1;
	}
	if (devno < 0 || devno >= gsimixer_num_devices [ devtype ]) {
		log_message(0, "set mixer: devno %d out of range (0...%d)", devno, gsimixer_num_devices [ devtype ] -1);

		return -1;
	}

	gsimixer_volume [ devtype ] [ devno ] = vol;

	switch (devtype) {
	case GSI_MASTER:
		hw_volume = vol;
		break;
	case GSI_PCM:
		hw_volume = vol;
		break;
	case GSI_SYNTH:
	case GSI_CD:
	case GSI_LINE_IN:
	case GSI_MICROPHONE:
		if (current == NULL) // Initialization (no applications yet)
			hw_volume = vol;
		else
			hw_volume = (vol * current->application_volume[ devtype ][ devno ]) >> 8;
		break;
	}

	return gsimixer_set_device_volume(devtype, devno, hw_volume);
}


int mixer_get_device_volume (int devtype, int devno) {
	return gsimixer_get_device_volume(devtype, devno);
}


/* Application volume: */
void mixer_set_application_volume (int devtype, int devno, int vol) {
	int  hw_volume = -1;

	current->application_volume [ devtype ][ devno ] = vol;
	switch (devtype) {
	case GSI_MASTER:
		/*
		 * There are 2 possibilities:
		 * 1. leave as a NOP (current choice).
		 * 2. adjust all other virtual volumes and thus virtualize the
		 *    master volume too.
		 */
		//for (i=1; i<GSIMIXER_NUM_DEVICE_TYPES; i++) {
		//	mixer_set_application_volume (i, gsi_application_volume [i]);
		//	// Note this needs a master volume multiplier fix in all the
		//	// functions below..
		//}

		return;

	case GSI_PCM:
		/* set volume of running sounds and the new default for new sounds */
//NEW	current->application_pcm_volume [ devtype ] = vol;
		mixer_set_internal_pcm_volume(vol);
//NEW		mixer_set_internal_pcm_volume(dev, vol);
		break;
	case GSI_SYNTH:
		if (synth_owner == current_client)
			hw_volume = (vol * gsimixer_volume [ devtype ] [ devno ]) >> 8;
		gsi_application_volume [ GSI_SYNTH ] [ devno ] = vol;
//NEW		if (synth_devices[ dev ].owner == current_client)
//NEW			hw_volume = (vol * gsi_synth_mixer_volume [ dev ]) >> 8;
//NEW		gsi_synth_application_volume [ dev ] = vol;
		break;
	case GSI_CD:
//YYY		if (cd_owner[devno] == current_client || (cd_owner[devno] >= 65536 && current->cd_request[devno] == GSI_CMD_SHARE_CD) )
		if (cd_owner[devno] == current_client || (cd_owner[devno] == -1) )
			hw_volume = (vol * gsimixer_volume [ devtype ] [ devno ]) >> 8;
		gsi_application_volume [ GSI_CD ] [ devno ] = vol;
		break;
	case GSI_LINE_IN:
		gsi_application_volume [ GSI_LINE_IN ] [ devno ] = vol;
		//if (device_owner[ GSI_LINE_IN ] == current_client)
			hw_volume = (vol * gsimixer_volume [ devtype ] [ devno ]) >> 8;
//NEW			hw_volume = (vol * gsi_linein_mixer_volume [ dev ]) >> 8;
		break;
	case GSI_MICROPHONE:
		gsi_application_volume [ GSI_MICROPHONE ] [ devno ] = vol;
		//if (device_owner[ GSI_MICROPHONE ] == current_client)
			hw_volume = (vol * gsimixer_volume [ devtype ] [ devno ]) >> 8;
//NEW			hw_volume = (vol * gsi_microphone_mixer_volume [ dev ]) >> 8;
		break;
	default:
		/* a recording volume */
		break;
	}
	if (hw_volume >= 0) {
//??		if (!gsimixer_muted)
			gsimixer_set_device_volume(devtype, devno, hw_volume);
//NEW			gsimixer_set_device_volume(devtype, dev, hw_volume);
	}
}


void mixer_set_internal_pcm_volume (int vol) {
	int  i;

	for (i=0; i<current->channel_playnum; i++) {
		current->channel[ current->channel_used[i] ]->mix_volume = (current->channel[ current->channel_used[i] ]->volume * vol) >> 8;
	}
}


/*---------------------------------------------------------------------------*
 *      IPC: send the client mixer settings
 *---------------------------------------------------------------------------*/
void mixer_send_status () {
	int  devtype, devno;

	for (devtype=0; devtype < GSIMIXER_NUM_DEVICE_TYPES; devtype++) {
		for (devno=0; devno < gsimixer_num_devices[ devtype ]; devno++) {
			mixer_send_status_of(devtype, devno);
		}
	}
}


void mixer_send_status_of (int devtype, int devno) {
	int  vol;

	if (devtype < 0 || devtype > GSIMIXER_NUM_DEVICE_TYPES || devno < 0 || devno >= gsimixer_num_devices[ devtype ]) {
		return;
	}

	vol = gsimixer_volume[ devtype ] [ devno ];
	send_list(current, 4, GSI_EV_MIXER_STATUS, devtype, devno, vol > 0 ? vol-1: 0);
}


/*      EOF mixer.c */
