/* GSI - general sound interface
 *
 * gsi_server.c - shared sound server, initializes stuff (signal handlers,
 * tables etc).
 *
 * Lower level API dependencies:
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <setjmp.h>
#include <sys/stat.h>
#include <stdarg.h>

#include <syslog.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSIS_"
//#include <gsi/gsi_debug.h>
#include <gsi/version.h>

#include <gsi/gsif.h>

#include <gsi/gsipcm.h>
#include <gsi/gsicd.h>
#include <gsi/gsimixer.h>

#include "gsi_server.h"
#include "client.h"
#include "client_data.h"
#include "conf.h"
#include "real_time/real_time.h"

#include "utils/tables.h"
#include "utils/p_signal.h"
#include "debug.h"

#include "log.h"


/* from pcm/pcm.c (TEMP): */
extern gsipcm_device_t *server_pcm;
extern gsipcm_devicedata_t *server_pcm_status;


/*__________________DEFINITIES___________________*/
#ifdef SYSCONFIG_LITTLE_ENDIAN_MACHINE
int server_byte_order = GSI_LITTLE_ENDIAN;
#else
int server_byte_order = GSI_BIG_ENDIAN;
#endif


/*__________________EXTERN___________________*/


/*__________________VARIABELEN___________________*/
char *server_name;

/* share the server? (if root starts it, then let others connect too for
 * example) */
int  share_server = False; /* I think I'm paranoid */

/* CONNECTION stuff: */
int  ls;

int  forced_connection = GSI_CONNECTION_ANY;
int  af_local_socket_created = False;

uid_t saved_uid;
gid_t saved_gid;


/*__________________FUNCTIE PROTOTYPES___________________*/
void use (void);

void server_loop (void);

static void close_server (void);
static void signal_handler (int msg);

void lib_message_handler (int msg_type, char *msg);


/*__________________PROGRAMMA___________________*/
void use(void) {
	fprintf(stderr, "\nUse of "SERVER_NAME":\n"
			"  "SERVER_NAME" -v\n"
			"or:\n"
			"  "SERVER_NAME" [-d] [-q] [-s] [-p protocol]\n"
			"where protocol is AF_LOCAL or AF_INET\n\n"
			"See "SERVER_NAME"(8) for more information.\n\n"
		);
	exit(0);
}

void set_real_user(void) {
	seteuid(
		getuid ());
	setegid(getgid());
}
void set_suid_user(void) {
	seteuid(saved_uid);
	setegid(saved_gid);
}

sigjmp_buf restart_server;

int daemon_init(void) {
	int  status;
	pid_t pid;

fprintf(stderr, "daemon init\n");
	if ((pid = fork()) < 0)
		return -1;
	else if (pid != 0)
		exit(0); /* parent quits, shell thinks command is done */

	/* child continues */
	status = setsid();
	if (status < 0) log_message(GSI_MESSAGE_ERROR, "Couldn't create a new session\n");

	chdir("/");
	/* Otherwise we can't unmount a filesystems if the server was started
	 * there (at least until another chdir). Note that this will not allow
	 * coredumping for non-root users... */

	umask(0);

	return 0;
}


int main (int argc, char **argv) {
	int  status;
	int  args = 0;
	int  debug_mode = 0;

	/* Handle all options before the fork */
	if ((argc-1)-args >= 1 && !strncmp(argv[ 1+args ], "-v", 3)) {
		printf (GSI_SERVER_STRING"\n");
		exit (0);
	}

	if ((argc-1)-args >= 1) {
		if (!strncmp(argv[ 1+args ], "-d", 3)) {
			debug_mode = 1;
			args++;
		}
	}

	while ( (argc-1)-args >= 1) {
		if (!strncmp(argv[ 1+args ], "-q", 3)) {
			server_be_quiet = 1;
			args++;
		}
		else if (!strncmp(argv[ 1+args ], "-s", 3)) {
			share_server = True;
			args++;
		}
		else if (argc >= 1+2+args && !strncmp(argv[ 1+args ], "-p", 3)) {
			/* protocol specification */
			if (!strncmp(argv[ 2+args ], "AF_LOCAL", 8))
				forced_connection = GSI_CONNECTION_AF_LOCAL;
			else if (!strncmp(argv[ 2+args ], "AF_INET", 8))
				forced_connection = GSI_CONNECTION_AF_INET;
			else use();
			args += 2;
		}
		else use();
	}

	server_name = argv[0];

//	if (!debug_mode) daemon_init();

	saved_uid = geteuid();
	saved_gid = getegid();

	set_real_user();

	/* handle signals */
	//p_signal(SIGSEGV, signal_handler);
//	p_signal(SIGFPE,  signal_handler);

	p_signal(SIGQUIT, signal_handler);
	p_signal(SIGHUP,  signal_handler);
	p_signal(SIGINT,  signal_handler);
	p_signal(SIGBUS,  signal_handler);
	p_signal(SIGTERM, signal_handler);

	p_signal(SIGPIPE, SIG_IGN);
	p_signal(SIGTTOU, SIG_IGN); /* don't suspend process if tty output for background process fails */

	openlog("GSI (server)", LOG_CONS, LOG_DAEMON);
	gsif_log_set_handler (lib_message_handler);
	gsimixer_log_set_handler (lib_message_handler);
	gsipcm_log_set_handler (lib_message_handler);
	gsicd_log_set_handler (lib_message_handler);
	//gsisynth_log_set_handler (lib_message_handler);

	atexit(close_server);

// To test core dumping:
// kill(0, SIGSEGV);


	/* For SIGHUP restarting: */
	sigsetjmp(restart_server, True);

	/* We need to do a setpgrp (=setpgid(0,0)) so that the daemon will no
	 * longer be associated with the user's control terminal.
	 *
	 * This is done before the fork, so that the child will not be a process
	 * group leader.  Otherwise, if the child were to open a terminal, it
	 * would become associated with that terminal as its control terminal.
	 * It is always best for the parent to do the setpgrp.
	 */
	setpgid(0,0);

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	status = read_server_config(); /* N.B. That's where devices are attached */

fprintf(stderr, "XYZ 0\n");
	if (status) {
fprintf(stderr, "XYZ 0.1\n");

		log_message (GSI_MESSAGE_ERROR, "Couldn't read config file, quitting");

		exit(1);
	}
fprintf(stderr, "XYZ 1\n");
	ls = init_connection();
fprintf(stderr, "XYZ 1\n");


	/* NOP macro or function depending on DEBUG: */
	debug_init();
	debug_show_devices();


	/* PCM module init: */
	/* init_pcm(); */
	make_volume_table();
	make_sinus();


	/* Should try RT only if set in the config file? And if RT succeeds we can
	 * make the 0.05s playahead smaller).
	 */
	set_suid_user(); init_real_time(); set_real_user();

	server_loop();

//	set_suid_user(); end_real_time(); set_real_user();

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Signal handler:                                                      *
 */
static void signal_handler(int sgn) {
	//int  save_errno = errno;
	char *sgntext = "?";

	switch(sgn) {
	case SIGFPE:  sgntext = "SIGFPE";  break;
	case SIGSEGV: sgntext = "SIGSEGV"; break;
	case SIGTERM: sgntext = "SIGTERM"; break;
	case SIGINT:  sgntext = "SIGINT";  break;
	case SIGQUIT: sgntext = "SIGQUIT"; break;
	case SIGHUP: {
		sgntext = "SIGHUP";
		/* In case of SIGHUP restart the server */
		/* XXX is this OK? syslog is not necessarily reentrant.. (possibly
		 * irrelevant as we don't restart any old syslog call as we
		 * siglongjmp). */
		syslog(LOG_NOTICE, "%s: restarted with signal %d (%s)\n", server_name, sgn, sgntext);
		//execve(argv[0], argv, NULL);
		// close all fd's ?
		siglongjmp(restart_server, 1);
	}

	}
	/* XXX is this OK? syslog is not necessarily reentrant.. (possibly
	 * irrelevant as we quit anyway). */
	syslog(LOG_NOTICE, "%s: killed with signal %d (%s)\n", server_name, sgn, sgntext);
	if (sgn == SIGFPE) {
		kill(0, SIGSEGV);
	}


	exit(sgn);
}


void lib_message_handler(int msg_type, char *msg) {
	log_message (msg_type, msg);
}


/*---------------------------------------------------------------------------*
 *     Server exits.
 *---------------------------------------------------------------------------*/
static void close_server(void) {
	if (connection_type == GSI_CONNECTION_AF_LOCAL && af_local_socket_created) {
		unlink(GSI_LOCAL_SOCKET_NAME);
	}
}


/*      EOF gsi_server.c */
