/* GSI - general sound interface
 *
 * io.c - server io
 *
 * Lower level API dependencies:
 *   -
 *
 *
 * Copyright (C) 1999 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include <gsi/whs_types.h>
//#define DEBUG_PREFIX "GSI_SERVER (IO)"
#define DEBUG_PREFIX "(IO)"
#include <gsi/gsi_debug.h>

#include "io.h"
#include "play/cd/cd.h"
#include "client.h"
#include "gsi_server.h"



/*---------------------------------------------------------------------------*
 *      Send status information/events.
 *---------------------------------------------------------------------------*/
void send_byte(gsi_client_t *c, int data) {
	byte b = data;

	flush_if_no_room_for (c, 1);

	/* should throw away earliest events if buffer still full.
	 * => store event length too in status_queue (separate?); no need to send
	 * the length to the client (but might do that later).
	 * Also need to keep track of partially sent data. */
	c->status_queue[ c->status_queue_len++ ] = b;
	DEB(DPRT("send_byte (client 0x%x): byte = %d\n", (int) c, data));
}


void send_int32(gsi_client_t *c, int data) {
	flush_if_no_room_for (c, 4);

	c->status_queue[ c->status_queue_len++ ] = (data >>24);
	c->status_queue[ c->status_queue_len++ ] = (data >>16);
	c->status_queue[ c->status_queue_len++ ] = (data >>8);
	c->status_queue[ c->status_queue_len++ ] = (data);
	DEB(DPRT("send_int32 (client 0x%x): int = %d\n", (int) c, data));
}


void send_list (gsi_client_t *c, int argc, ...) {
	va_list argp;

	va_start(argp, argc);
	flush_if_no_room_for (c, argc);
	while (argc--) {
		c->status_queue[ c->status_queue_len++ ] = va_arg(argp, byte);
		DEB(DPRT("send_list (client 0x%x): byte = %d\n", (int) c, c->status_queue[ c->status_queue_len-1 ]));
	}
//DEB(DPRT("send_list(): done\n"));
	va_end(argp);
}


void send_list_to_cd_owners (int cd, int argc, ...) {
	va_list argp;
	int  i, j;
	gsi_client_t *c;

	for (i=0; i<num_clients; i++) {
		if ( is_a_cd_owner(cd, i) ) {
			va_start(argp, argc);
			c = client_list[i];
			flush_if_no_room_for (c, argc);
			j = argc;
			while (j--) {
				c->status_queue[ c->status_queue_len++ ] = va_arg(argp, byte);
			}
		}
	}
	va_end(argp);
}


#if 0
void send_list_to (int send, int argc, ...) {
	int  i;
	va_list argp;

	va_start(argp, argc);

	if (send == SEND_CURRENT) {
		flush_if_no_room_for (current, argc);
		j = argc;
		while (j--) {
			current->status_queue[ current->status_queue_len++ ] = va_arg(argp, byte);
		}
	}
	else {
		for (i=0; i<num_clients; i++) {
			if ( ((send & SEND_CD_OWNERS) && is_a_cd_owner(cd, i)) || (send & SEND_ALL)) {
				c = client_list[i];
				va_start(argp, argc);
				j = argc;

				flush_if_no_room_for (c, argc);
				while (j--) {
					c->status_queue[ c->status_queue_len++ ] = va_arg(argp, byte);
				}
			}
		}
	}

	va_end(argp);
}
#endif


void flush_if_no_room_for (gsi_client_t *c, int bytes) {
	while (c->status_queue_len >= MAX_STATUS_QUEUE_LEN - bytes)
		flush (c);
	// should use select.
}


/*---------------------------------------------------------------------------*
 *      Soft flush (flush as many bytes as possible without blocking) to all
 *      clients.
 *---------------------------------------------------------------------------*/
void  flush (gsi_client_t *c) {
	int  j;

	if (c->status_queue_len > 0) {
		j = write (c->socket_fd, c->status_queue, c->status_queue_len);
		if (j > 0) {
			c->status_queue_len -= j;
			if (c->status_queue_len) memmove(c->status_queue, c->status_queue+j, c->status_queue_len);
		}
		else if (j < 0) {
			if (errno == EPIPE || errno == EBADF) {
				DEB1(DPRT("%s: writing to closed pipe or bad fd. Emptying queue (not yet removing the client)\n", server_name));
				c->status_queue_len = 0;
			}
		}
	}
}


/*---------------------------------------------------------------------------*
 *      Hard flush (flush everything, this may mean waiting for the
 *      application to read enough data so enough space in the OS' output
 *      buffer is available ..).
 *---------------------------------------------------------------------------*/
void flush_all () {
	int  i;
	gsi_client_t *c;

	DEB1(DPRT("("__FILE__"): *flush_all*\n"));
	for (i=0; i<num_clients; i++) {
		c = client_list[i];
		DEB2(DPRT("-> status flush of client %d (queue len=%d)\n", i, c->status_queue_len));
		flush(c);
		DEB2(DPRT("-> client %d done (now status_queue_len=%d)\n", i, c->status_queue_len));
	}
	DEB1(DPRT("#flush_all#\n"));
}


/*      EOF io.c */
