/* GSI - general sound interface
 *
 * slot.c - shared sound server, sample & list loading & converting samples
 * to internal format (using gsif_*).
 *
 * Lower level API dependencies: none (standard file system stuff)
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>

#include <signal.h>

#include <gsi/config.h>
#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_SERVER (play slot)"
#include <gsi/gsi_debug.h>
#include <gsi/gsi_commands.h>
#include <gsi/gsi_events.h>
#include <gsi/gsif.h>

#include "../gsi_server.h"
#include "../io.h"
#include "slot.h"
#include "../utils/tables.h"

#include <gsi/whs_defines.h>


/*__________________DEFINITIES___________________*/


/*__________________VARIABELEN___________________*/
//int free_slot(struct gsis_slot **s) {
int free_slot(struct gsi_client *c, int slot) {
	struct gsis_slot **s = &c->slot [ slot ];

	if (*s != NULL) {
		switch((*s)->type) {
		case _GSI_SLOTTYPE__UNDEFINED:
			break;

		case _GSI_SLOTTYPE__PCM_SAMPLE:
			if ((*s)->data.sample != NULL) {
				if ((*s)->data.sample->buffer != NULL) free((*s)->data.sample->buffer);

				switch((*s)->data.sample->input) {
				case _GSI_INPUT_FILE:
					gsif_free_sound((*s)->data.sample->data.file); break;
				case _GSI_INPUT_STREAM:
					free((*s)->data.sample->data.stream); break;
				case _GSI_INPUT_SHM:
					free((*s)->data.sample->data.shm); break;
				default:
					return -1;
				}
			}
			break;

		case _GSI_SLOTTYPE__MIDI_SONG:
			if ((*s)->data.midi != NULL) {
				gsif_free_song((*s)->data.midi->file);
			}
			kill(0, SIGSEGV); /* not finished */
			break;

		default:
			fprintf(stderr, "slot type %d\n", (*s)->type);
			kill(0, SIGSEGV); /* not implemented */	
			break;
		}
	}

	return 0;
}


int init_slot(struct gsis_slot **s, int type) {
	if (*s == NULL) {
		*s = calloc(1, sizeof(struct gsis_slot));
	}
	else {
//		free_slot(s);
//		free_slot(c, slot);
	}

	if (*s == NULL) return -1; /* malloc failed */

	switch((*s)->type) {
	case _GSI_SLOTTYPE__PCM_SAMPLE:
		(*s)->data.sample = calloc(1, sizeof(struct gsis_slot_sample));
		(*s)->data.sample->input = type;
		switch(type) {
		case _GSI_INPUT_FILE:
//			(*s)->sample->data.file = gsif_get_structure...
			break;
		case _GSI_INPUT_STREAM:
			(*s)->data.sample->data.stream = calloc(1, sizeof(struct gsis_stream));
			if ((*s)->data.sample->data.stream == NULL) return -1;
			break;
		case _GSI_INPUT_SHM:
			(*s)->data.sample->data.shm = calloc(1, sizeof(struct gsis_shm));
			if ((*s)->data.sample->data.shm == NULL) return -1;
			break;
		default:
			return -1;
		}
		break;

	case _GSI_SLOTTYPE__MIDI_SONG:
		(*s)->data.midi = calloc(1, sizeof(struct gsis_slot_midi));
		kill(0, SIGSEGV); /* not finished */
		break;

	default:
		kill(0, SIGSEGV); /* not implemented */	
		break;
	}

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Slot management.
 *      Returns True (1) if the slot is being played, False (0) if not.
 *      (assumes current = current client)
 *---------------------------------------------------------------------------*/
int slot_busy(struct gsi_client *c, int slot) {
	int  i;
	struct gsis_slot *s = c->slot[ slot ];

	if (s == NULL) return False;

	switch(s->type) {
	case _GSI_SLOTTYPE__PCM_SAMPLE:
		if (s->data.sample->buffer == NULL) return False;
		break;
	case _GSI_SLOTTYPE__MIDI_SONG:
		break;
	default:
		return 0;
	}

	for (i=0; i<current->channel_playnum; i++) {
		if (current->channel[ i ]->slot == slot) return True;
	}

	return False;
}


/*---------------------------------------------------------------------------*
 *      Unload a sound.
 *---------------------------------------------------------------------------*/
//void unload_slot(struct gsis_slot *s) {
void unload_slot(struct gsi_client *c, int slot) {
	struct gsis_slot *s = c->slot [ slot ];

	DEB3(DPRT("LOAD   ("__FILE__"): *unload_slot* (slot %d)\n", slot));

	if (s) {
		switch((s)->type) {
		case _GSI_SLOTTYPE__PCM_SAMPLE:
			if (s->data.sample->buffer != NULL) {
				free(s->data.sample->buffer);
				s->data.sample->buffer = NULL;
				s->data.sample->size = 0; /* used to test if sound was loaded in main loop */
			}
			break;
		case _GSI_SLOTTYPE__MIDI_SONG:
			break;
		default:
			break;
		}
	}
	DEB3(DPRT("LOAD   #unload slot#\n"));
}


/*---------------------------------------------------------------------------*
 *      Read all sound files from disk.
 *      (prevents hickups when a sound is first used)
 *---------------------------------------------------------------------------*/
void load_all_slots(struct gsi_client *c) {
	int i;

	for (i=0; i<c->max_slot_used; i++) {
//		struct gsis_slot *s;
//
//		s = c->slot [ i ];
//		if (!load_slot(s))
		if (!load_slot(c, i))
			; /* then what? */
	}
}


/*---------------------------------------------------------------------------*
 *      Unload all sounds loaded into slots.
 *---------------------------------------------------------------------------*/
void unload_all_slots(struct gsi_client *c) {
	int  i;

	for (i=0; i<c->max_slot_used; i++) {
//		struct gsis_slot *s;
//
//		s = c->slot [ i ];
//		unload_slot(s);
		unload_slot(c, i);
	}
}


void free_all_slots(struct gsi_client *c) {
	int i;

	for (i=0; i<c->max_slot_used; i++) {
//		struct gsis_slot *s;
//
//		s = c->slot [ i ];
//		free_slot(s);
		free_slot(c, i);
	}
}


/*---------------------------------------------------------------------------*
 *      Read a sound file from disk.
 *        If read_header /=0 check the file type at the given offset.
 *        If data_len=0, load the entire rest of the file (i.e. len=Inf).

 *     WHY??? Just use data_len=MAX_INT /MAX_UINT, whatever!
 *      ERRORS: should send_byte(c, GSI_EV...) be called from this function ?
 *---------------------------------------------------------------------------*/
int init_slot_with_sample(struct gsi_client *c, int slot, char *sound_file, int read_header, int data_offset, int data_len) {
	struct gsi_sound *new_snd;
	int  status;

fprintf(stderr, "init_slot_with_sample\n");
	unload_slot(c, slot);
	if (sound_file == NULL) {
		/* send an error (or in higher level code?) */
		return -1;
	}

	/* This should be at a higher level (no need to malloc/free each time) */
	free_slot(c, slot);
	if (c->slot[ slot ] == NULL) {
		c->slot[ slot ] = calloc(1, sizeof(struct gsis_slot));
	}

	c->slot [ slot ]->type = _GSI_SLOTTYPE__PCM_SAMPLE;
	init_slot(&c->slot [ slot ], _GSI_INPUT_FILE);
	
	new_snd = gsif_get_structure(sound_file, read_header, data_offset, data_len, 8000, 1 , GSI_8BIT, &status);
printf("new snd=%x\n", new_snd);
	if (new_snd == NULL) return -1; /* not a recognized sample */

	c->slot[ slot ]->data.sample->data.file = new_snd;

	c->max_slot_used = max(c->max_slot_used, slot);

	return 0;
}


int load_slot(struct gsi_client *c, int slot) {
	int  status;
	struct gsis_slot *s;

	if (c->slot[ slot ] == NULL) {
		return -1;
	}

	s = c->slot[ slot ];
	switch(s->type) {
	case _GSI_SLOTTYPE__PCM_SAMPLE:
	{
		struct gsis_slot_sample *s_sample;
		s_sample = s->data.sample;

		if (s_sample == NULL) return -1;

		switch(s_sample->input) {
		case _GSI_INPUT_FILE:
		{
			struct gsi_sound *new_snd;
			struct gsi_block new_blk;

			new_snd = s_sample->data.file;

			if (new_snd == NULL) {
				return -1;
			}

			/* Conversion into a single block: */
			status = gsif_load_sound_in_block (new_snd, &new_blk, -1 /*samplerate*/, -1 /*channels*/, -1 /*type*/);
			if (status < 0) {
				free_slot(c, slot);
				return -1;
			}

			s_sample->samplerate = new_blk.samplerate;
			s_sample->channels = new_blk.channels;
			s_sample->type = new_blk.type;
			s_sample->size = new_blk.data_length;
			DEB1(DPRT("LOAD   new_blk.data_length=%ld\n", new_blk.data_length));

			//padded_size = c->sound_size[ slot ];

			if (s_sample->samplerate < 1 ||
				s_sample->channels < 1 ||
				s_sample->channels > 2 ||
				s_sample->type < GSI_8BIT_MULAW ||
				s_sample->type > GSI_16BIT_LE
				) {
				s_sample->buffer = NULL;
				s_sample->size = 0;
				gsif_free_sound(new_snd);
				free(new_blk.data);

				return -1;
			}

			/* This should be at a higher level (no need to malloc/free each
			 * time. */
			s_sample->data.file = new_snd;

			s_sample->buffer = new_blk.data;
			new_blk.data = NULL;

			DEB1(DPRT("LOAD   new_blk.data_length=%ld, new_blk.samplerate=%d new_blk.channels=%d new_blk.type=%d\n", new_blk.data_length, new_blk.samplerate, new_blk.channels, new_blk.type));
#if 0
			// put in gsif_load_sound ?
			padded_size = size;
			/* padding: add silence for missing few bytes */
			pad = 1;
			if (s->type == GSI_16BIT) pad *= 2;
			pad *= s_sample->channels;

			padded_size= (size + (pad-1)) & ~(pad-1);

			s_sample->buffer = (sbyte *) malloc(padded_size);
			s_sample->size = padded_size;

			if (padded_size>size) {
				if (s_sample->type == GSI_8BIT) {
					s_sample->buffer[size] = 0;
					/* not needed when using callloc */
				}
			}
#endif

			convert_slotdata_to_internal(c, slot);
			s_sample->meta_sample_size = s_sample->channels * _server_gsi_bytes_per_sample[ s_sample->type ];
			s_sample->meta_samples = s_sample->size / s_sample->meta_sample_size;
			break;
		} /* INPUT FILE */

		case _GSI_INPUT_STREAM:
			return -1;

		case _GSI_INPUT_SHM:
			return -1;

		default:
			return -1;
		} /* _GSI_SLOTTYPE__PCM_SAMPLE */
		break;
	}

	case _GSI_SLOTTYPE__MIDI_SONG:
	{
//		struct gsi_song  *new_song;
		break;
	}

	default:
		return -1;
		break;
	}

	return 0;
}


/*---------------------------------------------------------------------------*
 *      Conversion from file format to internal format:
 *---------------------------------------------------------------------------*/
int convert_slotdata_to_internal(struct gsi_client *c, int slot) {
	struct gsis_slot *s = c->slot[ slot ];
	int  size;
	struct gsi_sample  new_sample;

	if (s->type != _GSI_SLOTTYPE__PCM_SAMPLE) return -1;

	size = s->data.sample->size;

	memcpy(&new_sample, s->data.sample, sizeof(struct gsi_sample));
	if (sample_internal_type_of (&new_sample)) return -1;
	convert_data_to_internal(s->data.sample->buffer, size, (struct gsi_sample *) s->data.sample);
	sample_internal_type_of((struct gsi_sample *) s->data.sample);

	return 0;
}


void convert_data_to_internal (byte *data, int size, struct gsi_sample *sample)
{
//	int  size = num_samples;
	int  i;

	/* Manipulate each sound type where needed */
	switch(sample->type) {
	case GSI_8BIT_SIGNED:
		/* no change, just internal notation (GSI_8BIT= signed...) */
		break;
	case GSI_8BIT:
		/* GSI_8BIT -> GSI_8BIT_SIGNED internal */
		for (i=0; i<size; i++) data[i] ^= 0x80;
		break;
	case GSI_8BIT_MULAW:
	case GSI_16BIT:
		/* = native */
		break;
	case GSI_16BIT_LE:
		//size *= 2;
		for (i=0; i<size; i+=2) {
			byte sbi    = data[ i ];
			data[ i ]   = data[ i+1 ];
			data[ i+1 ] = sbi;
		}
		break;
	default:
		DEB1(DPRT("(convert_data_to_internal): unhandled conversion from type %d to internal\n", sample->type));
	}
}


int sample_internal_type_of (struct gsi_sample *sample)
{
	if (!sample) return 1;
	switch(sample->type) {
	case GSI_8BIT_SIGNED:
		sample->type = GSI_8BIT;
		break;
	case GSI_8BIT:
		break;
	case GSI_8BIT_MULAW:
	case GSI_16BIT:
		break;
	case GSI_16BIT_LE:
		sample->type = GSI_16BIT;
		break;
	default:
		DEB1(DPRT("GSI (sample_internal_type_of): unknown sample type %d (no internal version known)\n", sample->type));
		return -1;
	}
	return 0;
}


/*---------------------------------------------------------------------------*
 *      Initialize a sample (set conversion factors for samplerate scaling).
 *---------------------------------------------------------------------------*/
void init_sample(struct gsi_client *c, int slot, int samplerate, int channels, int type, int size)
{
	struct gsis_slot *s = c->slot[ slot ];
	struct gsis_slot_sample *s_sample = s->data.sample;

	s_sample->samplerate = samplerate;
	s_sample->channels = channels;
	s_sample->type = type;
	s_sample->size = size;
	DEB1(DPRT("LOAD ("__FILE__"): *init_sample*\n"));
	DEB1(DPRT("LOAD    set slot %d to: samplerate=%d, channels=%d type=%d (%d bytes)\n", slot, samplerate, channels, type, size));

	convert_slotdata_to_internal(c, slot);
	//NOT NEEDED set_samplerate_conversion_of(slot);

	DEB1(DPRT("LOAD #init_sample#\n"));
}


int  gsimisc_type(char *type) {
	int  i;
	for (i=0; i<8; i++) {
		if (!strncmp(type, _server_gsi_type_strings[i], 10))
			return _server_gsi_types[i];
	}
	return -1;
}


/*      EOF slot.c */
