/* GSI - general sound interface
 *
 * main_loop.c - shared sound server, main loop; reads commands from
 * pipe/sockets, mixes channels, sends data to pcm, synth.
 *
 * Lower level API dependencies:
 *   - libgsipcm, unix file descriptors & their use in select.
 *
 *
 * Copyright (C) 1996 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/*__________________INCLUDES___________________*/
#include <gsi/config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>

#include <math.h>

#include <gsi/whs_types.h>
#define DEBUG_PREFIX "GSI_SERVER "
#include <gsi/gsi_debug.h>
#include <gsi/gsi_commands.h>
#include <gsi/gsi_events.h>
//#include <gsi/gsi_eventmasks.h>
#include <gsi/version.h>

#include <gsi/gsipcm.h>
#include <gsi/gsicd.h>


#include "gsi_server.h"
#include "client.h"
#include "conf.h"
#include "utils/tables.h"
#include "client_device_requests.h"

#include "device/mixer/mixer.h"
#include "device/pcm/pcm.h"

#include "load/slot.h"
#include "play/sample/mulaw.h"

#include "play/song/synth.h"
#include "play/song/song_play.h"

#include "play/cd/cd.h"


#include "io.h"

#include "log.h"

/* from pcm/pcm.c (TEMP): */
extern gsipcm_device_t *server_pcm;
extern gsipcm_devicedata_t *server_pcm_status;


/*__________________DEFINITIES___________________*/
#define PCM_PREMIX_SIZE  (sizeof(int)  * 1024*2)
#define PCM_FINAL_SIZE   (sizeof(byte) * 1024*2)



//#define SPEED_TEST
#ifdef SPEED_TEST
struct timeval tv1, tv2;
#  define SPEED_TEST__INIT_TIME() gettimeofday(&tv1, NULL)

#  define SPEED_TEST__PRINT_TIME(a) \
	gettimeofday(&tv2, NULL); \
	tv2.tv_usec -= tv1.tv_usec; \
	if (tv2.tv_usec < 0) { tv2.tv_usec += 1000000; } \
	fprintf(stderr, "dt=%3d s ("a")\n", tv2.tv_usec);
#else
#  define SPEED_TEST__INIT_TIME()

#  define SPEED_TEST__PRINT_TIME(a)
#endif

/* Number of fragments to write before closing the pcm device: */
//#define NUM_SILENCE_FRAGS 1
#define NUM_SILENCE_FRAGS 2147483647


/*__________________EXTERN___________________*/
/* commands.c: */
extern void  decode_command (gsi_client_t *c);
extern void  post_decode_command (gsi_client_t *c);


/*__________________VARIABLES___________________*/
extern int server_byte_order;

/* Device owners (client list numbers, -1 if none): */
/* OLD: int    pcm_owner   = -1;*/

int    synth_owner = -1; /* REMOVE/FIX for the multi device system */

int    cd_owner[ 16 ] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

int    mixer_owner = -1;



/* PCM: */
byte  *pcm_final;
/*     n_samples (default is 1024*2) */

int   *pcm_premix;
/*     n_samples*n_channels*n_bytes, i.e. <= n_samples*4 (default is 1024*8) */

int    total_playnum = 0;
/*     total number of channels playing (all clients) */

int    autorelease_pcm = False;
int    autoset_pcm = False;

int    pcm_request[ 64 ];
/*     numbers of the client */

int    pcm_requests;
/*     number of clients that currently want the use of the PCM */

int    write_silence_ct = NUM_SILENCE_FRAGS;
/*     nr. of frags to write after playing stops */

float  pcm_time;
/*     = pcm_samples/pcm_samplerate */

int    current_time;
/*     1/32s fragment counter for doppler effects */

int    select_timeout;
/*     PCM device timing */

/* SYNTH: */
int    num_songs_playing = 0;
/*     total number of songs playing (all clients) */

//int    song_client[ 64 ];
/*     each client can have a song playing, the song number gives the client
 *     number. */


/* CLIENT: */
int  current_client;
struct gsi_client *current = NULL;


/*__________________FUNCTION PROTOTYPES___________________*/
static void  play_fragment(struct gsi_client *c, int slot);



/*---------------------------------------------------------------------------*
 *      Main server loop. Decodes commands and plays samples.
 *---------------------------------------------------------------------------*/
void server_loop (void) {
	int  i;
	int  socket_fd;
	int  client;
	int  status;

	/* For select tests: */
	int max_fd; fd_set loop_fds; struct timeval loop_tv = { 0, 0 }; int n;

	pcm_premix = malloc( PCM_PREMIX_SIZE );
	pcm_final = malloc( PCM_FINAL_SIZE );






	/* Main loop that we only leave on exit. */
	for (;;) {
		DEB2(DPRT("main loop\n"));

		/* Note: doing a select takes 22s while accept (from add_client)
		 * takes 54s when no new client tries to connect!
		 * (timed under linux 2.0 on a K6-233).
		 *
		 * Perhaps it would be a good idea to only check for new connections
		 * every 0.10s or so.
		 */
		SPEED_TEST__INIT_TIME();

		FD_ZERO(&loop_fds);
		FD_SET(ls, &loop_fds); /* to catch connection requests (and disconnects) */
		max_fd = ls;

		/* => Should make a backup of loop_fds and memcpy it ? (actually not
		 *    needed, just don't use loop_fds elsewhere). Update loop_fds if
		 *    clients connect and/or disconnect). */
		for (n=0; n<num_clients; n++) {
			int fd_n = client_list[ n ]->socket_fd;

			FD_SET(fd_n, &loop_fds);
			max_fd = max(max_fd, fd_n);
		}
		status = select(max_fd + 1, &loop_fds, NULL, NULL, &loop_tv);

		/* select may have been interrupted by a timer signal for the CD(s) */
		//if (status < 0) {
			if (cd_track_alarm) {
				cd_track_alarm = False;
				//DEB(DPRT("CD: generating events\n"));
				gsicd_generate_events();
			}
		//}


		/* New client? */
		if (FD_ISSET(ls, &loop_fds)) {
			if (!add_client()) {
				/* Put this in add_client ?: */
				current = client_list[ num_clients -1 ];
				send_byte(current, GSI_EV_SERVER_PROTOCOL_VERSION);
				send_int32(current, GSI_PROTOCOL_VERSION_BCD);
				send_int32(current, GSI_PROTOCOL_COMPATIBLE_VERSION_BCD);

				mixer_send_status();
				flush (current);

				send_byte(current, GSI_EV_SHARE_PERMISSIONS);
				if (share_server)
					send_int32(current, af_local_permissions);
				else
					send_int32(current, 0700);
			}
		}
		SPEED_TEST__PRINT_TIME("add client");

		SPEED_TEST__INIT_TIME();

		for (client=0; client<num_clients; client++) {
			current_client = client;
			current = client_list[ client ];

			socket_fd = current->socket_fd;

			if (FD_ISSET(socket_fd, &loop_fds)) {

				/* Get all commands: */
				for (;;) {
					/* ring buffer:
					 *   if start <= end then read (BUFSIZE-end) bytes
					 *   else read (start-1-end) bytes. */
					int   max_read_len;
					void *read_addr;

					//label__io_read:
					/* If we're reading a block for a client, then read
					 * directly into it as we've malloced space for it... */
					if (current->input_read_block) {
						max_read_len = current->input_block_size - current->input_block_len;
						read_addr = current->input_block + current->input_block_len;
					}
					else {
						int  read_next = current->input_buffer_start + current->input_buffer_len;
						if (read_next >= RING_BUFFER_SIZE) {
							max_read_len = RING_BUFFER_SIZE - current->input_buffer_len;
							read_next -= RING_BUFFER_SIZE;
						}
						else {
							max_read_len = RING_BUFFER_SIZE - read_next;
						}

						read_addr = &(current->input_buffer[ read_next ]);
					}

					i = read(socket_fd, read_addr, max_read_len);
					DEB1(DPRT("read status=%d\n", i));

					if (i == 0) {
						/* EOF on the pipe means the parent has closed its end */
						DEB1(DPRT("EOF from client %d\n", client));
						remove_client(client);

						client--;
						break; /* Skip this loop and get the next client */
					}
					else if (i == -1) {
						DEB2(DPRT("no data for client %d\n", client));
						if (errno == EPIPE || errno == EBADF) {
							log_message(GSI_MESSAGE_ERROR, "EPIPE/EBADF: client %d dead", client);
							//remove_client(client);
							//client--;
						}
						break; /* next client */
					}

					/* When we get here, i is always > 0 */
					if (!current->input_read_block) {
						current->input_buffer_len += i;
					}
					else {
						current->input_block_len += i;
					}


					//if (i == max_read_len) {
					//	/* try to read more (at start of buffer, if there's
					//	 * space) */
					//	goto label__io_read;
					//}


				  label__next_command:
					if (current->input_nuke) {
						int nuked = min(current->input_buffer_len, current->input_nuke);
						current->input_buffer_start += nuked;
						if (current->input_buffer_start >= RING_BUFFER_SIZE)
							current->input_buffer_start -= RING_BUFFER_SIZE;
						current->input_nuke -= nuked;
						current->input_buffer_len -= nuked;
					}

					if (current->input_read_block) {
						if (current->input_block_len == current->input_block_size) {
							decode_command(current);
							post_decode_command(current);
							goto label__next_command;
						}
					}
					else {
						if (current->input_buffer_len > 0 && current->input_buffer_len >= current->input_cmd_len) {
							decode_command(current);
							post_decode_command(current);
							goto label__next_command;
						}
					}

				} /* end of loop to get all commands */
			} /* FDISSET */
		} /* loop to handle all clients */
		SPEED_TEST__PRINT_TIME("client commands");








		/* Send all callback info (only so much that we don't need to wait): */
		flush_all();

		DEB2(DPRT("pcm_fd = %d, autorelease_pcm = %d, autoset_pcm=%d\n", server_pcm_status->fd, autorelease_pcm, autoset_pcm));
		DEB2( { int clt; for (clt=0; clt< num_clients; clt++ ) DPRT("channel_playnum(%d)=%d\n", clt, client_list[clt]->channel_playnum); } );
		DEB2(DPRT("total_playnum=%d, write_silence_ct=%d\n", total_playnum, write_silence_ct));

		if ( ! ((total_playnum > 0) || (server_pcm_status->fd > 0 && write_silence_ct > 0)) ) {
//		if ( !( (autorelease_pcm && total_playnum > 0) || ((server_pcm_status->fd > 0) && (total_playnum > 0 || write_silence_ct > 0))) ) {

			fd_set read_fds;
			//fd_set write_fds;
			//struct timeval timeout;
			//int  tmp_synth_buffer_status = 0;

			DEB1(DPRT("= waiting =\n"));
			FD_ZERO(&read_fds);
			FD_SET(ls, &read_fds); /* To catch connection requests (and disconnects). */
			max_fd = ls;
			for (n=0; n<num_clients; n++) {
				int fd_n = client_list[ n ]->socket_fd;

				FD_SET(fd_n, &read_fds);
				max_fd = max(max_fd, fd_n);
			}

			select(max_fd + 1, &read_fds, NULL, NULL, NULL);
			DEB(DPRT("- continue\n"));
			continue;
		}

		DEB(DPRT("XYZ: server_pcm_status->fd=%d\n", server_pcm_status->fd));
		DEB(DPRT("XYZ: total_playnum=%d\n", total_playnum));
		DEB(DPRT("XYZ: autorelease_pcm=%d\n", autorelease_pcm));

		if (total_playnum == 0 && autorelease_pcm && server_pcm_status->fd > -1) {
			DEB2(DPRT("autoreleasing: pcm = -1\n"));
			gsipcm_close(server_pcm, server_pcm_status);
		}
		else if (total_playnum > 0 && autorelease_pcm && server_pcm_status->fd < 0) {
			current = client_list[ pcm_request[0] ];
			setup_pcm(server_pcm, server_pcm_status, current->new_pcm_samplerate, current->new_pcm_channels, current->new_pcm_type, current->new_pcm_log2_frag);
			DEB2(DPRT("autograbbing: pcm = %d\n", server_pcm_status->fd));
		}



		/* Now play all the active channels. */
		SPEED_TEST__INIT_TIME();
#if 0
		/* First check if we're running late or if we're too far ahead: */
		select_timeout = (int) (
			(1.0e6 * (double) (server_pcm_status->max_write_space - gsipcm_write_space(server_pcm, server_pcm_status) - server_pcm_status->fragsize))
			/
			((double) (server_pcm_status->samplerate * server_pcm_status->bps * server_pcm_status->channels))
			);
		if (select_timeout > 50000 + 10000) {
			/* Too far ahead, skip pcm fragment calculation and return to
			 * getting commands. */
			usleep(10000);
			continue;
		}
		if (select_timeout < 0) {
			/* PCM buffer underrun. */
			struct timeval cur_tv;

			gettimeofday(&cur_tv, NULL);
			dt = tv_diff(&cur_tv, &last_pcm_tv);
		}
#endif

		if (total_playnum) {
			write_silence_ct = NUM_SILENCE_FRAGS;

			DEB3(DPRT("play active channels\n"));
			/* Mix the sounds into the final buffer. */
			memset(pcm_premix, 0, PCM_PREMIX_SIZE);
			for (current_client=0; current_client<num_clients; current_client++) {
				DEB(DPRT("client %d\n", current_client));
				current = client_list[ current_client ];

				for (i=current->channel_playnum-1; i>=0; i--) {
					int  ch = current->channel_used[ i ];
					if (current->channel[ ch ]->type == _GSI_SOUNDTYPE_SAMPLE) {
						DEB(DPRT("client %d: play channel %d (sample)\n", current_client, ch));
						if (server_pcm_status->fd > 0)
							play_fragment(current, i);
//						else
//							silent_fragment(current, i);
					}
					else if (current->channel[ ch ]->type == _GSI_SOUNDTYPE_SONG) {
						DEB(DPRT("client %d: play channel %d (song)\n", current_client, ch));
						song_play_events(current_client, ch, 0.25);
					}
					else {
						log_message(GSI_MESSAGE_ERROR, "Bad slot type");
						//exit(1);
						kill(0, SIGSEGV);
					}
				}
			}
			SPEED_TEST__PRINT_TIME("play active");

			SPEED_TEST__INIT_TIME();
			/* How many samples do we calculate/write ? */
#include "play/sample/internal_to_pcm_format.c"
			SPEED_TEST__PRINT_TIME("buf->pcm format");

			SPEED_TEST__INIT_TIME();
		}
		else {
			DEB3(DPRT("Not using PCM, writing silence\n"));
			/* No sounds to play so fill the buffer with silence */
			/* Should do this too before releasing the PCM ? Is done now with
			 * NUM_SILENCE_FRAGS. */

			if (server_pcm_status->type == GSI_8BIT) {
				memset(pcm_final, 128, server_pcm_status->fragsize);
			}
			else if (server_pcm_status->type == GSI_8BIT_MULAW) {
				memset(pcm_final, 255, server_pcm_status->fragsize);
			}
			else { /* 16 & 24 bit (signed) */
				memset(pcm_final, 0, server_pcm_status->fragsize);
			}
			write_silence_ct --;
		}

		SPEED_TEST__PRINT_TIME("until pcm write");
		DEB1(DPRT("pcm_final =0x%x\n", pcm_final));
		/* Here we should: set non-blocking mode and read if we record
		 * gsipcm_set_mode(hw, hw_status, GSIPCM_MODE_NON_BLOCKING);
		 * // Read as much as the internal buffer can hold:
		 * bytes_read = gsipcm_read (hw, hw_status, record_buf, record_bufsize);
		 * // Set blocking mode if we use that (rely on blocking after writing a full fragment for timing)
		 * gsipcm_set_mode(hw, hw_status, GSIPCM_MODE_BLOCKING);
		 */

		if (server_pcm_status->fd > 0) {
			gsipcm_write_fragment(server_pcm, server_pcm_status);
		}

		/* Now write the recorded data to file or memory or socket. */

		/* For next select (or usleep?): 20000 = 0.02s (minimal repeated
		 * usleep time on most unices). */
#define PCM_PLAYAHEAD 40000
#define SYS_TIME      10000

		select_timeout = (int) ( (1.0e6 * (double) (server_pcm_status->max_write_space - gsipcm_write_space(server_pcm, server_pcm_status)))
			/ ((double) (server_pcm_status->samplerate * server_pcm_status->bps * server_pcm_status->channels)) ) - 40000;

		/* Could use select_timeout in select above, but then we must make sure
		 * we don't write too many fragments (select can return earlier than
		 * the timeout, usleep not)... So, for now: */

		//if (select_timeout > 0) usleep(select_timeout);
		if (select_timeout > 10000) {
			struct timeval sel_tv; sel_tv.tv_sec = 0; sel_tv.tv_usec = select_timeout;
			select (0, NULL, NULL, NULL, &sel_tv);
		}

		DEB(DPRT("select_timeout=%d\n", select_timeout));
		DEB2(DPRT("end of main loop\n\n"));
	}
}


/*---------------------------------------------------------------------------*
 *      Calculate one fragment of a sound playing on a given channel.
 *---------------------------------------------------------------------------*/
#include "play/sample/play_fragment.c"


/*      EOF main_loop.c */
