/* GSI - general sound interface
 *
 * shm.c - shared sound server, SYSV shared memory wrappers
 *
 * Lower level API dependencies:
 *   - SYSV SHM (shared memory)
 *
 *
 * Copyright (C) 1999 - 2001 W.H.Scholten
 *
 * Redistribution and use in source and binary forms, with or
 * without modification, are permitted provided that the following
 * conditions are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gsi/config.h>
#include <gsi/gsi_shm.h> /* SHM blocksize etc. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h> /* for rand */
#include <errno.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#define DEBUG_PREFIX "(SHM)"
#include <gsi/gsi_debug.h>

#include "shm.h"
//#include "gsi_server.h"
#include "../client.h"
#include "../log.h"


/*---------------------------------------------------------------------------*
 *      First function to be called.
 *---------------------------------------------------------------------------*/
int get_shmid(int size, struct gsis_shm *shm) {
#if USE_SHARED_MEMORY
#  if HAVE_SYSV_SHARED_MEMORY
	key_t ipc_key;
	int   shmid;
	char *segptr;
	int   ipc_permissions;

	ipc_key = rand();
	ipc_permissions = 0600;

	if (share_server) {
		ipc_permissions = af_local_permissions;
	}

	/* Open the shared memory segment - create if necessary */
	while ((shmid = shmget(ipc_key, size + GSI_SHM_BLOCK_SIZE, IPC_CREAT | IPC_EXCL | ipc_permissions)) == -1) {
		if (errno == EEXIST) {
			DEB1(DPRT("SHM  segment with id=%d exists - next key\n", shmid));
		}
		else {
			return -1;
		}
		ipc_key = rand();
	}

	/* Attach (map) the shared memory segment into the current process */
	if ((segptr = shmat(shmid, 0, 0)) == (void *) (-1)) {
		DEB1(perror("shmat"));
		log_message(GSI_MESSAGE_ERROR, "shmat: %s", strerror(errno));
		//syslog(LOG_ERR, "shmat: %s", strerror(errno));
		return -1;
	}

	shm->buffer = segptr;

	strcpy(shm->buffer+size, GSI_SHM_ID);
	return shmid;
#  endif
#  if HAVE_POSIX_SHARED_MEMORY
#  endif
#else
	return -1;
#endif
}


/*---------------------------------------------------------------------------*
 *      Second function to be called.
 *---------------------------------------------------------------------------*/
int attach_shm(int shmid) {
#if USE_SHARED_MEMORY
#  if HAVE_SYSV_SHARED_MEMORY
	char *segptr;

	if ((segptr = shmat(shmid, 0, 0)) == (void *) (-1)) {
		DEB1(perror("shmat"));
		log_message(GSI_MESSAGE_ERROR, "shmat: %s", strerror(errno));
		return -1;
	}
	return 0;
#  endif
#  if HAVE_POSIX_SHARED_MEMORY
#  endif
#else
	return -1;
#endif
}

int del_shm(char *shmaddr) {
#if USE_SHARED_MEMORY
#  if HAVE_SYSV_SHARED_MEMORY
	int status;

	status = shmdt(shmaddr);

	return status;
#  endif
#  if HAVE_POSIX_SHARED_MEMORY
#  endif
#else
	return -1;
#endif
}


/*---------------------------------------------------------------------------*
 *      Removing shm could be done in the server or the library...
 *---------------------------------------------------------------------------*/
#if 0
void remove_shm(int shmid)
{
	shmctl(shmid, IPC_RMID, 0);
	DEB1(DPRT("SHM  segment with id=%d marked for deletion\n", shmid));
}

void change_shm_mode(int shmid, char *mode)
{
	struct shmid_ds myshmds;

	DEB1(DPRT("SHM  Change permissions\n"));

	/* Get current values for internal data structure */
	shmctl(shmid, IPC_STAT, &myshmds);

	DEB1(DPRT("SHM  Old permissions were: %o\n", myshmds.shm_perm.mode));

	/* Update the mode */
	sscanf(mode, "%o", &myshmds.shm_perm.mode);
	shmctl(shmid, IPC_SET, &myshmds);

	DEB1(DPRT("SHM  New permissions are : %o\n", myshmds.shm_perm.mode));
}
#endif


/*      EOF shm.c */
