/* GSI - general sound interface
 *
 * convert:  converts instrument files from .pat/.ma/.mb -> .ma/.mb.
 *
 * By W.H.Scholten dd 2000
 *
 * This software is placed in the public domain and can be used freely
 * for any purpose. It comes without any kind of warranty, either
 * expressed or implied, including, but not limited to the implied
 * warranties of merchantability or fitness for a particular purpose.
 * Use it at your own risk. the author is not responsible for any damage
 * or consequences raised by use or inability to use this program.
 */

#include <gsi/kpmidi.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../lib/gsimidi_pcm/instrument/write_instrument.h"
#include "../../lib/gsimidi_pcm/patchmap.h"


void
use (char *program)
{
	fprintf (stderr, 
		 "Use: %s [-fixed_freq] [-fixed_velocity] [-release_loop]\n"
		 "\t[-scale <scale>] input_file output_file\n\n", program);
	exit (0);
}

#define INSTRUMENT_UNKNOWN   0
#define INSTRUMENT_MB        1
#define INSTRUMENT_MA        2
#define INSTRUMENT_GUS       3

char *instr_suffix[ 3 ] = {
	_GSIMIDIPCM_INSTRUMENT_NATIVE_BINARY__SUFFIX,
	_GSIMIDIPCM_INSTRUMENT_NATIVE_TEXT__SUFFIX,
	_GSIMIDIPCM_INSTRUMENT_GUS__SUFFIX
};


int main (int argc, char **argv)
{
	char    *program = argv[0];
	int      args_used = 0;
	int      status = 0;
	int      i;

	char    *instrument_in, *instrument_out;
	int      input_format, output_format;

	const MidiPatch *patch = NULL;
	unsigned long    flags = 0;
	float            scale = 1.0f;


	argv++;

	while (args_used < argc-1 && **argv == '-') {
		if (!strcmp (*argv, "-fixed_freq"))
			flags |= INSTRUMENT_FIXED_FREQUENCY;
		else if (!strcmp (*argv, "-fixed_velocity"))
			flags |= INSTRUMENT_FIXED_VELOCITY;
		else if (!strcmp (*argv, "-release_loop"))
			flags |= INSTRUMENT_RELEASE_LOOP;
		else if (!strcmp (*argv, "-scale")) {
			scale = atof (argv[1]);
			argv++;
			args_used++;
		}
		else {
			use (program);
		}
		argv++;
		args_used++;
	}

	input_format = INSTRUMENT_UNKNOWN;
	output_format = INSTRUMENT_UNKNOWN;

	if (argc - args_used - 1 < 2) use(program);

	instrument_in = argv[0];
	instrument_out = argv[1];

	for (i=0; i<3; i++) {
		if (strlen(instrument_in) > strlen(instr_suffix[i]) && (!strcmp(instrument_in+strlen(instrument_in)-strlen(instr_suffix[i]), instr_suffix[i]))) {
			input_format = 1+i;
			break;
		}
	}
	if (input_format == INSTRUMENT_UNKNOWN) {
		fprintf(stderr, "Couldn't determine input format\n");
		exit(1);
	}

#if 0
	if (args_used && (output_format != INSTRUMENT_MA)) {
		fprintf(stderr, "Arguments only useful for MA instruments\n");
		exit(1);
	}
#endif

	for (i=0; i<3; i++) {
		if (strlen(instrument_out) > strlen(instr_suffix[i]) && (!strcmp(instrument_out+strlen(instrument_out)-strlen(instr_suffix[i]), instr_suffix[i]))) {
			output_format = 1+i;
			break;
		}
	}
	if (output_format == INSTRUMENT_UNKNOWN) {
		fprintf(stderr, "Couldn't determine output format\n");
		exit(1);
	}

	switch (input_format) {
	case INSTRUMENT_MB:
		patch = MidiPatchMap (instrument_in);
		break;
	case INSTRUMENT_MA:
		patch = MidiPatchParse (instrument_in);
		break;
	case INSTRUMENT_GUS:
		patch = ReadGUSPatch (instrument_in);
		break;
	}
	if (!patch) {
		fprintf (stderr, "Can't read instrument %s\n", instrument_in);
		exit (1);
	}

	switch (output_format) {
	case INSTRUMENT_MB:
		status = InstrumWriteBinary (patch, scale, /* instrument */ instrument_out, instrument_out);
		break;
	case INSTRUMENT_MA:
		status = InstrumWriteText (patch, scale, /* instrument */ instrument_out, instrument_out);
		break;
	case INSTRUMENT_GUS:
		status = 0;
		break;
	}
	if (!status) {
		fprintf (stderr, "Can't write instrument %s\n", instrument_out);
		exit (1);
	}

	return 0;
}
